/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// Name: AutoPlane.h
// Desc: An AutoPlane is a plane that has the ability to fly automatically,
//       without the need for human intervention. This class provides
//       routines that let a plane fly towards a given point in the world 
//       whilst doing its best to avoid collisions with the landscape and
//       other planes. This class also provides routines that let a plane fly
//       home and refuel, these routines are used by both the PlayerPlane and
//       EnemyPlane classes.
//------------------------------------------------------------------------------

#ifndef _AUTOPLANE_H_
#define _AUTOPLANE_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include "CamelMath.h"
#include "Plane.h"

class AutoPlane : public Plane
{
public:
    static const int NEAR_WAY_POINT;
    static const int COLLISION_RANGE;

    // Determines the rate at which the plane checks to see if it will
    // collide with something when considering whether a given action
    // is dangerous.
    static const int COLLISION_CHECK_RATE;

    struct WAYPOINT {
        Point2D Location;
        uint NextPoint;
    };

    static const int NUM_WAYPOINTS_ON_ROUTE_HOME = 6;

    // These are used to calculate the location of the waypoints leading up
    // to the runway.
    static const Point2D HOME_WAYPOINT_OFFSETS[NUM_WAYPOINTS_ON_ROUTE_HOME-1];

protected:
    WAYPOINT RouteHome[NUM_WAYPOINTS_ON_ROUTE_HOME];

    bool FlyingHome;
    uint CurWayPoint;

    // Used when performing a loop
    int OldTargetDirection;
    bool ManoeuvreCompleted;

public:
    AutoPlane(Level *L, PLANE_STATE State, Point2D Pos, bool Inverted, 
        int Direction, ENT_SIDE Side);
    ~AutoPlane();

    void FlyHome();
    void StopFlyingHome();

    void ResetAutopilot();

    void FlyTo(Point2D p);
    bool IsDangerous(PLANE_ACTIONS AnAction);
    bool IsStallLikely(float CurVelocity, int CurDirection, bool CurInverted);
    bool IsCollisionLikely(Plane *APlane, Point2D *CollisionPoint, int *CollisionDirection);
};

#endif // _AUTOPLANE_H_
