/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: CamelMath.h
// Desc: Provides any mathematical constants or objects that are needed
//       throughout the program. This is done mainly with the standard C math
//       library and also with a few custom objects.
//
//       Included in this library are
//
//       CamelMath: This object provides a number of useful static functions
//           that are maths based and used throughout Camel. It also provides
//           constants and functions to translate from radians to discrete
//           directions used for moving some entites around the game world.
//       Point2D: Represents an integer point in 2D space.
//------------------------------------------------------------------------------

#ifndef _CAMELMATH_H_
#define _CAMELMATH_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#define _USE_MATH_DEFINES
#include <math.h>
#include "globals.h"

extern const float PI;
extern const float T_PI;
extern const float PI_2;
extern const float T_PI_8;
extern const float PI_4;
extern const float PI_8;

class Point2D 
{
public:
    int x, y;

    Point2D();
    Point2D(int x, int y);
    ~Point2D();

    float DistanceTo(Point2D p);
    float BearingTo(Point2D p);
};

struct XY_COMPONENT {
    float x, y;
};

class CamelMath
{
public:
    // Discrete directions are used in a number of areas of the game to reduce
    // the number of frames of animation that are needed. This requires a number
    // of helper functions to translate from continuous directions measured in 
    // Radians.
    static const int NUM_DIRECTIONS = 16;

    // Used to help calculate the general direction
    // It is assumed that direction 0 always points directly RIGHT
    static const int UP_DIRECTION;
    static const int DOWN_DIRECTION;

    enum GENERAL_DIRECTION {
        UP,
        DOWN,
        LEFT,
        RIGHT
    };

    // Half the difference between two directions in radians
    static const float HALF_DIRECTION_DIFF;         
    static const float DIRECTION_BEARINGS[NUM_DIRECTIONS];
    static const float AIR_THICKNESS;

    // Acceleration values are given in pixels per frame
    static const float MAX_GRAVITY_ACC;
    static const float GRAVITY_ACC[NUM_DIRECTIONS];
    static const XY_COMPONENT XY_COMPONENTS[NUM_DIRECTIONS];

protected:
    // This object cannot be created or destroyed, it is just a collection of
    // static functions and constants.
    CamelMath() {}
    ~CamelMath() {}

public:
    static int Random(int Max) { return Random(0, Max); }
    static int Random(int Start, int End);

    static int NormaliseDirection(int Direction);
    static GENERAL_DIRECTION GetGeneralDirection(int Direction);
    static int GetDirectionFromBearing(float Bearing);
};

#endif // _CAMELMATH_H_
