/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: CamelUtility.h
// Desc: Utility functions that I can use when programming. At the moment it
//       consists of a mechanism for printing debug messages, a system for
//       reporting errors and for WIN32 only, a very basic profiler
//
//       Printing Debugging Messages: CU_Dbg_Printf can be used to output
//         messages to a logfile or debugger. Only works when _DEBUG is 
//         defined. CU_DBG must be placed at the beginning of any new lines
//         if the call to CU_Dbg_Printf spreads across multiple lines
//
//       Reporting Errors: Batches of user defined errors can be specified
//         by using CU_AddErrorMessages. Errors can then be set using
//         the CU_SetLast###Error error functions and printed using 
//         CU_PrintLastError. _DEBUG must be defined to use error reporting
//
//       Profiling: This works by defining a series of nested frames and then
//         using macros to indicate when the program enters and exits these
//         frames. At the end of the program, som profiling data can be 
//         printed out. Profilng only works when _DEBUG, WIN32 and 
//         _CU_PROFILING are defined
//
// 2/11/03: 
//       Added a function to trim the whitespace from a  standard C++ 
//       library string.
//
// 27/11/03:
//       I'm probably missing something blatantly obvious, but there doesn't 
//       seem to be any way of listing the files in a directory using either
//       the C runtime library or the STL. I have therefore provided a 
//       function called CU_ListFiles which when given the name of a directory
//       returns a std::list of strings which contain the names of all of 
//       the files in the directory.
//------------------------------------------------------------------------------

#ifndef _CAMELUTILITY_H_
#define _CAMELUTILITY_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <list>
#include <string>
#include "globals.h"
#include "stdio.h"

//------------------------------------------------------------------------------
// Defines
//------------------------------------------------------------------------------
#define MAX_DEBUG_STRING_LENGTH     8192
#define USER_DEFINED_ERROR_MASK     0x20000000
           
//------------------------------------------------------------------------------
// Typedefs, Structs and Enums
//------------------------------------------------------------------------------

typedef unsigned long DWORD;
typedef long HRESULT;
union _LARGE_INTEGER;
typedef _LARGE_INTEGER LARGE_INTEGER;

#ifdef WIN32
namespace Gdiplus 
{
    enum Status;
}
#endif

typedef struct _CU_TIMINGFRAMEDATA {
    char *Name;
    int Children; // Head of list of children
    int Next;
} CU_TIMINGFRAMEDATA;

typedef struct _CU_TIMINGFRAME {
    char *Name;
    unsigned int HitCount;
#ifdef WIN32
    __int64 StartTicks;
    __int64 EndTicks;
    __int64 TotalTicks;
#else
    // james 
    unsigned int StartTicks;
    unsigned int EndTicks;
    unsigned int TotalTicks; //Uint32 
#endif
    struct _CU_TIMINGFRAME *Children;
    struct _CU_TIMINGFRAME *Next;
} CU_TIMINGFRAME;

typedef struct _CU_ERRORMESSAGES {
    uint ErrorBase;
    uint NumErrors;
    char **ErrorStrings;
} CU_ERRORMESSAGES;
                                    
//------------------------------------------------------------------------------
// Function Prototypes
//------------------------------------------------------------------------------

void CU_ShowMessage(char *title, char *fmt, ...);
std::string CU_TrimString(std::string AString);

#if defined(WIN32)
void CU_ListFilesWIN32(const char *dir, std::list<std::string>& FileList);
#define CU_ListFiles CU_ListFilesWIN32
#elif defined(LINUX) 
void CU_ListFilesLINUX(const char *dir, std::list<std::string>& FileList);
#define CU_ListFiles CU_ListFilesLINUX
#endif

// Debugging
#ifdef _DEBUG

void CU_Dbg_Printf(FILE *dst, char *fmt, ...);
void CU_AddErrorMessages(CU_ERRORMESSAGES ErrorMessages);

#if defined(WIN32) && defined(_GDIPLUS)
void CU_SetLastGdiPlusError(Gdiplus::Status Err);
#endif
#if defined(WIN32) && defined(_DIRECTX)
void CU_SetLastDirectXError(HRESULT Err);
#endif

#ifdef _SDL
void CU_SetLastSDLError();
void CU_SetLastMixError();
#endif 

void CU_SetLastError(DWORD Err);
void CU_PrintLastError(char *Filename, DWORD LineNum);

#define CU_DBG

#else // _DEBUG

#define CU_AddErrorMessages(x)
#define CU_SetLastGdiPlusError(x)
#define CU_SetLastDirectXError(x)
#define CU_SetLastSDLError()
#define CU_SetLastMixError()
#define CU_SetLastError(x)
#define CU_PrintLastError(x, y)

#define CU_Dbg_Printf /*james/##/*/
#define CU_DBG /*james/##/*/

#endif

#define ReturnWithGdiPlusError(x, Error) \
    { CU_SetLastGdiPlusError(Error); return x; }
#define ReturnWithDirectXError(x, Error) \
    { CU_SetLastDirectXError(Error); return x; }
#define ReturnWithSDLError(x) \
    { CU_SetLastSDLError(); return x; }
#define ReturnWithMixError(x) \
    { CU_SetLastMixError(); return x; }
#define ReturnWithError(x, Error) \
    { CU_SetLastError(Error); return x; }
                                    
#define FailWithGdiPlusError(Error) ReturnWithGdiPlusError(false, Error)
#define FailWithDirectXError(Error) ReturnWithDirectXError(false, Error)
#define FailWithSDLError() ReturnWithSDLError(false)
#define FailWithMixError() ReturnWithMixError(false)
#define FailWithError(Error) ReturnWithError(false, Error)

// Set, display and then return x
#define SDR_GdiplusError(x, Error, File, Line) \
    { CU_SetLastGdiPlusError(Error); \
      CU_PrintLastError(File, Line); \
      return x; }
#define SDR_DirectXError(x, Error, File, Line) \
    { CU_SetLastDirectXError(Error); \
      CU_PrintLastError(File, Line); \
      return x; }
#define SDR_SDLError(x, File, Line) \
    { CU_SetLastSDLError(); \
      CU_PrintLastError(File, Line); \
      return x; }
#define SDR_MixError(x, File, Line) \
    { CU_SetLastMixError(); \
      CU_PrintLastError(File, Line); \
      return x; }
#define SDR_Error(x, Error, File, Line) \
    { CU_SetLastError(Error); \
      CU_PrintLastError(File, Line); \
      return x; }
      
// Set, display and then fail (return false)
#define SDF_GdiplusError(Error, File, Line) \
    SDR_GdiplusError(false, Error, File, Line)
#define SDF_DirectXError(Error, File, Line) \
    SDR_DirectXError(false, Error, File, Line)
#define SDF_SDLError(File, Line) \
    SDR_SDLError(false, File, Line)
#define SDF_MixError(File, Line) \
    SDR_MixError(false, File, Line)
#define SDF_Error(Error, File, Line) \
    SDR_Error(false, Error, File, Line)

// Set, display and then exit the program
#define SDE_GdiplusError(Error, File, Line) \
    { CU_SetLastGdiPlusError(Error); \
      CU_PrintLastError(File, Line); \
      exit(-1); }
#define SDE_DirectXError(Error, File, Line) \
    { CU_SetLastDirectXError(Error); \
      CU_PrintLastError(File, Line); \
      exit(-1); }
#define SDE_SDLError(File, Line) \
    { CU_SetLastSDLError(); \
      CU_PrintLastError(File, Line); \
      exit(-1); }
#define SDE_MixError(File, Line) \
    { CU_SetLastMixError(); \
      CU_PrintLastError(File, Line); \
      exit(-1); }
#define SDE_Error(Error, File, Line) \
    { CU_SetLastError(Error); \
      CU_PrintLastError(File, Line); \
      exit(-1); }

// Profiling
#if defined(_DEBUG) && defined(_CU_PROFILING)

extern CU_TIMINGFRAME *TFArray;

void CU_InitialiseProfiling(CU_TIMINGFRAMEDATA *TFData);
void CU_PrintProfilingResults();
void CU_ResetFrames();

// Store the starting ticks from the performance counter
#define CU_STARTFRAME(x) { \
    TFArray[x].HitCount++; \
    QueryPerformanceCounter((LARGE_INTEGER *)&TFArray[x].StartTicks); }
// Work out the number of ticks that have passed and add to the 
// frames total ticks.
#define CU_ENDFRAME(x) { \
    QueryPerformanceCounter((LARGE_INTEGER *)&TFArray[x].EndTicks); \
    TFArray[x].TotalTicks += TFArray[x].EndTicks - TFArray[x].StartTicks; }

#else

#define CU_InitialiseProfiling(x)
#define CU_PrintProfilingResults()
#define CU_ResetFrames()
#define CU_STARTFRAME(x)
#define CU_ENDFRAME(x)

#endif

#endif // _CAMELUTILITY_H_
