/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: ControlledEntity.h
// Desc: A ControlledEntity is a FightingEntity that can be viewed through
//       the in game OrdersPanel. The ControlledEntity provides details of its
//       current orders, the number of entities it contains and an icon in the
//       form of a sprite. It must also respond to the predefined orders
//       ADVANCE, HOLD and WITHDRAW.
//
// Note: A ControlledEntity is usually viewed as a unit containing a number
//       of subordinate entities. Therefore it must also maintain a count of
//       the number of subordinate entities that it contains and how many of
//       them are left alive.
//------------------------------------------------------------------------------

#ifndef _CONTROLLED_ENTITY_H_
#define _CONTROLLED_ENTITY_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <string>
#include "FightingEntity.h"

class OrdersPanel;

class ControlledEntity : public FightingEntity
{
public:
    class Order
    {
        std::string OrderString;
    public:
        Order(const char *OrderString);

        const char *GetOrderString() const;
    };

    static const Order ADVANCE;
    static const Order HOLD;
    static const Order WITHDRAW;
protected:
    OrdersPanel *OP;

    int IconFrame;  // The sprite frame used to represent the ControlledEntity
    int NumSubordinatesAlive;   
    int DestinationX;

    // Variables related to the flag that is displayed to represent the 
    // ControlledEntity destination
    int FlagFrame;
    Point2D FlagPos;
    int FlagWidth, FlagHeight;

    const Order *CurOrders;
    bool OrdersHaveChanged;

    inline void UpdateFlagPos();
    virtual bool UpdateDestinationX(const Order *NewOrders, int NewDestinationX);
public:
    ControlledEntity(Level *L, ENT_SIDE Side);
    virtual ~ControlledEntity();

    int GetIconFrame() { return IconFrame; }
    int GetNumSubordinatesAlive() { return NumSubordinatesAlive; }
    int GetDestinationX() { return DestinationX; }
    virtual int GetSubordinateDestinationX() = 0;
    // This returns the distance at which a ControlledEntity is 'near' to something
    virtual int GetNearDestinationX() = 0;   

    void SetOrders(const Order *NewOrders, int NewDestinationX = 0);
    const Order *GetOrders() { return CurOrders; }

    void Draw(SDL_Surface *ScreenBuffer);
};

#endif // _CONTROLLED_ENTITY_H_
