/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Entity.h
// Desc: The abstract base class from which everything that moves, shoots 
//       and/or blows up in the game is derived.
//------------------------------------------------------------------------------

#ifndef _ENTITY_H_
#define _ENTITY_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include "globals.h"
#include "CamelMath.h"
#include "SDL_Interface.h"

class Game;
class InputHandler;
class Sprites;
class Level;
class Landscape;

class Entity
{
public:
    // This is the default value for the MaxCollisions variable, most entities
    // will use this but some may want to change it. A soldiers bullet for 
    // example can only collide with one person. The MaxCollisions variable must
    // not be greater than this however.
    static const int ABSOLUTE_MAX_COLLISIONS = 12;

    enum ENT_SIDE {
        NEUTRAL = -1,
        GOOD,
        EVIL,
        NUM_SIDES
    };
    static const char *ENT_SIDE_STRINGS[NUM_SIDES];

    enum ENT_TYPE {
        NO_TYPE = -1,
        BULLET,
        FACTORY,
        PLANE,
        SOLDIER,
        TANK,
        BOMB,
        EXPLOSION,
        PARTICLE,
        FLAG,
        PLATOON,
        TANKSQUAD,
        PLANESPAWNPOINT,
        PLATOONSPAWNPOINT,
        TANKSQUADSPAWNPOINT,
        NUM_ENT_TYPES
    };

    typedef Entity *COLLISIONS_ARRAY[ABSOLUTE_MAX_COLLISIONS];
protected:
    struct ACTION {
        ACTION(uint Delay, uint Rate, bool Active, void (*DoAction)(void *));
        void Reset();

        uint Delay;
        // Number of milliseconds needed for the action to occur. This should be in
        // multiples of 30 to keep everything consistent
        uint Rate;  
        bool Active;
        void (*DoAction)(void *);
    };

    ENT_SIDE Side;
    ENT_TYPE Type;
    bool NeedsDeletion;
    // The maximum number of entities that an entity can collide with at any
    // one time. 
    int MaxCollisions;

    // Pointers to objects used by the entity for easy reference.
    // These will not change throughout the life of an entity so are safe, but
    // I would like to now of a cleaner more OO way of doing this.
    Game *G;
    SDL_Interface *SDLI;
    InputHandler *IH;
    Sprites *SPR;
    Level *L;       // The level to which this entity belongs
    Landscape *LS;

    // RealX and RealY don't necessarily have to be kept up to date
    float RealX, RealY; 
    Point2D WorldPos;
    int Width, Height;
    int Frame;

    int NumVictims;
    ENT_TYPE Victims[NUM_ENT_TYPES];

    static void UpdateAction(ACTION *AnAction, bool Command, uint dt, void *Object);
    
    bool TestForCollision(int E1_Frame, Point2D E1Pos, int E2_Frame, Point2D E2Pos);

public:
    Entity(Level *L);
    virtual ~Entity();

    ENT_SIDE GetSide() { return Side; }
    ENT_TYPE GetType() { return Type; }
    virtual void Update(uint dt) = 0;
    virtual void Draw(SDL_Surface *ScreenBuffer);
    virtual int FindCollisions(COLLISIONS_ARRAY *Collisions);
    virtual bool HasCollided(int OtherEntityFrame, Point2D OtherEntityFramePos);
    virtual void FindAndHandleCollisionWithLandscape();
    virtual void HandleCollision(Entity *AnEntity);

    bool GetNeedsDeletion() { return NeedsDeletion; }
    int GetWorldX() { return WorldPos.x; }
    int GetWorldY() { return WorldPos.y; }
    Point2D GetWorldPos() { return WorldPos; }
    int GetWidth() { return Width; }
    int GetHeight() { return Height; }
    int GetCentreX() { return WorldPos.x + (Width >> 1); }
    int GetCentreY() { return WorldPos.y + (Height >> 1); }
    Point2D GetCentrePos() { return Point2D(WorldPos.x + (Width >> 1), WorldPos.y + (Height >> 1)); }

    // Useful way of dealing with the Entities bounding box
    int Top() { return WorldPos.y; }
    int Left() { return WorldPos.x; }
    int Bottom() { return WorldPos.y + Height; }
    int Right() { return WorldPos.x + Width; }

    int GetFrame() { return Frame; }
};

#endif // _ENTITY_H_
