/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// Name: FightingEntity.h
// Desc: A FightingEntity is an enity that can become involved in combat upon
//       the ground. A FightingEntity needs to be able to give some indication
//       of how far it can currently fire and whether or not it is fleeing
//       (fleeing uints cannot be engaged in combat).
//
//       This entity can be used to provide a subtle distinction between
//       entities that fighting units will attack and entities that will just
//       take damage as the result of a combat going on nearby. For example,
//       a Factory is not a FightingEntity (although it could be) and is 
//       therefore not targeted specifically by enemy units. However, AA guns
//       if they were to be implemented could be implemented as a 
//       FightingEntity. They would have a range of 0 and never flee, but 
//       enemy units would see them as a FightingEntity and specifically 
//       attack them.
//------------------------------------------------------------------------------

#ifndef _FIGHTING_ENTITY_H_
#define _FIGHTING_ENTITY_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include "CamelMath.h"
#include "Entity.h"

class FightingEntity : public Entity
{
public:
    // Morale can range between 0 and 100, these constants are available to 
    // improve code readability.
    static const float MAX_MORALE;
    static const float MIN_MORALE;
protected:
    float Morale;       // How brave the entity is feeling

    FightingEntity *FindClosestEnemyInRange();
    bool FightingEntityInRange(FightingEntity *EnemyFE, bool& OnTheLeft,
        int& RelativeDistance, int& MaxRange, int& NearestX, Point2D& TargetPos);

    virtual int GetAvailableRange(int HeightDiff) = 0;
    virtual int GetTargetHeight() = 0;
    static Point2D GetShotPos(FightingEntity *FE, int x);
public:
    FightingEntity(Level *L);
    virtual ~FightingEntity();

    float GetMorale() { return Morale; }
    virtual bool IsFleeing() = 0;
    virtual bool IsWithdrawing() = 0;
};

#endif // _FIGHTING_ENTITY_H_
