/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Game.h
// Desc: This object is created on start up to run the game, it contains all
//       of the objects used by the game as well as game initialisation,
//       shutdown and the main game loop
//------------------------------------------------------------------------------

#ifndef _GAME_H_
#define _GAME_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <list>
#include "globals.h"
#include "CamelMath.h"
#include "SDL_Interface.h"
#include "Level.h"
#include "Entity.h"

class PG_Application;
class InputHandler;
class MenuController;
class Sprites;
class OrdersPanel;
class Level;
class Landscape;
class PlayerPlane;
class EnemyPlane;
class Platoon;

class Game 
{
public:
    // The time in milliseconds that music takes to fade in or out
    static const int MUSIC_FADE_TIME;

    // The height of the user interface which is displayed at the bottom of the screen
    static const int UI_HEIGHT;

    // Constants used to draw the needles on the user interface gauges
    static const double UI_GAUGE_NEEDLE_START_ANGLE;
    static const double UI_GAUGE_NEEDLE_END_ANGLE;
    static const double UI_GAUGE_NEEDLE_RANGE;
    static const int UI_GAUGE_NEEDLE_LENGTH;

    enum GAME_STATE {
	STOPPED,
	PAUSED,
	PLAYING,
        ISSUING_ORDERS,
        ENDING_LEVEL,
        EXITING
    };

    enum GAME_OPTIONS {
        NO_OPTION = -1,
        FULLSCREEN,
        SCREENMODE,
        BITDEPTH,
        NUM_PARALLAX_LAYERS,
        KEY_CLIMB,
        KEY_DIVE,
        KEY_INVERT,
        KEY_FIRE,
        KEY_START_ENGINE,
        KEY_DROP_BOMB,
        KEY_ISSUE_ORDERS,
        KEY_FLY_HOME,
        CLOCK_CONTROLS,
        SFX_VOLUME,
        MUSIC_VOLUME,
        REVERSE_STEREO,
        MENU_MUSIC,
        DISPLAY_FPS,
        USE_DEBUG_CAMERA,
        NUM_OPTIONS
    };

    struct OPTION_INFO {
        char *String;
        char *Comment;
        char *Tooltip;
    };

    static const int MAX_HIGH_SCORE_NAME_LENGTH = 10;
    struct HIGH_SCORE {
        char Name[MAX_HIGH_SCORE_NAME_LENGTH+1];
        char LevelName[Level::MAX_LEVEL_NAME_LENGTH+1];
        uint Score;
    };
    static const int NUM_HIGH_SCORES = 10;
    
protected:
    // Time in milliseconds that the 'Game Over' or 'You Win!' banner is displayed for
    static const int END_OF_LEVEL_DELAY;

    static const char *OPTIONS_DELIMITER;
    static const int DEFAULT_OPTIONS[NUM_OPTIONS];
    static const OPTION_INFO OPTION_DETAILS[NUM_OPTIONS];

    static const HIGH_SCORE DEFAULT_HIGH_SCORES[NUM_HIGH_SCORES];

    static Game *Instance;

    // Subsystems
    PG_Application *PGApp;
    SDL_Interface *SDLI;
    InputHandler *IH;
    MenuController *MC;
    Sprites *SPR;
    OrdersPanel *OP;
    
    Level *CurLevel;
    Landscape *LS;
    int Options[NUM_OPTIONS];
    HIGH_SCORE HighScores[NUM_HIGH_SCORES];
    std::string LevelName;
    uint FinalScore;

    std::string MenuMusicFile;
    Mix_Music *MenuMusic;
        
	GAME_STATE GameState;
    GAME_STATE OldState;

    // These variables are used to keep track of the players plane
    PlayerPlane *PPlane; 
    Point2D PPlanePos;
    int PPlaneBombsLeft;
    int PPlaneBulletsLeft;
    int PPlaneFuelLeft;
    float PPlaneVelocity;

    Point2D DebugCameraPos;

    Game();

    void SortHighScores();
public:
    ~Game();
    static Game *GetInstance();

    void RunMainLoop();

    Level *GetCurLevel() { return CurLevel; }
    std::string GetLevelName() { return LevelName; }
    void LoadLevel(const char *Filename);
	void StartNewGame();
    void PauseGame();
	void ContinueGame();
	void QuitGame();
    void Exit();
    void OpenOrdersPanel();
    void CloseOrdersPanel();
	GAME_STATE GetGameState() { return GameState; }

    void LoadOptions();
    void SaveOptions();
    int GetOption(int AnOption);
    void SetOption(int AnOption, int OptionValue);
    const char *GetOptionTooltip(int AnOption);

    uint GetFinalScore() { return FinalScore; }
    void LoadHighScores();
    void SaveHighScores();
    HIGH_SCORE GetHighScore(int Index);
    void SetHighScore(int Index, HIGH_SCORE NewHighScore);
    
    void DrawGameFrame(SDL_Surface *Surface);
    void DrawUserInterface(SDL_Surface *Surface);
    void DrawGaugeNeedle(SDL_Surface *Surface, Point2D NeedleStartPos, 
        int RangeStart, int RangeEnd, int Value);
};

#endif // _GAME_H_

