/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Landscape.h
// Desc: This file provides functions to draw the landscape and move the
//       viewport around.
//------------------------------------------------------------------------------

#ifndef _LANDSCAPE_H_
#define _LANDSCAPE_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include "globals.h"
#include "Camel.h"
#include "CamelMath.h"
#include "SDL_Interface.h"
#include "Entity.h"

#include <string>
#include <fstream> 

class Game;
class Sprites;

class Landscape
{
public:
    // These are used to help store the contour of the landscape so that units
    // on the ground can quickly find out what they are standing on.
    enum SLOPE_TYPE {
        SLOPE_UP,
        FLAT,
        SLOPE_DOWN,
        NUM_SLOPE_TYPES
    };

    // Constants related to the tiles that make up a landscape
    static const int TILE_WIDTH = 10;
    static const int TILE_HEIGHT = 10;
    static const int TILE_NUM_PIXELS = TILE_WIDTH * TILE_HEIGHT;
    static const int TILESET_WIDTH = 16;
    static const int TILESET_HEIGHT = 16;
    static const int TILESET_NUM_TILES = TILESET_WIDTH * TILESET_HEIGHT;
    static const int TILESET_IMAGE_WIDTH = TILESET_WIDTH * TILE_WIDTH;
    static const int TILESET_IMAGE_HEIGHT = TILESET_HEIGHT * TILE_HEIGHT;
    static const int TILESET_IMAGE_BPP = CAM_INTERNAL_SURFACE_BPP;
    static const int NO_TILE = 0xFF;

    // Constraints on the size of the landscape
    static const int MIN_LEVEL_WORLDWIDTH = 80;
    static const int MAX_LEVEL_WORLDWIDTH = 800;
    static const int MIN_LEVEL_WORLDHEIGHT = 30;
    static const int MAX_LEVEL_WORLDHEIGHT = 120;
    static const int MIN_LEVEL_LANDHEIGHT = 10;
    static const int MAX_LEVEL_LANDHEIGHT = 30;
	static const int MIN_LEVEL_BACKGROUNDHEIGHT = 0;

    static const int NUM_PARALLAX_LAYERS = 2;
    // The rate at which a layer moves horizontally compared to the foreground
    static const float PL_H_SPEED[NUM_PARALLAX_LAYERS];
    // The rate of vertical parallax motion compared to horizontal parallax 
    // motion. The reason that this isn't one to one is so that the height of
    // the parallax layers can be smaller.
    static const float PL_V_SPEED;

protected:
    struct PARALLAX_LAYER {
        SDL_Surface *Image;
        uint Height;
    };

    struct CONTOUR {
        // The height is in tiles and takes the bottom of the screen as 0
        int Height;  
        // Height * TILE_HEIGHT, precalculated to save time later
        int BasicPixelHeight;
        SLOPE_TYPE SlopeType;
    };

    SDL_Rect TileRects[TILESET_NUM_TILES];

    Game *G;
    SDL_Interface *SDLI;
    Sprites *SPR;

    // All measurents are in tiles which have a width and height of 
    // Landscape::TILE_WIDTH and Landscape::TILE_HEIGHT 
    std::string Tileset;
    uint WorldWidth;         
    uint WorldHeight;        
    uint LandHeight;
	uint BackgroundHeight;
    COLOUR_RGBA SkyColour0, SkyColour1;
    int ParallaxLayersLoaded;
    PARALLAX_LAYER ParallaxLayers[NUM_PARALLAX_LAYERS];
    CONTOUR *Contours;
    uchar *LandscapeData;
	uchar *BackgroundData;

    SDL_Surface *LandImage;         
    SDL_Surface *LandImageRLE;   // Compressed land image for fast blitting
    SDL_Surface *BackgroundImageRLE;   // Background image is always compressed

    // Variables that define the viewport that looks into the gameworld
    Point2D ViewportPos;
    int ViewportWidth, ViewportHeight;
    Point2D CurWorldPos;
    Point2D MaxWorldPos;
    SDL_Rect StoredClippingRect;

    // Dimensions of the world in pixels
    int WorldPixelWidth;
    int WorldPixelHeight;

    void DrawLandscape();

    void LoadLevelFile(const char *Filename);
    bool ReadInLevelFileHeader(std::ifstream *LevelFile);
    bool ReadInLandscapeData(std::string *CurLine, uchar *Data);
    void NormaliseContours();

    void CalculateTileRects(SDL_Rect *RectArray);
    SDL_Surface *LoadTileset(const char *Filename);
public:
    Landscape(const char *Filename);
    ~Landscape();

    void SetViewport(Point2D ScreenPos, int Width, int Height);
    void EnableViewportClippingRect();
    void DisableViewportClippingRect();
    bool SetViewportWorldPos(Point2D Pos);
    void CentreViewportAt(Point2D Pos);    
    void MoveViewport(int dx, int dy);
    bool DrawViewportBackground(SDL_Surface *ScreenBuffer);
    bool DrawViewportForeground(SDL_Surface *ScreenBuffer);
    bool HasCollided(uint Frame, Point2D p, bool UseContours);

    int GetViewportWorldX() { return CurWorldPos.x; }
    int GetViewportWorldY() { return CurWorldPos.y; }
    Point2D GetViewportWorldPos() { return CurWorldPos; }
    int GetViewportCentreWorldX() { return CurWorldPos.x + ViewportWidth / 2; }
    int GetViewportCentreWorldY() { return CurWorldPos.y + ViewportHeight / 2; }
    Point2D GetViewportCentreWorldPos() 
        { return Point2D(CurWorldPos.x + ViewportWidth / 2, 
            CurWorldPos.y + ViewportHeight / 2); }
    int GetViewportMaxWorldX() { return MaxWorldPos.x; }
    int GetViewportMaxWorldY() { return MaxWorldPos.y; }
    Point2D GetViewportMaxWorldPos() { return MaxWorldPos; }
    int GetViewportX() { return ViewportPos.x; }
    int GetViewportY() { return ViewportPos.y; }
    Point2D GetViewportPos() { return ViewportPos; }
    int GetViewportWidth() { return ViewportWidth; }
    int GetViewportHeight() { return ViewportHeight; }

    int GetWorldPixelWidth() { return WorldPixelWidth; }
    int GetWorldPixelHeight() { return WorldPixelHeight; }

    uint CalculatePixelHeight(int PixelX);
    bool LineOfSightExists(Point2D StartPos, Point2D EndPos);
    SLOPE_TYPE GetPixelSlopeType(int PixelX);
};

#endif // _LANDSCAPE_H_
