/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// Name: Level.h
// Desc: A level is loaded in from a '.lev' file and is the thing that a player
//       actually plays. A level contains a 2D landscape for the player to
//       explore along with entites (planes, soldiers, bombs etc) that can be
//       controlled and/or interacted with.
//------------------------------------------------------------------------------

#ifndef _LEVEL_H_
#define _LEVEL_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <list>
#include <fstream> 
#include <string>
#include "globals.h"
#include "CamelMath.h"
#include "SDL_Interface.h"
#include "Entity.h"
#include "EnemyPlane.h"
#include "Flag.h"

class Landscape;
class PlaneSpawnPoint;
class PlayerPlane;

class Level
{
public:
    static const int FILE_MAGIC_NUMBER;
    static const char *FILE_FIELD_DELIMITER;
    static const char *FILE_SECTION_DELIMITER;

    static const uint MAX_LEVEL_NAME_LENGTH = 20;

    // The number of entity types whose position can be defined by the level
    // definition script at the end of a '.lev' file.
    static const int NUM_PLACEABLE_ENTITIES = 8;

    // These constants are related to how the game plays. They may need to be
    // modified and it might be a good idea to move them out to the LD script.
    static const int NUM_SOLDIERS_IN_PLATOON;
    static const int NUM_TANKS_IN_TANKSQUAD;

    static const int STARTING_LIVES;
    static const uint STARTING_SCORE;
    static const uint FIRST_NEW_LIFE_SCORE;

protected:
    static const Entity::ENT_TYPE PLACEABLE_ENTITIES[NUM_PLACEABLE_ENTITIES];
    static const char *PLACEABLE_ENTITY_STRINGS[NUM_PLACEABLE_ENTITIES];
    static const char *LDS_DELIMITER;

    Landscape *LS;
    
    std::list<Entity *> NewEntities;
    PlaneSpawnPoint *PlayerPlaneSpawnPoint;
    Flag *Flags[Entity::NUM_SIDES];

    std::string Name;

    int PlayerLives;
    uint PlayerScore;
    uint OldPlayerScore;
    uint ScoreNeededForNewLife;
    bool GameOver;
    bool PlayerVictory;

    void ReadInLevelDetails(const char *Filename);
    void ReadInLevelDefinitionScript(const char *Filename);
    bool ReadInPatrolPoints(std::ifstream *LevelFile, 
        EnemyPlane::WAYPOINT PatrolPoints[EnemyPlane::MAX_NUM_PATROL_POINTS]);
    void AddAllNewEntities();
public:
    SDL_Surface *IntroScreen;
    Mix_Music *Music;
    bool NewTroopsReady;
    std::list<Entity *> GameEntities[Entity::NUM_ENT_TYPES];

    Level(const char *Filename);
    ~Level();

    Landscape *GetLandscape() { return LS; }

    void Update(uint dt);

    bool DrawBackground(SDL_Surface *ScreenBuffer);
    void DrawEntities(SDL_Surface *ScreenBuffer);
    bool DrawForeground(SDL_Surface *ScreenBuffer);

    void AddNewEntity(Entity *AnEntity);
    bool IsEntityAlive(Entity *AnEntity);
    PlayerPlane *GetPlayerPlane();
    static float GetAirResistance(Entity::ENT_TYPE Type, float Velocity);

    std::string GetName() { return Name; }
    int GetPlayerLives() { return PlayerLives; }
    void AddPlayerLife() { PlayerLives++; }
    void RemovePlayerLife() { PlayerLives--; }
    uint GetPlayerScore() { return PlayerScore; }
    void AddToPlayerScore(uint s) { PlayerScore += s; }
    bool GetGameOver() { return GameOver; }
    bool GetPlayerVictory() { return PlayerVictory; }
    Flag *GetFlag(Entity::ENT_SIDE Side) { return Flags[Side]; }
    static bool ReadNameFromLevelFile(const char *Filename, std::string& LevelName);
};

#endif // _LEVEL_H_
