/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: MenuController.h
// Desc: This static class provides the menus for the games. A menu is a 
//       collection of ParaGUI widgets linked to event handlers that
//       perform various actions such as starting a new game or changing the
//       input keys.
//
//       In the future, it might be desirable to allow an arbitrary menu
//       structure to be submitted to the menu controller, but for now the
//       menus are just going to be hardwired in.
//------------------------------------------------------------------------------

#ifndef _MENUCONTROLLER_H_
#define _MENUCONTROLLER_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <list>
#include <string>
#include <globals.h>
#include "BFont.h"
#include "Game.h"
#include <assert.h>
#include "SDL_Interface.h"
#include "InputHandler.h"
#include "paragui.h"
#include "pgeventobject.h"
#include "pgapplication.h"
#include "pgslider.h"
#include "pgcheckbutton.h"
#include "pgdropdown.h"

class Game;
class OrdersPanel;

class MenuController : PG_EventObject
{
public:
    static const char *TITLE_MENU_PIC;
    static const int NUM_HELP_SCREENS = 5;
    static const char *HELP_SCREEN_PICS[NUM_HELP_SCREENS];

    static const PG_Rect DEFAULT_MENU_AREA_RECT;

    enum MENUS {
        NO_MENU = -1,
        START_MENU,
        PICK_LEVEL_MENU,
        INTROSCREEN_MENU,
        OPTIONS_MENU,
        VIDEO_OPTIONS_MENU,
        AUDIO_OPTIONS_MENU,
        CONTROLS_OPTIONS_MENU,
        HELP_MENU,
        INGAME_MENU,
        NEW_HIGH_SCORE_MENU,
        HIGH_SCORES_MENU,
        NUM_MENUS
    };

    // Start menu rects
    static const PG_Rect SM_START_GAME_RECT;
    static const PG_Rect SM_OPTIONS_RECT;
    static const PG_Rect SM_HELP_RECT;
    static const PG_Rect SM_HIGH_SCORES_RECT;
    static const PG_Rect SM_EXIT_RECT;

    // Pick Level menu rects
    static const PG_Rect PLM_LEVEL_LIST_RECT;
    static const PG_Rect PLM_LEVEL_BUTTON_RECT;
    static const PG_Rect PLM_CANCEL_RECT;

    // IntroScreen menu rects
    static const PG_Rect ISM_CONTINUE_RECT;

    // Options menu rects
    static const PG_Rect OM_VIDEO_RECT;
    static const PG_Rect OM_AUDIO_RECT;
    static const PG_Rect OM_CONTROLS_RECT;
    static const PG_Rect OM_DONE_RECT;

    // Video Options menu rects
    static const PG_Rect VOM_FULLSCREEN_RECT;
    static const PG_Rect VOM_FULLSCREEN_CHECK_RECT;
    static const PG_Rect VOM_RESOLUTION_RECT;
    static const PG_Rect VOM_RESOLUTION_COMBO_RECT;
    static const PG_Rect VOM_BITDEPTH_RECT;
    static const PG_Rect VOM_BITDEPTH_COMBO_RECT;
    static const PG_Rect VOM_PARALLAX_RECT;
    static const PG_Rect VOM_PARALLAX_SLIDER_RECT;
    static const PG_Rect VOM_PARALLAX_LABEL0_RECT;
    static const PG_Rect VOM_PARALLAX_LABEL1_RECT;
    static const PG_Rect VOM_PARALLAX_LABEL2_RECT;
    static const PG_Rect VOM_DONE_RECT;
    static const PG_Rect VOM_TOOLTIP_RECT;

    // Audio Options menu rects
    static const PG_Rect AOM_SOUND_VOLUME_RECT;
    static const PG_Rect AOM_SOUND_VOLUME_SLIDER_RECT;
    static const PG_Rect AOM_MUSIC_VOLUME_RECT;
    static const PG_Rect AOM_MUSIC_VOLUME_SLIDER_RECT;
    static const PG_Rect AOM_REVERSE_STEREO_RECT;
    static const PG_Rect AOM_REVERSE_STEREO_CHECK_RECT;
    static const PG_Rect AOM_DONE_RECT;
    static const PG_Rect AOM_TOOLTIP_RECT;

    // Controls Options menu rects
    static const PG_Rect COM_OPTIONS_LIST_RECT;
    static const PG_Rect COM_TEMP_WIDGET_RECT;
    static const PG_Rect COM_OPTION_RECT;
    static const PG_Rect COM_OPTION_CHANGE_RECT;
    static const PG_Rect COM_CLOCK_CONTROLS_CHECK_RECT;
    static const PG_Rect COM_DONE_RECT;
    static const PG_Rect COM_TOOLTIP_RECT;

    // Help menu rects
    static const PG_Rect HM_PREV_RECT;
    static const PG_Rect HM_DONE_RECT;
    static const PG_Rect HM_NEXT_RECT;

    // InGame menu rects
    static const PG_Rect IGM_CONTINUE_GAME_RECT;
    static const PG_Rect IGM_OPTIONS_RECT;
    static const PG_Rect IGM_HELP_RECT;
    static const PG_Rect IGM_QUIT_GAME_RECT;
    static const PG_Rect IGM_EXIT_RECT;

    // New High Score menu rects
    static const PG_Rect NHSM_NAME_ENTRY_FORM_RECT;
    static const PG_Rect NHSM_CONGRATS_L1_RECT;
    static const PG_Rect NHSM_CONGRATS_L2_RECT;
    static const PG_Rect NHSM_NAME_REQUEST_RECT;
    static const PG_Rect NHSM_NAME_TEXTBOX_RECT;

    // High Score menu rects
    static const PG_Rect HSM_NAMETITLE_RECT;
    static const PG_Rect HSM_LEVELTITLE_RECT;
    static const PG_Rect HSM_SCORETITLE_RECT;
    static const PG_Rect HSM_FIRST_ENTRY_RECT;
    static const PG_Rect HSM_ENTRY_NAME_RECT;
    static const PG_Rect HSM_ENTRY_LEVELNAME_RECT;
    static const PG_Rect HSM_ENTRY_SCORE_RECT;
    static const PG_Rect HSM_BACK_TO_START_RECT;
    static const int HSM_ENTRY_STEP;

protected:
    static MenuController *Instance;

    Game *G;
    PG_Application *PGApp;
    InputHandler *IH;
    SDL_Interface *SDLI;
    OrdersPanel *OP;

    bool Activated;
    PG_Widget *MenuArea;
    PG_Button *btnChangeCommand[InputHandler::NUM_COMMANDS];
    PG_Button *btnTooltip;
    SDL_Surface *BackgroundImage;
    int CurHelpScreen;
    std::list<std::string> LevelFileList;     // Used to store the filenames of levels found

    MENUS CurMenu;
    MENUS NewMenu;
    bool CurMenuInvalid;

    MenuController(Game *G, PG_Application *PGApp);
   
    void SetCurMenu(MENUS AMenu);
    void ActivateStartMenu();
    void ActivatePickLevelMenu();
    void ActivateIntroScreenMenu();
    void ActivateOptionsMenu();
    void ActivateVideoOptionsMenu();
    void ActivateAudioOptionsMenu();
    void ActivateControlsOptionsMenu();
    void ActivateHelpMenu();
    void ActivateInGameMenu();
    void ActivateNewHighScoreMenu();
    void ActivateHighScoresMenu();

    void UpdateBackgroundImage();
    SDL_Surface *CreateBackgroundImage();
    const char *GetKeyName(int i);
public:
    ~MenuController();
    static MenuController *GetInstance(Game *G, PG_Application *PGApp);

    void DrawBackground();
    
    void SetActivated(bool Activated);
    bool GetActivated() { return Activated; }
    MENUS GetCurMenu() { return CurMenu; }

    void Update();

    // Menu Actions
    PARAGUI_CALLBACK(btnStartGame_Click);
    PARAGUI_CALLBACK(btnLevel_Click);
    PARAGUI_CALLBACK(btnOptions_Click);
    PARAGUI_CALLBACK(btnVideo_Click);
    PARAGUI_CALLBACK(AlterFullscreen);
    PARAGUI_CALLBACK(AlterResolution);
    PARAGUI_CALLBACK(AlterBitdepth);
    PARAGUI_CALLBACK(AlterNumParallaxLayers);
    PARAGUI_CALLBACK(btnAudio_Click);
    PARAGUI_CALLBACK(AlterSoundVolume);
    PARAGUI_CALLBACK(AlterMusicVolume);
    PARAGUI_CALLBACK(AlterReverseStereo);
    PARAGUI_CALLBACK(btnControls_Click);
    PARAGUI_CALLBACK(PickNewKey);
    PARAGUI_CALLBACK(AlterClockControls);
    PARAGUI_CALLBACK(btnContinue_Click);
    PARAGUI_CALLBACK(btnHelp_Click);
    PARAGUI_CALLBACK(btnContinueGame_Click);
    PARAGUI_CALLBACK(btnQuitGame_Click);
    PARAGUI_CALLBACK(btnExit_Click);
    PARAGUI_CALLBACK(btnBackToStart_Click);
    PARAGUI_CALLBACK(btnHighScores_Click);
    PARAGUI_CALLBACK(tbxName_EditEnd);
    PARAGUI_CALLBACK(btnPrevHelp_Click);
    PARAGUI_CALLBACK(btnDoneWithGameMenu_Click);
    PARAGUI_CALLBACK(btnNextHelp_Click);
};

//------------------------------------------------------------------------------
// Custom widgets
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Name: OptionsSlider
// Desc: The options slider sets the text on a tooltip display when the mouse
//       is over it and clears it when the mouse leaves.
//
// Note: This has had to be implemented in a rather odd way. The problem is
//       that when a scrollbar's drag button has the mouse move over it, it
//       does not tell it's scrollbar. The way I have got around this is to
//       use a containing widget which checks to see if the mouse is 
//       inside it's scrollbar.
//------------------------------------------------------------------------------
class OptionsSlider : public PG_Widget
{
public:
    static const int BORDER_SIZE;

    OptionsSlider(PG_Widget *parent, int id, const PG_Rect& r, 
        PG_Widget *TooltipDisplay, int direction = PG_SB_VERTICAL, 
        const char* style="Slider");

    ~OptionsSlider();

    PG_Slider *TheSlider;
protected:
    Game *G;

    PG_Widget *TooltipDisplay;

    void eventMouseEnter();
    void eventMouseLeave();

};

//------------------------------------------------------------------------------
// Name: OptionsCheckButton
// Desc: The options checkbutton sets the text on a tooltip display when the 
//       mouse is over it and clears it when the mouse leaves.
//------------------------------------------------------------------------------
class OptionsCheckButton : public PG_CheckButton
{
public:
    OptionsCheckButton(PG_Widget *parent, Game::GAME_OPTIONS id, const PG_Rect& r, 
        PG_Widget *TooltipDisplay, const char* text = "", 
        const char* style="CheckButton");

    ~OptionsCheckButton();

protected:
    Game *G;

    PG_Widget *TooltipDisplay;

    void eventMouseEnter();
    void eventMouseLeave();

};

//------------------------------------------------------------------------------
// Name: OptionsDropDown
// Desc: The options dropdown sets the text on a tooltip display when the 
//       mouse is over it and clears it when the mouse leaves.
//
// Note: The support that PG_DropDown offers for themes, seems to be a bit 
//       incomplete at the moment. Therefore this class adds theme support
//       but probably not in exactly the same way as other parts of ParaGUI.
//------------------------------------------------------------------------------
class OptionsDropDown : public PG_DropDown
{
    static const int SCROLL_LIMIT;
public:
    OptionsDropDown(PG_Widget* parent, int id, const PG_Rect& r, 
        PG_Widget *TooltipDisplay, const char* style="DropDown");

    ~OptionsDropDown();

    void AddItem(const char* Text, void* UserData = NULL);
protected:
    Game *G;
    std::string Style;

    PG_Widget *TooltipDisplay;

    void eventMouseEnter();
    void eventMouseLeave();

    PARAGUI_CALLBACK(NewSelectHandler);
};

//------------------------------------------------------------------------------
// Name: OptionsButton
// Desc: The options button sets the text on a tooltip display when the 
//       mouse is over it and clears it when the mouse leaves.
//------------------------------------------------------------------------------
class OptionsButton : public PG_Button
{
public:
    OptionsButton(PG_Widget *parent, int id, const PG_Rect& r, 
        PG_Widget *TooltipDisplay, const char* text = "", 
        const char* style="Button");

    ~OptionsButton();

protected:
    Game *G;

    PG_Widget *TooltipDisplay;

    void eventMouseEnter();
    void eventMouseLeave();

};

#endif // _MENUCONTROLLER_H_
