/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// Name: OrdersPanel.h
// Desc: This is the means by which the player can control the troops in his
//       army that are not under his direct control, i.e. everything other than
//       his plane. It is called up during play by pressing the 'Issue Orders'
//       key and then presents a GUI that can be controlled with the mouse.
//------------------------------------------------------------------------------

#ifndef _ORDERSPANEL_H_
#define _ORDERSPANEL_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include "CamelMath.h"
#include "SDL_Interface.h"
#include "paragui.h"
#include "pgeventobject.h"
#include "pgapplication.h"
#include "pgbutton.h"

class Game;
class SDL_Interface;
class InputHandler;
class Landscape;
class ControlledEntity;

class PG_Window;
class PG_Widget;
class PG_WidgetList;

class OrdersPanel : PG_EventObject
{
public:
    // Constants used to determine the size of various components
    static const int PANEL_WIDTH;
    static const int PANEL_HEIGHT;
    static const int HORIZONTAL_MARGIN;
    static const int VERTICAL_MARGIN;

    static const int GOOD_TAB_ID;
    static const int EVIL_TAB_ID;
    static const PG_Rect GOOD_TAB_BUTTON_RECT;
    static const PG_Rect EVIL_TAB_BUTTON_RECT;

    static const PG_Rect UNIT_LIST_RECT;
    static const int UNIT_LIST_H_MARGIN;
    static const int UNIT_LIST_V_MARGIN;
    static const PG_Rect UNIT_BUTTON_RECT;
    static const PG_Rect UNIT_INFO_RECT;
    static const PG_Rect BUTTON_CLOSE_RECT;

    static const PG_Rect UI_TITLE_LABEL_RECT;
    static const PG_Rect UI_NUMALIVE_LABEL_RECT;
    static const PG_Rect UI_NUMALIVE_RECT;
    static const PG_Rect UI_MORALE_LABEL_RECT;
    static const PG_Rect UI_MORALEBAR_RECT;
    static const PG_Rect UI_ORDERS_LABEL_RECT;
    static const PG_Rect UI_ORDERS_RECT;
    static const PG_Rect UI_HOLD_BUTTON_RECT;

    // The speed at which the viewport scrolls (pixels per second)
    static const float VP_SCROLL_SPEED;
    // The distance from a screen edge where the mouse induces scrolling
    static const int SCROLL_MARGIN;

protected:
    static OrdersPanel *Instance;

    Game *G;
    PG_Application *PGApp;
    SDL_Interface *SDLI;
    InputHandler *IH;

    ControlledEntity *CurUnit;

    // Whilst the orders panel is active, it provides an alternative viewpoint
    float RealX;
    Point2D ViewportCentrePos;

    // Variables related to the OrdersPanelWindow
    PG_Rect DisplayRect;
    PG_Window *Display;
    PG_Button *btnGoodTab;
    PG_Button *btnEvilTab;
    PG_WidgetList *UnitList;
    PG_Widget *UnitInfo;
    PG_Button *btnClose;
    bool NeedToUpdateUnitInfo;

    bool Activated;
    bool Visible;

    OrdersPanel();
    OrdersPanel(Game *G, PG_Application *PGApp);

    void RefreshUnitList();
    void RefreshUnitInfo();
public:
    ~OrdersPanel();
    static OrdersPanel *GetInstance(Game *G, PG_Application *PGApp);

    void Update(uint dt);

    void SetActivated(bool Activated);
    bool GetActivated() { return Activated; }
    void SetVisible(bool Visible);
    bool GetVisible() { return Visible; }

    void SetCurUnit(ControlledEntity *AUnit);
    ControlledEntity *GetCurUnit() { return CurUnit; }
    Point2D GetViewportCentrePos();
    int ValidateViewportX(int x);

    // Callback functions
    PARAGUI_CALLBACK(btnTab_Click);
    PARAGUI_CALLBACK(btnClose_Click);
    PARAGUI_CALLBACK(btnUnit_Click);
    PARAGUI_CALLBACK(btnHold_Click);
};

//------------------------------------------------------------------------------
// Name: UnitButton
// Desc: A unit button is a button that draws a sprite on its drawing surface
//       after the underlying button has been drawn. This can be used to 
//       create a button which represents a ControlledEntity. The 
//       ControlledEntity provides the sprite frame to draw.
//------------------------------------------------------------------------------
class Sprites;

class UnitButton : public PG_Button
{
public:
    UnitButton(PG_Widget *parent, int id, const PG_Rect& r, 
        int Frame, const char* text = "", 
        const char* style="Button");

    ~UnitButton();

protected:
    Sprites *SPR;

    int Frame;
    int FrameWidth, FrameHeight;

    void eventBlit(SDL_Surface* srf, const PG_Rect& src, const PG_Rect& dst);
};

#endif // _ORDERSPANEL_H_
