/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Plane.h
// Desc: The plane entity is vital to the whole game, it is used in the 
//       creation of the players plane. Also it is used to create the enemy 
//       planes that try to thwart the players actions.
//------------------------------------------------------------------------------

#ifndef _PLANE_H_
#define _PLANE_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include "Entity.h"

class Plane : public Entity
{
public:
    static const int POINTS_VALUE;

    enum PLANE_STATE {
        LANDED,
        FLYING,
        SPIRALING,
        CRASHING
    };

    static const int STARTING_BOMBS;
    static const int STARTING_BULLETS;
    static const int STARTING_FUEL;
protected:
    enum PLANE_ACTIONS {
        NO_ACTION = -1,
        CLIMB,
        DIVE,
        INVERT,
        FIRE,
        HALT_GUNFIRE,
        START_ENGINE,
        SPIRAL,
        EMIT_SMOKE,
		DROP_BOMB,
        RESET_BOMB,
        CONSUME_FUEL,
        LOAD_FUEL,
        LOAD_BULLETS,
        LOAD_BOMBS,
        NUM_ACTIONS
    };

    // Amount of time player must struggle with the stick before recovering 
    // from a stall
    static const int STALL_RECOVERY_THRESHOLD;

    // Rates for various actions are measured in terms of frames
    static const int CLIMBING_RATE;
    static const int DIVING_RATE;
    static const int INVERTING_RATE;
    static const int FIRING_RATE;
    static const int HALTING_GUNFIRE_RATE;
    static const int STARTING_ENGINE_RATE;
    static const int SPIRALING_RATE;
    static const int EMITING_SMOKE_RATE;
	static const int DROPPING_BOMB_RATE;
    static const int RESETTING_BOMB_RATE;
    static const int CONSUMING_FUEL_RATE;
    static const int LOADING_FUEL_RATE;
    static const int LOADING_BULLETS_RATE;
    static const int LOADING_BOMBS_RATE;

    static const float MINIMUM_AIRSPEED;

    // Acceleration values are given in pixels per frame
    static const float MAX_ENGINE_ACC;
    static const float ENGINE_ACC_ACC;

    static const int BULLET_RANGE;

    PLANE_STATE State;
    bool EngineOn;
    float EngineAcc;

    bool Inverted;
    int Direction;
    float Velocity;             // Pixels per second
    int BombsLeft;
    int BulletsLeft;
    int FuelLeft;
    int CurEngineSample;
    int NewEngineSample;

    int StartFrame;             // First frame of the sprite used for the plane
    int EngineChannel;          // Channels used for playing sounds
    int GunfireChannel;
    int SFX_Volume;
    uchar LeftVolume, RightVolume;  // Stereo volumes

    bool Commands[NUM_ACTIONS];
    ACTION *Actions[NUM_ACTIONS];

    bool Refuelling;
    bool HaltingGunfire;
    bool ResetingBomb;
    bool FiringMultipleShots;
    int StallRecovery;  // Used to measure the planes recovery from a stall                       

    int CalcFrame(int CurDirection, bool Inverted);
    float CalcNewVelocity(float CurVelocity, int CurDirection, uint dt);

    static void Climb(void *Object);
    static void Dive(void *Object);
    static void Invert(void *Object);
    static void Fire(void *Object);
    static void HaltGunfire(void *Object);
    static void StartEngine(void *Object);
    static void Spiral(void *Object);
    static void EmitSmoke(void *Object);
	static void DropBomb(void *Object);
    static void ResetBomb(void *Object);
    static void ConsumeFuel(void *Object);
    static void LoadFuel(void *Object);
    static void LoadBullets(void *Object);
    static void LoadBombs(void *Object);
public:
    Plane(Level *L, PLANE_STATE State, Point2D Pos, bool Inverted, int Direction,
        ENT_SIDE Side);
    ~Plane();

    void Update(uint dt);
    void FindAndHandleCollisionWithLandscape();
	void HandleCollision(Entity *AnEntity);
    bool GetInverted() { return Inverted; }
    int GetDirection() { return Direction; }
    float GetVelocity() { return Velocity; }
    int GetBombsLeft() { return BombsLeft; }
    int GetBulletsLeft() { return BulletsLeft; }
    int GetFuelLeft() { return FuelLeft; }
};

#endif // _PLANE_H_
