/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Platoon.h
// Desc: This object provides a way to group together soldier entities. The
//       platoon can be given orders as a whole and will move and fight as one.
//       It also provides a group that can be affected by morale so they can
//       also run away as one. :P
//
//       This class has been made an entity so that it can be easily kept 
//       updated with the rest of the game. I am not sure whether this is
//       actually a good design decision or not however, as a platoon lacks
//       some of the elements of a traditional entity, such as a frame and a
//       width and height
//------------------------------------------------------------------------------

#ifndef _PLATOON_H_
#define _PLATOON_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <vector>
#include "SDL_Interface.h"
#include "ControlledEntity.h"

class Soldier;

class Platoon : public ControlledEntity
{
public:
    static const int MAX_NUM_SOLDIERS = 100;

    // The higher this number is, the more spread out the platoon will be be
    // when it is placed
    static const int SOLDIER_SPREAD;
    static const int SOLDIERS_PER_RANK;
    static const int DISENGAGE_RANGE;
    static const int MAX_ENTANGLEMENT;

    static const int NEAR_DESTINATION_X;
    
    // If morale drops to below this point then the platoon runs away
    static const float BREAKPOINT;
    // If morale rises back up above this point then the platoon rallies
    static const float RALLYPOINT;
    static const int MIN_NUM_SOLDIERS_NEEDED_TO_RALLY;

    static const int RANK_SEPERATION;
    static const int RANK_PLACEMENT_RANGE;
    static const int RANK_PLACEMENT_RANGE_2;

    static const float SECOND_RANK_REPLACEMENT_CHANCE;

    enum BATTLE_POS {
        FRONT_RANK,
        SECOND_RANK,
        RESERVE,
        NUM_OF_BATTLE_POSITIONS
    };

    static const Order FLEE;
protected:
    enum PLATOON_ACTIONS {
        UPDATE_MORALE,
        NUM_ACTIONS
    };

    // Rates for various actions are measured in milliseconds
    static const int UPDATING_MORALE_RATE;

    ACTION *Actions[NUM_ACTIONS];
    int NumSoldiers;
    int SoldiersAliveLastMoraleCheck;
    Soldier *Soldiers[MAX_NUM_SOLDIERS];

    // The following variables are used when the Platoon is fighting other
    // FightingEntities. They only contain meaningful values when 
    // CurEnemy != NULL and can be updated by using FindClosestEnemyInRange
    // and FightingEntityInRange from the FightingEntity abstract base class.
    // The reason that they have not been abstracted up to FightingEntity is that
    // I am not sure a similar method would work for all FightingEntities. 
    FightingEntity *CurEnemy; 
    bool OnTheLeft;     // Is the Platoon on the left of its current enemy?       
    int NearestX;       // This will either be Left() or Right(), but stored to save time.
    int MaxRange;       // +ve value indicating how far a soldier at NearestX could fire
    // Distance from NearestX to the nearest point of CurEnemy (always +ve)
    int RelativeDistance;   
    // The point the soldiers are being told to shoot at on the current enemy
    Point2D TargetPos;

    // These determine how spread out various parts of the platoon are
    int UnitPlacementRange;
    int ReservePlacementRange;

    // Variables related to the position of the platoon
    int BattlePosX[NUM_OF_BATTLE_POSITIONS];
    std::vector<int> BattleGroups[NUM_OF_BATTLE_POSITIONS];
    int FallbackX;

    Platoon();

    void UpdateBoundingBox();
    int GetPlacementPos(int x, int PlacementRange);
    void ChangeSoldiersOrders();
    bool UpdateDestinationX(const Order *NewOrders, int NewDestinationX);
    
    void FormulateBattlePlan();
    void FormulatePlanHOLD();
    void FormulatePlanADVANCE();
    void UpdatePlanHOLD();
    void UpdatePlanADVANCE();
    void DrawUpBattleGroups();
    void RemoveSoldierFromBattleGroup(BATTLE_POS BattleGroup, int SoldierNum);

    int GetAvailableRange(int HeightDiff);
    int GetTargetHeight();
    static void UpdateMorale(void *Object);
public:
    Platoon(Level *L, int NumSoldiers, int x, ENT_SIDE Side);
    ~Platoon();

    int GetUnitPlacementRange() { return UnitPlacementRange; }
    int GetRankPlacementRange() { return RANK_PLACEMENT_RANGE; }
    int GetReservePlacementRange() { return ReservePlacementRange; }

    void PlaceAt(int x);
    void Update(uint dt);
    void SoldierIsDying(int SoldierNum);

    int GetSubordinateDestinationX();
    int GetNearDestinationX() { return NEAR_DESTINATION_X; }
    int GetBattlePosX(BATTLE_POS BattlePos);
    bool IsFighting() { return CurEnemy != NULL; }
    bool IsFleeing() { return CurOrders == &FLEE; }
    bool IsWithdrawing() { return CurOrders == &WITHDRAW; }
    
    // Should only be called when the platoon is fighting
    Point2D GetTargetPos() { return TargetPos; }
   
    void FindAndHandleCollisionWithLandscape() {}
    void HandleCollision(Entity *AnEntity) {}
};

#endif // _PLATOON_H_
