/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: SDL_Interface.h
// Desc: Provides access to the SDL library, include this file instead of
//       SDL.h. The SDL_Interface object sets up and provides access to the 
//       screen as well as providing some limited (no scaling etc) software 
//       blitting functions which seem to give better performance under Win32
//       windowed mode.
//------------------------------------------------------------------------------

#ifndef _SDL_INTERFACE_H_
#define _SDL_INTERFACE_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <assert.h>
#include "globals.h"
#include "CamelMath.h"
#include "SDL.h"
#include "SDL_image.h"
#include "SDL_mixer.h"
#include "BFont.h"

class Game;

class SDL_Interface 
{
public:
    enum SCREEN_MODES {
        _640x480,
        _800x600,
        _1024x768,
        NUM_SCREEN_MODES
    };

    struct SCREEN_MODE_INFO {
        char *String;
        int Width;
        int Height;
    };

    enum SCREEN_BITDEPTHS {
        _16,
        _32,
        NUM_SCREEN_BITDEPTHS
    };

    static const SCREEN_MODE_INFO SCREEN_MODE_DETAILS[NUM_SCREEN_MODES];
    static const int SCREEN_BITDEPTH_DETAILS[NUM_SCREEN_BITDEPTHS];

    static const int FREQUENCY;
    static const int FORMAT;
    static const int NUM_SOUND_CHANNELS;
    static const int CHUNKSIZE;

    // This gives the size pixels of a unit of distance used to measure how far 
    // an entity producing sound effects is from the centre of the viewport.
    // The volume of sound is proportional to the inverse square of the distance
    // from the sound source.
    static const int SOUND_UNIT_OF_DISTANCE = 400;

    // Volume of a sound channel (left or right) when a sound is centred
    static const uchar CENTRAL_VOLUME = 127;

    // Constants related to the number of mix channels and their uses
    enum CHANNELS {
        NO_CHANNEL = -1,
        NUM_CHANNELS = 64
    };

    // Samples used for the games sound effects
    enum SAMPLES {
        NO_SAMPLE = -1,
        PLANE_ENGINE_SPEED_0,
        PLANE_ENGINE_SPEED_1,
        PLANE_ENGINE_SPEED_2,
        PLANE_ENGINE_SPEED_3,
        PLANE_ENGINE_SPEED_4,
        PLANE_ENGINE_SPEED_5,
        PLANE_ENGINE_SPEED_6,
        PLANE_ENGINE_SPEED_7,
        PLANE_ENGINE_SPEED_8,
        PLANE_ENGINE_SPEED_9,
        PLANE_ENGINE_SPEED_10,
        PLANE_ENGINE_SPEED_11,
        PLANE_ENGINE_SPEED_12,
        PLANE_ENGINE_SPEED_13,
        PLANE_ENGINE_SPEED_14,
        PLANE_ENGINE_SPEED_15,
        PLANE_ENGINE_SPEED_16,
        PLANE_GUNFIRE,
        BOMB_EXPLOSION,
        MENU_CLICK,
        SOLDIER_GUN_0,
        SOLDIER_GUN_1,
        SOLDIER_GUN_2,
        SOLDIER_GUN_3,
        SOLDIER_GUN_4,
        SOLDIER_GUN_5,
        SOLDIER_GUN_6,
        SOLDIER_SCREAM_0,
        SOLDIER_SCREAM_1,
        SOLDIER_SCREAM_2,
        SOLDIER_SCREAM_3,
        SOLDIER_SCREAM_4,
        SOLDIER_SCREAM_5,
        SOLDIER_SCREAM_6,
        SOLDIER_SCREAM_7,
        GAME_OVER,
        YOU_WIN,
        TANK_LOOP_0,
        TANK_LOOP_1,
        TANK_LOOP_2,
        TANK_GUN,
        NUM_SAMPLES
    };

    // These constants are used to define groups of samples
    static const int SOLDIER_GUN_BASE;
    static const int SOLDIER_GUN_END;
    static const int SOLDIER_SCREAM_BASE;
    static const int SOLDIER_SCREAM_END;
protected:
    struct SOUND_FILE_INFO {
        char *FileName;
        double ScalingFactor;   // Only really makes sense for a sample
    };

    // Used to store information about the files from which to retrieve 
    // the samples and pieces of music
    static SOUND_FILE_INFO SampleFiles[NUM_SAMPLES];

    static SDL_Interface *Instance;

    Game *G;

    SDL_Surface *Screen;
    bool Fullscreen;
    SCREEN_MODES ScreenMode;
    SCREEN_BITDEPTHS Bitdepth;
    int SFX_Volume;
    int Music_Volume;
    int CurChannelsUsed;
    int MaxChannelsUsed;
    bool ReverseStereo;
    bool MusicPaused;

    SDL_Interface();
    SDL_Interface(Game *G);

    void SetupVideo();
    bool ScreenModeAccepted(SCREEN_MODES F_ScreenMode, bool& M_Fullscreen, SCREEN_BITDEPTHS& M_Bitdepth);
    bool FullscreenAccepted(SCREEN_MODES F_ScreenMode, bool F_Fullscreen, SCREEN_BITDEPTHS& M_Bitdepth);
public:
    ~SDL_Interface();
    static SDL_Interface *GetInstance(Game *G);

    BFont_Info *MainFont;
    Mix_Chunk *Samples[NUM_SAMPLES];

    Mix_Chunk *LoadScaledSample(const char *FileName, double SFactor);
    SDL_Surface *GetScreen() { return Screen; }

    bool IsFullScreen() { return Fullscreen; }
    int GetScreenWidth() { return SCREEN_MODE_DETAILS[ScreenMode].Width; }
    int GetScreenHeight() { return SCREEN_MODE_DETAILS[ScreenMode].Height; }
    int GetScreenBitdepth() { return SCREEN_BITDEPTH_DETAILS[Bitdepth]; }
    int GetSFX_Volume() { return SFX_Volume; }
    void SetSFX_Volume(int NewVolume);
    int GetMusic_Volume() { return Music_Volume; }
    void SetMusic_Volume(int NewVolume);
    bool GetReverseStereo() { return ReverseStereo; }
    void SetReverseStereo(bool ReverseStereo);
    bool SW_FillRect(SDL_Surface *dst, SDL_Rect *dstrect, Uint32 colour);
    bool SW_BlitSurface(SDL_Surface *src, SDL_Rect *srcrect, 
        SDL_Surface *dst, SDL_Rect *dstrect);
    bool AdjustRects(SDL_Surface *SrcSurface, SDL_Rect *SrcRectInfo, 
        SDL_Rect *SrcRect, SDL_Surface *DstSurface, SDL_Rect *DstRectInfo, 
        SDL_Rect *DstRect);
    bool AdjustRect(SDL_Surface *DstSurface, SDL_Rect *DstRectInfo, 
        SDL_Rect *DstRect);
    bool SameSurfaceFormat(SDL_Surface *S1, SDL_Surface *S2);
    void DrawLine(SDL_Surface *Surface, uint Colour, 
        Point2D StartPos, Point2D EndPos);

    // Methods for monitoring the use of channels
    int PlayChannel(int Channel, Mix_Chunk *Chunk, int Loops);
    static void ChannelFinished(int Channel);

    int VolumeOverDistance(float Distance);
    void GetStereoVolumes(int ViewportWidth_2, int X_DistanceFromCentre,
        uchar *LeftChannel, uchar *RightChannel);

    void PlayMusic(Mix_Music *Melody, int Loops, int FadeIn);
    void PauseMusic();
    void ResumeMusic();
    void StopMusic(int FadeOut);
};

#endif // _SDL_INTERFACE_H_
