/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Tank.h
// Desc: Tanks are big metal monsters that should hopefully inspire fear and 
//       dread in the player when they face them, and joy and excitement when
//       they have them on their side. Tanks are constructed by a different
//       factory type and at a far slower rate than infantry platoons.
//       Having said that, tanks take a lot of effort to kill and can cause
//       a fair bit of damage. They fire explosive shells and can only be 
//       killed by a lot of bullets or a direct hit from a bomb. An explosion
//       will not cause any damage to tank.
//------------------------------------------------------------------------------

#ifndef _TANK_H_
#define _TANK_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include "CamelMath.h"
#include "Entity.h"

class TankSquad;

class Tank : public Entity
{
    // The points earned for an EVIL tanks demise
    static const int POINTS_VALUE;

    // How high above the ground our enemies should aim
    static const int HEIGHT_ABOVE_GROUND;

    // The hitpoints that a tank starts with
    static const int INITIAL_HITPOINTS;
    // The damage that various collisions cause
    static const int BOMB_DAMAGE;
    static const int PLANE_DAMAGE;
    static const int PLANE_BULLET_DAMAGE;
    static const int TANK_BULLET_DAMAGE;
    static const int SOLDIER_BULLET_DAMAGE;

    // The speed of a tank is measured in pixels per second. However, the 
    // actual speed of a tank is random and is in the range
    //              TANK_SPEED +/- TANK_SPEED_RANGE/2
    static const int TANK_SPEED;
    static const int TANK_SPEED_RANGE;
    static const int TANK_SPEED_RANGE_2;

    // As this game is 2D, the tank probably 'flips' rather than turning in a 
    // circle. However, this constant is used so that if a tank has to go 
    // backwards a distance less than TURNING_CIRCLE it reverses rather than
    // turning first.
    static const int TURNING_CIRCLE;

    static const int RANGE_OF_GUN;

    // The tanks position is matched to the contours of the land by placing its
    // tracks and then finding the coordinates of the sprites from the location
    // of the tracks.
    struct TRACK_INFO {
        int x, y;
        int Width, Height;
    };

    enum TANK_FACING {
        LEFT,
        RIGHT,
        NUM_FACINGS
    };

    // When looking at tank directions and turret directions, 0 is taken as the
    // horizontal (regardless of which way you are facing) and other directions
    // are relative to that. We will never have to deal with a direction outside
    // of the range +/- 3 so that is the range used here.
    enum OFFSET_FROM_HORIZONTAL {
        MINUS_3 = 0,
        MINUS_2,
        MINUS_1,
        HORIZONTAL,
        PLUS_1,
        PLUS_2,
        PLUS_3,
        NUM_OFFSETS
    };

    // This is used to help determine when a tanks direction goes to PLUS_1
    // or MINUS_1 as it travels between HORIZONTAL and PLUS_2 or MINUS_2
    // and back again.
    static const int TILT_RANGE;
    static const int TILT_RAISE;

    // Description of the tank's tracks for any given facing and offset
    static const TRACK_INFO TRACK_DETAILS[NUM_FACINGS][NUM_OFFSETS];

    // Constants to do with the frames used to animate the tank
    static const int NUM_DRIVING_FRAMES;
    static const int FACING_OFFSET[NUM_FACINGS];

    // The direction offsets are taken from the Tanks 'BaseFrame' plus a facing offset.
    static const int DIRECTION_OFFSET[NUM_OFFSETS];
    // The turret offset is taken from the 'BaseFrame' plus a facing offset 
    // and a direction offset.
    static const int TURRET_OFFSET[NUM_OFFSETS];

    struct FD_TO {  // Used to store the TurretOffset for a given FiringDirection
        int FiringDirection;
        OFFSET_FROM_HORIZONTAL TurretOffset;
    };

    // This struct is used to store the allowed firing directions and corresponding
    // turret offset all of the possible facings and direction offsets that the
    // tank may adopt
    struct F_DO_TO {    
        int NumAllowedFiringDirs;
        FD_TO AllowedFiringDirs[NUM_OFFSETS];
    };

    // Used to find if there is a valid turret offset for a given firing direction
    static const F_DO_TO TURRET_OFFSET_DIRECTORY[NUM_FACINGS][NUM_OFFSETS];

    enum TANK_STATE {
        WAITING,
        DRIVING,
        FIRING
    };
protected:
    enum TANK_ACTIONS {
        CHANGE_SPEED,
        MOVE_TRACKS,
        FIRE,
        NUM_ACTIONS
    };

    // Rates for various actions are measured in milliseconds
    static const int CHANGING_SPEED_RATE;
    static const int MOVING_TRACKS_RATE;
    static const int FIRING_RATE;
    static const int FIRING_RATE_VARIATION;

    ACTION *Actions[NUM_ACTIONS];

    int HitPoints;
    Point2D TrackPos;
    float RealTrackX;

    float Velocity;     // Measured in pixels per second
    int DestinationX;   // Where the tank wants to go
    int TankNum;
    TankSquad *CurTankSquad;
    TANK_STATE State;

    int TrackNoiseChannel;  // The tank makes a noise as it moves

    // These variables are used to help determine the correct frame to use
    int BaseFrame;
    int DrivingFrame;

    TANK_FACING Facing;
    OFFSET_FROM_HORIZONTAL DirectionOffset;
    OFFSET_FROM_HORIZONTAL TurretOffset;

    // When the tank is firing, this is the point to fire from
    Point2D ShotPos;
    bool CanFire;

    Tank();
    void CalcFrame();
    bool CalcTurretOffset(TANK_FACING Facing, 
        OFFSET_FROM_HORIZONTAL DirectionOffset, int FiringDirection,
        OFFSET_FROM_HORIZONTAL& TurretOffset);
    void UpdateTurretOffset();

    static void ChangeSpeed(void *Object);
    static void MoveTracks(void *Object);
    static void Fire(void *Object);

    void CalculatePosition();
public:
    bool OrdersHaveChanged;

    Tank(Level *L, int x, int TankNum, TankSquad *CurTankSquad);
    ~Tank();

    void Update(uint dt);
    void FindAndHandleCollisionWithLandscape() {}
    void HandleCollision(Entity *AnEntity);

    void PlaceAt(int x);

    static int GetAvailableRange(int HeightDiff);
    static int GetTargetHeight() { return HEIGHT_ABOVE_GROUND; };
};

#endif // _TANK_H_
