/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: TankSquad.h
// Desc: A TankSquad is to a Tank, what a Platoon is to a Soldier. A tank 
//       squad contains far fewer tanks than a Platoon contains soldiers 
//       but then again tanks cause a bit more damage and are immune to 
//       psychology ;).
//------------------------------------------------------------------------------

#ifndef _TANKSQUAD_H_
#define _TANKSQUAD_H_

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include "SDL_Interface.h"
#include "ControlledEntity.h"

class Tank;

class TankSquad : public ControlledEntity
{
public:
    static const int MAX_NUM_TANKS = 3;

    // This determines how spread out the tank squad is
    static const int UNIT_PLACEMENT_RANGE;
    static const int BATTLE_PLACEMENT_RANGE;
    static const int DISENGAGE_RANGE;
    static const int MAX_ENTANGLEMENT;

    static const int NEAR_DESTINATION_X;
    
protected:
    
    int InitialNumTanks;
    Tank *Tanks[MAX_NUM_TANKS];
    
    // The following variables are used when the TankSquad is fighting other
    // FightingEntities. They only contain meaningful values when 
    // CurEnemy != NULL and can be updated by using FindClosestEnemyInRange
    // and FightingEntityInRange from the FightingEntity abstract base class.
    // The reason that they have not been abstracted up to FightingEntity is that
    // I am not sure a similar method would work for all FightingEntities. 
    FightingEntity *CurEnemy; 
    bool OnTheLeft;     // Is the TankSquad on the left of its current enemy?       
    int NearestX;       // This will either be Left() or Right(), but stored to save time.
    int MaxRange;       // +ve value indicating how far a tank at NearestX could fire
    // Distance from NearestX to the nearest point of CurEnemy (always +ve)
    int RelativeDistance;   
    // The point the tanks are being told to shoot at on the current enemy
    Point2D TargetPos;

    // Where the TankSquad is positioned during combat
    int BattlePosX;

    void UpdateBoundingBox();
    int GetPlacementPos(int x, int PlacementRange);
    void ChangeTanksOrders();
    
    void FormulateBattlePlan();
    void FormulatePlanHOLD();
    void FormulatePlanADVANCE();
    void UpdatePlanHOLD();
    void UpdatePlanADVANCE();

    int GetAvailableRange(int HeightDiff);
    int GetTargetHeight();
public:
    TankSquad(Level *L, int InitialNumTanks, int x, ENT_SIDE Side);
    ~TankSquad();

    void PlaceAt(int x);
    void Update(uint dt);
    void TankIsDying(int TankNum);

    int GetSubordinateDestinationX();
    int GetNearDestinationX() { return NEAR_DESTINATION_X; }
    int GetBattlePosX();
    bool IsFighting() { return CurEnemy != NULL; }
    bool IsFleeing() { return false; }
    bool IsWithdrawing() { return CurOrders == &WITHDRAW; }
    
    // Should only be called when the TankSquad is fighting
    Point2D GetTargetPos() { return TargetPos; }
   
    void FindAndHandleCollisionWithLandscape() {}
    void HandleCollision(Entity *AnEntity) {}
};

#endif // _TANKSQUAD_H_
