/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Game.cpp
// Desc: This object is created on start up to run the game, it contains all
//       of the objects used by the game as well as game initialisation,
//       shutdown and the main game loop
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <fstream> 
#include <sstream>
#include <string>
#include "globals.h"
#include "CamelUtility.h"
#include "Camel.h"
#include "SDL_Interface.h"
#include "InputHandler.h"
#include "MenuController.h"
#include "Sprites.h"
#include "OrdersPanel.h"
#include "Level.h"
#include "Landscape.h"
#include "Plane.h"
#include "PlayerPlane.h"
#include "EnemyPlane.h"
#include "Factory.h"
#include "Platoon.h"
#include "Soldier.h"
#include "Game.h"

// ParaGUI includes
#include "pgapplication.h"
#include "pgradiobutton.h"

using namespace std;

//------------------------------------------------------------------------------
// Game Implementation
//------------------------------------------------------------------------------

// The height of the user interface which is displayed at the bottom of the screen
const int Game::UI_HEIGHT = 140;

// Constants used to draw the needles on the user interface gauges
const double Game::UI_GAUGE_NEEDLE_START_ANGLE = 0.13255153229667402126991744370427;
const double Game::UI_GAUGE_NEEDLE_END_ANGLE = 2.8724251608040927630361075420064;
const double Game::UI_GAUGE_NEEDLE_RANGE = 
    UI_GAUGE_NEEDLE_END_ANGLE - UI_GAUGE_NEEDLE_START_ANGLE;
const int Game::UI_GAUGE_NEEDLE_LENGTH = 31;

// The time in milliseconds that music takes to fade in or out
const int Game::MUSIC_FADE_TIME = 1000;

// Time in milliseconds that the 'Game Over' or 'You Win!' banner is displayed for
const int Game::END_OF_LEVEL_DELAY = 4000;

const char *Game::OPTIONS_DELIMITER = ":";
const int Game::DEFAULT_OPTIONS[NUM_OPTIONS] = {
    0,
    SDL_Interface::_800x600,
    SDL_Interface::_32,
    2,
    SDLK_a,
    SDLK_d,
    SDLK_s,
    SDLK_SPACE,
    SDLK_e,
	SDLK_b,
    SDLK_o,
    SDLK_h,
    0,
    32,
    24,
    0,
    0,
    0,
    0
};

const Game::OPTION_INFO Game::OPTION_DETAILS[NUM_OPTIONS] = {
    {
        "FULLSCREEN", 
        "# Set to 1 for fullscreen, 0 for windowed mode",
        "Clear this to select windowed mode (You must restart the game for a change to take effect)"
    },{ 
        "SCREENMODE", 
        "\n# Supported modes are 640x480, 800x600 and 1024x768",
        "Select the screen resolution (You must restart the game for a change to take effect)"
    },{ 
        "BITDEPTH", 
        "\n# Supported bitdepths are 16 bits and 32 bits",
        "In windowed mode this must match your system settings "
            "(You must restart the game for a change to take effect)"
    },{ 
        "NUM_PARALLAX_LAYERS", 
        "\n# Can be 0, 1 or 2. Use fewer layers for a faster game",
        "Use fewer layers for a faster game"
    },{ 
        "KEY_CLIMB", 
        "\n# See the readme.txt for a list of valid keys",
        "Pull the nose of the plane up"
    },{ 
        "KEY_DIVE", 
        "",
        "Push the nose of the plane down"
    },{ 
        "KEY_INVERT", 
        "",
        "Turn the plane upside down"
    },{ 
        "KEY_FIRE", 
        "",
        "Fire your gun"
    },{ 
        "KEY_START_ENGINE", 
        "",
        "Start your engine"
    },{ 
        "KEY_DROP_BOMB", 
        "",
        "Kaboom!"
    },{ 
        "KEY_ISSUE_ORDERS", 
        "",
        "Give orders to the troops under your command"
    },{
        "KEY_FLY_HOME",
        "",
        "Fly back home to refuel and get more ammo"
    },{
        "CLOCK_CONTROLS", 
        "\n# Set to 0 for normal controls. Set to 1 so that climb and dive mean"
            "\n# turn clockwise and turn anti-clockwise.",
        "If this is checked then climb becomes 'Turn Clockwise' and dive becomes "
            "'Turn Anti-Clockwise'"
    },{ 
        "SFX_VOLUME", 
        "\n# Valid music and sound volumes range from 0 (quietest) to 128 (loudest)",
        "Sound Volume"
    },{ 
        "MUSIC_VOLUME", 
        "",
        "Music Volume"
    },{ 
        "REVERSE_STEREO", 
        "\n# Set to 1 to swap your speakers around and 0 to leave them as they are",
        "Swaps your speakers over"
    },{
        "MENU_MUSIC",
        "",
        ""
    },{
        "DISPLAY_FPS",
        "\n# Set to 1 to display the frames per second",
        ""
    },{
        "USE_DEBUG_CAMERA",
        "\n# Set this to 1 and you will be able to control the camera with the mouse",
        ""
    }
};

const Game::HIGH_SCORE Game::DEFAULT_HIGH_SCORES[NUM_HIGH_SCORES] = {
    { "Alan", "Trial by Fire (Easy)", 1000 },
    { "Alan", "Trial by Fire (Easy)", 900 },
    { "Alan", "Trial by Fire (Easy)", 800 },
    { "Alan", "Trial by Fire (Easy)", 700 },
    { "Weirdo", "Trial by Fire (Easy)", 642 },
    { "Alan", "Trial by Fire (Easy)", 500 },
    { "Alan", "Trial by Fire (Easy)", 400 },
    { "Alan", "Trial by Fire (Easy)", 300 },
    { "Alan", "Trial by Fire (Easy)", 200 },
    { "Alan", "Trial by Fire (Easy)", 100 }
};

Game *Game::Instance = NULL;

//------------------------------------------------------------------------------
// Name: Game
// Desc: Set up the main Game object. Also start up all of the major components
//       used in the game.
//------------------------------------------------------------------------------
Game::Game()
{
    string FullFilename;

    // Start up the necessary SDL subsystems
    if (SDL_FAILED(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_AUDIO))) 
        SDE_SDLError(__FILE__, __LINE__);
    // Seed the random number generator
    srand(SDL_GetTicks());

    // Load in game options and then save them back out to properly format file
    LoadOptions();
    SaveOptions();
    // Load in High scores and then save them back out to ensure correct ordering
    LoadHighScores();
    SaveHighScores();

    // Start up the subsystems
    PGApp = new PG_Application();

    SDLI = SDL_Interface::GetInstance(this);
    IH = InputHandler::GetInstance(this);
    SPR = Sprites::GetInstance();

    // Continue with the setup of ParaGUI
    PGApp->SetScreen(SDLI->GetScreen());
    PGApp->LoadTheme("data/camelgui");
    PGApp->SetBulkMode(true);

    MC = MenuController::GetInstance(this, PGApp);
    OP = OrdersPanel::GetInstance(this, PGApp);

	// Set the initial state of the game
	GameState = STOPPED;   
    CurLevel = NULL;
    LevelName = "";
    LS = NULL;

    // Load in the menu music if it exists
    if (GetOption(MENU_MUSIC) != 0) {
        FullFilename = string(CAM_MUSIC_DIR) + "/" + MenuMusicFile;
    } else {
        FullFilename = string(CAM_MUSIC_DIR) + "/" + string(DEFAULT_MENU_MUSIC);
    }
    MenuMusic = Mix_LoadMUS(FullFilename.c_str());
}

//------------------------------------------------------------------------------
// Name: ~Game
// Desc: Shutdown and cleanup after the main Game object
//------------------------------------------------------------------------------
Game::~Game()
{
    // Close the current level if it is still open
    if (CurLevel != NULL) {
        delete CurLevel;
        CurLevel = NULL;
    }

    // Free the memory used for the menu music
    if (MenuMusic != NULL) Mix_FreeMusic(MenuMusic);

    // Save current game options and highscores
    SaveHighScores();
    SaveOptions();

    // Shutdown subsystems
    delete OP;
    delete MC;
    delete SPR;
    delete IH;
    delete SDLI;

    delete PGApp;

    // Shutdown SDL
    SDL_Quit();
}

//------------------------------------------------------------------------------
// Name: GetInstance
// Desc: Returns or creates the single instance of this class
//------------------------------------------------------------------------------
Game *Game::GetInstance() 
{
    if (Instance == NULL) {
        Instance = new Game();
    }
    return Instance; 
}

//------------------------------------------------------------------------------
// Name: RunMainLoop
// Desc: Here is where it all happens, this function finishes off anything else
//       that needs to be done to setup the game and then enters into the
//       main game loop. This main loop is locked to a length of at least 30ms
//       in order to give a constant frame rate.
//------------------------------------------------------------------------------
void Game::RunMainLoop()
{
    uint LastTicks;
    uint CurTicks;
    uint TimePassed = 0;
    int FPS = 0;
    SDL_Event event;
    int EndOfLevelDelay;

    // ParaGUI stuff
    PGApp->SetCursor(PGApp->GetTheme()->FindSurface("Pointer", "Pointer", "normal"));
    PGApp->ShowCursor(PG_CURSOR_SOFTWARE);
    
    MC->SetActivated(true);

    // Start the music
    if (MenuMusic != NULL) SDLI->PlayMusic(MenuMusic, -1, MUSIC_FADE_TIME);

    LastTicks = SDL_GetTicks();
    // Main Game Loop: All the action takes place in here
    do {
        int dx = 0;
        int dy = 0;

        CU_ENDFRAME(CAM_P_IDLE_TIME);
        CU_ENDFRAME(CAM_P_MAIN_LOOP);

        CU_STARTFRAME(CAM_P_MAIN_LOOP);
        CU_STARTFRAME(CAM_P_ACTIVE_TIME);

        CurTicks = SDL_GetTicks();
        TimePassed = CurTicks - LastTicks;
        LastTicks = CurTicks;

        if (TimePassed > 0) {
            FPS = 1000 / TimePassed;

            // This is mainly so that the game can be more easily debugged
            if (TimePassed > 100) TimePassed = 100;
        }

        IH->UpdateInput();

        if (IH->KeyPressed[SDLK_ESCAPE] 
            && (GameState == PLAYING || GameState == ISSUING_ORDERS)) {  
            // Catch the ESCAPE key
            PauseGame();
        }

        if (IH->Commands[InputHandler::ISSUE_ORDERS] 
            && GameState == PLAYING) {
            OpenOrdersPanel();
        }

        if (GameState == ENDING_LEVEL) {
            EndOfLevelDelay -= TimePassed;
            if (EndOfLevelDelay <= 0) QuitGame();
        }

		// If the game is stopped then we need only draw the background before 
		// going onto the menu. If it is paused then we only need to draw the
		// current frame, if it is playing then we also need to update the
		// entities

        if (GameState == STOPPED || GameState == PAUSED) {
			// Draw the background
            MC->DrawBackground();
        } else {
            if (GameState == PLAYING) {
			    CurLevel->Update(TimePassed);
                if (CurLevel->GetGameOver()) {
                    GameState = ENDING_LEVEL;
                    EndOfLevelDelay = END_OF_LEVEL_DELAY;

                    // Stop all of the level's sounds and music
                    Mix_HaltChannel(-1);
                    SDLI->StopMusic(MUSIC_FADE_TIME);

                    // Play a sound effect to signify either the players 
                    // victory or defeat.
                    if (CurLevel->GetPlayerVictory()) {
                        SDLI->PlayChannel(-1, 
                            SDLI->Samples[SDL_Interface::YOU_WIN], 0);
                    } else {
                        SDLI->PlayChannel(-1, 
                            SDLI->Samples[SDL_Interface::GAME_OVER], 0);
                    }
                }
            } 

            // Update the buffered variables
            // We keep a copy of some of the player's planes variables in case
            // the original is killed
            PPlane = CurLevel->GetPlayerPlane();
            if (PPlane) {
                PPlanePos = PPlane->GetWorldPos();
                PPlaneBombsLeft = PPlane->GetBombsLeft();
                PPlaneBulletsLeft = PPlane->GetBulletsLeft();
                PPlaneFuelLeft = PPlane->GetFuelLeft();
                PPlaneVelocity = PPlane->GetVelocity();
            }

            // Update the DebugCamera if it is in use
            if (GetOption(USE_DEBUG_CAMERA) != 0) {
                Point2D MousePos;
                float SecondsPassed = (float)TimePassed / 1000.0f;

                // Get the state of the mouse
                SDL_GetMouseState(&MousePos.x, &MousePos.y);

                // Scroll the viewport if the mouse is near a vertical edge of the screen
                if (MousePos.x <= OrdersPanel::SCROLL_MARGIN) {
                    DebugCameraPos.x -= (int)(OrdersPanel::VP_SCROLL_SPEED*SecondsPassed);
                } else if (MousePos.x >= LS->GetViewportWidth() - OrdersPanel::SCROLL_MARGIN) {
                    DebugCameraPos.x += (int)(OrdersPanel::VP_SCROLL_SPEED*SecondsPassed);
                }
                DebugCameraPos.x = OP->ValidateViewportX(DebugCameraPos.x);
                DebugCameraPos.y = LS->CalculatePixelHeight(DebugCameraPos.x);
            }

            // Now draw the current game frame
            DrawGameFrame(SDLI->GetScreen());
            DrawUserInterface(SDLI->GetScreen());
        }
    
        if (GetOption(DISPLAY_FPS) != 0) {
            BFont_LeftPrintStringFont(SDLI->GetScreen(), 
                SDLI->MainFont, 10, "  FPS: %i", FPS);
        }

        CU_ENDFRAME(CAM_P_ACTIVE_TIME);
        CU_STARTFRAME(CAM_P_IDLE_TIME);

        if (MC->GetActivated() || OP->GetActivated()) {
            if (OP->GetActivated() && OP->GetVisible()) {
                // OrdersPanel must be active on its own
                OP->Update(TimePassed);
            } else {
                MC->Update();
            }
            // As a part of the program that uses ParaGUI is activated, we
            // need to pass messages to ParaGUI so that it can do its funky thing.
            // This is adapted from PG_Application::RunEventLoop and designed to
            // take its place.

            // pull motion events (may flood the eventqueue)
		    while(SDL_PeepEvents(&event, 1, SDL_GETEVENT, SDL_MOUSEMOTIONMASK) > 0)
			    ;
		
            // Handle any events that are waiting for us
		    while (SDL_PollEvent(&event) != 0) {
			    PG_MessageObject::PumpIntoEventQueue(&event);
		    }

            SDL_Surface* screen = PG_Application::GetScreen();
            PG_Rect r(0, 0, SDLI->GetScreenWidth(), SDLI->GetScreenHeight());

            // TODO: In windows, when the user switches away from fullscreen mode
            // I think that some of ParaGUI surfaces are lost. However, the surfaces
            // that I have stored on hardware surfaces are not lost (at least on this
            // computer) and I don't know why that should be. Anyway, this next if
            // statement just makes sure that no ParaGUI widgets are drawn when the game
            // is in fullscreen mode and hasn't got th ecomplete attention of the user.
            // It does not stop graphical corruption of the ParaGUI widgets however so
            // I may need to reload them, I am leaving this for now though until I have done
            // further research.
            if (!SDLI->IsFullScreen() 
                || SDL_GetAppState() == (SDL_APPMOUSEFOCUS | SDL_APPINPUTFOCUS | SDL_APPACTIVE)) {
            
                // Do an update of the entire screen
	            PG_Application::LockScreen();
	            SDL_SetClipRect(screen, (PG_Rect*)&r);
	            PG_Widget::GetWidgetList()->Intersect((PG_Rect*)&r).Blit(r);
	            SDL_SetClipRect(screen, NULL);
	            PG_Application::UnlockScreen();

                PG_Application::DrawCursor();
            }
        }

        SDL_Flip(SDLI->GetScreen());

    } while(GameState != EXITING);
}

//------------------------------------------------------------------------------
// Name: LoadLevel
// Desc: This routine loads a level into the game but does not start playing 
//       it. This routine should only be called when the game is in the state
//       STOPPED.
//------------------------------------------------------------------------------
void Game::LoadLevel(const char *Filename) 
{
    assert(GameState == STOPPED);

    // Get rid of any level that was loaded in before
    if (CurLevel != NULL) delete CurLevel;
    CurLevel = new Level(Filename);

    // Set other objects in the game to their initial state
    LevelName = CurLevel->GetName();
    LS = CurLevel->GetLandscape();

    PPlane = CurLevel->GetPlayerPlane();
    if (PPlane != NULL) {
        PPlanePos = PPlane->GetWorldPos();
        PPlaneBombsLeft = PPlane->GetBombsLeft();
        PPlaneBulletsLeft = PPlane->GetBulletsLeft();
        PPlaneFuelLeft = PPlane->GetFuelLeft();
        PPlaneVelocity = PPlane->GetVelocity();
    } else {
        PPlanePos = Point2D(0, 600);
        PPlaneBombsLeft = Plane::STARTING_BOMBS;
        PPlaneBulletsLeft = Plane::STARTING_BULLETS;
        PPlaneFuelLeft = Plane::STARTING_FUEL;
        PPlaneVelocity = 0;
    }
    
    // Define the area of the screen that the viewport is to occupy
    LS->SetViewport(Point2D(0, 0), 
        SDLI->GetScreenWidth(), SDLI->GetScreenHeight());
}

//------------------------------------------------------------------------------
// Name: StartNewGame
// Desc: Allows the player to start playing the currently loaded level
//------------------------------------------------------------------------------
void Game::StartNewGame() 
{
    assert(GameState == STOPPED && CurLevel != NULL);

    MC->SetActivated(false);
    SDLI->StopMusic(MUSIC_FADE_TIME);
    if (CurLevel->Music != NULL) {
        SDLI->PlayMusic(CurLevel->Music, -1, MUSIC_FADE_TIME);
    }
    GameState = PLAYING;

    // Position the DebugCamera
    DebugCameraPos.x = SDLI->GetScreenWidth() / 2;
}

//------------------------------------------------------------------------------
// Name: PauseGame
// Desc: Pauses the game and opens up the menu
//------------------------------------------------------------------------------
void Game::PauseGame()
{
    OldState = GameState;
    GameState = PAUSED;
    MC->SetActivated(true);
    if (OP->GetActivated()) OP->SetVisible(false);
      
    IH->KeyPressed[SDLK_ESCAPE] = false;
    Mix_Pause(-1);
}

//------------------------------------------------------------------------------
// Name: ContinueGame
// Desc: This subroutine resumes a game that was paused
//------------------------------------------------------------------------------
void Game::ContinueGame() 
{
    MC->SetActivated(false);
    GameState = OldState;
    if (OP->GetActivated()) OP->SetVisible(true);

    if (GameState == PLAYING) Mix_Resume(-1);
}

//------------------------------------------------------------------------------
// Name: QuitGame
// Desc: Quits the current game and returns the player to the main menu
//------------------------------------------------------------------------------
void Game::QuitGame() 
{
    FinalScore = CurLevel->GetPlayerScore();

    // If this routine is called from the game then the menu controller will 
    // need to be activated.
    if (!MC->GetActivated()) MC->SetActivated(true);
    // Close down the orders panel if it is open
    OP->SetActivated(false);

    // Halt the levels music if it is still playing
    SDLI->StopMusic(MUSIC_FADE_TIME);
    SDL_Delay(MUSIC_FADE_TIME);

    // Delete the current level and set the pointers to it and its 
    // landscape to NULL
    delete CurLevel;
    CurLevel = NULL;
    LS = NULL;

    // Free up any channels that may have reamained paused
    Mix_Resume(-1);
    Mix_HaltChannel(-1);

    
    if (MenuMusic != NULL) SDLI->PlayMusic(MenuMusic, -1, MUSIC_FADE_TIME);

    GameState = STOPPED;
}

//------------------------------------------------------------------------------
// Name: Exit
// Desc: Exits the game
//------------------------------------------------------------------------------
void Game::Exit()
{
    SDLI->StopMusic(MUSIC_FADE_TIME);
    GameState = EXITING;
}

//------------------------------------------------------------------------------
// Name: OpenOrdersPanel
// Desc: Opens up the orders panel so that the player can issue orders to his
//       or her units.
//------------------------------------------------------------------------------
void Game::OpenOrdersPanel()
{
    GameState = ISSUING_ORDERS;
    Mix_Pause(-1);

    OP->SetActivated(true);
}

//------------------------------------------------------------------------------
// Name: CloseOrdersPanel
// Desc: Closes the orders panel and gets on with the game.
//------------------------------------------------------------------------------
void Game::CloseOrdersPanel()
{
    OP->SetActivated(false);

    Mix_Resume(-1);
    GameState = PLAYING;
}

//------------------------------------------------------------------------------
// Name: LoadOptions
// Desc: Loads in various game options from an external text file. If an
//       invalid value is given from an option then it is ignored and a
//       default value used in its stead.
//------------------------------------------------------------------------------
void Game::LoadOptions()
{
    int i;
    int CurOption;
    string CurLine;
    string Name;
    string StringValue;
    int NumericValue;
    string Delimiter(OPTIONS_DELIMITER);
    string::size_type DelimiterPos;
    
    ifstream OptionFile(CAM_OPTIONS_FILE, ios::in);

    memcpy(Options, DEFAULT_OPTIONS, NUM_OPTIONS*sizeof(int));

    if (!OptionFile) {
        fprintf(stderr, "Unable to access file %s\n", CAM_OPTIONS_FILE);
        return;
    }

    getline(OptionFile, CurLine);
    bool ScreenModeFound;
    bool BitdepthFound;
    while (OptionFile) {    
        DelimiterPos = CurLine.find(Delimiter);
        if (DelimiterPos == 0 || DelimiterPos == string::npos) 
            goto GetNextLine;       // No option on this line

        Name = CU_TrimString(CurLine.substr(0, DelimiterPos));
        StringValue = CU_TrimString(CurLine.substr(DelimiterPos + Delimiter.length()));
        stringstream(StringValue) >> NumericValue;

        CurOption = NO_OPTION;
        for (i = 0; i < NUM_OPTIONS; i++) {
            if (Name == OPTION_DETAILS[i].String) {
                CurOption = i;
                break;
            }
        }

        // Translate the current option into the approporiate format
        ScreenModeFound = false;
        BitdepthFound = false;
        switch (CurOption) {
            case NO_OPTION:
                goto GetNextLine;       // Option not recognised 
            case SCREENMODE:
                for (i = 0; i < SDL_Interface::NUM_SCREEN_MODES; i++) {
                    if (StringValue == SDL_Interface::SCREEN_MODE_DETAILS[i].String) {
                        ScreenModeFound = true;
                        NumericValue = i;
                        break;
                    }
                }
                if (!ScreenModeFound) goto GetNextLine;   // Invalid screenmode
                break;
            case BITDEPTH:
                for (i = 0; i < SDL_Interface::NUM_SCREEN_BITDEPTHS; i++) {
                    if (NumericValue == SDL_Interface::SCREEN_BITDEPTH_DETAILS[i]) {
                        NumericValue = i;
                        BitdepthFound = true;
                        break;
                    }
                }
                if (!BitdepthFound) goto GetNextLine;   // Invalid bitdepth
                break;
            case NUM_PARALLAX_LAYERS:
                if (NumericValue < 0 || NumericValue > Landscape::NUM_PARALLAX_LAYERS) {
                    goto GetNextLine;   // Invalid number of parallax layers
                }
                break;
            case KEY_CLIMB:
            case KEY_DIVE:
            case KEY_INVERT:
            case KEY_FIRE:
            case KEY_START_ENGINE:
            case KEY_DROP_BOMB:
            case KEY_ISSUE_ORDERS:
            case KEY_FLY_HOME:
                if (StringValue.length() == 0) {
                    NumericValue = SDLK_ESCAPE;
                } else if (!IH->IsKeyValid(StringValue.c_str(), &NumericValue)) {
                    goto GetNextLine;   // An invalid key was specified
                }
        
                if (NumericValue == SDLK_ESCAPE) break;

                // Check that this key has not been used already
                for (i = KEY_CLIMB; i < CurOption; i++) {
                    if (Options[i] == NumericValue)
                        goto GetNextLine;   // Key has already been used
                }

                break;
            case SFX_VOLUME:
                if (NumericValue < 0 || NumericValue > MIX_MAX_VOLUME) {
                    goto GetNextLine;   // Not a supported volume
                }
                break;
            case MUSIC_VOLUME:
                if (NumericValue < 0 || NumericValue > MIX_MAX_VOLUME) {
                    goto GetNextLine;   // Not a supported volume
                }
                break;
            case FULLSCREEN:
            case CLOCK_CONTROLS:
            case REVERSE_STEREO:
            case DISPLAY_FPS:
            case USE_DEBUG_CAMERA:
                NumericValue = (NumericValue ? 1 : 0);
                break;
            case MENU_MUSIC:
                NumericValue = 1;
                MenuMusicFile = StringValue;
                break;
        }
        Options[CurOption] = NumericValue;  // Set the option

GetNextLine:
        getline(OptionFile, CurLine);       // Get the next line
    } 
}

//------------------------------------------------------------------------------
// Name: SaveOptions
// Desc: This outputs various game options to an external text file
//------------------------------------------------------------------------------
void Game::SaveOptions()
{
    int i;
    ofstream OptionFile(CAM_OPTIONS_FILE, ios::out);

    if (!OptionFile) {
        fprintf(stderr, "Unable to access or create options file %s\n", 
            CAM_OPTIONS_FILE);
        return;
    }

    for (i = 0; i < NUM_OPTIONS; i++) {
        if (string(OPTION_DETAILS[i].Comment).length() > 0) {
            OptionFile << OPTION_DETAILS[i].Comment << endl;
        }
        OptionFile << OPTION_DETAILS[i].String << OPTIONS_DELIMITER << " ";    

        // Format the options and write them to the option file
        switch (i) {
            case SCREENMODE:
                OptionFile << SDL_Interface::SCREEN_MODE_DETAILS[Options[i]].String << endl;
                break;
            case BITDEPTH:
                OptionFile << SDL_Interface::SCREEN_BITDEPTH_DETAILS[Options[i]] << endl;
                break;
            case KEY_CLIMB:
            case KEY_DIVE:
            case KEY_INVERT:
            case KEY_FIRE:
            case KEY_START_ENGINE:
            case KEY_DROP_BOMB:
            case KEY_ISSUE_ORDERS:
            case KEY_FLY_HOME:
                if (Options[i] == SDLK_ESCAPE) {
                    OptionFile << endl;
                } else {
                    OptionFile << SDL_GetKeyName((SDLKey)Options[i]) << endl;
                }
                break;
            case MENU_MUSIC:
                if (Options[i] == 0) {
                    OptionFile << DEFAULT_MENU_MUSIC << endl;
                } else {
                    OptionFile << MenuMusicFile << endl;
                }
                break;
            default:
                OptionFile << Options[i] << endl;
        }
    }
    OptionFile.close();
}

//------------------------------------------------------------------------------
// Name: GetOption
// Desc: Returns the current value of a game option
//------------------------------------------------------------------------------
int Game::GetOption(int AnOption)
{
    assert(AnOption >= 0 && AnOption < NUM_OPTIONS);
    return Options[AnOption];
}

//------------------------------------------------------------------------------
// Name: SetOption
// Desc: Sets the value of a game option
//------------------------------------------------------------------------------
void Game::SetOption(int AnOption, int OptionValue)
{
    assert(AnOption >= 0 && AnOption < NUM_OPTIONS);
    Options[AnOption] = OptionValue;
}

//------------------------------------------------------------------------------
// Name: GetOptionTooltip
// Desc: Used by the options menus to display a tooltip for a particular
//       option.
//------------------------------------------------------------------------------
const char *Game::GetOptionTooltip(int AnOption)
{
    assert(AnOption >= 0 && AnOption < NUM_OPTIONS);
    return OPTION_DETAILS[AnOption].Tooltip;
}

//------------------------------------------------------------------------------
// Name: LoadHighScores
// Desc: Loads in the high scores from an external text file. If the high
//       scores file isn't found, or if the format of the file is invalid then
//       default high scores are used instead.
//
// Note: The file must be opened with ios::binary as there may be EOF 
//       characters in the file.
//------------------------------------------------------------------------------
void Game::LoadHighScores()
{
    int i;
    ifstream HighScoresFile(CAM_HIGH_SCORES_FILE, ios::in | ios::binary);
    uchar *FileData = NULL;
    int TableSize = sizeof(HIGH_SCORE)*NUM_HIGH_SCORES;
    int FileSize = TableSize + 1;

    memcpy(HighScores, DEFAULT_HIGH_SCORES, TableSize);
    if (!HighScoresFile) {
        fprintf(stderr, "Unable to access file %s\n", CAM_HIGH_SCORES_FILE);
        return;
    }

    FileData = (uchar *)malloc(FileSize);
    if (FileData == NULL) goto FinishedWithFile;
    memset(FileData, 0, FileSize);  // Clear the memory
    HighScoresFile.read((char *)FileData, FileSize);
    
    if (!HighScoresFile) {
        CU_ShowMessage("Error", "Incomplete high score table");
        goto FinishedWithFile;      // Not enough data in the file
    }
    
    // Unencrypt the high score data and copy it into the highscores table
    for (i = 1; i < FileSize; i++) FileData[i] ^= FileData[0];
    memcpy(HighScores, FileData + 1, TableSize);

FinishedWithFile:
    HighScoresFile.close();
}

//------------------------------------------------------------------------------
// Name: SaveHighScores
// Desc: This subroutine saves the current contents of the high score table
//       to an external text file. Before the high score table is saved, it is
//       sorted using a simple bubble sort and then 'encrypted'.
//
//       The encryption is very simple and only meant to stop casual cheaters.
//       A number is chosen from the range 1 to 255 and XOR'd with every
//       byte in the data to be saved. This 'key' is then written to the front
//       of the file and the original data can then be decoded by XORing every
//       byte with the key again. This works because of the following 
//       properties of the XOR function.
//
//              A XOR A = 0
//          and A XOR 0 = A
//
//       So the key cancels itself out.
//------------------------------------------------------------------------------
void Game::SaveHighScores()
{
    int i;
    ofstream HighScoresFile(CAM_HIGH_SCORES_FILE, ios::out | ios::binary);
    uchar *FileData = NULL;
    int TableSize = sizeof(HIGH_SCORE)*NUM_HIGH_SCORES;
    int FileSize = TableSize + 1;

    if (!HighScoresFile) {
        fprintf(stderr, "Unable to access or create high scores file %s\n", 
            CAM_HIGH_SCORES_FILE);
        return;
    }
    SortHighScores();

    FileData = (uchar *)malloc(FileSize);
    if (FileData == NULL) goto FinishedWithFile;
    // Pick an encryption key
    FileData[0] = (uchar)CamelMath::Random(1, 255);
    // Copy over the high scores data
    memcpy(FileData + 1, HighScores, TableSize);
    // 'Encrypt' the high score data
    for (i = 1; i < FileSize; i++) FileData[i] ^= FileData[0];
    // Save it to the file
    HighScoresFile.write((char *)FileData, FileSize);

FinishedWithFile:
    HighScoresFile.close();
}

//------------------------------------------------------------------------------
// Name: SortHighScores
// Desc: Sorts the highscores table using a simple bubble sort.
//------------------------------------------------------------------------------
void Game::SortHighScores()
{
    int i, j;
    HIGH_SCORE SwapSpace;

    // Sort the high scores, with score 0 being the highest and score
    // NUM_HIGH_SCORES-1 being the lowest.
    for (i = 0; i < NUM_HIGH_SCORES; i++) {
        for (j = i + 1; j < NUM_HIGH_SCORES; j++) {
            if (HighScores[j].Score > HighScores[i].Score) {
                SwapSpace = HighScores[i];
                HighScores[i] = HighScores[j];
                HighScores[j] = SwapSpace;
            }
        }
    }
}

//------------------------------------------------------------------------------
// Name: GetHighScore
// Desc: Gets the high score located at 'Index' in the high score table.
//------------------------------------------------------------------------------
Game::HIGH_SCORE Game::GetHighScore(int Index)
{
    assert(Index >= 0 && Index < NUM_HIGH_SCORES);
    return HighScores[Index];
}

//------------------------------------------------------------------------------
// Name: SetHighScore
// Desc: Replaces the high score located at 'Index' in the high score table
//       with 'NewHighScore'.
//------------------------------------------------------------------------------
void Game::SetHighScore(int Index, HIGH_SCORE NewHighScore)
{
    assert(Index >= 0 && Index < NUM_HIGH_SCORES);
    HighScores[Index] = NewHighScore;

    SortHighScores();
}

//------------------------------------------------------------------------------
// Name: DrawGameFrame
// Desc: Draws the current game frame onto the given surface.
//------------------------------------------------------------------------------
void Game::DrawGameFrame(SDL_Surface *Surface)
{
    assert(Surface != NULL);

    LS->EnableViewportClippingRect();

    // TODO: This next section of code could be tidier
    if (GameState == ENDING_LEVEL) {
        if (CurLevel->GetPlayerLives() <= 0) {
            Point2D CentrePos = PPlanePos;
            CentrePos.y += 150;
            LS->CentreViewportAt(CentrePos);
        } else if (CurLevel->GetPlayerVictory()) {
            LS->CentreViewportAt(CurLevel->GetFlag(Entity::EVIL)->GetCentrePos());
        } else {
            LS->CentreViewportAt(CurLevel->GetFlag(Entity::GOOD)->GetCentrePos());
        }
    } else {
        if (OP->GetActivated()) {
            LS->CentreViewportAt(OP->GetViewportCentrePos());
        } else if (GetOption(USE_DEBUG_CAMERA) != 0) {
            LS->CentreViewportAt(DebugCameraPos);
        } else {
            Point2D CentrePos = PPlanePos;
            CentrePos.y += 150;
            LS->CentreViewportAt(CentrePos);
        }
    }

	CU_STARTFRAME(CAM_P_DRAW_BACKGROUND);
	CurLevel->DrawBackground(SDLI->GetScreen());
    CU_ENDFRAME(CAM_P_DRAW_BACKGROUND);

	CU_STARTFRAME(CAM_P_DRAW_ENTITIES);
	CurLevel->DrawEntities(SDLI->GetScreen());
	CU_ENDFRAME(CAM_P_DRAW_ENTITIES);

    CU_STARTFRAME(CAM_P_DRAW_FOREGROUND);
	CurLevel->DrawForeground(SDLI->GetScreen());
    CU_ENDFRAME(CAM_P_DRAW_FOREGROUND);

    LS->DisableViewportClippingRect();
}

//------------------------------------------------------------------------------
// Name: DrawUserInterface
// Desc: This routine draws the user interface, containing information about
//       lives, bullets and bombs left onto the screen.
//------------------------------------------------------------------------------
void Game::DrawUserInterface(SDL_Surface *Surface)
{
    assert(Surface != NULL);

    int ScreenWidth = SDLI->GetScreenWidth();
    int ScreenHeight = SDLI->GetScreenHeight();
    int BannerFrame;
    Point2D BannerPos;
    int BannerWidth, BannerHeight;

    // Draw the user interface. This isn't particularly pretty code, but fortunately it
    // doesn't do anything complicated.
    SPR->DrawFrame(Sprites::UI_SPEEDOMETER_SPRITE, Surface, 0, ScreenHeight - 135);
    DrawGaugeNeedle(Surface, Point2D(40, ScreenHeight - 95), 0, 580, (int)PPlaneVelocity);
    SPR->DrawFrame(Sprites::UI_FUEL_GUAGE_SPRITE, Surface, 30, ScreenHeight - 80);
    DrawGaugeNeedle(Surface, Point2D(70, ScreenHeight - 40), 0, Plane::STARTING_FUEL, PPlaneFuelLeft);
    SPR->DrawFrame(Sprites::UI_BOMB_SPRITE, Surface, ScreenWidth - 150, ScreenHeight - 40);
    BFont_PrintStringFont(Surface, SDLI->MainFont, ScreenWidth - 140, ScreenHeight - 30, "%i", PPlaneBombsLeft);
    SPR->DrawFrame(Sprites::UI_BULLET_SPRITE, Surface, ScreenWidth - 95, ScreenHeight - 35);
    BFont_PrintStringFont(Surface, SDLI->MainFont, ScreenWidth - 90, ScreenHeight - 30, "%i", PPlaneBulletsLeft);
    SPR->DrawFrame(Sprites::UI_HEART_SPRITE, Surface, ScreenWidth - 40, ScreenHeight - 40);
    BFont_PrintStringFont(Surface, SDLI->MainFont, ScreenWidth - 25, ScreenHeight - 30, "%i", CurLevel->GetPlayerLives());

    BFont_RightPrintStringFont(Surface, SDLI->MainFont, 10, "Score: %u  ", CurLevel->GetPlayerScore());

    if (CurLevel->GetGameOver()) {
        if (CurLevel->GetPlayerVictory()) {
            // Draw the victory banner
            BannerFrame = Sprites::YOUWIN_BANNER_SPRITE;
        } else {
            // Draw the game over banner
            BannerFrame = Sprites::GAMEOVER_BANNER_SPRITE;
        }
        SPR->GetFrameDims(BannerFrame, &BannerWidth, &BannerHeight);

        BannerPos.x = (SDLI->GetScreenWidth() / 2) - (BannerWidth / 2);
        BannerPos.y = (SDLI->GetScreenHeight() / 2) - BannerHeight;
    
        SPR->DrawFrame(BannerFrame, SDLI->GetScreen(), BannerPos.x, BannerPos.y);
    }
}

//------------------------------------------------------------------------------
// Name: DrawGaugeNeedle
// Desc: Draws a needle in black at the point given by 'CentrePos'. The
//       needles value can range from RangeStart to RangeEnd and RangeStart 
//       must be less than RangeEnd
//------------------------------------------------------------------------------
void Game::DrawGaugeNeedle(SDL_Surface *Surface, Point2D NeedleStartPos, 
                           int RangeStart, int RangeEnd, int Value)
{
    double NeedleAngle;
    Point2D NeedleEndPos;
    int NeedleX, NeedleY;

    assert(Surface != NULL);
    assert(RangeStart < RangeEnd);

    if (Value < RangeStart) {
        Value = RangeStart;
    } else if (Value > RangeEnd) {
        Value = RangeEnd;
    }

    // Convert the value to an angle within the needles range
    NeedleAngle = UI_GAUGE_NEEDLE_START_ANGLE + 
        UI_GAUGE_NEEDLE_RANGE * (double)(Value - RangeStart)/(RangeEnd - RangeStart);

    // Work out where the needle end point will be
    NeedleX = (int)((double)UI_GAUGE_NEEDLE_LENGTH * cos(NeedleAngle));
    NeedleY = (int)((double)UI_GAUGE_NEEDLE_LENGTH * sin(NeedleAngle));
    // NeedleX needs to be subtracted as angle 0 is taken as left, and NeedleY
    // needs to be subtracted due to the screen coordinate system.
    NeedleEndPos.x = NeedleStartPos.x - NeedleX;
    NeedleEndPos.y = NeedleStartPos.y - NeedleY;

    // Draw the needle
    SDLI->DrawLine(SDLI->GetScreen(), SDL_MapRGB(SDLI->GetScreen()->format, 0, 0, 0),
        NeedleStartPos, NeedleEndPos);
}
