/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: InputHandler.cpp
// Desc: Each frame, a member of this class is called to examine the state of 
//       the mouse and keyboard and determine what commands the user is trying
//       to exit. This 'input' can then be retrieved, either to be used by the
//       rest of the game, or to be recorded, or both.
//
//       This class also provides functionality that allows user-defined keys
//       to be assigned to the commands
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <string>
#include "CamelUtility.h"
#include "SDL.h"
#include "Game.h"
#include "InputHandler.h"

using namespace std;

//------------------------------------------------------------------------------
// InputHandler Implementation
//------------------------------------------------------------------------------
const InputHandler::COMMAND_INFO InputHandler::COMMAND_DETAILS[NUM_COMMANDS] = {
    { "Climb", Game::KEY_CLIMB },
    { "Dive", Game::KEY_DIVE },
    { "Invert", Game::KEY_INVERT },
    { "Fire", Game::KEY_FIRE },
    { "Start Engine", Game::KEY_START_ENGINE },
    { "Drop Bomb", Game::KEY_DROP_BOMB },
    { "Issue Orders", Game::KEY_ISSUE_ORDERS },
    { "Fly Home", Game::KEY_FLY_HOME }
};

InputHandler *InputHandler::Instance = NULL;
InputHandler::InputHandler() {}

InputHandler::InputHandler(Game *G) 
{
    int i;

    assert(G != NULL);
    this->G = G;

    KeyBoard = NULL;
    OldKeyBoard = NULL;
    KeyPressed = NULL;

    KeyBoard = SDL_GetKeyState(&NumKeys);
    OldKeyBoard = (uchar *)malloc(NumKeys * sizeof(uchar));
    KeyPressed = (bool *)malloc(NumKeys * sizeof(bool));
    assert(KeyBoard != NULL && OldKeyBoard != NULL && KeyPressed != NULL);

    OldMouseState = 0;
    memset(OldKeyBoard, 0, NumKeys * sizeof(uchar));
    memset(KeyPressed, false, NumKeys * sizeof(bool));
    memset(MouseButtonPressed, false, 2 * sizeof(bool));
    memset(Commands, false, InputHandler::NUM_COMMANDS);

    // Work out which keys are used for each command
    for (i = 0; i < NUM_COMMANDS; i++) {
        CommandKeys[i] = G->GetOption(COMMAND_DETAILS[i].Option);
    }
}

InputHandler::~InputHandler()
{
    NumKeys = 0;
    memset(Commands, false, InputHandler::NUM_COMMANDS);
    if (OldKeyBoard != NULL) free(OldKeyBoard);
    if (KeyPressed != NULL) free(KeyPressed);
}

InputHandler *InputHandler::GetInstance(Game *G) 
{
    if (Instance == NULL) {
        Instance = new InputHandler(G);
    }
    return Instance; 
}

void InputHandler::UpdateInput()
{
    int i;
    int x, y;
    uchar MouseState = SDL_GetMouseState(&x, &y);

    SDL_PumpEvents();
    
    for (i = 0; i < NUM_MOUSE_BUTTONS; i++) {
        MouseButtonPressed[i] = ((OldMouseState&SDL_BUTTON(i+1) 
            && !(MouseState&SDL_BUTTON(i+1))) ? true : false);
    }
    for (i = 0; i < NumKeys; i++) {
        KeyPressed[i] = ((OldKeyBoard[i] && !KeyBoard[i]) ? true : false);
    }
    for (i = 0; i < InputHandler::NUM_COMMANDS; i++) {
        if (CommandKeys[i] == SDLK_ESCAPE) {    // Catch the ESCAPE key
            Commands[i] = false;
        } else {
            Commands[i] = (KeyBoard[CommandKeys[i]] ? true : false);
        }
    }
    OldMouseState = MouseState;
    memcpy(OldKeyBoard, KeyBoard, NumKeys * sizeof(uchar));
}

int InputHandler::GetCommandKey(uint ACommand)
{
    assert(ACommand < InputHandler::NUM_COMMANDS);

    return CommandKeys[ACommand];
}

void InputHandler::SetCommandKey(uint ACommand)
{
    int i, j;

    assert(ACommand < InputHandler::NUM_COMMANDS);

    for (;;) {
        SDL_PumpEvents();

        // Don't assign a key if the ESCAPE key is pressed
        if (KeyBoard[SDLK_ESCAPE]) return;

        // The ESCAPE key is used as a blank. Any command it is attached to will
        // never actually be called as the key press will have been caught 
        // before then
        for (i = 0; i < NumKeys; i++) {
            if (KeyBoard[i]     // Ignore the 'lock' keys
                && i != SDLK_NUMLOCK && i != SDLK_CAPSLOCK && i != SDLK_SCROLLOCK) {
                if (CommandKeys[ACommand] == i) {
                    return;     // No change
                } else {
                    // Clear any commands that already use this key
                    for (j = 0; j < InputHandler::NUM_COMMANDS; j++) {
                        if (CommandKeys[j] == i) {
                            CommandKeys[j] = SDLK_ESCAPE;
                            G->SetOption(COMMAND_DETAILS[j].Option, SDLK_ESCAPE);
                        }
                    }
                    CommandKeys[ACommand] = i;
                    G->SetOption(COMMAND_DETAILS[ACommand].Option, i);
                    return;
                }
            }
        }
    }
}

bool InputHandler::IsKeyValid(const char *KeyName, int *KeyNumber)
{
    int i;
    string TestKey;
    string SDL_Key;

    assert(KeyName != NULL && KeyNumber != NULL);

    TestKey = KeyName;
    for (i = 0; i < SDLK_LAST; i++) {
        SDL_Key = SDL_GetKeyName((SDLKey)i);
        if (SDL_Key != "unknown key" && SDL_Key == TestKey) {
            *KeyNumber = i;
            return true;
        }
    }
    return false;
}
