/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Landscape.cpp
// Desc: This file provides functions to draw the landscape and move the
//       viewport around.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <string.h>
#include <string>
#include <fstream> 
#include <sstream>
#include <assert.h>
#include "globals.h"
#include "CamelUtility.h"
#include "Camel.h"
#include "CamelMath.h"
#include "Game.h"
#include "SDL_Interface.h"
#include "Sprites.h"
#include "Level.h"
#include "Landscape.h"

using namespace std;

//------------------------------------------------------------------------------
// Landscape Implementation
//------------------------------------------------------------------------------

// The rate at which a layer moves horizontally compared to the foreground
const float Landscape::PL_H_SPEED[NUM_PARALLAX_LAYERS] = {
    0.5f, 1.0f / 3.0f
};
// The rate of vertical parallax motion compared to horizontal parallax 
// motion. The reason that this isn't one to one is so that the height of
// the parallax layers can be smaller.
const float Landscape::PL_V_SPEED = 0.25f;

//------------------------------------------------------------------------------
// Name: Landscape
// Desc: Performs general setup tasks and loads in the landscape from the 
//       supplied .lev file.
//
// Note: Loading in a landscape involves loading in the landscape
//       data and then drawing the landscape onto an SDL surface that is kept
//       in memory. This takes up a lot of space in memory, but allows for the
//       landscape to be drawn quickly to the screen and makes scrolling very
//       easy to implement.
//
//       Also, two drawings of the landscape are made. One, made on an RLE
//       surface is used for blitting to the screen due to its speed. The
//       other is stored on an uncompressed surface and used for collision
//       detection. Two surfaces are used as locking a RLE surface seem to
//       cause a huge slowdown.
//------------------------------------------------------------------------------
Landscape::Landscape(const char *Filename) 
{
    int i;

	assert(Filename != NULL);

    // Get pointers to global interfaces
    G = Game::GetInstance();
    SDLI = SDL_Interface::GetInstance(NULL);
    SPR = Sprites::GetInstance();
    
    // Give variables an initial value
    WorldWidth = WorldHeight = LandHeight = BackgroundHeight = 0;
    Contours = NULL;
    LandscapeData = BackgroundData = NULL;

    ViewportPos = Point2D(0, 0);
    ViewportWidth = SDLI->GetScreenWidth();
    ViewportHeight = SDLI->GetScreenHeight();
    CurWorldPos = MaxWorldPos = Point2D(0, 0);
    ParallaxLayersLoaded = 0;
    for (i = 0; i < NUM_PARALLAX_LAYERS; i++) {
        ParallaxLayers[i].Image = NULL;
        ParallaxLayers[i].Height = 0;
    }
    LandImage = NULL;
    LandImageRLE = NULL;
    BackgroundImageRLE = NULL;

    CalculateTileRects(TileRects);

    // Load in the landscape from the supplied .lev file
    LoadLevelFile(Filename);
    DrawLandscape();
    // Work out the dimensions of the landscape in pixels
    WorldPixelWidth = WorldWidth*TILE_WIDTH;
    WorldPixelHeight = WorldHeight*TILE_HEIGHT;
    // Work out the contours of this landscape
    NormaliseContours();

    // Update current viewport to account for this new landscape
    SetViewport(ViewportPos, ViewportWidth, ViewportHeight);
}

//------------------------------------------------------------------------------
// Name: ~Landscape
// Desc: Cleans up after the object when it is no longer needed.
//------------------------------------------------------------------------------
Landscape::~Landscape() 
{
    int i;

    // Frees the memory used to hold the landscape data and images
    if (Contours != NULL) free(Contours);
    if (LandscapeData != NULL) free(LandscapeData);
	if (BackgroundData != NULL) free(BackgroundData);
    if (LandImage != NULL) SDL_FreeSurface(LandImage);
    if (LandImageRLE != NULL) SDL_FreeSurface(LandImageRLE);
    if (BackgroundImageRLE != NULL) SDL_FreeSurface(BackgroundImageRLE);
    for (i = 0; i < NUM_PARALLAX_LAYERS; i++) {
        if (ParallaxLayers[i].Image != NULL) {
            SDL_FreeSurface(ParallaxLayers[i].Image);
            ParallaxLayers[i].Image = NULL;
        }
    }
}

//------------------------------------------------------------------------------
// Name: CalculateTileRects
// Desc: This subroutine populates an array of rectangles that can be used for
//       finding the location of a tile on the tileset image
//------------------------------------------------------------------------------
void Landscape::CalculateTileRects(SDL_Rect *RectArray)
{
    int CurTile;
    int TileX, TileY;
    SDL_Rect EmptyRect = { 0, 0, 0, 0 };
    
    assert(RectArray != NULL);

    for (CurTile = 0; CurTile < TILESET_NUM_TILES; CurTile++) {
        TileX = CurTile%TILESET_WIDTH;
        TileY = (CurTile - TileX) / TILESET_WIDTH;
        RectArray[CurTile].x = TileX * TILE_WIDTH;
        RectArray[CurTile].y = TileY * TILE_HEIGHT;
        RectArray[CurTile].w = TILE_WIDTH;
        RectArray[CurTile].h = TILE_HEIGHT;
    }
    RectArray[NO_TILE] = EmptyRect;
}

//------------------------------------------------------------------------------
// Name: LoadTileset
// Desc: Loads a tileset from a given file onto an SDL surface. The image must
//       conform to the expected dimensions and bitdepth.
//------------------------------------------------------------------------------
SDL_Surface *Landscape::LoadTileset(const char *Filename)
{
    SDL_Surface *Result;

    assert(Filename != NULL);

    Result = IMG_Load(Filename);

    if (Result == NULL) SDE_SDLError(__FILE__, __LINE__);
    if (Result->w != TILESET_IMAGE_WIDTH || Result->h != TILESET_IMAGE_HEIGHT
        || Result->format->BitsPerPixel != TILESET_IMAGE_BPP) {
        SDL_FreeSurface(Result);
        CU_ShowMessage("Error", "The tileset %s is in an incorrect format. It "
            "should be in the format %ix%ix%i.", Filename, TILESET_IMAGE_WIDTH,
            TILESET_IMAGE_HEIGHT, TILESET_IMAGE_BPP);
        exit(1);
    }
    if (SDL_FAILED(SDL_SetAlpha(Result, SDL_SRCALPHA, 0))) {
        SDL_FreeSurface(Result);
        SDE_SDLError(__FILE__, __LINE__);
    }

    return Result;
}

//------------------------------------------------------------------------------
// Name: LoadLevelFile
// Desc: Loads a .lev file and places it into a LANDSCAPE_DATA struct.
//------------------------------------------------------------------------------
void Landscape::LoadLevelFile(const char *Filename)
{
    int i;
    uint Offset, x, y;
    bool ContoursLoaded;
    bool LevelLoaded;
	bool BackgroundLoaded;
    string CurLine;  
    ifstream *LevelFile;

    LevelFile = new ifstream(Filename, ios::in);

    // Deal with the trivial case where the file doesn't exist
    if (!LevelFile) goto InvalidLevelFile;

    // Read in and check the file header
    if (!ReadInLevelFileHeader(LevelFile)
        || WorldWidth < MIN_LEVEL_WORLDWIDTH 
        || WorldWidth > MAX_LEVEL_WORLDWIDTH 
        || WorldHeight < MIN_LEVEL_WORLDHEIGHT
        || WorldHeight > MAX_LEVEL_WORLDHEIGHT
        || LandHeight < MIN_LEVEL_LANDHEIGHT
        || LandHeight > MAX_LEVEL_LANDHEIGHT
		|| BackgroundHeight < MIN_LEVEL_BACKGROUNDHEIGHT
		|| BackgroundHeight > LandHeight) {

        goto InvalidLevelFile;
    }
           
    // Read forward 3 sections to get to the Parallax Layers section
    for (i = 0; i < 3; i++) {
        do {
            getline(*LevelFile, CurLine);
        } while (*LevelFile && CurLine != Level::FILE_SECTION_DELIMITER);
        if (!(*LevelFile)) break;
    }

    // Read in up to NUM_PARALLAX_LAYERS if they exist
    for (i = 0; i < NUM_PARALLAX_LAYERS; i++) {
        string Delimiter(Level::FILE_FIELD_DELIMITER);
        string::size_type DelimiterPos;
        uint MaxHeight = WorldHeight * TILE_HEIGHT;
        string ParallaxFilename;

        getline(*LevelFile, CurLine);
        if (*LevelFile) {
            // Stop reading section if the line does not match our expectations
            if ((DelimiterPos = CurLine.find(Delimiter, 1)) == string::npos)
                goto FinishedReadingParallaxSection;    // Line not in correct format

            ParallaxFilename = string(CAM_PARALLAX_DIR) + "/" + CurLine.substr(0, DelimiterPos);
            ParallaxLayers[i].Image = IMG_Load(ParallaxFilename.c_str());
            if (ParallaxLayers[i].Image == NULL)
                goto FinishedReadingParallaxSection;    // Cannot load image

            SDL_SetAlpha(ParallaxLayers[i].Image, SDL_SRCALPHA | SDL_RLEACCEL, SDL_ALPHA_TRANSPARENT);

            stringstream(CurLine.substr(DelimiterPos + Delimiter.length())) 
                >> ParallaxLayers[i].Height;
            if (ParallaxLayers[i].Height >= MaxHeight)
                ParallaxLayers[i].Height = MaxHeight - 1;

            ParallaxLayersLoaded++;
        }
    }
FinishedReadingParallaxSection:

    // Allocate the space for the data
    Contours = (CONTOUR *)malloc(WorldWidth * sizeof(CONTOUR));
    LandscapeData =(uchar *)malloc(WorldWidth * LandHeight * sizeof(uchar));
	BackgroundData =(uchar *)malloc(WorldWidth * BackgroundHeight * sizeof(uchar));
	if (Contours == NULL ||
        LandscapeData == NULL || BackgroundData == NULL) {
		CU_ShowMessage("Error", "Out of memory");
		exit(-1);
	}

    // Set up variables so that we know how far we have got with reading the file
    ContoursLoaded = false;
    LevelLoaded = false;
	BackgroundLoaded = false;
      
    // Read forward to the contours section
    while (*LevelFile && CurLine != Level::FILE_SECTION_DELIMITER)
	    getline(*LevelFile, CurLine);
    // Read in the contours data
    x = 0;
    getline(*LevelFile, CurLine);
    while (*LevelFile && ContoursLoaded == false) {
        do {
            uchar Data;

            if (!ReadInLandscapeData(&CurLine, &Data)) {
                goto InvalidLevelFile;
            }
            Contours[x].Height = Data;
            
            x++;
            if (x >= WorldWidth) {
                ContoursLoaded = true;
                break;
            }
        } while (CurLine != "");
        getline(*LevelFile, CurLine);
    }
    if (!ContoursLoaded) goto InvalidLevelFile;

    // Now read forward until we find a LEVEL_FILE_SECTION_DELIMITER and read in
    // the landscape data
    while (*LevelFile && CurLine != Level::FILE_SECTION_DELIMITER)
	    getline(*LevelFile, CurLine);
    if (!(*LevelFile)) goto InvalidLevelFile;

    // Read in the landscape data
	Offset = x = y = 0;
    getline(*LevelFile, CurLine);
    while (*LevelFile && LevelLoaded == false) {
        do {
            if (!ReadInLandscapeData(&CurLine, &LandscapeData[Offset])) {
                goto InvalidLevelFile;
            }
                
            Offset++;
            x++;
            if (x >= WorldWidth) {
                x = 0;
                y++;
                if (y >= LandHeight) {
                    LevelLoaded = true;
                    break;
                }
            }
        } while (CurLine != "");
        getline(*LevelFile, CurLine);
    }

    if (!LevelLoaded) goto InvalidLevelFile;
        
	// Now read forward until we find a LEVEL_FILE_SECTION_DELIMITER and read in
    // the background data
	while (*LevelFile && CurLine != Level::FILE_SECTION_DELIMITER)
	    getline(*LevelFile, CurLine);
	if (!(*LevelFile)) goto InvalidLevelFile;

	if (BackgroundHeight <= 0) BackgroundLoaded = true;
	Offset = x = y = 0;
	getline(*LevelFile, CurLine);
	while (*LevelFile && BackgroundLoaded == false) {
	    do {
	        if (!ReadInLandscapeData(&CurLine, &BackgroundData[Offset])) {
                goto InvalidLevelFile;
            }
                
            Offset++;
            x++;
            if (x >= WorldWidth) {
                x = 0;
                y++;
                if (y >= BackgroundHeight) {
                    BackgroundLoaded = true;
                    break;
                }
            }
        } while (CurLine != "");
        getline(*LevelFile, CurLine);
    }

    if (!BackgroundLoaded) goto InvalidLevelFile;
    LevelFile->close();
    delete LevelFile;
    return;

InvalidLevelFile:
    CU_ShowMessage("Error", "Unable to read in the level file %s. Please ensure "
        "that the file is present and that its format is correct.", Filename);
    exit(1);
}
    
//------------------------------------------------------------------------------
// Name: ReadInLevelFileHeader
// Desc: Attempts to read in the header of a level file from the provided 
//       data stream. If successful then it will return true along with values
//       from the header. It will also leave the data stream positioned at the
//       start of the data section of the level file.
//------------------------------------------------------------------------------
bool Landscape::ReadInLevelFileHeader(ifstream *LevelFile)
{
    string CurLine;
    string Delimiter(Level::FILE_FIELD_DELIMITER);
    string::size_type DelimiterPos;
    int MagicNumber;
    int RawColour;

    // Check for the 'magic number'
    getline(*LevelFile, CurLine); if (!(*LevelFile)) return false;
    stringstream(CurLine) >> MagicNumber;
    if (MagicNumber != Level::FILE_MAGIC_NUMBER) return false;

    // Skip the level name
    getline(*LevelFile, CurLine); if (!(*LevelFile)) return false;

    // Read in the tileset used
    getline(*LevelFile, Tileset); if (!(*LevelFile)) return false;

    // Read in the level dimensions
    getline(*LevelFile, CurLine); if (!(*LevelFile)) return false;
    if ((DelimiterPos = CurLine.find(Delimiter, 1)) == string::npos) return false;
    stringstream(CurLine.substr(0, DelimiterPos)) >> WorldWidth;
    CurLine = CurLine.substr(DelimiterPos + Delimiter.length());

    if ((DelimiterPos = CurLine.find(Delimiter, 1)) == string::npos) return false;
    stringstream(CurLine.substr(0, DelimiterPos)) >> WorldHeight;
    CurLine = CurLine.substr(DelimiterPos + Delimiter.length());

    if ((DelimiterPos = CurLine.find(Delimiter, 1)) == string::npos) return false;
    stringstream(CurLine.substr(0, DelimiterPos)) >> LandHeight;
    stringstream(CurLine.substr(DelimiterPos + Delimiter.length())) >> 
        BackgroundHeight;

    // Read in the colours used for the sky
    getline(*LevelFile, CurLine); if (!(*LevelFile)) return false;
    if ((DelimiterPos = CurLine.find(Delimiter, 1)) == string::npos) return false;

    // Colours come in as an integer and are then broken up into RGB components
    stringstream(CurLine.substr(0, DelimiterPos)) >> RawColour;
    SkyColour0.r = (RawColour >> 16) & 0xFF;
    SkyColour0.g = (RawColour >> 8) & 0xFF;
    SkyColour0.b = (RawColour) & 0xFF;
    SkyColour0.a = SDL_ALPHA_OPAQUE;

    stringstream(CurLine.substr(DelimiterPos + Delimiter.length())) >> RawColour;
    SkyColour1.r = (RawColour >> 16) & 0xFF;
    SkyColour1.g = (RawColour >> 8) & 0xFF;
    SkyColour1.b = (RawColour) & 0xFF;
    SkyColour1.a = SDL_ALPHA_OPAQUE;

    return true;
}

//------------------------------------------------------------------------------
// Name: ReadInLandscapeData
// Desc: Reads in a value from a comma seperated line of variables and removes
//       that value from the line
//------------------------------------------------------------------------------
bool Landscape::ReadInLandscapeData(string *CurLine, uchar *Data)
{
    int Buffer;
    string Delimiter(Level::FILE_FIELD_DELIMITER);
    string::size_type DelimiterPos;

    assert(Data != NULL);

    // Cope with values at the end of lines
    if ((DelimiterPos = CurLine->find(Delimiter, 1)) == string::npos) {
        stringstream(*CurLine) >> Buffer;
            
    } else {
        stringstream(CurLine->substr(0, DelimiterPos)) >> Buffer;
        *CurLine = CurLine->substr(DelimiterPos + Delimiter.length());
    }
    *Data = (uchar)Buffer;
    // If Data != Buffer there has been an overflow
    return (*Data == Buffer);
}

//------------------------------------------------------------------------------
// Name: SetViewport
// Desc: Defines the viewport that is currently looking into the gameworld and
//       sets a corresponding clipping area on the game screen. The viewport 
//       must fit onto the game screen that has been created. After the 
//       viewport has been created it is moved to the top left corner of the
//       game world.
//------------------------------------------------------------------------------
void Landscape::SetViewport(Point2D ScreenPos, int Width, int Height)
{
    SDL_Rect ClippingRect = { ScreenPos.x, ScreenPos.y, Width, Height };

    // Test assumptions about object state
    assert(ScreenPos.x >= 0 && ScreenPos.x <= SDLI->GetScreen()->w);
    assert(ScreenPos.y >= 0 && ScreenPos.y <= SDLI->GetScreen()->h);
    assert(Width >= 0 && ScreenPos.x + Width <= SDLI->GetScreen()->w);
    assert(Height >= 0 && ScreenPos.y + Height <= SDLI->GetScreen()->h);

    ViewportPos = ScreenPos;
    ViewportWidth = Width; 
    ViewportHeight = Height;

    // Set bounds regarding where the viewport can be placed in the world
    MaxWorldPos.x = WorldWidth*TILE_WIDTH - ViewportWidth;
    if (MaxWorldPos.x < 0) MaxWorldPos.x = 0;
    MaxWorldPos.y = WorldHeight*TILE_HEIGHT - ViewportHeight;
    if (MaxWorldPos.y < 0) MaxWorldPos.y = 0;

    // Move viewport to a known position
    SetViewportWorldPos(Point2D(0, 0));
}

//------------------------------------------------------------------------------
// Name: EnableViewportClippingRect
// Desc: This routine sets the screens clipping area to the area of the screen 
//       occupied by the viewport. The current clipping rect of the screen
//       is stored ready for restoration later on.
//------------------------------------------------------------------------------
void Landscape::EnableViewportClippingRect()
{
    SDL_Rect ClippingRect = 
    { ViewportPos.x, ViewportPos.y, ViewportWidth, ViewportHeight };

    // Save the old clipping rectangle and then set the new one
    SDL_GetClipRect(SDLI->GetScreen(), &StoredClippingRect);
    SDL_SetClipRect(SDLI->GetScreen(), &ClippingRect);
}

//------------------------------------------------------------------------------
// Name: DisableViewportClippingRect
// Desc: Restores the screens clipping area, should be called after all items
//       in the viewport have been drawn.
//------------------------------------------------------------------------------
void Landscape::DisableViewportClippingRect()
{
    SDL_SetClipRect(SDLI->GetScreen(), &StoredClippingRect);
}

//------------------------------------------------------------------------------
// Name: SetViewportWorldPos
// Desc: Moves the viewport that looks into the game world to a given set of
//       coordinates. There are limits to where the viewport can be moved,
//       so this function returns false if the move is impossible or true
//       otherwise.
//------------------------------------------------------------------------------
bool Landscape::SetViewportWorldPos(Point2D Pos)
{
    if (Pos.x < 0 || Pos.x > MaxWorldPos.x 
        || Pos.y < 0 || Pos.y > MaxWorldPos.y) {
        return false;
    } else {
        CurWorldPos = Pos;
        return true;
    }
}

//------------------------------------------------------------------------------
// Name: CentreViewportAt
// Desc: Centres the viewport that looks into the game world on a given pair of
//       coordinates. If it is not possible to centre exactly on the 
//       coordinates then this function tries to get as close as possible.
//------------------------------------------------------------------------------
void Landscape::CentreViewportAt(Point2D Pos)
{
    int ViewportWidth_2 = ViewportWidth / 2;
    int ViewportHeight_2 = ViewportHeight / 2;

    if (Pos.x < ViewportWidth_2) {
        CurWorldPos.x = 0;
    } else if (Pos.x > MaxWorldPos.x + ViewportWidth_2) {
        CurWorldPos.x = MaxWorldPos.x;
    } else {
        CurWorldPos.x = Pos.x - ViewportWidth_2;
    }
    
    if (Pos.y < ViewportHeight_2) {
        CurWorldPos.y = 0;
    } else if (Pos.y > MaxWorldPos.y + ViewportHeight_2) {
        CurWorldPos.y = MaxWorldPos.y;
    } else {
        CurWorldPos.y = Pos.y - ViewportHeight_2;
    }
}

//------------------------------------------------------------------------------
// Name: MoveViewport
// Desc: Moves the viewport that looks onto the game world by given distances
//       on the x and y axis.
//------------------------------------------------------------------------------
void Landscape::MoveViewport(int dx, int dy)
{
    CurWorldPos.x += dx;
    CurWorldPos.y += dy;

    if (CurWorldPos.x < 0) {
        CurWorldPos.x = 0;
    } else if (CurWorldPos.x > MaxWorldPos.x) {
        CurWorldPos.x = MaxWorldPos.x;
    }

    if (CurWorldPos.y < 0) {
        CurWorldPos.y = 0;
    } else if (CurWorldPos.y > MaxWorldPos.y) {
        CurWorldPos.y = MaxWorldPos.y;
    }
}

//------------------------------------------------------------------------------
// Name: DrawViewportBackground
// Desc: This draws all of the background that can be seen in the viewport.
//       The background consists of the sky, any parallax layers that are to
//       be drawn and the background layer of the landscape. Once this has 
//       been drawn the entities can be drawn on top before the foreground is
//       drawn.
//------------------------------------------------------------------------------
bool Landscape::DrawViewportBackground(SDL_Surface *ScreenBuffer)
{
    int i;
    // Variables for drawing the sky
    float dr = (float)(SkyColour1.r - SkyColour0.r) / WorldPixelHeight;  // delta r, g, and b
    float dg = (float)(SkyColour1.g - SkyColour0.g) / WorldPixelHeight;
    float db = (float)(SkyColour1.b - SkyColour0.b) / WorldPixelHeight;
    float r = SkyColour0.r + dr*CurWorldPos.y;
    float g = SkyColour0.g + dg*CurWorldPos.y;
    float b = SkyColour0.b + db*CurWorldPos.y;
    SDL_Rect SkyLine = { ViewportPos.x, ViewportPos.y, ViewportWidth, 1 };
    int EndY;

    // Variables for drawing the parallax layers
    int NumPLsToDraw;   // The number of layers to draw
    Point2D PLPos[NUM_PARALLAX_LAYERS];

    // Variables for drawing the background
    int TargetStartY;
    SDL_Rect SourceRect = { CurWorldPos.x, 0, ViewportWidth, 0 }; 
    SDL_Rect TargetRect = { ViewportPos.x, ViewportPos.y, ViewportWidth, 0 };
    int BackgroundPixelHeight;

    assert(ScreenBuffer != NULL);

    // Work out how many Parallax Layers we should draw and where we should
    // draw them.
    if (ParallaxLayersLoaded < G->GetOption(Game::NUM_PARALLAX_LAYERS)) {
        NumPLsToDraw = ParallaxLayersLoaded;
    } else {
        NumPLsToDraw = G->GetOption(Game::NUM_PARALLAX_LAYERS);
    }
    for (i = 0; i < NumPLsToDraw; i++) {
        PLPos[i].x = (int)(CurWorldPos.x * PL_H_SPEED[i]);
        PLPos[i].y = (int)(ParallaxLayers[i].Height -
            (1.0f - PL_H_SPEED[i])*(MaxWorldPos.y - CurWorldPos.y)*PL_V_SPEED);
    }

    // Draw the sky, either down to the bottom of the viewport or the 
    // bottom of the highest parallax layer, whichever comes first.
    if (NumPLsToDraw > 0) {
    
        EndY = PLPos[NumPLsToDraw-1].y + ParallaxLayers[NumPLsToDraw-1].Image->h;

        if (CurWorldPos.y + ViewportHeight < EndY) {
            EndY = CurWorldPos.y + ViewportHeight;
        }
    } else {
        EndY = CurWorldPos.y + ViewportHeight;
    }
    for (i = CurWorldPos.y; i < EndY; i++) {

        SDL_FillRect(ScreenBuffer, &SkyLine,  
            SDL_MapRGB(ScreenBuffer->format, (uchar)r, (uchar)g, (uchar)b));

        SkyLine.y++;
        r += dr; g += dg; b += db;
    }

    // Now draw the parallax layers from back to front
    for (i = NumPLsToDraw-1; i >= 0; i--) {
        SDL_Rect PLSR;  // Source and target rects
        SDL_Rect PLTR;

        // Find the starting height for the layer in the viewport
        PLSR.y = 0;
        TargetStartY = PLPos[i].y - CurWorldPos.y;
        if (TargetStartY < 0) {
            PLSR.y = -TargetStartY;     // Find the point at the top of the viewport
            TargetStartY = 0;           // Cannot start from less than 0
        }
        if (TargetStartY < ViewportHeight) {

            PLTR.y = ViewportPos.y + TargetStartY;
            if (ViewportHeight - TargetStartY < ParallaxLayers[i].Image->h) {
                PLSR.h = PLTR.h = ViewportHeight - TargetStartY;
            } else {
                PLSR.h = PLTR.h = ParallaxLayers[i].Image->h;
            }

            // Draw chunks of the parallax layer image until we have a strip that runs
            // across the screen.
            // Work out where the first chunk comes from
            PLSR.x = PLPos[i].x % ParallaxLayers[i].Image->w;
            PLTR.x = 0;

            while (PLTR.x < ViewportWidth) {
                if (ParallaxLayers[i].Image->w - PLSR.x > ViewportWidth - PLTR.x) {
                    PLSR.w = PLTR.w = ViewportWidth - PLTR.x;
                } else {
                    PLSR.w = PLTR.w = ParallaxLayers[i].Image->w - PLSR.x;
                }

                if (SDL_BlitSurface(ParallaxLayers[i].Image, &PLSR, ScreenBuffer, &PLTR) == -2) {
                    fprintf(stderr, "Parallax Layer %i lost\n", i);
                }
                PLSR.x = 0;
                PLTR.x += PLTR.w;
            }
        }
    }
    
    // Now draw the background
    // Find the starting height for the background in the viewport
    TargetStartY = (WorldHeight - LandHeight)*TILE_HEIGHT - CurWorldPos.y;
    if (TargetStartY < 0) {
        SourceRect.y = -TargetStartY;   // Find the land at the top of the viewport
        TargetStartY = 0;               // Cannot start from less than 0
    }
    if (TargetStartY < ViewportHeight) {
        BackgroundPixelHeight = BackgroundHeight*TILE_HEIGHT;

        TargetRect.y = ViewportPos.y + TargetStartY;

        if (ViewportHeight - TargetStartY < BackgroundPixelHeight) {
            SourceRect.h = TargetRect.h = ViewportHeight - TargetStartY;
        } else {
            SourceRect.h = TargetRect.h = BackgroundPixelHeight;
        }

        SDL_BlitSurface(BackgroundImageRLE, &SourceRect, ScreenBuffer, &TargetRect);
    }

    return true;
}

//------------------------------------------------------------------------------
// Name: DrawViewportForeground
// Desc: Draws the foreground of the gameworld onto the supplied 'ScreenBuffer'
//------------------------------------------------------------------------------
bool Landscape::DrawViewportForeground(SDL_Surface *ScreenBuffer)
{
    int TargetStartY;
    SDL_Rect SourceRect = { CurWorldPos.x, 0, ViewportWidth, 0 }; 
    SDL_Rect TargetRect = { ViewportPos.x, ViewportPos.y, ViewportWidth, 0 };

    assert(ScreenBuffer != NULL);

    // Find the starting height for the land in the viewport
    TargetStartY = (WorldHeight - LandHeight)*TILE_HEIGHT - CurWorldPos.y;
    if (TargetStartY < 0) {
        SourceRect.y = -TargetStartY;   // Find the land at the top of the viewport
        TargetStartY = 0;               // Cannot start from less than 0
    }
    if (TargetStartY < ViewportHeight) {
        TargetRect.y = ViewportPos.y + TargetStartY;
        SourceRect.h = TargetRect.h = ViewportHeight - TargetStartY;

        SDL_BlitSurface(LandImageRLE, &SourceRect, ScreenBuffer, &TargetRect);
    }

    return true;
}

//------------------------------------------------------------------------------
// Name: HasCollided
// Desc: This function carries out a test to see whether or not a given
//       frame from the sprites object positioned at the point 'p' has 
//       collided with the landscape. The basic test for collision is to ask
//       whether any non transparent pixels from the sprite overlap with any
//       non transparent pixels in the foreground image (LandImage) of the
//       landscape. This is of course speeded up by using test to reject or
//       accept trivial cases.
//
//       Also, a routine calling this functions can request that the function
//       'UseContours'. Here the algorithm checks first against the contours
//       of the landscape before moving onto a pixel by pixel test. This can
//       by used so that some entities avoid parts of the landscape which are
//       above ground (such as bombs not colliding with trees for example)
//       whereas other entities do (planes crash into trees).
//------------------------------------------------------------------------------
bool Landscape::HasCollided(uint Frame, Point2D p, bool UseContours)
{
    int FrameTop = p.y;
    int FrameLeft = p.x;
    int FrameWidth, FrameHeight;
    int FrameBottom, FrameRight;
    int ROITop, ROILeft, ROIBottom, ROIRight;   // Region of interest
    int LeftBorder = 0;
    int RightBorder = WorldWidth*TILE_WIDTH;
    int BottomBorder = WorldHeight*TILE_HEIGHT;
    int TopBorder = BottomBorder - LandHeight*TILE_HEIGHT;
    SDL_Rect FrameArea, ImageArea;
    
    SPR->GetFrameDims(Frame, &FrameWidth, &FrameHeight);
    FrameBottom = FrameTop + FrameHeight;
    FrameRight = FrameLeft + FrameWidth;

    // Reject trivial cases
    if (FrameBottom < TopBorder || FrameTop > BottomBorder) return false;

    // Is the sprite on screen? If so then the check needs to be detailed
    if (FrameRight > LeftBorder && FrameLeft < RightBorder) {
        if (UseContours) {
            // When checking against the contours, we look at all of the contours
            // that the frame crosses. Working out what would be the highest possible
            // piece of ground that the frame would encounter. If the frame bottom
            // is below this point then the test continues.
            int CurTileX;
            int CurContourHeight;
            int StartTileX = FrameLeft / TILE_WIDTH;
            int EndTileX = FrameRight / TILE_WIDTH;
            int MaxHeight = 0;
            int FrameBottomHeight = WorldPixelHeight - FrameBottom;

            StartTileX = FrameLeft / TILE_WIDTH;
            EndTileX = FrameRight / TILE_WIDTH;
            if (StartTileX < 0) StartTileX = 0;
            if ((uint)EndTileX >= WorldWidth) EndTileX = WorldWidth - 1;

            for (CurTileX = StartTileX; CurTileX <= EndTileX; CurTileX++) {
                CurContourHeight = Contours[CurTileX].BasicPixelHeight;
                if (Contours[CurTileX].SlopeType != FLAT) 
                    CurContourHeight += TILE_HEIGHT;

                if (CurContourHeight > MaxHeight) MaxHeight = CurContourHeight;
            }
            // An extra tiles height is used to make sure that we catch even 
            // small fast moving objects such as bombs.
            if (FrameBottomHeight >= MaxHeight+TILE_HEIGHT) return false;
        }
        // Calculate the region of interest, this is the exact area of the
        // sprite that is on screen
        ROITop = (FrameTop < TopBorder ? TopBorder : FrameTop);
        ROIBottom = (FrameBottom > BottomBorder ? BottomBorder : FrameBottom);
        ROILeft = (FrameLeft < LeftBorder ? LeftBorder : FrameLeft);
        ROIRight = (FrameRight > RightBorder ? RightBorder : FrameRight);

        FrameArea.x = ROILeft - FrameLeft;
        FrameArea.y = ROITop - FrameTop;
        FrameArea.w = ROIRight - ROILeft;
        FrameArea.h = ROIBottom - ROITop;

        ImageArea.x = ROILeft;
        ImageArea.y = ROITop - TopBorder;
        ImageArea.w = FrameArea.w;
        ImageArea.h = FrameArea.h;

        // Check for a collision
        if (SPR->FindCollision(Frame, FrameArea, LandImage, ImageArea))
            return true;
    }

    // Now do a rough check if all of the sprite is offscreen to the 
    // right or left
    if (FrameLeft < LeftBorder) {
        if (FrameBottom > BottomBorder - Contours[0].BasicPixelHeight)
            return true;
    }
    if (FrameRight > RightBorder) {
        if (FrameBottom > BottomBorder 
            - Contours[WorldWidth-1].BasicPixelHeight)
            return true;
    }

    return false;
}

//------------------------------------------------------------------------------
// Name: DrawLandscape
// Desc: Here, an image of the landscape is drawn onto an SDL surface using
//       the tileset loaded in when the object was initialised and a
//       LANDSCAPE_DATA struct. If anything goes wrong then NULL is returned,
//       otherwise the function returns a pointer to the newly created SDL
//       surface.
//------------------------------------------------------------------------------
void Landscape::DrawLandscape()
{
    uint i, j;
    int Offset;
    SDL_Surface *TilesetImage;
    SDL_Rect TargetRect = { 0, 0, TILE_WIDTH, TILE_HEIGHT };
    SDL_PixelFormat *SurfaceFmt;
    string TilesetFileName;
    
    TilesetFileName = string(CAM_TILESETS_DIR) + "/" + string(Tileset);
    TilesetImage = LoadTileset(TilesetFileName.c_str());
    SurfaceFmt = TilesetImage->format;

    LandImage = NULL;
    LandImageRLE = NULL;
    BackgroundImageRLE = NULL;
    // Create an SDL surface to hold the image
    LandImage = SDL_CreateRGBSurface(SDL_SWSURFACE, WorldWidth*TILE_WIDTH,
        LandHeight*TILE_HEIGHT, CAM_INTERNAL_SURFACE_BPP, 
        SurfaceFmt->Rmask, SurfaceFmt->Gmask, SurfaceFmt->Bmask, SurfaceFmt->Amask);
    if (LandImage == NULL) SDE_SDLError(__FILE__, __LINE__);

    CU_Dbg_Printf(NULL, "%f megabytes used to store landscape\n",
    CU_DBG    (float)(WorldWidth*TILE_WIDTH*LandHeight
    CU_DBG        *TILE_HEIGHT*CAM_INTERNAL_SURFACE_BPP/8)/(1024*1024));

    // Enable alpha blending
    SDL_SetAlpha(LandImage, SDL_SRCALPHA, SDL_ALPHA_TRANSPARENT);

    // Fill surface with a transparent colour
    SDL_FillRect(LandImage, NULL, SDL_MapRGBA(LandImage->format, 
        TRANSPARENT_COLOUR.r, TRANSPARENT_COLOUR.g, 
        TRANSPARENT_COLOUR.b, TRANSPARENT_COLOUR.a));

    // Now draw the land
    TargetRect.x = 0;
    for (i = 0; i < WorldWidth; i++) {
        TargetRect.y = 0;
        Offset = i;
        for (j = 0; j < LandHeight; j++) {
            SDLI->SW_BlitSurface(TilesetImage, &TileRects[LandscapeData[Offset]],
                LandImage, &TargetRect);

            TargetRect.y+=TILE_HEIGHT;
            Offset += WorldWidth;
        }
        TargetRect.x+=TILE_WIDTH;
    }

    // Create an SDL surface to hold the background image
    BackgroundImageRLE = SDL_CreateRGBSurface(SDL_SWSURFACE, WorldWidth*TILE_WIDTH,
        BackgroundHeight*TILE_HEIGHT, CAM_INTERNAL_SURFACE_BPP, 
        SurfaceFmt->Rmask, SurfaceFmt->Gmask, SurfaceFmt->Bmask, SurfaceFmt->Amask);
    if (BackgroundImageRLE == NULL) SDE_SDLError(__FILE__, __LINE__);

    CU_Dbg_Printf(NULL, "%f megabytes used to store background\n",
    CU_DBG    (float)(WorldWidth*TILE_WIDTH*BackgroundHeight
    CU_DBG        *TILE_HEIGHT*CAM_INTERNAL_SURFACE_BPP/8)/(1024*1024));

    // Enable alpha blending
    SDL_SetAlpha(BackgroundImageRLE, SDL_SRCALPHA | SDL_RLEACCEL, SDL_ALPHA_TRANSPARENT);

    // Fill surface with a transparent colour
    SDL_FillRect(BackgroundImageRLE, NULL, SDL_MapRGBA(BackgroundImageRLE->format, 
        TRANSPARENT_COLOUR.r, TRANSPARENT_COLOUR.g, 
        TRANSPARENT_COLOUR.b, TRANSPARENT_COLOUR.a));

    // Now draw the background
    TargetRect.x = 0;
    for (i = 0; i < WorldWidth; i++) {
        TargetRect.y = 0;
        Offset = i;
        for (j = 0; j < BackgroundHeight; j++) {
            SDLI->SW_BlitSurface(TilesetImage, &TileRects[BackgroundData[Offset]],
                BackgroundImageRLE, &TargetRect);

            TargetRect.y+=TILE_HEIGHT;
            Offset += WorldWidth;
        }
        TargetRect.x+=TILE_WIDTH;
    }
    
    // Create a compressed copy of the landscape
    LandImageRLE = SDL_ConvertSurface(LandImage, LandImage->format, 
        SDL_SWSURFACE | SDL_SRCALPHA  | SDL_RLEACCEL);
    SDL_SetAlpha(LandImageRLE, SDL_SRCALPHA | SDL_RLEACCEL, SDL_ALPHA_TRANSPARENT);
    SDL_FreeSurface(TilesetImage);
}

//------------------------------------------------------------------------------
// Name: NormaliseContours
// Desc: This routine goes through the contours of the landscape and makes 
//       sure that they are acceptable. It first makes sure that each contour
//       is in range (i.e. between 0 and LandHeight), it then makes sure
//       that there is only at most a stepwise change between each contour.
//       Finally, it loops through the contours and classifies the slope at 
//       each point.
//------------------------------------------------------------------------------
void Landscape::NormaliseContours()
{
    uint i;
    int LastContour;
    int MaxHeight;

    MaxHeight = LandHeight - 1;

    // Normalise the first contour
    if (Contours[0].Height < 0) {
        Contours[0].Height = 0;
    } else if (Contours[0].Height > MaxHeight) {
        Contours[0].Height = MaxHeight;
    }
    Contours[0].BasicPixelHeight = Contours[0].Height * TILE_HEIGHT;
    Contours[0].SlopeType = FLAT;

    // Loop through all of the columns, finding out their height and setting
    // the slope to flat
    for (i = 1; i < WorldWidth; i++) {
        LastContour = i-1;
        if (Contours[i].Height > Contours[LastContour].Height) {
            if (Contours[LastContour].Height == MaxHeight) {
                Contours[i].Height = MaxHeight;
            } else {
                Contours[i].Height = Contours[LastContour].Height + 1;
            }
        } else if (Contours[i].Height < Contours[LastContour].Height) {
            if (Contours[LastContour].Height == 0) {
                Contours[i].Height = 0;
            } else {
                Contours[i].Height = Contours[LastContour].Height - 1;
            }
        }
        Contours[i].BasicPixelHeight = Contours[i].Height * TILE_HEIGHT;
        Contours[i].SlopeType = FLAT;
    }

    // Now determine the slope types
    if (Contours[0].Height > Contours[1].Height) {
        Contours[0].SlopeType = SLOPE_DOWN;
    } else if (Contours[0].Height < Contours[1].Height) {
        Contours[0].SlopeType = SLOPE_UP;
    }
    for (i = 1; i < WorldWidth - 1; i++) {
        if (Contours[i-1].Height > Contours[i].Height) {
            Contours[i].SlopeType = SLOPE_DOWN;
        } else if (Contours[i+1].Height > Contours[i].Height) {
            Contours[i].SlopeType = SLOPE_UP;
        }
    }
    if (Contours[WorldWidth - 1].Height 
        > Contours[WorldWidth - 2].Height) {
        Contours[WorldWidth - 1].SlopeType = SLOPE_UP;
    } else if (Contours[WorldWidth - 1].Height 
        < Contours[WorldWidth - 2].Height) {
        Contours[WorldWidth - 1].SlopeType = SLOPE_DOWN;
    }
}

//------------------------------------------------------------------------------
// Name: CalculatePixelHeight
// Desc: Given the x value of a pixel in the world, this function will return
//       the corresponding y value so that the pixel is on top of the ground
//       at this point. During this calculation, the tiles used to make up the 
//       landscape are idealised to be either flat or sloping at 45 degrees.
//------------------------------------------------------------------------------
uint Landscape::CalculatePixelHeight(int PixelX)
{
    uint Result;
    int TileX = PixelX / TILE_WIDTH;
    uint Offset = PixelX - (TileX * TILE_WIDTH);
    uint BaseY;

    if (TileX < 0) {
        TileX = 0;
    } else if ((uint)TileX >= WorldWidth) {
        TileX = WorldWidth-1;
    }

    // Calculate the starting point that we'll use for y
    // This would be the pixel height if the ground was FLAT
    BaseY = WorldPixelHeight - Contours[TileX].BasicPixelHeight;

    // For the sloped tiles, we are assuming that TILE_WIDTH == TILE_HEIGHT
    // to speed up the calculation. This would have to be altered if the tile
    // engine was later generalised.
    assert(TILE_WIDTH == TILE_HEIGHT);
    if (Contours[TileX].SlopeType == FLAT) {
        Result = BaseY;
    } else if (Contours[TileX].SlopeType == SLOPE_UP) {
        Result = BaseY - Offset;
    } else {    // SLOPE_DOWN
        Result = BaseY - (TILE_HEIGHT - Offset);
    }

    return Result;
}

//------------------------------------------------------------------------------
// Name: LineOfSightExists
// Desc: This function will find out if a line of sight exists between two 
//       points on the landscape. This is done by checking the line at intervals
//       of TILE_WIDTH to see if it has collided with the landscape.
//
// Note: Like the CalculatePixelHeight function, this function uses the 
//       assumption that TILE_WIDTH == TILE_HEIGHT to speed up the code. This 
//       would have to be altered if the tile engine was later generalised.
//------------------------------------------------------------------------------
bool Landscape::LineOfSightExists(Point2D StartPos, Point2D EndPos)
{
    const int HEIGHT_ABOVE_GROUND = 10;
    // Variables used to tell us where we are
    Point2D TempPos;
    float CurPosX, CurPosY;
    int StartTileX, EndTileX, CurTileX;
    // Variables that measure how x and y change along the firing line
    const float dx_Tile = (float)TILE_WIDTH;
    float dy_Pixel, dy_Tile, dy_Offset;
    // The exact height of the ground at a given point
    int TestY;
    // Offsets from the top left corner of the current tile to the current point
    int OffsetX;    
    int OffsetY[NUM_SLOPE_TYPES];

    if (StartPos.x > EndPos.x) {
        // Reverse the start and end point
        TempPos = StartPos;
        StartPos = EndPos;
        EndPos = TempPos;
    } 
    if (StartPos.x < 0 || EndPos.x >= WorldPixelWidth) return false;

    StartTileX = StartPos.x / TILE_WIDTH;
    EndTileX = EndPos.x / TILE_WIDTH;

    // Work out how y will change from pixel to pixel and tile to tile
    dy_Pixel = (float)(EndPos.y - StartPos.y) / (EndPos.x - StartPos.x);
    dy_Tile = dy_Pixel * TILE_WIDTH;

    // Work out the offsets needed to calculate the exact slope height at
    // each point along the line
    OffsetX = StartPos.x - StartTileX*TILE_WIDTH;
    OffsetY[FLAT] = 0;
    OffsetY[SLOPE_UP] = OffsetX;
    OffsetY[SLOPE_DOWN] = TILE_HEIGHT - OffsetX;
    // Work out how y will change across OffsetX
    dy_Offset = dy_Pixel * OffsetX;

    CurPosX = (float)StartPos.x;
    CurPosY = (float)StartPos.y;
    // Note: Here 'CurTileX' is incremented upto and including 'EndTileX'.
    //       Now, the value of 'OffsetX' for 'EndPos.x' will probably not be the
    //       same as the value of 'OffsetX' for 'StartPos.x'. However it has been
    //       assumed (although not rigorously) that this does not matter for
    //       all practical situations that will occur in the game.
    for (CurTileX = StartTileX; CurTileX <= EndTileX; CurTileX++) {
        SLOPE_TYPE CurSlope = Contours[CurTileX].SlopeType;

        CurPosX += dx_Tile;
        CurPosY += dy_Tile;

        TestY = (WorldPixelHeight - Contours[CurTileX].BasicPixelHeight)
            - OffsetY[CurSlope];

        // Test first against the height of the ground at CurPos
        if (CurPosY > TestY) goto CollisionWithGround;
        // Then test if the line came in through an obstructed edge
        // This is equivilant to testing against the ground height at the beginning
        // of the tile.
        switch (CurSlope) {
            case FLAT:
                // The bottom edge is obstructed
                if (CurPosY - dy_Offset > TestY) goto CollisionWithGround;
                break;
            case SLOPE_DOWN:
                // Both the left and bottom edges are obstructed
                if (CurPosY - dy_Offset > 
                    (WorldPixelHeight - Contours[CurTileX].BasicPixelHeight) - TILE_HEIGHT) 
                    goto CollisionWithGround;
        }
    }
    // The route from StartPos to EndPos has been checked and no 
    // obstructions were found.
    return true;

CollisionWithGround:    // Collsion with ground so no line of sight
    return false;
}

Landscape::SLOPE_TYPE Landscape::GetPixelSlopeType(int PixelX)
{
    int TileX = PixelX / TILE_WIDTH;

    if (TileX < 0) {
        TileX = 0;
    } else if ((uint)TileX >= WorldWidth) {
        TileX = WorldWidth-1;
    }

    return Contours[TileX].SlopeType;
}
