/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// Name: Level.cpp
// Desc: A level is loaded in from a '.lev' file and is the thing that a player
//       actually plays. A level contains a 2D landscape for the player to
//       explore along with entites (planes, soldiers, bombs etc) that can be
//       controlled and/or interacted with.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <list>
#include <string>
#include <fstream> 
#include <sstream>
#include <assert.h>
#include "globals.h"
#include "CamelUtility.h"
#include "Landscape.h"
#include "Entity.h"
#include "Platoon.h"
#include "TankSquad.h"
#include "Factory.h"
#include "PlaneSpawnPoint.h"
#include "PlatoonSpawnPoint.h"
#include "TankSquadSpawnPoint.h"
#include "EnemyPlane.h"
#include "Flag.h"
#include "Level.h"

using namespace std;

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------
const int Level::FILE_MAGIC_NUMBER = 0xC0FFEE;
const char *Level::FILE_FIELD_DELIMITER = ",";
const char *Level::FILE_SECTION_DELIMITER = "###";

// Constants relating to entities whose position can be determined by the level
// definition script found at the end of '.lev' files.
const Entity::ENT_TYPE Level::PLACEABLE_ENTITIES[NUM_PLACEABLE_ENTITIES] = {
    Entity::PLATOON,
    Entity::TANKSQUAD,
    Entity::FACTORY,
    Entity::FACTORY,
    Entity::PLANESPAWNPOINT,
    Entity::PLATOONSPAWNPOINT,
    Entity::TANKSQUADSPAWNPOINT,
    Entity::FLAG
};

const char *Level::PLACEABLE_ENTITY_STRINGS[NUM_PLACEABLE_ENTITIES] = {
    "PLATOON",
    "TANKSQUAD",
    "PLATOONFACTORY",
    "TANKSQUADFACTORY",
    "PLANESPAWNPOINT",
    "PLATOONSPAWNPOINT",
    "TANKSQUADSPAWNPOINT",
    "FLAG"
};

const char *Level::LDS_DELIMITER = ":";

// These constants are related to how the game plays. They may need to be
// modified and it might be a good idea to move them out to the LD script.
const int Level::NUM_SOLDIERS_IN_PLATOON = 50;
const int Level::NUM_TANKS_IN_TANKSQUAD = 3;

const int Level::STARTING_LIVES = 3;
const uint Level::STARTING_SCORE = 0;
const uint Level::FIRST_NEW_LIFE_SCORE = 5000;

//------------------------------------------------------------------------------
// Level Implementation
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Name: Level
// Desc: Loads in a level from a given '.lev' file. This includes the landscape
//       and the entities that make up the level.
//------------------------------------------------------------------------------
Level::Level(const char *Filename)
{
    assert(Filename != NULL);

    IntroScreen = NULL;
    Music = NULL;

    // Set the 'unique entity' pointers to NULL
    PlayerPlaneSpawnPoint = NULL;
    Flags[Entity::GOOD] = NULL;
    Flags[Entity::EVIL] = NULL;

    ReadInLevelDetails(Filename);
    // Load in the landscape
    LS = new Landscape(Filename);
    // Load in the rest of the level
    ReadInLevelDefinitionScript(Filename);
    // Add in the entities that were created by the LD script
    AddAllNewEntities();

    // Check that all required entities were created
    if (PlayerPlaneSpawnPoint == NULL 
        || Flags[Entity::GOOD] == NULL || Flags[Entity::EVIL] == NULL) {
        CU_ShowMessage("Error", "The level file provided is invalid as it does not "
            "specify all of the entities needed for a complete");
        exit(-1);
    }

    // Set the initial state of the level
    PlayerLives = STARTING_LIVES;
    PlayerScore = STARTING_SCORE;
    OldPlayerScore = STARTING_SCORE;
    ScoreNeededForNewLife = FIRST_NEW_LIFE_SCORE;
    GameOver = false;
    PlayerVictory = false;
}

//------------------------------------------------------------------------------
// Name: ~Level
// Desc: Closes down and cleans up after the level
//------------------------------------------------------------------------------
Level::~Level() 
{
    int i;
    list<Entity *>::iterator Iter;

    // Delete all entities
    for (i = 0; i < Entity::NUM_ENT_TYPES; i++) {
        for (Iter = GameEntities[i].begin(); Iter != GameEntities[i].end(); Iter++) {
            delete *Iter;
        }
        GameEntities[i].clear();
    }
    for (Iter = NewEntities.begin(); Iter != NewEntities.end(); Iter++) {
        delete *Iter;
    }
    NewEntities.clear();

    // Delete the landscape
    delete LS;

    // Free the introscreen and/or music that may have been used by the level.
    if (IntroScreen != NULL) SDL_FreeSurface(IntroScreen);
    if (Music != NULL) Mix_FreeMusic(Music);
}

//------------------------------------------------------------------------------
// Name: ReadInLevelDetails
// Desc: This routine reads through a supplied level file and extracts its 
//       name. It also loads in a piece of music and an IntroScreen for the
//       level if they are specified in the file.
//------------------------------------------------------------------------------
void Level::ReadInLevelDetails(const char *Filename)
{
    ifstream LevelFile(Filename, ios::in);
    string CurLine;
    string ResourceFile;
    int MagicNumber;

    // Deal with the trivial case where the file doesn't exist
    if (!LevelFile) goto InvalidLevelFile;

    // Check for the 'magic number'
    getline(LevelFile, CurLine); if (!LevelFile) goto InvalidLevelFile;
    stringstream(CurLine) >> MagicNumber;
    if (MagicNumber != Level::FILE_MAGIC_NUMBER) goto InvalidLevelFile;

    // Read in the level's name, and truncate it if needed
    getline(LevelFile, Name); if (!LevelFile) goto InvalidLevelFile;
    if (Name.length() > MAX_LEVEL_NAME_LENGTH) 
        Name = Name.substr(0, MAX_LEVEL_NAME_LENGTH);

    // Go to the IntroScreen section and read in the IntroScreen if it exists
	while (LevelFile && CurLine != Level::FILE_SECTION_DELIMITER)
	    getline(LevelFile, CurLine);
	if (!LevelFile) goto InvalidLevelFile;

    getline(LevelFile, CurLine);
    ResourceFile = string(CAM_INTROS_DIR) + "/" + CurLine;
    IntroScreen = IMG_Load(ResourceFile.c_str());
    if (IntroScreen != NULL) {
        SDL_SetAlpha(IntroScreen, SDL_SRCALPHA, SDL_ALPHA_TRANSPARENT);
    } else {
        CU_Dbg_Printf(NULL, "No introscreen found\n");
    }

    // Go to the Music section and read in the music for the level if it exists
	while (LevelFile && CurLine != Level::FILE_SECTION_DELIMITER)
	    getline(LevelFile, CurLine);
	if (!LevelFile) goto InvalidLevelFile;

    getline(LevelFile, CurLine);
    ResourceFile = string(CAM_MUSIC_DIR) + "/" + CurLine;
    Music = Mix_LoadMUS(ResourceFile.c_str());
    if (Music == NULL) {
        CU_Dbg_Printf(NULL, "No music found\n");
    }

    return;
InvalidLevelFile:
    CU_ShowMessage("Error", "Unable to read in the level file %s. Please ensure "
        "that the file is present and that its format is correct.", Filename);
    exit(1);
}

//------------------------------------------------------------------------------
// Name: ReadInLevelDefinitionScript
// Desc: Reads in the level definition script (if it exists) from the supplied
//       '.lev' file. At the moment, level definition files simply define
//       the position of various 'placeable' (yes I know that's not a word)
//       entities in the game world. 
//
//       The format for an entry in the level definition is as follows
//
//          ENTITY_NAME ': ' X_POS ',' ENTITY_SIDE
//
//       So a valid entry that defines the position of a factory on the players
//       side is given as follows
//
//          FACTORY: 400,GOOD
//
// Note: Although NEUTRAL is a valid alignment that is used at some points in 
//       the game, it is not at the moment allowed in the level definition
//       script. This is because a lot of entities don't work with a NEUTRAL
//       alignment or become rather judgemental and assume that everything
//       that is not GOOD is therefore EVIL. ;)
//------------------------------------------------------------------------------
void Level::ReadInLevelDefinitionScript(const char *Filename)
{
    int i;
    ifstream LevelFile(Filename, ios::in);
    string LDS_Delimiter(Level::LDS_DELIMITER);
    string Field_Delimiter(Level::FILE_FIELD_DELIMITER);
    string::size_type DelimiterPos;
    string CurLine;
    int MagicNumber;
    Entity *NewEntity;

    // Data found on a line of the level definition script 
    string strEntityType;
    string strEntityX;
    string strEntitySide;
    Entity::ENT_TYPE EntityType;
    int EntityX;
    Entity::ENT_SIDE EntitySide;
    EnemyPlane::WAYPOINT PatrolPoints[EnemyPlane::MAX_NUM_PATROL_POINTS];

    // Deal with the trivial case where the file doesn't exist
    if (!LevelFile) goto InvalidLevelFile;

    // Check for the 'magic number'
    getline(LevelFile, CurLine); if (!LevelFile) goto InvalidLevelFile;
    stringstream(CurLine) >> MagicNumber;
    if (MagicNumber != Level::FILE_MAGIC_NUMBER) goto InvalidLevelFile;

    // Now read through the file looking for the level definition script.
    // The script is in the seventh section
    for (i = 0; i < 7; i++) {
        do {
            getline(LevelFile, CurLine);
        } while (LevelFile && CurLine != Level::FILE_SECTION_DELIMITER);
        if (!LevelFile) break;
    }

    // read in the level definition script if it has been found
    if (CurLine == Level::FILE_SECTION_DELIMITER) {
        getline(LevelFile, CurLine);
        while (LevelFile) {
            // Extract the data from the current line if invalid data is found
            // or it cannot be read, simply skip to the next line
            DelimiterPos = CurLine.find(LDS_Delimiter);
            if (DelimiterPos == 0 || DelimiterPos == string::npos) 
                goto GetNextLine;       // No valid entry on this line

            strEntityType = CurLine.substr(0, DelimiterPos);
            CurLine = CurLine.substr(DelimiterPos + LDS_Delimiter.length());
            DelimiterPos = CurLine.find(Field_Delimiter);
            if (DelimiterPos == 0 || DelimiterPos == string::npos) 
                goto GetNextLine;       // No valid entry on this line

            strEntityX = CurLine.substr(0, DelimiterPos);
            strEntitySide = CurLine.substr(DelimiterPos + Field_Delimiter.length());

            // Translate the data from strings to the internal representations
            strEntityType = CU_TrimString(strEntityType);
            strEntityX = CU_TrimString(strEntityX);
            strEntitySide = CU_TrimString(strEntitySide);
            EntityType = Entity::NO_TYPE;
            for (i = 0; i < NUM_PLACEABLE_ENTITIES; i++) {
                if (strEntityType == PLACEABLE_ENTITY_STRINGS[i]) {
                    EntityType = PLACEABLE_ENTITIES[i];
                    break;
                }
            }
            stringstream(strEntityX) >> EntityX;
            EntitySide = Entity::NEUTRAL;
            for (i = 0; i < Entity::NUM_SIDES; i++) {
                if (strEntitySide == Entity::ENT_SIDE_STRINGS[i]) {
                    EntitySide = (Entity::ENT_SIDE)i;
                    break;
                }
            }
            if (EntitySide == Entity::NEUTRAL) goto GetNextLine;    // NEUTRAL not allowed

            // Create the corresponding entity
            switch (EntityType) {
                case Entity::PLATOON:
                    NewEntity = new Platoon(this, NUM_SOLDIERS_IN_PLATOON, EntityX, EntitySide);
                    break;
                case Entity::TANKSQUAD:
                    NewEntity = new TankSquad(this, NUM_TANKS_IN_TANKSQUAD, EntityX, EntitySide);
                    break;
                case Entity::FACTORY:
                    if (strEntityType == string("PLATOONFACTORY")) {
                        NewEntity = new Factory(this, EntityX, Factory::PLATOON_FACTORY, EntitySide);
                    } else {
                        NewEntity = new Factory(this, EntityX, Factory::TANKSQUAD_FACTORY, EntitySide);
                    }
                    break;
                case Entity::PLANESPAWNPOINT:
                    if (EntitySide == Entity::GOOD) {
                        if (PlayerPlaneSpawnPoint != NULL) {
                            goto GetNextLine;   // The player can only have one plane!
                        } else {
                            NewEntity = new PlaneSpawnPoint(this, EntityX, EntitySide, NULL);
                            PlayerPlaneSpawnPoint = (PlaneSpawnPoint *)NewEntity;
                        }
                    } else {
                        if (!ReadInPatrolPoints(&LevelFile, PatrolPoints)) {
                            goto GetNextLine;   // No patrol points found
                        }
                            
                        NewEntity = new PlaneSpawnPoint(this, EntityX, 
                            EntitySide, PatrolPoints);
                    }
                    break;
                case Entity::PLATOONSPAWNPOINT:
                    NewEntity = new PlatoonSpawnPoint(this, EntityX, EntitySide);
                    break;
                case Entity::TANKSQUADSPAWNPOINT:
                    NewEntity = new TankSquadSpawnPoint(this, EntityX, EntitySide);
                    break;
                case Entity::FLAG:
                    if (Flags[EntitySide] != NULL) {
                        goto GetNextLine;   // Can only have one flag per side
                    } else {
                        NewEntity = new Flag(this, EntityX, EntitySide);
                        Flags[EntitySide] = (Flag *)NewEntity;
                    }
                    break;
                default:
                    goto GetNextLine;   // Invalid entity 
            }
            AddNewEntity(NewEntity);
            
GetNextLine:
            getline(LevelFile, CurLine);       // Get the next line
        }
    }
    LevelFile.close();

    return;

InvalidLevelFile:
    CU_ShowMessage("Error", "Unable to read in the level file %s. Please ensure "
        "that the file is present and that its format is correct.", Filename);
    exit(1);
}

//------------------------------------------------------------------------------
// Name: ReadInPatrolPoints
// Desc: An EVIL plane spawn point comes with a set of patrol points written
//       directly below the PLANESPAWNPOINT declaration. This function reads
//       in these points and returns 'true' or 'false' if no patrol points were
//       found. There can be up to EnemyPlane::MAX_NUM_PATROL_POINTS patrol 
//       points and the format of a patrol point is
//
//              PointX, PointY, NextPoint
//
//       When working out the value of NextPoint it must be noted that the
//       numbering of points starts from 0. An example of a patrol route
//       to make a plane take off and fly around the game world is as follows
//
//              1800, 500, 1
//              1200, 500, 2
//              100, 500, 3
//              100, 100, 4
//              2000, 500, 2
//
//------------------------------------------------------------------------------
bool Level::ReadInPatrolPoints(ifstream *LevelFile, 
    EnemyPlane::WAYPOINT PatrolPoints[EnemyPlane::MAX_NUM_PATROL_POINTS])
{
    int PointsFound = 0;
    int LastLinePos;
    string Field_Delimiter(Level::FILE_FIELD_DELIMITER);
    string::size_type DelimiterPos;
    string CurLine;
    string strPointX;
    string strPointY;
    string strNextPoint;

    assert(LevelFile != NULL && PatrolPoints != NULL);

    LastLinePos = LevelFile->tellg();
    getline(*LevelFile, CurLine);
    while (*LevelFile && PointsFound < EnemyPlane::MAX_NUM_PATROL_POINTS) {
        DelimiterPos = CurLine.find(Field_Delimiter);
        if (DelimiterPos == 0 || DelimiterPos == string::npos) 
            goto FinishedReadingPoints;     // No valid entry on this line

        strPointX = CurLine.substr(0, DelimiterPos);
        CurLine = CurLine.substr(DelimiterPos + Field_Delimiter.length());
        DelimiterPos = CurLine.find(Field_Delimiter);
        if (DelimiterPos == 0 || DelimiterPos == string::npos) 
            goto FinishedReadingPoints;       // No valid entry on this line

        strPointY = CurLine.substr(0, DelimiterPos);
        strNextPoint = CurLine.substr(DelimiterPos + Field_Delimiter.length());

        // Translate the data from strings to the internal representations
        strPointX = CU_TrimString(strPointX);
        strPointY = CU_TrimString(strPointY);
        strNextPoint = CU_TrimString(strNextPoint);

        stringstream(strPointX) >> PatrolPoints[PointsFound].Location.x;
        stringstream(strPointY) >> PatrolPoints[PointsFound].Location.y;
        stringstream(strNextPoint) >> PatrolPoints[PointsFound].NextPoint;

        if (PatrolPoints[PointsFound].NextPoint < 0 
            || PatrolPoints[PointsFound].NextPoint >= EnemyPlane::MAX_NUM_PATROL_POINTS) {
            goto FinishedReadingPoints;     // points to an invalid next point
        } else {
            PointsFound++;
        }
        // Get the next line
        LastLinePos = LevelFile->tellg();
        getline(*LevelFile, CurLine);
    }
FinishedReadingPoints:

    // Put back the last line we read as we didn't need it
    LevelFile->seekg(LastLinePos);

    return (PointsFound > 0);
}

//------------------------------------------------------------------------------
// Name: AddAllNewEntities
// Desc: This subroutine goes through all of the entities in the NewEntities
//       list and allocates them to the appropriate list.
//------------------------------------------------------------------------------
void Level::AddAllNewEntities()
{
    list<Entity *>::iterator Iter;

    if (!NewEntities.empty()) {
        for (Iter = NewEntities.begin(); Iter != NewEntities.end(); Iter++) {
            assert((*Iter)->GetType() != Entity::NO_TYPE);
    	    GameEntities[(*Iter)->GetType()].push_front(*Iter);
	    }
		NewEntities.clear();
	}
}

//------------------------------------------------------------------------------
// Name: Update
// Desc: This updates the status of the level. This involves updating all of 
//       the entities in the level given that 'dt' milliseconds have passed and
//       checking to see whether the conditions for victory or defeat have been
//       met. This routine handles updating the entities, checks to see if any 
//       entities have collided with each other and checks to see if any 
//       entities have collided with the landscape.
//------------------------------------------------------------------------------
void Level::Update(uint dt)
{
    int i, j;
    list<Entity *>::iterator Iter, NextIter;
	int NumCollisions;
	Entity::COLLISIONS_ARRAY Collisions;    

    CU_STARTFRAME(CAM_P_UPDATE_ENTITIES);

    // Call general update routine
    for (i = 0; i < Entity::NUM_ENT_TYPES; i++) {
        for (Iter = GameEntities[i].begin(); Iter != GameEntities[i].end(); Iter++) {
    	    (*Iter)->Update(dt);
	    	if ((*Iter)->GetNeedsDeletion()) {
		        NextIter = Iter;
			    NextIter++;

				delete *Iter;
			    GameEntities[i].erase(Iter);

    		    Iter = NextIter;
	    		if (Iter == GameEntities[i].end()) break;
		    }
	    }
    }
    
	// Collision detection with other entities
    for (i = 0; i < Entity::NUM_ENT_TYPES; i++) {
	    for (Iter = GameEntities[i].begin(); Iter != GameEntities[i].end(); Iter++) {
		    // If the entity has already been destroyed in a previous collision
    	    // then delete it
	    	if ((*Iter)->GetNeedsDeletion()) {
		        NextIter = Iter;
			    NextIter++;

                delete *Iter;
			    GameEntities[i].erase(Iter);

                Iter = NextIter;
	    		if (Iter == GameEntities[i].end()) break;
		    }
			// Find all the collisions this entity is involved with
		    NumCollisions = (*Iter)->FindCollisions(&Collisions);
    		for (j = 0; j < NumCollisions; j++) {
	    	    Collisions[j]->HandleCollision(*Iter);
		    	(*Iter)->HandleCollision(Collisions[j]);
			    // Delete any entities in previous lists
				if (Collisions[j]->GetNeedsDeletion() 
					&& Collisions[j]->GetType() < (*Iter)->GetType()) {
				    GameEntities[Collisions[j]->GetType()].remove(Collisions[j]);
    				delete Collisions[j];
	    		}
		    }
			if ((*Iter)->GetNeedsDeletion()) {
		        NextIter = Iter;
				NextIter++;

    			delete *Iter;
	    		GameEntities[i].erase(Iter);

				Iter = NextIter;
    			if (Iter == GameEntities[i].end()) break;
	    	}
        }
	}

	// Collision Detection with the landscape
    for (i = 0; i < Entity::NUM_ENT_TYPES; i++) {
	    for (Iter = GameEntities[i].begin(); Iter != GameEntities[i].end(); Iter++) {
	        (*Iter)->FindAndHandleCollisionWithLandscape();
		    if ((*Iter)->GetNeedsDeletion()) {
			    NextIter = Iter;
				NextIter++;

    			delete *Iter;
	    		GameEntities[i].erase(Iter);

		    	Iter = NextIter;
	            if (Iter == GameEntities[i].end()) break;
		    }
        }
    }

    CU_ENDFRAME(CAM_P_UPDATE_ENTITIES);

    // Now update the level itself
    // Allocate any new entities to the appropriate list
    AddAllNewEntities();

    // Check to see if the player has earnt a new life
    if (OldPlayerScore < ScoreNeededForNewLife 
        && PlayerScore >= ScoreNeededForNewLife) {

        PlayerLives++;
        // Make it progressively harder for a player to get a new life
        ScoreNeededForNewLife *= 2;
    }
    OldPlayerScore = PlayerScore;

    // Check for game over conditions
    if (Flags[Entity::EVIL]->GetCaptured() || Flags[Entity::GOOD]->GetCaptured() 
        || PlayerLives <= 0) {
        GameOver = true;
        PlayerVictory = Flags[Entity::EVIL]->GetCaptured();
    }
}

//------------------------------------------------------------------------------
// Name: DrawBackground
// Desc: Draws the landscape background to the given SDL_Surface
//------------------------------------------------------------------------------
bool Level::DrawBackground(SDL_Surface *ScreenBuffer)
{
    return LS->DrawViewportBackground(ScreenBuffer);
}

//------------------------------------------------------------------------------
// Name: DrawEntities
// Desc: Loops through all of the entities in the game world and draws them.
//------------------------------------------------------------------------------
void Level::DrawEntities(SDL_Surface *ScreenBuffer)
{
    int i;
    list<Entity *>::iterator Iter;

    for (i = 0; i < Entity::NUM_ENT_TYPES; i++) {
        for (Iter = GameEntities[i].begin(); Iter != GameEntities[i].end(); Iter++) {
            (*Iter)->Draw(ScreenBuffer);
		}
	}
}

//------------------------------------------------------------------------------
// Name: DrawForeground
// Desc: Draws the landscape foreground to the given SDL_Surface
//------------------------------------------------------------------------------
bool Level::DrawForeground(SDL_Surface *ScreenBuffer)
{
    return LS->DrawViewportForeground(ScreenBuffer);
}

//------------------------------------------------------------------------------
// Name: AddNewEntity
// Desc: Adds a new entity to the NewEntities list. These entities are then 
//       added to the appropriate lists at the begining of the next frame
//       when update entities is called.
//------------------------------------------------------------------------------
void Level::AddNewEntity(Entity *AnEntity)
{
    NewEntities.push_front(AnEntity);
}

//------------------------------------------------------------------------------
// Name: IsEntityAlive
// Desc: Although entities are created and live in the entity lists exposed
//       by the Game object. Some parts of the program may maintain direct
//       pointers to entities to keep track of them. This can cause problems
//       however if an entity is deleted and the pointer left hanging. To
//       avoid this problem, a pointer to an entity should be checked before
//       use if there is a chance that the entity has been deleted since last
//       the pointer was used.
//------------------------------------------------------------------------------
bool Level::IsEntityAlive(Entity *AnEntity)
{
    int i;
    list<Entity *>::iterator Iter;
    
    // We must search through all entity types as we cannot use
    // AnEntity->GetType() because AnEntity may be an invalid pointer
    for (i = 0; i < Entity::NUM_ENT_TYPES; i++) {
        for (Iter = GameEntities[i].begin();
            Iter != GameEntities[i].end(); Iter++) {
            if (*Iter == AnEntity) return true;
        }
    }

    return false;
}

//------------------------------------------------------------------------------
// Name: GetPlayerPlane
// Desc: If one exists, this function returns a pointer to the entity that
//       represents the players plane.
//------------------------------------------------------------------------------
PlayerPlane *Level::GetPlayerPlane()
{
    if (PlayerPlaneSpawnPoint == NULL) {
        return NULL;
    } else {
        return (PlayerPlane *)PlayerPlaneSpawnPoint->GetCurPlane();
    }
}

//------------------------------------------------------------------------------
// Name: GetAirResistance
// Desc: Works out the air resistance that a flying entity experiences based
//       upon its type and velocity. 
//
// TODO: This function has been placed here mainly because I didn't know where
//       else to put it. It causes a bad circular reference if placed in the 
//       CamelMath object and is only used by a few entities. More thought 
//       should be given to how it fits into the overall game and it should be
//       altered accordingly.
//------------------------------------------------------------------------------
float Level::GetAirResistance(Entity::ENT_TYPE Type, float Velocity)
{
    // Air resistance increases as something goes faster
    switch (Type) {
        case Entity::PLANE:
            return Velocity * CamelMath::AIR_THICKNESS;
        case Entity::BOMB:
            return (Velocity*CamelMath::AIR_THICKNESS) / 4;
        default:
            return 0.0f;
    }
}

//------------------------------------------------------------------------------
// Name: ReadNameFromLevelFile
// Desc: Takes a name of a file and treating it as a level file tries to read
//       the name from it. If the file is not a valid level file then the
//       function returns 'false'. Otherwise the function returns 'true' and 
//       places the name of the level in 'LevelName'.
//------------------------------------------------------------------------------
bool Level::ReadNameFromLevelFile(const char *Filename, string& LevelName)
{
    ifstream LevelFile(Filename, ios::in);
    string CurLine;
    int MagicNumber;

    // Deal with the trivial case where the file doesn't exist
    if (!LevelFile) goto InvalidLevelFile;

    // Check for the 'magic number'
    getline(LevelFile, CurLine); if (!LevelFile) goto InvalidLevelFile;
    stringstream(CurLine) >> MagicNumber;
    if (MagicNumber != Level::FILE_MAGIC_NUMBER) goto InvalidLevelFile;

    // Read in the level's name, and truncate it if needed
    getline(LevelFile, LevelName); if (!LevelFile) goto InvalidLevelFile;
    if (LevelName.length() > MAX_LEVEL_NAME_LENGTH) 
        LevelName = LevelName.substr(0, MAX_LEVEL_NAME_LENGTH);

    return true;
InvalidLevelFile:
    return false;
}
