/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: MenuController.cpp
// Desc: This static class provides the menus for the games. A menu is a 
//       collection of ParaGUI widgets linked to event handlers that
//       perform various actions such as starting a new game or changing the
//       input keys.
//
//       In the future, it might be desirable to allow an arbitrary menu
//       structure to be submitted to the menu controller, but for now the
//       menus are just going to be hardwired in.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <assert.h>
#include <list>
#include <string>
#include <sstream>
#include "globals.h"
#include "Camel.h"
#include "CamelUtility.h"
#include "Game.h"
#include "SDL_Interface.h"
#include "InputHandler.h"
#include "OrdersPanel.h"
#include "Level.h"
#include "Landscape.h"
#include "BFont.h"
#include "MenuController.h"

// Paragui includes
#include "paragui.h"
#include "pgapplication.h"
#include "pgwidget.h"
#include "pgwidgetlist.h"
#include "pgwindow.h"
#include "pgbutton.h"
#include "pglabel.h"
#include "pglineedit.h"
#include "pgimage.h"
#include "pgslider.h"
#include "pgcheckbutton.h"
#include "pgdropdown.h"

using namespace std;

//------------------------------------------------------------------------------
// MenuController Implementation
//------------------------------------------------------------------------------
const char *MenuController::TITLE_MENU_PIC = CAM_GRAPHICS_DIR "/titlemenupic.png";
const char *MenuController::HELP_SCREEN_PICS[NUM_HELP_SCREENS] = {
    CAM_GRAPHICS_DIR "/helpscreen01.png",
    CAM_GRAPHICS_DIR "/helpscreen02.png",
    CAM_GRAPHICS_DIR "/helpscreen03.png",
    CAM_GRAPHICS_DIR "/helpscreen04.png",
    CAM_GRAPHICS_DIR "/helpscreen05.png"
};

const PG_Rect MenuController::DEFAULT_MENU_AREA_RECT = PG_Rect(0, 130, 640, 350);

// When changing the rects for the menu widgets. The dimensions of the 
// DEFAULT_MENU_AREA_RECT must be kept in mind. Ideally, it would be cleaner
// to write the dimensions of the rects in terms of the dimensions of
// DEFAULT_MENU_AREA_RECT but life really is too short. :)

// Start menu rects
const PG_Rect MenuController::SM_START_GAME_RECT = PG_Rect(220, 20, 200, 28);
const PG_Rect MenuController::SM_OPTIONS_RECT = PG_Rect(220, 50, 200, 28);
const PG_Rect MenuController::SM_HELP_RECT = PG_Rect(220, 80, 200, 28);
const PG_Rect MenuController::SM_HIGH_SCORES_RECT = PG_Rect(220, 110, 200, 28);
const PG_Rect MenuController::SM_EXIT_RECT = PG_Rect(220, 140, 200, 28);

// Pick Level menu rects
const PG_Rect MenuController::PLM_LEVEL_LIST_RECT = PG_Rect(170, 18, 300, 240);
const PG_Rect MenuController::PLM_LEVEL_BUTTON_RECT = PG_Rect(50, 2, 200, 28);
const PG_Rect MenuController::PLM_CANCEL_RECT = PG_Rect(220, 290, 200, 28);

// IntroScreen menu rects
const PG_Rect MenuController::ISM_CONTINUE_RECT = PG_Rect(220, 290, 200, 28);

// Options menu rects
const PG_Rect MenuController::OM_VIDEO_RECT = PG_Rect(220, 20, 200, 28);
const PG_Rect MenuController::OM_AUDIO_RECT = PG_Rect(220, 50, 200, 28);
const PG_Rect MenuController::OM_CONTROLS_RECT = PG_Rect(220, 80, 200, 28);
const PG_Rect MenuController::OM_DONE_RECT = PG_Rect(220, 110, 200, 28);

// Video Options menu rects
const PG_Rect MenuController::VOM_FULLSCREEN_RECT = PG_Rect(175, 20, 140, 28);
const PG_Rect MenuController::VOM_FULLSCREEN_CHECK_RECT = PG_Rect(325, 20, 30, 30);
const PG_Rect MenuController::VOM_RESOLUTION_RECT = PG_Rect(175, 50, 140, 28);
const PG_Rect MenuController::VOM_RESOLUTION_COMBO_RECT = PG_Rect(325, 54, 140, 20);
const PG_Rect MenuController::VOM_BITDEPTH_RECT = PG_Rect(175, 80, 140, 28);
const PG_Rect MenuController::VOM_BITDEPTH_COMBO_RECT = PG_Rect(325, 84, 60, 20);
const PG_Rect MenuController::VOM_PARALLAX_RECT = PG_Rect(175, 110, 140, 28);
const PG_Rect MenuController::VOM_PARALLAX_SLIDER_RECT = PG_Rect(325, 114, 140, 20);
const PG_Rect MenuController::VOM_PARALLAX_LABEL0_RECT = PG_Rect(325, 132, 20, 20);
const PG_Rect MenuController::VOM_PARALLAX_LABEL1_RECT = PG_Rect(393, 132, 20, 20);
const PG_Rect MenuController::VOM_PARALLAX_LABEL2_RECT = PG_Rect(455, 132, 20, 20);
const PG_Rect MenuController::VOM_DONE_RECT = PG_Rect(220, 170, 200, 28);
const PG_Rect MenuController::VOM_TOOLTIP_RECT = PG_Rect(0, 320, 640, 28);

// Audio Options menu rects
const PG_Rect MenuController::AOM_SOUND_VOLUME_RECT = PG_Rect(175, 20, 140, 28);
const PG_Rect MenuController::AOM_SOUND_VOLUME_SLIDER_RECT = PG_Rect(325, 24, 140, 20);
const PG_Rect MenuController::AOM_MUSIC_VOLUME_RECT = PG_Rect(175, 50, 140, 28);
const PG_Rect MenuController::AOM_MUSIC_VOLUME_SLIDER_RECT = PG_Rect(325, 54, 140, 20);
const PG_Rect MenuController::AOM_REVERSE_STEREO_RECT = PG_Rect(175, 80, 140, 28);
const PG_Rect MenuController::AOM_REVERSE_STEREO_CHECK_RECT = PG_Rect(325, 80, 30, 30);
const PG_Rect MenuController::AOM_DONE_RECT = PG_Rect(220, 140, 200, 28);
const PG_Rect MenuController::AOM_TOOLTIP_RECT = PG_Rect(0, 320, 640, 28);

// Controls Options menu rects
const PG_Rect MenuController::COM_OPTIONS_LIST_RECT = PG_Rect(150, 20, 340, 180);
const PG_Rect MenuController::COM_TEMP_WIDGET_RECT = PG_Rect(0, 0, 340, 30);
const PG_Rect MenuController::COM_OPTION_RECT = PG_Rect(25, 0, 140, 28);
const PG_Rect MenuController::COM_OPTION_CHANGE_RECT = PG_Rect(175, 0, 140, 28);
const PG_Rect MenuController::COM_CLOCK_CONTROLS_CHECK_RECT = PG_Rect(175, 0, 30, 30);
const PG_Rect MenuController::COM_DONE_RECT = PG_Rect(220, 230, 200, 28);
const PG_Rect MenuController::COM_TOOLTIP_RECT = PG_Rect(0, 320, 640, 28);

// Help menu rects
const PG_Rect MenuController::HM_PREV_RECT = PG_Rect(200, 290, 80, 28);
const PG_Rect MenuController::HM_DONE_RECT = PG_Rect(280, 290, 80, 28);
const PG_Rect MenuController::HM_NEXT_RECT = PG_Rect(360, 290, 80, 28);

// InGame menu rects
const PG_Rect MenuController::IGM_CONTINUE_GAME_RECT = PG_Rect(220, 20, 200, 28);
const PG_Rect MenuController::IGM_OPTIONS_RECT = PG_Rect(220, 50, 200, 28);
const PG_Rect MenuController::IGM_HELP_RECT = PG_Rect(220, 80, 200, 28);
const PG_Rect MenuController::IGM_QUIT_GAME_RECT = PG_Rect(220, 110, 200, 28);
const PG_Rect MenuController::IGM_EXIT_RECT = PG_Rect(220, 140, 200, 28);

// New High Score menu rects
const PG_Rect MenuController::NHSM_NAME_ENTRY_FORM_RECT = PG_Rect(170, 40, 300, 150);
const PG_Rect MenuController::NHSM_CONGRATS_L1_RECT = PG_Rect(10, 5, 280, 20);
const PG_Rect MenuController::NHSM_CONGRATS_L2_RECT = PG_Rect(10, 25, 280, 25);
const PG_Rect MenuController::NHSM_NAME_REQUEST_RECT = PG_Rect(10, 80, 280, 25);
const PG_Rect MenuController::NHSM_NAME_TEXTBOX_RECT = PG_Rect(20, 110, 260, 25);

// High Score menu rects
const PG_Rect MenuController::HSM_NAMETITLE_RECT = PG_Rect(120, 20, 115, 28);
const PG_Rect MenuController::HSM_LEVELTITLE_RECT = PG_Rect(235, 20, 170, 28);
const PG_Rect MenuController::HSM_SCORETITLE_RECT = PG_Rect(405, 20, 115, 28);
const PG_Rect MenuController::HSM_FIRST_ENTRY_RECT = PG_Rect(120, 50, 400, 22);
const PG_Rect MenuController::HSM_ENTRY_NAME_RECT = PG_Rect(10, 2, 100, 20);
const PG_Rect MenuController::HSM_ENTRY_LEVELNAME_RECT = PG_Rect(120, 2, 160, 20);
const PG_Rect MenuController::HSM_ENTRY_SCORE_RECT = PG_Rect(290, 2, 100, 20);
const PG_Rect MenuController::HSM_BACK_TO_START_RECT = PG_Rect(220, 290, 200, 28);
const int MenuController::HSM_ENTRY_STEP = 24;

MenuController *MenuController::Instance = NULL;

//------------------------------------------------------------------------------
// Name: MenuController
// Desc: When this object is created, a container widget for all of the games
//       menus is created and the current menu set to NO_MENU.
//------------------------------------------------------------------------------
MenuController::MenuController(Game *G, PG_Application *PGApp) 
{
    PG_Rect MenuAreaRect = DEFAULT_MENU_AREA_RECT;
    assert(G != NULL && PGApp != NULL);

    this->G = G;
    this->PGApp = PGApp;
    SDLI = SDL_Interface::GetInstance(G);
    IH = InputHandler::GetInstance(G);
    OP = OrdersPanel::GetInstance(G, PGApp);

    Activated = false;
    
    // This piece of code determines the size of, and creates the MenuArea
    // widget. It is assumed that the smallest possible screen area is 640x480
    // and that the default screen size is 800x600.
    MenuAreaRect.my_xpos = (SDLI->GetScreenWidth() - MenuAreaRect.my_width) / 2;
    if (SDLI->GetScreenHeight() > 600) {
        MenuAreaRect.my_ypos += (SDLI->GetScreenHeight() - 600)/2;
    }
    MenuArea = new PG_Widget(NULL, MenuAreaRect);
    btnTooltip = NULL;

    BackgroundImage = NULL;
    CurHelpScreen = 0;
    CurMenuInvalid = false;
}

//------------------------------------------------------------------------------
// Name: ~MenuController
// Desc: All cleanup done here
//------------------------------------------------------------------------------
MenuController::~MenuController() {}

//------------------------------------------------------------------------------
// Name: SetCurMenu
// Desc: This routine cleans up after the current menu and then activates the
//       menu given by 'AMenu'.
//------------------------------------------------------------------------------
void MenuController::SetCurMenu(MENUS AMenu)
{
    // Clear out the current menu
    MenuArea->RemoveAllChilds();
    btnTooltip = NULL;

    CurMenu = AMenu;
    NewMenu = AMenu;
    UpdateBackgroundImage();
    CurMenuInvalid = false;

    switch (CurMenu) {
        case START_MENU:
            ActivateStartMenu();
            break;
        case PICK_LEVEL_MENU:
            ActivatePickLevelMenu();
            break;
        case INTROSCREEN_MENU:
            ActivateIntroScreenMenu();
            break;
        case OPTIONS_MENU:
            ActivateOptionsMenu();
            break;
        case VIDEO_OPTIONS_MENU:
            ActivateVideoOptionsMenu();
            break;
        case AUDIO_OPTIONS_MENU:
            ActivateAudioOptionsMenu();
            break;
        case CONTROLS_OPTIONS_MENU:
            ActivateControlsOptionsMenu();
            break;
        case HELP_MENU:
            ActivateHelpMenu();
            break;
        case INGAME_MENU:
            ActivateInGameMenu();
            break;
        case NEW_HIGH_SCORE_MENU:
            ActivateNewHighScoreMenu();
            break;
        case HIGH_SCORES_MENU:
            ActivateHighScoresMenu();
            break;
    }
    MenuArea->Hide();
    MenuArea->Show();
}


//------------------------------------------------------------------------------
// Name: ActivateStartMenu
// Desc: Activates the start menu. The start menu is the first menu that the
//       player sees when they enter the game.
//------------------------------------------------------------------------------
void MenuController::ActivateStartMenu()
{
    PG_Button *btnStartGame;
    PG_Button *btnOptions;
    PG_Button *btnHelp;
    PG_Button *btnHighScores;
    PG_Button *btnExit;

    btnStartGame = new PG_Button(MenuArea, 0, SM_START_GAME_RECT, 
        "Start Game", "MENU_BUTTON");
    btnStartGame->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnStartGame_Click);
    //btnStartGame->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)MenuController::btnStartGame_Click);
    btnOptions = new PG_Button(MenuArea, 0, SM_OPTIONS_RECT, "Options", "MENU_BUTTON");
    btnOptions->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnOptions_Click);
    btnHelp = new PG_Button(MenuArea, 0, SM_HELP_RECT, "Help", "MENU_BUTTON");
    btnHelp->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnHelp_Click);
    btnHighScores = new PG_Button(MenuArea, 0, SM_HIGH_SCORES_RECT, 
        "High Scores", "MENU_BUTTON");
    btnHighScores->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnHighScores_Click);
    btnExit = new PG_Button(MenuArea, 0, SM_EXIT_RECT, "Exit", "MENU_BUTTON");
    btnExit->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnExit_Click);
}
 
//------------------------------------------------------------------------------
// Name: ActivatePickLevelMenu
// Desc: This menu displays all of the valid levels it can find in the
//       CAM_LEVELS_DIR directory. The user can either select a level to play
//       or press the 'Cancel' button to return to the start menu.
//------------------------------------------------------------------------------
void MenuController::ActivatePickLevelMenu()
{
    PG_Button *btnLevel;
    PG_WidgetList *LevelList;
    PG_Button *btnCancel;
    string Filename;
    string LevelName;
    list<string>::iterator Iter; 

    // Update the list of files
    CU_ListFiles(CAM_LEVELS_DIR, LevelFileList);

    LevelList = new PG_WidgetList(MenuArea, PLM_LEVEL_LIST_RECT);
    LevelList->SetTransparency(255);		
	LevelList->EnableScrollBar(true, PG_SB_VERTICAL);

    // Create a button for each level file found
    for (Iter = LevelFileList.begin(); Iter != LevelFileList.end(); Iter++) {

        // Construct the full file name and try to read the level name from it
        Filename = string(CAM_LEVELS_DIR) + "/" + *Iter;
        if (Level::ReadNameFromLevelFile(Filename.c_str(), LevelName)) {

            if (LevelName.length() > 0) {
                btnLevel = new PG_Button(NULL, 0, PLM_LEVEL_BUTTON_RECT, 
                    LevelName.c_str(), "MENU_BUTTON");
            } else {
                btnLevel = new PG_Button(NULL, 0, PLM_LEVEL_BUTTON_RECT, 
                    (*Iter).c_str(), "MENU_BUTTON");
            }
            // Pass on a pointer to the string storing the files name
            btnLevel->SetEventObject(MSG_BUTTONCLICK, this, 
                (MSG_CALLBACK_OBJ)&MenuController::btnLevel_Click, &(*Iter));
            LevelList->AddWidget(btnLevel);
        }
    }

    btnCancel = new PG_Button(MenuArea, 0, PLM_CANCEL_RECT, "Cancel", "MENU_BUTTON");
    btnCancel->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnBackToStart_Click);
}

//------------------------------------------------------------------------------
// Name: ActivateIntroScreenMenu
// Desc: If a level has an IntroScreen then it is shown here to provide a 
//       breifing to the player before they play the mission.
//------------------------------------------------------------------------------
void MenuController::ActivateIntroScreenMenu()
{
    PG_Button *btnContinue;

    btnContinue = new PG_Button(MenuArea, 0, ISM_CONTINUE_RECT, "Continue", "MENU_BUTTON");
    btnContinue->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnContinue_Click);
}

//------------------------------------------------------------------------------
// Name: ActivateOptionsMenu
// Desc: The options menu lets the player configure using a user friendly gui
//       rather than mucking around with the options file.
//------------------------------------------------------------------------------
void MenuController::ActivateOptionsMenu()
{
    PG_Button *btnVideo;
    PG_Button *btnAudio;
    PG_Button *btnControls;
    PG_Button *btnDone;

    btnVideo = new PG_Button(MenuArea, 0, OM_VIDEO_RECT, "Video", "MENU_BUTTON");
    btnVideo->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnVideo_Click);
    btnAudio = new PG_Button(MenuArea, 0, OM_AUDIO_RECT, "Audio", "MENU_BUTTON");
    btnAudio->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnAudio_Click);
    btnControls = new PG_Button(MenuArea, 0, OM_CONTROLS_RECT, "Controls", "MENU_BUTTON");
    btnControls->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnControls_Click);
    btnDone = new PG_Button(MenuArea, 0, OM_DONE_RECT, "Done", "MENU_BUTTON");
    btnDone->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnDoneWithGameMenu_Click);
}

//------------------------------------------------------------------------------
// Name: ActivateVideoOptionsMenu
// Desc: Here the player can change the screen resolution and bitdepth as well
//       as being able to decide to run the game in windowed mode or reduce
//       the number of parallax layers used in order to get a faster game.
//       In order to avoid the complications of having to reload the video
//       surfaces changes other than the number of parallax layers require
//       the game to be restarted before they take affect.
//------------------------------------------------------------------------------
void MenuController::ActivateVideoOptionsMenu()
{
    int i;
    char BitdepthBuffer[10];
    PG_Button *btnFullscreen;
    PG_Button *btnResolution;
    PG_Button *btnBitdepth;
    PG_Button *btnParallax;
    PG_Button *btnDone;
    OptionsCheckButton *cbnFullscreen;
    OptionsDropDown *cbxResolution;
    OptionsDropDown *cbxBitdepth;
    OptionsSlider *sldParallax;
    PG_Label *lblParallax0;
    PG_Label *lblParallax1;
    PG_Label *lblParallax2;

    // Set up the buttons
    btnFullscreen = new PG_Button(MenuArea, 0, VOM_FULLSCREEN_RECT, 
        "Fullscreen", "MENU_BUTTON");
    btnFullscreen->EnableReceiver(false);
    btnResolution = new PG_Button(MenuArea, 0, VOM_RESOLUTION_RECT, 
        "Resolution", "MENU_BUTTON");
    btnResolution->EnableReceiver(false);
    btnBitdepth = new PG_Button(MenuArea, 0, VOM_BITDEPTH_RECT, 
        "Bitdepth", "MENU_BUTTON");
    btnBitdepth->EnableReceiver(false);
    btnParallax = new PG_Button(MenuArea, 0, VOM_PARALLAX_RECT, 
        "Parallax", "MENU_BUTTON");
    btnParallax->EnableReceiver(false);
    btnDone = new PG_Button(MenuArea, 0, VOM_DONE_RECT, "Done", "MENU_BUTTON");
    btnDone->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnOptions_Click);
    btnTooltip = new PG_Button(MenuArea, 0, VOM_TOOLTIP_RECT, 
        "", "TOOLTIP_BUTTON");
    btnTooltip->EnableReceiver(false);

    // Set up the controls that let the user actually change the options
    cbnFullscreen = new OptionsCheckButton(MenuArea, Game::FULLSCREEN,
        VOM_FULLSCREEN_CHECK_RECT, btnTooltip, "");
    if (G->GetOption(Game::FULLSCREEN) != 0) {
        cbnFullscreen->SetPressed();
    } else {
        cbnFullscreen->SetUnpressed();
    }
    cbnFullscreen->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::AlterFullscreen);

    cbxResolution = new OptionsDropDown(MenuArea, Game::SCREENMODE,
        VOM_RESOLUTION_COMBO_RECT, btnTooltip, "OPTIONS_DROPDOWN");
    cbxResolution->SetEditable(false);
    for (i = 0; i < SDL_Interface::NUM_SCREEN_MODES; i++) {
        cbxResolution->AddItem(SDL_Interface::SCREEN_MODE_DETAILS[i].String);
    }
    cbxResolution->SetText(SDL_Interface::SCREEN_MODE_DETAILS[
        G->GetOption(Game::SCREENMODE)].String);
    cbxResolution->SetEventObject(MSG_SELECTITEM, this, (MSG_CALLBACK_OBJ)&MenuController::AlterResolution);
        

    cbxBitdepth = new OptionsDropDown(MenuArea, Game::BITDEPTH,
        VOM_BITDEPTH_COMBO_RECT, btnTooltip, "OPTIONS_DROPDOWN");
    cbxBitdepth->SetEditable(false);
    for (i = 0; i < SDL_Interface::NUM_SCREEN_BITDEPTHS; i++) {
        sprintf(BitdepthBuffer, "%i", SDL_Interface::SCREEN_BITDEPTH_DETAILS[i]);
        cbxBitdepth->AddItem(BitdepthBuffer);
    }
    sprintf(BitdepthBuffer, "%i", SDL_Interface::SCREEN_BITDEPTH_DETAILS[
        G->GetOption(Game::BITDEPTH)]);
    cbxBitdepth->SetText(BitdepthBuffer);
    cbxBitdepth->SetEventObject(MSG_SELECTITEM, this, (MSG_CALLBACK_OBJ)&MenuController::AlterBitdepth);

    lblParallax0 = new PG_Label(MenuArea, VOM_PARALLAX_LABEL0_RECT, "0", "PARALLAX_SLIDER_LABEL");
    lblParallax1 = new PG_Label(MenuArea, VOM_PARALLAX_LABEL1_RECT, "1", "PARALLAX_SLIDER_LABEL");
    lblParallax2 = new PG_Label(MenuArea, VOM_PARALLAX_LABEL2_RECT, "2", "PARALLAX_SLIDER_LABEL");

    sldParallax = new OptionsSlider(MenuArea, Game::NUM_PARALLAX_LAYERS, 
        VOM_PARALLAX_SLIDER_RECT, btnTooltip, PG_SB_HORIZONTAL);
    sldParallax->TheSlider->SetRange(0, 2);
    sldParallax->TheSlider->SetPosition(G->GetOption(Game::NUM_PARALLAX_LAYERS));
    sldParallax->TheSlider->SetEventObject(MSG_SLIDEEND, this, 
        (MSG_CALLBACK_OBJ)&MenuController::AlterNumParallaxLayers);
}

//------------------------------------------------------------------------------
// Name: ActivateAudioOptionsMenu
// Desc: Lets the player change the volume of the game's sound and music and
//       also lets the player reverse the stereo.
//------------------------------------------------------------------------------
void MenuController::ActivateAudioOptionsMenu()
{
    PG_Button *btnSoundVolume;
    PG_Button *btnMusicVolume;
    PG_Button *btnReverseStereo;
    PG_Button *btnDone;
    OptionsSlider *sldSoundVolume;
    OptionsSlider *sldMusicVolume;
    OptionsCheckButton *cbnReverseStereo;

    // Set up the buttons
    btnSoundVolume = new PG_Button(MenuArea, 0, AOM_SOUND_VOLUME_RECT, 
        "Sound Volume", "MENU_BUTTON");
    btnSoundVolume->EnableReceiver(false);
    btnMusicVolume = new PG_Button(MenuArea, 0, AOM_MUSIC_VOLUME_RECT, 
        "Music Volume", "MENU_BUTTON");
    btnMusicVolume->EnableReceiver(false);
    btnReverseStereo = new PG_Button(MenuArea, 0, AOM_REVERSE_STEREO_RECT, 
        "Reverse Stereo", "MENU_BUTTON");
    btnReverseStereo->EnableReceiver(false);
    btnDone = new PG_Button(MenuArea, 0, AOM_DONE_RECT, "Done", "MENU_BUTTON");
    btnDone->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnOptions_Click);
    btnTooltip = new PG_Button(MenuArea, 0, AOM_TOOLTIP_RECT, 
        "", "TOOLTIP_BUTTON");
    btnTooltip->EnableReceiver(false);

    // Set up the controls that let the user actually change the options
    sldSoundVolume = new OptionsSlider(MenuArea, Game::SFX_VOLUME, 
        AOM_SOUND_VOLUME_SLIDER_RECT, btnTooltip, PG_SB_HORIZONTAL);
    sldSoundVolume->TheSlider->SetRange(0, 128);
    sldSoundVolume->TheSlider->SetPosition(SDLI->GetSFX_Volume());
    sldSoundVolume->TheSlider->SetEventObject(MSG_SLIDEEND, this, (MSG_CALLBACK_OBJ)&MenuController::AlterSoundVolume);

    sldMusicVolume = new OptionsSlider(MenuArea, Game::MUSIC_VOLUME, 
        AOM_MUSIC_VOLUME_SLIDER_RECT, btnTooltip, PG_SB_HORIZONTAL);
    sldMusicVolume->TheSlider->SetRange(0, 128);
    sldMusicVolume->TheSlider->SetPosition(SDLI->GetMusic_Volume());
    sldMusicVolume->TheSlider->SetEventObject(MSG_SLIDEEND, this, (MSG_CALLBACK_OBJ)&MenuController::AlterMusicVolume);

    cbnReverseStereo = new OptionsCheckButton(MenuArea, Game::REVERSE_STEREO,
        AOM_REVERSE_STEREO_CHECK_RECT, btnTooltip, "");
    if (SDLI->GetReverseStereo()) {
        cbnReverseStereo->SetPressed();
    } else {
        cbnReverseStereo->SetUnpressed();
    }
    cbnReverseStereo->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::AlterReverseStereo);
}

//------------------------------------------------------------------------------
// Name: ActivateControlsOptionsMenu
// Desc: This menu allows the player to change the controls that they use to
//       play the game.
//------------------------------------------------------------------------------
void MenuController::ActivateControlsOptionsMenu()
{
    int i;
    PG_Widget *TempWidget;
    PG_WidgetList *OptionsList;
    PG_Button *btnDisplay;
    OptionsCheckButton *cbnClockControls;
    PG_Button *btnDone;

    OptionsList = new PG_WidgetList(MenuArea, COM_OPTIONS_LIST_RECT);
    OptionsList->SetTransparency(255);		
	OptionsList->EnableScrollBar(true, PG_SB_VERTICAL);

    // Set up the buttons that will just display information
    btnDone = new PG_Button(MenuArea, 0, COM_DONE_RECT, "Done", "MENU_BUTTON");
    btnDone->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnOptions_Click);
    btnTooltip = new PG_Button(MenuArea, 0, COM_TOOLTIP_RECT, 
        "", "TOOLTIP_BUTTON");
    btnTooltip->EnableReceiver(false);

    for (i = 0; i < InputHandler::NUM_COMMANDS; i++) {
        TempWidget = new PG_Widget(NULL, COM_TEMP_WIDGET_RECT);
        btnDisplay = new PG_Button(TempWidget, 0, COM_OPTION_RECT, 
            InputHandler::COMMAND_DETAILS[i].Name, "MENU_BUTTON");
        btnDisplay->EnableReceiver(false);
        btnChangeCommand[i] = new OptionsButton(TempWidget, 
            InputHandler::COMMAND_DETAILS[i].Option,
            COM_OPTION_CHANGE_RECT, btnTooltip, 
            GetKeyName(InputHandler::COMMAND_DETAILS[i].Option), "MENU_BUTTON");
        btnChangeCommand[i]->SetEventObject(MSG_BUTTONCLICK, this, 
            (MSG_CALLBACK_OBJ)&MenuController::PickNewKey, (void *)i);
        OptionsList->AddWidget(TempWidget);
    }
 
    TempWidget = new PG_Widget(NULL, COM_TEMP_WIDGET_RECT);
    TempWidget->MoveWidget(0, 20);
    btnDisplay = new PG_Button(TempWidget, 0, COM_OPTION_RECT, "Clock Controls", "MENU_BUTTON");
    btnDisplay->EnableReceiver(false);
    cbnClockControls = new OptionsCheckButton(TempWidget, Game::CLOCK_CONTROLS,
        COM_CLOCK_CONTROLS_CHECK_RECT, btnTooltip);
    if (G->GetOption(Game::CLOCK_CONTROLS) != 0) {
        cbnClockControls->SetPressed();
    } else {
        cbnClockControls->SetUnpressed();
    }
    cbnClockControls->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::AlterClockControls);
    OptionsList->AddWidget(TempWidget);
}

const char *MenuController::GetKeyName(int i)
{
    const char *EMPTY_STRING = "";

    if (G->GetOption(i) == SDLK_ESCAPE) {
        return EMPTY_STRING;
    } else {
        return SDL_GetKeyName((SDLKey)G->GetOption(i));
    }
}

//------------------------------------------------------------------------------
// Name: ActivateHelpMenu
// Desc: The help menu provides a breif overview on how to play the game and
//       offers some very simple hints and tips to the player.
//------------------------------------------------------------------------------
void MenuController::ActivateHelpMenu()
{
    PG_Button *btnPrev;
    PG_Button *btnDone;
    PG_Button *btnNext;
    SDL_Color DisabledColour = { 128, 128, 128, 0 };

    btnPrev = new PG_Button(MenuArea, 0, HM_PREV_RECT, "Prev", "MENU_BUTTON");
    btnPrev->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnPrevHelp_Click);
    if (CurHelpScreen == 0) {
        btnPrev->EnableReceiver(false);
        btnPrev->SetFontColor(DisabledColour);
    }
    btnDone = new PG_Button(MenuArea, 0, HM_DONE_RECT, "Done", "MENU_BUTTON");
    btnDone->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnDoneWithGameMenu_Click);
    btnNext = new PG_Button(MenuArea, 0, HM_NEXT_RECT, "Next", "MENU_BUTTON");
    btnNext->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnNextHelp_Click);
    if (CurHelpScreen >= NUM_HELP_SCREENS-1) {
        btnNext->EnableReceiver(false);
        btnNext->SetFontColor(DisabledColour);
    }
}

//------------------------------------------------------------------------------
// Name: ActivateInGameMenu
// Desc: Activates the in game menu. When the player pauses the game by
//       pressing the escape key. This is the menu they are taken to.
//------------------------------------------------------------------------------
void MenuController::ActivateInGameMenu()
{
    PG_Button *btnContinueGame;
    PG_Button *btnOptions;
    PG_Button *btnHelp;
    PG_Button *btnQuitGame;
    PG_Button *btnExit;

    btnContinueGame = new PG_Button(MenuArea, 0, IGM_CONTINUE_GAME_RECT, 
        "Continue Game", "MENU_BUTTON");
    btnContinueGame->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnContinueGame_Click);
    btnOptions = new PG_Button(MenuArea, 0, IGM_OPTIONS_RECT, "Options", "MENU_BUTTON");
    btnOptions->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnOptions_Click);
    btnHelp = new PG_Button(MenuArea, 0, IGM_HELP_RECT, "Help", "MENU_BUTTON");
    btnHelp->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnHelp_Click);
    btnQuitGame = new PG_Button(MenuArea, 0, IGM_QUIT_GAME_RECT, 
        "Quit Game", "MENU_BUTTON");
    btnQuitGame->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnQuitGame_Click);
    btnExit = new PG_Button(MenuArea, 0, IGM_EXIT_RECT, "Exit", "MENU_BUTTON");
    btnExit->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::btnExit_Click);
}

//------------------------------------------------------------------------------
// Name: ActivateNewHighScoreMenu
// Desc: If the player has got a high score in the game then they are presented
//       with a text box to enter their name into. Once they have entered their
//       name, or if they didn't get a high score at all, they are taken back
//       to the start menu.
//------------------------------------------------------------------------------
void MenuController::ActivateNewHighScoreMenu()
{
    SDL_Event event;
    PG_Window *wndNameEntryForm;
    PG_Label *lblCongratsL1;
    PG_Label *lblCongratsL2;
    PG_Label *lblNameRequest;
    PG_LineEdit *tbxName;

    if (G->GetFinalScore() > G->GetHighScore(Game::NUM_HIGH_SCORES-1).Score) {
        wndNameEntryForm = new PG_Window(MenuArea, NHSM_NAME_ENTRY_FORM_RECT, "", 0, "Window", 0);
        lblCongratsL1 = new PG_Label(wndNameEntryForm, NHSM_CONGRATS_L1_RECT,
            "Congratulations!", "MC_TEXT_LABEL");
        lblCongratsL2 = new PG_Label(wndNameEntryForm, NHSM_CONGRATS_L2_RECT,
            "You have got a new high score!", "MC_TEXT_LABEL");
        lblNameRequest = new PG_Label(wndNameEntryForm, NHSM_NAME_REQUEST_RECT,
            "Please enter your name:", "MC_TEXT_LABEL");

        // Clear any superfluous key presses from the event queue as otherwise
        // they will appear as garbage in the textbox.
        while(SDL_PeepEvents(&event, 1, SDL_GETEVENT, SDL_KEYDOWNMASK) > 0)
            ;
        tbxName = new PG_LineEdit(wndNameEntryForm, NHSM_NAME_TEXTBOX_RECT, 
            "LineEdit", Game::MAX_HIGH_SCORE_NAME_LENGTH);
        tbxName->SetEventObject(MSG_EDITEND, this, (MSG_CALLBACK_OBJ)&MenuController::tbxName_EditEnd);
        tbxName->EditBegin();
    } else {
        // No high score obtained so go back to the start menu
        NewMenu = START_MENU;
        CurMenuInvalid = true;
    }
}

//------------------------------------------------------------------------------
// Name: ActivateHighScoresMenu
// Desc: Not so much a menu as a table, this menu displays the contents of the 
//       high score table.
//------------------------------------------------------------------------------
void MenuController::ActivateHighScoresMenu()
{
    int i;
    PG_Rect CurRect = HSM_FIRST_ENTRY_RECT;
    PG_Button *btnNameTitle;
    PG_Button *btnLevelTitle;
    PG_Button *btnScoreTitle;
    PG_Button *btnHighScore;
    PG_Button *btnBackToStart;
    Game::HIGH_SCORE CurHighScore;
    char strScore[20];
    PG_Label *lblName, *lblLevelName, *lblScore;

    // Create the title bar
    btnNameTitle = new PG_Button(MenuArea, 0, HSM_NAMETITLE_RECT, 
        "Name", "HIGHSCORE_LEFT_TITLE_BUTTON");
    btnNameTitle->EnableReceiver(false);
    btnLevelTitle = new PG_Button(MenuArea, 0, HSM_LEVELTITLE_RECT, 
        "Level", "HIGHSCORE_CENTRE_TITLE_BUTTON");
    btnLevelTitle->EnableReceiver(false);
    btnScoreTitle = new PG_Button(MenuArea, 0, HSM_SCORETITLE_RECT, 
        "Score", "HIGHSCORE_RIGHT_TITLE_BUTTON");
    btnScoreTitle->EnableReceiver(false);

    // Create the highscores
    for (i = 0; i < Game::NUM_HIGH_SCORES; i++) {
        CurHighScore = G->GetHighScore(i);
        sprintf(strScore, "%u", CurHighScore.Score);
        btnHighScore = new PG_Button(MenuArea, 0, CurRect, "", "MENU_BUTTON");
        btnHighScore->EnableReceiver(false);

        // Create labels for the name and score
        lblName = new PG_Label(btnHighScore, HSM_ENTRY_NAME_RECT, CurHighScore.Name);
        lblName->SetFontSize(14);
        lblName->SetAlignment(PG_TA_CENTER);
        lblLevelName = new PG_Label(btnHighScore, 
            HSM_ENTRY_LEVELNAME_RECT, CurHighScore.LevelName);
        lblLevelName->SetFontSize(14);
        lblLevelName->SetAlignment(PG_TA_CENTER);
        lblScore = new PG_Label(btnHighScore, HSM_ENTRY_SCORE_RECT, strScore);
        lblScore->SetFontSize(14);
        lblScore->SetAlignment(PG_TA_CENTER);

        CurRect.my_ypos += HSM_ENTRY_STEP;
    }
    
    btnBackToStart = new PG_Button(MenuArea, 0, HSM_BACK_TO_START_RECT, "Done", "MENU_BUTTON");
    btnBackToStart->SetEventObject(MSG_BUTTONCLICK, this, 
        (MSG_CALLBACK_OBJ)&MenuController::btnBackToStart_Click);
}

//------------------------------------------------------------------------------
// Name: UpdateBackgroundImage
// Desc: This routine updates the background image dependant upon the current
//       state of the menu controller and the current state of the game. If
//       a level is currently being played then the background shows the 
//       paused game screen, otherwise the background will display the
//       TITLE_MENU_PIC image.
//------------------------------------------------------------------------------
void MenuController::UpdateBackgroundImage()
{
    SDL_Surface *SubScreenImage;
    SDL_Rect SubScreenRect;

    if (BackgroundImage != NULL) {
        SDL_FreeSurface(BackgroundImage);
        BackgroundImage = NULL;
    }

    if (Activated) {
        if (G->GetGameState() == Game::STOPPED) {
            BackgroundImage = IMG_Load(TITLE_MENU_PIC);
        } else if (G->GetGameState() == Game::PAUSED 
            || G->GetGameState() == Game::ENDING_LEVEL) {
            BackgroundImage = CreateBackgroundImage();
        } else {
            assert(false);
        }

        // Some menus use external image files to display arrangements of non interactive
        // text and pictures called subscreens. These images should be of the 
        // same dimensions as the MenuArea.

        // Work out the position and size of a subscreen
        SubScreenRect.y = 130;
        SubScreenRect.w = MenuArea->my_width;
        SubScreenRect.h = MenuArea->my_height;
        SubScreenRect.x = (BackgroundImage->w - SubScreenRect.w) / 2;
        if (BackgroundImage->h > 600) {
            SubScreenRect.y += (BackgroundImage->h - 600) / 2;
        }

        if (CurMenu == HELP_MENU) {
            // If we are currently displaying help to the user add the appropriate
            // help screen to the background.

            SubScreenImage = IMG_Load(HELP_SCREEN_PICS[CurHelpScreen]);
            if (SubScreenImage == NULL) return;
            assert(SubScreenImage->w == SubScreenRect.w 
                && SubScreenImage->h == SubScreenRect.h);
            
            SDL_SetAlpha(SubScreenImage, SDL_SRCALPHA, SDL_ALPHA_TRANSPARENT);
            SDL_BlitSurface(SubScreenImage, NULL, BackgroundImage, &SubScreenRect);
            SDL_FreeSurface(SubScreenImage);
        } else if (CurMenu == INTROSCREEN_MENU 
            && G->GetCurLevel()->IntroScreen != NULL) {

            SubScreenImage = G->GetCurLevel()->IntroScreen;
            // If a level has an introscreen then display it to the user
            assert(SubScreenImage->w == SubScreenRect.w 
                && SubScreenImage->h == SubScreenRect.h);

            SDL_BlitSurface(SubScreenImage, NULL, BackgroundImage, &SubScreenRect);
        }
    }
}

//------------------------------------------------------------------------------
// Name: CreateBackgroundImage
// Desc: This routine creates the background image from the current game frame.
//------------------------------------------------------------------------------
SDL_Surface *MenuController::CreateBackgroundImage()
{
    SDL_Surface *BackgroundImage;
    SDL_Surface *Filter;
    PG_Rect TargetRect(0, 0, SDLI->GetScreenWidth(), SDLI->GetScreenHeight());
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    uint rmask = 0xff000000;
    uint gmask = 0x00ff0000;
    uint bmask = 0x0000ff00;
    uint amask = 0x000000ff;
#else
    uint rmask = 0x000000ff;
    uint gmask = 0x0000ff00;
    uint bmask = 0x00ff0000;
    uint amask = 0xff000000;
#endif

    BackgroundImage = SDL_CreateRGBSurface(SDL_SWSURFACE, 
        SDLI->GetScreenWidth(), SDLI->GetScreenHeight(), 
        CAM_INTERNAL_SURFACE_BPP, rmask, gmask, bmask, amask);
    if (BackgroundImage != NULL) {
        // The background image is drawn to the screen first because ParaGUI 
        // objects use static variables to store the screen and don't seem to 
        // respond to the screen being changed. So I couldn't just tell ParaGUI to 
        // draw straight to BackgroundImage.
        G->DrawGameFrame(SDLI->GetScreen());
        G->DrawUserInterface(SDLI->GetScreen());

        // Draw the orders panel if it is active
        if (OP->GetActivated()) {
            OP->SetVisible(true);
	        PG_Application::LockScreen();
	        PG_Widget::GetWidgetList()->Intersect((PG_Rect*)&TargetRect).Blit(TargetRect);
	        PG_Application::UnlockScreen();
            OP->SetVisible(false);
        }

        // Fade out the background a bit
        Filter = SDL_CreateRGBSurface(SDL_SWSURFACE, 
            SDLI->GetScreenWidth(), SDLI->GetScreenHeight(), 
            CAM_INTERNAL_SURFACE_BPP, rmask, gmask, bmask, amask);
        if (Filter != NULL) {
            SDL_FillRect(Filter, NULL, SDL_MapRGBA(Filter->format, 0, 0, 0, 128));
            SDL_BlitSurface(Filter, NULL, SDLI->GetScreen(), NULL);
            SDL_FreeSurface(Filter);
        }
        SDL_BlitSurface(SDLI->GetScreen(), NULL, BackgroundImage, NULL);
        SDL_SetAlpha(BackgroundImage, SDL_RLEACCEL, SDL_ALPHA_OPAQUE);
    }
   
    return BackgroundImage;
}

//------------------------------------------------------------------------------
// Name: GetInstance
// Desc: Returns or creates the single instance of this class
//------------------------------------------------------------------------------
MenuController *MenuController::GetInstance(Game *G, PG_Application *PGApp)
{
    if (Instance == NULL) {
        Instance = new MenuController(G, PGApp);
    }
    return Instance; 
}

//------------------------------------------------------------------------------
// Name: DrawBackground
// Desc: Draw some background for the menu screen. For now this subroutine
//       just blanks the screen, but hopefully in the future it will show
//       something a bit more interesting.
//------------------------------------------------------------------------------
void MenuController::DrawBackground()
{
    if (BackgroundImage == NULL) {
        SDL_FillRect(SDLI->GetScreen(), NULL, 
            SDL_MapRGB(SDLI->GetScreen()->format, 192, 192, 192));
    } else {
        SDL_Rect SourceRect = BackgroundImage->clip_rect;
        SDL_Rect TargetRect = SDLI->GetScreen()->clip_rect;

        if (SourceRect.w < TargetRect.w || SourceRect.h < TargetRect.h) {
            SDL_FillRect(SDLI->GetScreen(), NULL, 
                SDL_MapRGB(SDLI->GetScreen()->format, 192, 192, 192));
        }

        if (SourceRect.w > TargetRect.w) {
            SourceRect.x = (SourceRect.w - TargetRect.w) / 2;
            SourceRect.w = TargetRect.w;
        } else if (SourceRect.w < TargetRect.w) {
            TargetRect.x = (TargetRect.w - SourceRect.w) / 2;
            TargetRect.w = SourceRect.w;
        }
        if (SourceRect.h > TargetRect.h) {
            SourceRect.h = TargetRect.h;
        } else if (SourceRect.h < TargetRect.h) {
            TargetRect.y = (TargetRect.h - SourceRect.h) / 2;
            TargetRect.h = SourceRect.h;
        }

        SDL_BlitSurface(BackgroundImage, &SourceRect, 
            SDLI->GetScreen(), &TargetRect);
    }
}

//------------------------------------------------------------------------------
// Name: SetActivated
// Desc: Either activates or deactivates the menu controller. When the menu is
//       activated an appropriate menu is chosden depending upon the current
//       state of the game.
//------------------------------------------------------------------------------
void MenuController::SetActivated(bool Activated)
{
    this->Activated = Activated;
    
    UpdateBackgroundImage();
    if (Activated) {
        if (G->GetGameState() == Game::STOPPED) {
            NewMenu = START_MENU;
        } else if (G->GetGameState() == Game::PAUSED) {
            NewMenu = INGAME_MENU;
        } else if (G->GetGameState() == Game::ENDING_LEVEL) {
            // This rather strange bit of code, gives us the faded out game
            // screen as the background to the new high score menu and
            // allows the title screen to be drawn as the background if there
            // is no new high score.
            //SetCurMenu(INGAME_MENU);
            NewMenu = NEW_HIGH_SCORE_MENU;
        } else {
            assert(false);
        }
        CurMenuInvalid = true;
        Update();
    } else {
        // Hide the current menu
        MenuArea->Hide();
    }
}

//------------------------------------------------------------------------------
// Name: Update
// Desc: This small routine simply checks to see if a new menu has been 
//       requested. if so then the current menu is set to match the new menu.
//       This is a bit of a botch, the reason that the event handler for a 
//       menu item cannot change the current menu is that this would cause the 
//       destruction of the ParaGUI widget that called the event handler. This 
//       understandably causes ParaGUI to panic. 
//------------------------------------------------------------------------------
void MenuController::Update()
{
    if (CurMenuInvalid) SetCurMenu(NewMenu);

    if (CurMenu == INTROSCREEN_MENU 
        && G->GetCurLevel()->IntroScreen == NULL) {

        // If there is no introscreen to display then just go straight to the game.
        G->StartNewGame();
    }
    if (btnTooltip != NULL && strcmp(btnTooltip->GetText(), "") == 0) {
        btnTooltip->SetVisible(false);
    }
}

//------------------------------------------------------------------------------
// Event Handlers
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Name: btnStartGame_Click
// Desc: Lets the player pick the level that they want to play from the files
//       in the CAM_LEVELS_DIR directory.
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::btnStartGame_Click)
{
    NewMenu = PICK_LEVEL_MENU;
    CurMenuInvalid = true;
    return true;
}

//------------------------------------------------------------------------------
// Name: btnLevel_Click
// Desc: Once a level has been chosen, it is loaded up and its 'intro screen'
//       (if it has one) is displayed.
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::btnLevel_Click)
{
    string Filename = string(CAM_LEVELS_DIR) + "/" + *((string *)clientdata);

    G->LoadLevel(Filename.c_str());
    NewMenu = INTROSCREEN_MENU;
    CurMenuInvalid = true;
    return true;
}

//------------------------------------------------------------------------------
// Name: btnOptions_Click
// Desc: Opens up the options menu
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::btnOptions_Click)
{
    NewMenu = OPTIONS_MENU;
    CurMenuInvalid = true;
    return true;
}

//------------------------------------------------------------------------------
// Name: btnVideo_Click
// Desc: Opens up the video options menu
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::btnVideo_Click)
{
    NewMenu = VIDEO_OPTIONS_MENU;
    CurMenuInvalid = true;
    return true;
}

//------------------------------------------------------------------------------
// Name: AlterFullscreen
// Desc: Updates the FULLSCREEN option
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::AlterFullscreen)
{
    if (((OptionsCheckButton *)widget)->GetPressed()) {
        G->SetOption(Game::FULLSCREEN, 1);
    } else {
        G->SetOption(Game::FULLSCREEN, 0);
    }
    return true;
}

//------------------------------------------------------------------------------
// Name: AlterResolution
// Desc: Updates the SCREENMODE option
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::AlterResolution)
{
    int i;
    string NewResolution = ((OptionsDropDown *)widget)->GetText();

    for (i = 0; i < SDL_Interface::NUM_SCREEN_MODES; i++) {
        if (SDL_Interface::SCREEN_MODE_DETAILS[i].String == NewResolution) {
            G->SetOption(Game::SCREENMODE,  i);
            return true;
        }
    }

    assert(false);  // Resolution not found!?!
    return false;
}

//------------------------------------------------------------------------------
// Name: AlterBitdepth
// Desc: Updates the BITDEPTH option
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::AlterBitdepth)
{
    int i;
    int NewBitdepth;

    assert("augh -a james");    
    //stringstream(string(((OptionsDropDown *)widget)->GetText())) >> NewBitdepth;

    for (i = 0; i < SDL_Interface::NUM_SCREEN_BITDEPTHS; i++) {
        if (SDL_Interface::SCREEN_BITDEPTH_DETAILS[i] == NewBitdepth) {
            G->SetOption(Game::BITDEPTH,  i);
            return true;
        }
    }

    assert(false);  // Bitdepth not found!?!
    return false;
}

//------------------------------------------------------------------------------
// Name: AlterNumParallaxLayers
// Desc: Updates the NUM_PARALLAX_LAYERS option
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::AlterNumParallaxLayers)
{
    G->SetOption(Game::NUM_PARALLAX_LAYERS, 
        ((PG_Slider *)widget)->GetPosition());
    ((PG_Slider *)widget)->SetPosition(((PG_Slider *)widget)->GetPosition());
    return true;
}

//------------------------------------------------------------------------------
// Name: btnAudio_Click
// Desc: Opens up the audio options menu
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::btnAudio_Click)
{
    NewMenu = AUDIO_OPTIONS_MENU;
    CurMenuInvalid = true;
    return true;
}

//------------------------------------------------------------------------------
// Name: AlterSoundVolume
// Desc: Updates the SFX_VOLUME option
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::AlterSoundVolume)
{
    SDLI->SetSFX_Volume(((PG_Slider *)widget)->GetPosition());
    return true;
}

//------------------------------------------------------------------------------
// Name: AlterMusicVolume
// Desc: Updates the MUSIC_VOLUME option
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::AlterMusicVolume)
{
    SDLI->SetMusic_Volume(((PG_Slider *)widget)->GetPosition());
    return true;
}

//------------------------------------------------------------------------------
// Name: AlterReverseStereo
// Desc: Updates the REVERSE_STEREO option
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::AlterReverseStereo)
{
    SDLI->SetReverseStereo(((OptionsCheckButton *)widget)->GetPressed());
    return true;
}

//------------------------------------------------------------------------------
// Name: btnControls_Click
// Desc: Opens up the controls options menu
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::btnControls_Click)
{
    NewMenu = CONTROLS_OPTIONS_MENU;
    CurMenuInvalid = true;
    return true;
}

//------------------------------------------------------------------------------
// Name: PickNewKey
// Desc: Lets the user pick a new key for a control
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::PickNewKey)
{
    int i;

    IH->SetCommandKey((int)clientdata);
    for (i = 0; i < InputHandler::NUM_COMMANDS; i++) {
        btnChangeCommand[i]->SetText(
            GetKeyName(InputHandler::COMMAND_DETAILS[i].Option));
    }
    return true;
}

//------------------------------------------------------------------------------
// Name: AlterClockControls
// Desc: Updates the CLOCK_CONTROLS option
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::AlterClockControls)
{
    if (((OptionsCheckButton *)widget)->GetPressed()) {
        G->SetOption(Game::CLOCK_CONTROLS, 1);
    } else {
        G->SetOption(Game::CLOCK_CONTROLS, 0);
    }
    return true;
}

PARAGUI_CALLBACK(MenuController::btnContinue_Click)
{
    G->StartNewGame();
    return true;
}

PARAGUI_CALLBACK(MenuController::btnHelp_Click)
{
    NewMenu = HELP_MENU;
    CurHelpScreen = 0;
    CurMenuInvalid = true;
    
    return true;
}

PARAGUI_CALLBACK(MenuController::btnContinueGame_Click)
{
    G->ContinueGame();
    return true;
}

PARAGUI_CALLBACK(MenuController::btnQuitGame_Click)
{
    G->QuitGame();
    NewMenu = NEW_HIGH_SCORE_MENU;
    CurMenuInvalid = true;
    return true;
}

PARAGUI_CALLBACK(MenuController::btnExit_Click)
{
    G->Exit();
    return true;
}

PARAGUI_CALLBACK(MenuController::btnBackToStart_Click) 
{
    NewMenu = START_MENU;
    CurMenuInvalid = true;
    return true;
}

PARAGUI_CALLBACK(MenuController::btnHighScores_Click)
{
    NewMenu = HIGH_SCORES_MENU;
    CurMenuInvalid = true;
    return true;
}

PARAGUI_CALLBACK(MenuController::tbxName_EditEnd)
{
    Game::HIGH_SCORE NewHighScore;
    
    // Create a new high score and add it to the table
    strncpy(NewHighScore.Name, widget->GetText(), Game::MAX_HIGH_SCORE_NAME_LENGTH+1);
    strncpy(NewHighScore.LevelName, 
        G->GetLevelName().c_str(), Level::MAX_LEVEL_NAME_LENGTH+1);
    NewHighScore.Score = G->GetFinalScore();
    G->SetHighScore(Game::NUM_HIGH_SCORES-1, NewHighScore);

    NewMenu = HIGH_SCORES_MENU;
    CurMenuInvalid = true;
    return true;
}

PARAGUI_CALLBACK(MenuController::btnPrevHelp_Click)
{
    CurHelpScreen--;
    CurMenuInvalid = true;
    return true;
}
    
//------------------------------------------------------------------------------
// Name: btnDoneWithGameMenu_Click
// Desc: Called by menus that need to return to either the start menu or the 
//       ingame menu when they have finished doing their funky thing.
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(MenuController::btnDoneWithGameMenu_Click)
{
    if (G->GetGameState() == Game::STOPPED) {
        NewMenu = START_MENU;
    } else if (G->GetGameState() == Game::PAUSED) {
        NewMenu = INGAME_MENU;
    } else {
        assert(false);
    }
    CurMenuInvalid = true;
    return true;
}

PARAGUI_CALLBACK(MenuController::btnNextHelp_Click)
{
    CurHelpScreen++;
    CurMenuInvalid = true;
    return true;
}

//------------------------------------------------------------------------------
// OptionsSlider Implementation
//------------------------------------------------------------------------------
const int OptionsSlider::BORDER_SIZE = 1;

OptionsSlider::OptionsSlider(PG_Widget *parent, int id, const PG_Rect& r, 
    PG_Widget *TooltipDisplay, int direction,  
    const char* style) : PG_Widget(parent, r, false)
{
    PG_Rect SliderRect = r;

    assert(TooltipDisplay != NULL);

    G = Game::GetInstance();

    SetID(id);
    // Change the slider rect so that it will fit inside this widget but with
    // a BORDER_SIZE pixel border.
    SliderRect.my_xpos = SliderRect.my_ypos = BORDER_SIZE;
    SliderRect.my_width -= 2*BORDER_SIZE;
    SliderRect.my_height -= 2*BORDER_SIZE;
    TheSlider = new PG_Slider(this, 0, SliderRect, direction, style);
    this->TooltipDisplay = TooltipDisplay;
}

OptionsSlider::~OptionsSlider() {}

void OptionsSlider::eventMouseEnter()
{
    TooltipDisplay->SetText(G->GetOptionTooltip(GetID()));
    TooltipDisplay->SetVisible(true);
    PG_Widget::eventMouseEnter();
}

void OptionsSlider::eventMouseLeave()
{
    if (!TheSlider->IsMouseInside()) {
        TooltipDisplay->SetText("");
        TooltipDisplay->SetVisible(false);
    }
    PG_Widget::eventMouseLeave();
}

//------------------------------------------------------------------------------
// OptionsCheckButton Implementation
//------------------------------------------------------------------------------
OptionsCheckButton::OptionsCheckButton(PG_Widget *parent, Game::GAME_OPTIONS id, const PG_Rect& r, 
    PG_Widget *TooltipDisplay, const char* text,
    const char* style) : PG_CheckButton(parent, id, r, text, style)
{
    assert(TooltipDisplay != NULL);

    G = Game::GetInstance();
    this->TooltipDisplay = TooltipDisplay;
}

OptionsCheckButton::~OptionsCheckButton() {}

void OptionsCheckButton::eventMouseEnter()
{
    TooltipDisplay->SetText(G->GetOptionTooltip(GetID()));
    TooltipDisplay->SetVisible(true);
    PG_CheckButton::eventMouseEnter();
}

void OptionsCheckButton::eventMouseLeave()
{
    TooltipDisplay->SetText("");
    TooltipDisplay->SetVisible(false);
    PG_CheckButton::eventMouseLeave();
}

//------------------------------------------------------------------------------
// OptionsDropDown Implementation
//------------------------------------------------------------------------------
const int OptionsDropDown::SCROLL_LIMIT = 4;

OptionsDropDown::OptionsDropDown(PG_Widget *parent, int id, const PG_Rect& r, 
    PG_Widget *TooltipDisplay, const char* style) : PG_DropDown(parent, id, r, style)
{
    assert(TooltipDisplay != NULL);

    G = Game::GetInstance();
    this->TooltipDisplay = TooltipDisplay;

    Style = style;
    LoadThemeStyle(style);
    my_DropList->SetEventObject(MSG_SELECTITEM, this, (MSG_CALLBACK_OBJ)&OptionsDropDown::NewSelectHandler);
    //my_DropButton->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&MenuController::DropDownButtonPressed);
}

OptionsDropDown::~OptionsDropDown() {}

//------------------------------------------------------------------------------
// Name: AddItem
// Desc: Adapted from PG_DropDown::AddItem. The main change it makes is to
//       alter the height of the listbox used to display the items.
//------------------------------------------------------------------------------
void OptionsDropDown::AddItem(const char* Text, void* UserData) 
{
	Uint16 NewHeight;
    PG_ListBoxItem *NewItem = new PG_ListBoxItem(10, Text, NULL, UserData, Style.c_str());

    // This is a messy way of getting round a bug in ParaGUI
    NewHeight = NewItem->GetFontHeight();
    AddChild(NewItem);
    RemoveChild(NewItem);
    delete NewItem;
    NewItem = new PG_ListBoxItem(NewHeight, Text, NULL, UserData, Style.c_str());
    
	my_DropList->AddItem(NewItem);
    
    // Do some very rough sizing of the DropList, assuming that all items are
    // of equal height
    if (my_DropList->GetWidgetCount() >= SCROLL_LIMIT) {
        my_DropList->SizeWidget(my_DropList->Width(), NewHeight*SCROLL_LIMIT);
    } else {
        my_DropList->SizeWidget(my_DropList->Width(), 
            NewHeight*my_DropList->GetWidgetCount());
    }
}

void OptionsDropDown::eventMouseEnter()
{
    TooltipDisplay->SetText(G->GetOptionTooltip(GetID()));
    TooltipDisplay->SetVisible(true);
    PG_DropDown::eventMouseEnter();
}

void OptionsDropDown::eventMouseLeave()
{
    TooltipDisplay->SetText("");
    TooltipDisplay->SetVisible(false);
    PG_DropDown::eventMouseLeave();
}

//------------------------------------------------------------------------------
// Name: NewSelectHandler
// Desc: This allows objects external to the dropdown to know that an item has
//       been selected.
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(OptionsDropDown::NewSelectHandler)
{
    select_handler(id, widget, data, clientdata);
    SendMessage(NULL, MSG_SELECTITEM, GetID(), (unsigned long)this);
    return true;
}

//------------------------------------------------------------------------------
// OptionsButton Implementation
//------------------------------------------------------------------------------
OptionsButton::OptionsButton(PG_Widget *parent, int id, const PG_Rect& r, 
    PG_Widget *TooltipDisplay, const char* text,
    const char* style) : PG_Button(parent, id, r, text, style)
{
    assert(TooltipDisplay != NULL);

    G = Game::GetInstance();
    this->TooltipDisplay = TooltipDisplay;
}

OptionsButton::~OptionsButton() {}

void OptionsButton::eventMouseEnter()
{
    TooltipDisplay->SetText(G->GetOptionTooltip(GetID()));
    TooltipDisplay->SetVisible(true);
    PG_Button::eventMouseEnter();
}

void OptionsButton::eventMouseLeave()
{
    TooltipDisplay->SetText("");
    TooltipDisplay->SetVisible(false);
    PG_Button::eventMouseLeave();
}
