/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// Name: OrdersPanel.cpp
// Desc: This is the means by which the player can control the troops in his
//       army that are not under his direct control, i.e. everything other than
//       his plane. It is called up during play by pressing the 'Issue Orders'
//       key and then presents a GUI that can be controlled with the mouse.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <list>
#include <assert.h>
#include "globals.h"
#include "CamelUtility.h"
#include "Camel.h"
#include "CamelMath.h"
#include "Game.h"
#include "InputHandler.h"
#include "Sprites.h"
#include "Level.h"
#include "Landscape.h"
#include "Entity.h"
#include "ControlledEntity.h"
#include "OrdersPanel.h"

// Paragui includes
#include "paragui.h"
#include "pgapplication.h"
#include "pgwindow.h"
#include "pgeventobject.h"
#include "pgwidget.h"
#include "pgwidgetlist.h"
#include "pgbutton.h"
#include "pglabel.h"
#include "pgprogressbar.h"

using namespace std;

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------

// Constants used to determine the size of various components
const int OrdersPanel::PANEL_WIDTH = 320;
const int OrdersPanel::PANEL_HEIGHT = 200;
const int OrdersPanel::HORIZONTAL_MARGIN = 20;
const int OrdersPanel::VERTICAL_MARGIN = 20;

const int OrdersPanel::GOOD_TAB_ID = 0;
const int OrdersPanel::EVIL_TAB_ID = 1;
const PG_Rect OrdersPanel::GOOD_TAB_BUTTON_RECT = PG_Rect(10, 10, 70, 25);
const PG_Rect OrdersPanel::EVIL_TAB_BUTTON_RECT = PG_Rect(80, 10, 70, 25);

const PG_Rect OrdersPanel::UNIT_LIST_RECT = PG_Rect(10, 40, 76, 150);
const int OrdersPanel::UNIT_LIST_H_MARGIN = 5;
const int OrdersPanel::UNIT_LIST_V_MARGIN = 5;
const PG_Rect OrdersPanel::UNIT_BUTTON_RECT = 
    PG_Rect(UNIT_LIST_H_MARGIN, UNIT_LIST_V_MARGIN, 50, 50);
const PG_Rect OrdersPanel::UNIT_INFO_RECT = PG_Rect(90, 40, 220, 120);
const PG_Rect OrdersPanel::BUTTON_CLOSE_RECT = PG_Rect(245, 165, 65, 25);

const PG_Rect OrdersPanel::UI_TITLE_LABEL_RECT = PG_Rect(5, 5, 210, 20);
const PG_Rect OrdersPanel::UI_NUMALIVE_LABEL_RECT = PG_Rect(5, 35, 100, 20);
const PG_Rect OrdersPanel::UI_NUMALIVE_RECT = PG_Rect(85, 35, 100, 20);
const PG_Rect OrdersPanel::UI_MORALE_LABEL_RECT = PG_Rect(5, 60, 100, 20);
const PG_Rect OrdersPanel::UI_MORALEBAR_RECT = PG_Rect(85, 60, 100, 20);
const PG_Rect OrdersPanel::UI_ORDERS_LABEL_RECT = PG_Rect(5, 85, 100, 20);
const PG_Rect OrdersPanel::UI_ORDERS_RECT = PG_Rect(85, 85, 100, 20);
const PG_Rect OrdersPanel::UI_HOLD_BUTTON_RECT = PG_Rect(155, 95, 65, 25);

// The speed at which the viewport scrolls (pixels per second)
const float OrdersPanel::VP_SCROLL_SPEED = 600.0f;
// The distance from a screen edge where the mouse induces scrolling
const int OrdersPanel::SCROLL_MARGIN = 20;

//------------------------------------------------------------------------------
// OrdersPanel Implementation
//------------------------------------------------------------------------------
OrdersPanel *OrdersPanel::Instance = NULL;

//------------------------------------------------------------------------------
// Name: OrdersPanel
// Desc: This sets up the OrdersPanel object and creates all the Paragui 
//       widgets that will be used to display the panel.
//------------------------------------------------------------------------------
OrdersPanel::OrdersPanel(Game *G, PG_Application *PGApp)
{
    assert(G != NULL && PGApp != NULL);

    this->G = G;
    this->PGApp = PGApp;
    SDLI = SDL_Interface::GetInstance(G);
    IH = InputHandler::GetInstance(G);

    CurUnit = NULL;

    // Place the panel in the top right corner of the screen
    DisplayRect.my_xpos = SDLI->GetScreenWidth() - (PANEL_WIDTH + HORIZONTAL_MARGIN);
    DisplayRect.my_ypos = VERTICAL_MARGIN;
    DisplayRect.my_width = PANEL_WIDTH;
    DisplayRect.my_height = PANEL_HEIGHT;
    Display = new PG_Window(NULL, DisplayRect, "", 0, "Window", 0);

    // Code to set up the panel

    // Create tab buttons to switch between the two sides
    btnGoodTab = new PG_Button(Display, GOOD_TAB_ID, GOOD_TAB_BUTTON_RECT, "Good", "ORDERSPANEL_BUTTON");
    btnGoodTab->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&OrdersPanel::btnTab_Click);
    btnGoodTab->SetToggle(true);
    btnGoodTab->SetPressed(true);
    btnEvilTab = new PG_Button(Display, EVIL_TAB_ID, EVIL_TAB_BUTTON_RECT, "Evil", "ORDERSPANEL_BUTTON");
    btnEvilTab->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&OrdersPanel::btnTab_Click);
    btnEvilTab->SetToggle(true);
    btnEvilTab->SetPressed(false);

    // Create a close button
    btnClose = new PG_Button(Display, 0, BUTTON_CLOSE_RECT, "Close", "ORDERSPANEL_BUTTON");
    btnClose->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&OrdersPanel::btnClose_Click);

    // The unit list displays all the controllabe units on the players side
    UnitList = new PG_WidgetList(Display, UNIT_LIST_RECT);
	UnitList->SetDirtyUpdate(false);
	UnitList->SetBackgroundBlend(0);
	// james UnitList->SetSimpleBackground(true);
			
	UnitList->EnableScrollBar(true, PG_SB_VERTICAL);
	UnitList->EnableScrollBar(true, PG_SB_HORIZONTAL);

    // The unit info area is where information about the current unit is displayed
    UnitInfo = new PG_Widget(Display, UNIT_INFO_RECT);

    Activated = false;
}

//------------------------------------------------------------------------------
// Name: ~OrdersPanel
// Desc: All cleanup done here
//------------------------------------------------------------------------------
OrdersPanel::~OrdersPanel() {}

//------------------------------------------------------------------------------
// Name: Update
// Desc: Handles anything that isn't taken care of by the ParaGUI window that
//       the orders panel displays. At the moment this mainly involves 
//       scrolling the viewpoint left and right when the player moves the mouse
//       near the edge of the screen.
//------------------------------------------------------------------------------
void OrdersPanel::Update(uint dt)
{
    Landscape *LS = G->GetCurLevel()->GetLandscape();
    float SecondsPassed = (float)dt / 1000.0f;
    uchar MouseState;
    Point2D MousePos;
    int MouseWorldX;

    // Get the state of the mouse
    MouseState = SDL_GetMouseState(&MousePos.x, &MousePos.y);

    // Scroll the viewport if the mouse is near a vertical edge of the screen
    if (MousePos.x <= SCROLL_MARGIN) {
        RealX -= VP_SCROLL_SPEED*SecondsPassed;
    } else if (MousePos.x >= LS->GetViewportWidth() - SCROLL_MARGIN) {
        RealX += VP_SCROLL_SPEED*SecondsPassed;
    }
    RealX = (float)ValidateViewportX((int)RealX);
    ViewportCentrePos.x = (int)RealX;
    ViewportCentrePos.y = LS->CalculatePixelHeight(ViewportCentrePos.x);

    // If a unit is currently selected and the user clicks outside the OrdersPanel
    // then the click is interpreted as an order to that unit. Units can however,
    // only be given orders when they are not fleeing.
    if (CurUnit != NULL && IH->MouseButtonPressed[0]
        && (MousePos.x < DisplayRect.my_xpos || MousePos.y < DisplayRect.my_ypos 
            || MousePos.x > DisplayRect.my_xpos + DisplayRect.my_width 
            || MousePos.y > DisplayRect.my_ypos + DisplayRect.my_height)) {

        MouseWorldX = LS->GetViewportWorldX() + MousePos.x;
        
        if (!CurUnit->IsFleeing()) {
            if (CurUnit->GetSide() == Entity::GOOD) {
                if (MouseWorldX < CurUnit->GetCentreX()) {
                    CurUnit->SetOrders(&ControlledEntity::WITHDRAW, MouseWorldX);
                } else {
                    CurUnit->SetOrders(&ControlledEntity::ADVANCE, MouseWorldX);
                }
            } 
#ifdef _DEBUG
            else {
                if (MouseWorldX > CurUnit->GetCentreX()) {
                    CurUnit->SetOrders(&ControlledEntity::WITHDRAW, MouseWorldX);
                } else {
                    CurUnit->SetOrders(&ControlledEntity::ADVANCE, MouseWorldX);
                }
            }
#endif // _DEBUG
            NeedToUpdateUnitInfo = true;
        }
    }
    if (NeedToUpdateUnitInfo) {
        RefreshUnitInfo();
        NeedToUpdateUnitInfo = false;
    }
}

//------------------------------------------------------------------------------
// Name: SetActivated
// Desc: Activates or deactivates the OrdersPanel. 
//------------------------------------------------------------------------------
void OrdersPanel::SetActivated(bool Activated)
{
    Landscape *LS = G->GetCurLevel()->GetLandscape();

    this->Activated = Activated;
    if (Activated) {
        // Update the state of the orders panel window
        SetCurUnit(NULL);
        RefreshUnitList();
        RefreshUnitInfo();

        // Sort out an alternative viewpoint
        ViewportCentrePos.x = LS->GetViewportCentreWorldX();
        RealX = (float)ViewportCentrePos.x;
        ViewportCentrePos.y = LS->CalculatePixelHeight(ViewportCentrePos.x);

        // Show the window
        SetVisible(true);
    } else {
        SetVisible(false);
    }
}

//------------------------------------------------------------------------------
// Name: SetVisible
// Desc: Hides or unhides the orders panel. This should only be called when
//       the order panel is active.
//------------------------------------------------------------------------------
void OrdersPanel::SetVisible(bool Visible)
{
    this->Visible = Visible;

    if (Activated && Visible) {
        Display->Show();
    } else {
        Display->Hide();
    }
}

//------------------------------------------------------------------------------
// Name: GetInstance
// Desc: Returns or creates the single instance of this class
//------------------------------------------------------------------------------
OrdersPanel *OrdersPanel::GetInstance(Game *G, PG_Application *PGApp)
{
    if (Instance == NULL) {
        Instance = new OrdersPanel(G, PGApp);
    }
    return Instance; 
}

//------------------------------------------------------------------------------
// Name: RefreshUnitList
// Desc: Fills the unit list with all of the units that the player controls.
//------------------------------------------------------------------------------
void OrdersPanel::RefreshUnitList()
{
    int i;
    Level *L = G->GetCurLevel();
    Entity::ENT_SIDE CurSide;
    list<Entity *>::iterator Iter;
    UnitButton *NewUnitButton;
    PG_Rect NewUnitButtonRect = UNIT_BUTTON_RECT;

    // Clear out widgets representing old units
    UnitList->DeleteAll();

    CurSide = (btnGoodTab->GetPressed() ? Entity::GOOD : Entity::EVIL);

    // Loop through all of the entity lists looking for lists that contain
    // ControlledEntities
    for (i = 0; i < Entity::NUM_ENT_TYPES; i++) {
        if (L->GameEntities[i].size() > 0 
            && dynamic_cast<ControlledEntity *>(*(L->GameEntities[i].begin()))) {
    
            for (Iter = L->GameEntities[i].begin();
                Iter != L->GameEntities[i].end(); Iter++) {
            
                assert((*Iter)->GetSide() != Entity::NEUTRAL);

                if ((*Iter)->GetSide() == CurSide) {
                    NewUnitButton = new UnitButton(NULL, 0, 
                        NewUnitButtonRect, ((ControlledEntity *)*Iter)->GetIconFrame());
                    NewUnitButton->SetToggle(true);
                    NewUnitButton->SetEventObject(MSG_BUTTONCLICK, 
                        this, (MSG_CALLBACK_OBJ)&OrdersPanel::btnUnit_Click, *Iter);

                    UnitList->AddWidget(NewUnitButton);
                    NewUnitButtonRect.my_ypos = 0;
                }
            }
        }
    }
    // Adds a dummy widget to make space at the bottom of the list
    UnitList->AddWidget(new PG_Widget(NULL, 
        PG_Rect(UNIT_LIST_H_MARGIN, UNIT_LIST_V_MARGIN, 0, 0)));
}

//------------------------------------------------------------------------------
// Name: RefreshUnitInfo
// Desc: Fills the UnitInfo widget with information about the currently
//       selected unit if there is one.
//------------------------------------------------------------------------------
void OrdersPanel::RefreshUnitInfo()
{
    PG_ProgressBar *MoraleBar;
    PG_Button *HoldButton;
    char Buffer[50];
    SDL_Color DisabledColour = { 128, 128, 128, 0 };

    UnitInfo->RemoveAllChilds();
    UnitInfo->AddChild(new PG_Label(UnitInfo, UI_TITLE_LABEL_RECT, 
        "Unit Information", "UI_TitleLabel"));

    if (CurUnit != NULL) {
        sprintf(Buffer, "%i", CurUnit->GetNumSubordinatesAlive());

        UnitInfo->AddChild(new PG_Label(UnitInfo, UI_NUMALIVE_LABEL_RECT, 
            "No. Alive:", "UI_InfoLabel"));
        UnitInfo->AddChild(new PG_Label(UnitInfo, UI_NUMALIVE_RECT, 
            Buffer, "UI_InfoLabel"));
        UnitInfo->AddChild(new PG_Label(UnitInfo, UI_MORALE_LABEL_RECT, 
            "Morale:", "UI_InfoLabel"));
        MoraleBar = new PG_ProgressBar(UnitInfo, UI_MORALEBAR_RECT);
        MoraleBar->SetProgress(CurUnit->GetMorale());
        UnitInfo->AddChild(MoraleBar);
        UnitInfo->AddChild(new PG_Label(UnitInfo, UI_ORDERS_LABEL_RECT, 
            "Orders:", "UI_InfoLabel"));
        UnitInfo->AddChild(new PG_Label(UnitInfo, UI_ORDERS_RECT, 
            CurUnit->GetOrders()->GetOrderString(), "UI_InfoLabel"));

        HoldButton = new PG_Button(UnitInfo, 0, UI_HOLD_BUTTON_RECT, "Hold", "ORDERSPANEL_BUTTON");
        HoldButton->SetEventObject(MSG_BUTTONCLICK, this, (MSG_CALLBACK_OBJ)&OrdersPanel::btnHold_Click);

        // Check to see if the hold button needs to be disabled
#ifdef _DEBUG
        if (CurUnit->IsFleeing()) {
#else
        if (CurUnit->IsFleeing() || CurUnit->GetSide() == Entity::EVIL) {
#endif
            HoldButton->EnableReceiver(false);
            HoldButton->SetFontColor(DisabledColour);
        }
        UnitInfo->AddChild(HoldButton);
    }
    UnitInfo->Hide();
    UnitInfo->Show();
}

//------------------------------------------------------------------------------
// Name: ValidateViewportX
// Desc: This small helper function checks that the proposed viewport x
//       value will be able to appear on the centre of the screen. If 'x' is
//       valid then it is returned unchanged, otherwise, the nearest valid x is
//       returned.
//------------------------------------------------------------------------------
int OrdersPanel::ValidateViewportX(int x)
{
    Landscape *LS = G->GetCurLevel()->GetLandscape();
    int ViewportWidth_2 = LS->GetViewportWidth() / 2;
    int MinViewportX = ViewportWidth_2;
    int MaxViewportX = LS->GetWorldPixelWidth() - ViewportWidth_2;

    if (x < MinViewportX) {
        return MinViewportX;
    } else if (x > MaxViewportX) {
        return MaxViewportX;
    } else {
        return x;
    }
}

//------------------------------------------------------------------------------
// Name: SetCurUnit
// Desc: Selects the provided unit and centres the screen on it.
//------------------------------------------------------------------------------
void OrdersPanel::SetCurUnit(ControlledEntity *AUnit)
{
    Landscape *LS = G->GetCurLevel()->GetLandscape();
    CurUnit = AUnit;

    if (AUnit != NULL) {
        ViewportCentrePos.x = ValidateViewportX(AUnit->GetCentreX());
        RealX = (float)ViewportCentrePos.x;
        ViewportCentrePos.y = LS->CalculatePixelHeight(ViewportCentrePos.x);
    }
    RefreshUnitInfo();
}

//------------------------------------------------------------------------------
// Name: GetViewportCentrePos
// Desc: When the orders panel is active, it provides an alternative viewpoint
//       which allows the player to scroll around the battle field and review
//       their troops. The main loop of the game object uses this viewpoint
//       when the orders panel is active but it should not be used at any other
//       time.
//------------------------------------------------------------------------------
Point2D OrdersPanel::GetViewportCentrePos()
{
    assert(Activated);
    return ViewportCentrePos;
}

//------------------------------------------------------------------------------
// Name: btnTab_Click
// Desc: The tab buttons at the top of the form switch the unit list between
//       the forces of good and evil. The player may control the good army but
//       only look at the evil army.
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(OrdersPanel::btnTab_Click)
{
    switch (id) {
        case GOOD_TAB_ID:
            btnGoodTab->SetPressed(true);
            btnEvilTab->SetPressed(false);
            RefreshUnitList();
            if (CurUnit != NULL && CurUnit->GetSide() == Entity::EVIL) {
                SetCurUnit(NULL);
            }
            break;
        case EVIL_TAB_ID:
            btnGoodTab->SetPressed(false);
            btnEvilTab->SetPressed(true);
            RefreshUnitList();
            if (CurUnit != NULL && CurUnit->GetSide() == Entity::GOOD) {
                SetCurUnit(NULL);
            }
            break;
    }
    return true;
}

//------------------------------------------------------------------------------
// Name: btnClose_Click
// Desc: When the close button is clicked, the orders panel is closed down and
//       gameplay resumes.
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(OrdersPanel::btnClose_Click)
{
    G->CloseOrdersPanel();
    return true;
}

//------------------------------------------------------------------------------
// Name: btnUnit_Click
// Desc: This is called when a unit button is clicked. If the unit button is
//       on before the click is made then it is turned off and its unit
//       deselected. Otherwise, the unit is selected, information about the 
//       unit is displayed and the viewport is centred on the unit.
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(OrdersPanel::btnUnit_Click)
{
    int i;
    PG_Widget *CurWidget;

    // Loop through all of the unit buttons in the unit list
    for (i = 0; i < UnitList->GetWidgetCount() - 1; i++) {

        CurWidget = UnitList->FindWidget(i);
        if (CurWidget == widget) {
            if (CurUnit == (ControlledEntity *)clientdata) {
                // If already pressed then unpress the button
                SetCurUnit(NULL);
                ((PG_Button *)widget)->SetPressed(false);
            } else {
                SetCurUnit((ControlledEntity *)clientdata);
                ((PG_Button *)widget)->SetPressed(true);
            }
        } else {
            // Unpress all other buttons
            ((PG_Button *)CurWidget)->SetPressed(false);
        }
    }

    return true;
}

//------------------------------------------------------------------------------
// Name: btnHold_Click
// Desc: This instructs a unit to hold its ground rather than wandering off
//       somewhere else.
//------------------------------------------------------------------------------
PARAGUI_CALLBACK(OrdersPanel::btnHold_Click)
{
    assert(CurUnit != NULL);
    
    CurUnit->SetOrders(&ControlledEntity::HOLD);
    // This variable needs to be used (as opposed to just calling RefreshUnitInfo)
    // because calling RefreshUnitInfo here would delete the button that this 
    // function is attached to. This causes bad things to happen.
    NeedToUpdateUnitInfo = true;

    return true;
}

//------------------------------------------------------------------------------
// UnitButton Implementation
//------------------------------------------------------------------------------
UnitButton::UnitButton(PG_Widget *parent, int id, const PG_Rect& r, 
    int Frame, const char* text, const char* style) 
    : PG_Button(parent, id, r, text, style)
{
    assert(Frame >= 0 && Frame < Sprites::NUM_FRAMES);

    SPR = Sprites::GetInstance();
    this->Frame = Frame;
    SPR->GetFrameDims(Frame, &FrameWidth, &FrameHeight);
}

UnitButton::~UnitButton() {}

void UnitButton::eventBlit(SDL_Surface* srf, const PG_Rect& src, const PG_Rect& dst)
{
    int Shift;
    int dx, dy;
    PG_Rect ButtonRect = *this;
    PG_Rect DisplayRect;
    PG_Rect SrcRect, DstRect;

    PG_Button::eventBlit(srf, src, dst);

    // This code adapted from PG_Button::eventBlit
    Shift = (GetPressed() ? 1 : 0) * my_pressShift;

    dx = (ButtonRect.my_width - FrameWidth) >> 1;
    dy = (ButtonRect.my_height - FrameHeight) >> 1;

	DisplayRect.my_xpos = ButtonRect.my_xpos + dx + Shift;
	DisplayRect.my_ypos = ButtonRect.my_ypos + dy + Shift;
    DisplayRect.my_width = FrameWidth;
    DisplayRect.my_height = FrameHeight;

    GetClipRects(SrcRect, DstRect, DisplayRect);

	SPR->DrawFrame(Frame, my_srfScreen, DstRect.my_xpos, DstRect.my_ypos, SrcRect);
}
