/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: SDL_Interface.cpp
// Desc: Provides access to the SDL library, include this file instead of
//       SDL.h. The SDL_Interface object sets up and provides access to the 
//       screen as well as providing some limited (no scaling etc) software 
//       blitting functions which seem to give better performance under Win32
//       windowed mode.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <assert.h>
#include <math.h>
#include "globals.h"
#include "Camel.h"
#include "Game.h"
#include "CamelUtility.h"
#include "SDL_Interface.h"

//------------------------------------------------------------------------------
// Global Variables and Constants
//------------------------------------------------------------------------------
const int SDLI_ERROR_BASE = 0x20000100;

const int SDLI_ERROR_INVALID_PARAMETERS = SDLI_ERROR_BASE;
const int SDLI_ERROR_INCOMPATIBLE_SURFACES = SDLI_ERROR_BASE + 1;
const int SDLI_ERROR_CANNOT_ALLOCATE_MEMORY = SDLI_ERROR_BASE + 2;

const int SDLI_NUM_ERRORS = 3;

static char *SDLI_ErrorStrings[SDLI_NUM_ERRORS] = {
    "SDLI_ERROR_INVALID_PARAMETERS",
    "SDLI_ERROR_INCOMPATIBLE_SURFACES",
    "SDLI_ERROR_CANNOT_ALLOCATE_MEMORY"
};

static CU_ERRORMESSAGES SDLI_ErrorMessages = {
    SDLI_ERROR_BASE, SDLI_NUM_ERRORS, SDLI_ErrorStrings
};

const SDL_Interface::SCREEN_MODE_INFO 
SDL_Interface::SCREEN_MODE_DETAILS[NUM_SCREEN_MODES] = {
    { "640x480", 640, 480 },
    { "800x600", 800, 600 },
    { "1024x768", 1024, 768 }
};

const int SDL_Interface::SCREEN_BITDEPTH_DETAILS[NUM_SCREEN_BITDEPTHS] = {
    16, 32
};

const int SDL_Interface::FREQUENCY = MIX_DEFAULT_FREQUENCY;
const int SDL_Interface::FORMAT = MIX_DEFAULT_FORMAT;
const int SDL_Interface::NUM_SOUND_CHANNELS = 2;      // Stereo sound
const int SDL_Interface::CHUNKSIZE = 1024;

SDL_Interface::SOUND_FILE_INFO 
SDL_Interface::SampleFiles[SDL_Interface::NUM_SAMPLES] = {
    { CAM_PLANE_ENGINE_WAV, 6.0 },
    { CAM_PLANE_ENGINE_WAV, 5.0 },
    { CAM_PLANE_ENGINE_WAV, 4.0 },
    { CAM_PLANE_ENGINE_WAV, 3.0 },
    { CAM_PLANE_ENGINE_WAV, 2.5 },
    { CAM_PLANE_ENGINE_WAV, 2.0 },
    { CAM_PLANE_ENGINE_WAV, 1.5 },
    { CAM_PLANE_ENGINE_WAV, 1.0 },    
    { CAM_PLANE_ENGINE_WAV, 0.85 },      
    { CAM_PLANE_ENGINE_WAV, 0.75 },
    { CAM_PLANE_ENGINE_WAV, 0.65 },
    { CAM_PLANE_ENGINE_WAV, 0.5 }, 
    { CAM_PLANE_ENGINE_WAV, 0.4 },
    { CAM_PLANE_ENGINE_WAV, 0.35 },
    { CAM_PLANE_ENGINE_WAV, 0.3125 },
    { CAM_PLANE_ENGINE_WAV, 0.3125 },
    { CAM_PLANE_ENGINE_WAV, 0.25 },
    { CAM_PLANE_GUNFIRE_WAV, 1.0 },
    { CAM_BOMB_EXPLOSION_WAV, 1.0 },
    { CAM_MENU_CLICK_WAV, 1.0 },
    { SOLDIER_GUN_0_WAV, 1.0 },
    { SOLDIER_GUN_1_WAV, 1.0 },
    { SOLDIER_GUN_2_WAV, 1.0 },
    { SOLDIER_GUN_3_WAV, 1.0 },
    { SOLDIER_GUN_4_WAV, 1.0 },
    { SOLDIER_GUN_5_WAV, 1.0 },
    { SOLDIER_GUN_6_WAV, 1.0 },
    { SOLDIER_SCREAM_0_WAV, 1.0 },
    { SOLDIER_SCREAM_1_WAV, 1.0 },
    { SOLDIER_SCREAM_2_WAV, 1.0 },
    { SOLDIER_SCREAM_3_WAV, 1.0 },
    { SOLDIER_SCREAM_4_WAV, 1.0 },
    { SOLDIER_SCREAM_5_WAV, 1.0 },
    { SOLDIER_SCREAM_6_WAV, 1.0 },
    { SOLDIER_SCREAM_7_WAV, 1.0 },
    { GAME_OVER_WAV, 1.0 },
    { YOU_WIN_WAV, 1.0 },
    { TANK_LOOP_0_WAV, 1.0 },
    { TANK_LOOP_1_WAV, 1.0 },
    { TANK_LOOP_2_WAV, 1.0 },
    { TANK_GUN_WAV, 1.0 }
};

// These constants are used to define groups of samples
const int SDL_Interface::SOLDIER_GUN_BASE = SOLDIER_GUN_0;
const int SDL_Interface::SOLDIER_GUN_END = SOLDIER_GUN_6;
const int SDL_Interface::SOLDIER_SCREAM_BASE = SOLDIER_SCREAM_0;
const int SDL_Interface::SOLDIER_SCREAM_END = SOLDIER_SCREAM_7;

//------------------------------------------------------------------------------
// SDL_Interface Implementation
//------------------------------------------------------------------------------
SDL_Interface *SDL_Interface::Instance = NULL;
SDL_Interface::SDL_Interface() {}

//------------------------------------------------------------------------------
// Name: SDL_Interface
// Desc: Sets up the required video mode and loads any fonts used in the
//       game. Also initialises SDL_mixer so that we can play sounds and make
//       beautiful music...
//------------------------------------------------------------------------------
SDL_Interface::SDL_Interface(Game *G) 
{
    int i;
    int BytesUsed = 0;

    CU_AddErrorMessages(SDLI_ErrorMessages);

    assert(G != NULL);
    this->G = G;

    if (SDL_FAILED(Mix_OpenAudio(FREQUENCY, FORMAT, NUM_SOUND_CHANNELS, CHUNKSIZE)))
        SDE_MixError(__FILE__, __LINE__);
    atexit(Mix_CloseAudio);

    // Allocate channels and load in samples
    Mix_AllocateChannels(NUM_CHANNELS);
    memset(Samples, NULL, NUM_SAMPLES*sizeof(Mix_Chunk *));
    for (i = 0; i < NUM_SAMPLES; i++) {
        Samples[i] = LoadScaledSample(SampleFiles[i].FileName,
            SampleFiles[i].ScalingFactor);
        // If we are unable to load in a particular sample then shutdown
        if (Samples[i] == NULL) SDE_SDLError(__FILE__, __LINE__);

        BytesUsed += Samples[i]->alen;
    }
    Mix_ChannelFinished(ChannelFinished);
    CU_Dbg_Printf(NULL, "%f megabytes used to store samples\n",
    CU_DBG    (float)BytesUsed / (1024*1024));

    // Set the initial values of the monitoring variables
    CurChannelsUsed = 0;
    MaxChannelsUsed = 0;

    // Set the volume of the channels and centre them
    SFX_Volume = G->GetOption(Game::SFX_VOLUME);
    Mix_Volume(-1, SFX_Volume);
    Mix_SetPanning(-1, CENTRAL_VOLUME, CENTRAL_VOLUME);

    ReverseStereo = G->GetOption(Game::REVERSE_STEREO) != 0;
    Mix_SetReverseStereo(MIX_CHANNEL_POST, ReverseStereo);

    // Set the music volume
    Music_Volume = G->GetOption(Game::MUSIC_VOLUME);
    Mix_VolumeMusic(Music_Volume);

    SetupVideo();

    // Load in the font used for non ParaGUI text
    MainFont = BFont_LoadFont(CAM_MAIN_FONT);
    if (MainFont == NULL) SDE_SDLError(__FILE__, __LINE__);
     
    SDL_SetAlpha(MainFont->Surface, 0, SDL_ALPHA_TRANSPARENT);
    SDL_SetColorKey(MainFont->Surface, SDL_SRCCOLORKEY, 
        SDL_MapRGBA(MainFont->Surface->format, 255, 255, 255, SDL_ALPHA_OPAQUE));

    // Turn off the hardware mouse
    SDL_ShowCursor(SDL_DISABLE);
}

//------------------------------------------------------------------------------
// Name: ~SDL_Interface
// Desc: Shuts down the SDL_Interface class
//------------------------------------------------------------------------------
SDL_Interface::~SDL_Interface() 
{
    int i;

    if (Screen != NULL) {
        SDL_FreeSurface(Screen);
        Screen = NULL;
    }

    // Stop any playback that may be going on
    Mix_HaltChannel(-1);
    Mix_HaltMusic();
    for (i = 0; i < NUM_SAMPLES; i++) {
        if (Samples[i] != NULL) {
            Mix_FreeChunk(Samples[i]);
            Samples[i] = NULL;
        }
    }
    
    // Turn the hardware mouse back on
    SDL_ShowCursor(SDL_ENABLE);

    CU_Dbg_Printf(NULL, "Maximum number of channels used: %i\n", MaxChannelsUsed);
}

//------------------------------------------------------------------------------
// Name: GetInstance
// Desc: Returns or creates the single instance of this class
//------------------------------------------------------------------------------
SDL_Interface *SDL_Interface::GetInstance(Game *G) 
{
    if (Instance == NULL) {
        Instance = new SDL_Interface(G);
    }
    return Instance; 
}

//------------------------------------------------------------------------------
// Name: SetupVideo
// Desc: Sets the video mode to one that is as close as possible to the one
//       requested in the options file. If a requested mode is not available
//       then this routine tests out other modes to see if they are suitable.
//       If an alternative is found then the user is informed and the game is 
//       run. If no alternative is found then an error message is displayed
//       and the game is halted.
//------------------------------------------------------------------------------
void SDL_Interface::SetupVideo()
{
    int i;
    bool VideoModeFound;
    SCREEN_MODES DesiredScreenMode;
    bool DesiredFullscreen;
    SCREEN_BITDEPTHS DesiredBitdepth;
    char VideoWarningMessageBuffer[1024];
    char *VideoWarningMessage = "Unable to use the video settings in your options "
        "file which requests %s mode at a resolution of %s and a bitdepth of %i "
        "bits. Using the closest alternative mode which is %s mode at a "
        "resolution of %s and a bitdepth of %i bits instead.";

    // Get the desired characteristics for the video mode
    DesiredScreenMode = ScreenMode = (SCREEN_MODES)G->GetOption(Game::SCREENMODE);
    DesiredFullscreen = Fullscreen = G->GetOption(Game::FULLSCREEN) != 0;
    DesiredBitdepth = Bitdepth = (SCREEN_BITDEPTHS)G->GetOption(Game::BITDEPTH);

    VideoModeFound = false;
    if (ScreenModeAccepted(ScreenMode, Fullscreen, Bitdepth)) {
        VideoModeFound = true;
    } else {
        for (i = 0; i < NUM_SCREEN_MODES; i++) {
            if (i != ScreenMode) {
                if (ScreenModeAccepted((SCREEN_MODES)i, Fullscreen, Bitdepth)) {
                    ScreenMode = (SCREEN_MODES)i;
                    VideoModeFound = true;
                    break;
                }
            }
        }
    }
    if (!VideoModeFound) {
        CU_ShowMessage("Error", "Unable to find a suitable video mode");
        exit(1);
    }
        
    if (Bitdepth != DesiredBitdepth 
        || Fullscreen != DesiredFullscreen
        || ScreenMode != DesiredScreenMode) {

        sprintf(VideoWarningMessageBuffer, VideoWarningMessage, 
            (DesiredFullscreen ? "fullscreen" : "windowed"),
            SCREEN_MODE_DETAILS[DesiredScreenMode].String,
            SCREEN_BITDEPTH_DETAILS[DesiredBitdepth],
            (Fullscreen ? "fullscreen" : "windowed"),
            SCREEN_MODE_DETAILS[ScreenMode].String,
            SCREEN_BITDEPTH_DETAILS[Bitdepth]);
        CU_ShowMessage("Warning", VideoWarningMessageBuffer);

        G->SetOption(Game::BITDEPTH, Bitdepth);
        G->SetOption(Game::FULLSCREEN, Fullscreen);
        G->SetOption(Game::SCREENMODE, ScreenMode);
    }

    Screen = SDL_SetVideoMode(
        SCREEN_MODE_DETAILS[ScreenMode].Width, 
        SCREEN_MODE_DETAILS[ScreenMode].Height, 
        SCREEN_BITDEPTH_DETAILS[Bitdepth], 
        SDL_HWSURFACE | SDL_DOUBLEBUF | (Fullscreen ? SDL_FULLSCREEN : 0));
    if (Screen == NULL) SDE_SDLError(__FILE__, __LINE__);

    SDL_WM_SetCaption(CAM_APPLICATION_TITLE, NULL);
}

//------------------------------------------------------------------------------
// Name: ScreenModeAccepted
// Desc: This function takes a fixed screenmode as well as desired values for
//       fullscreen mode and bitdepth. It will then try to find the closest
//       configuration that allows the screenmode.
//------------------------------------------------------------------------------
bool SDL_Interface::ScreenModeAccepted(SCREEN_MODES F_ScreenMode, 
    bool& M_Fullscreen, SCREEN_BITDEPTHS& M_Bitdepth)
{
    if (FullscreenAccepted(F_ScreenMode, M_Fullscreen, M_Bitdepth)) {
        return true;
    } else if (FullscreenAccepted(F_ScreenMode, !M_Fullscreen, M_Bitdepth)) {
        M_Fullscreen = !M_Fullscreen;
        return true;
    } else {
        return false;
    }
}

//------------------------------------------------------------------------------
// Name: FullscreenAccepted
// Desc: This function takes a fixed screenmode and fullscreen mode as well 
//       as a desired value for bitdepth. It will then try to find the closest
//       configuration that allows the screenmode.
//------------------------------------------------------------------------------
bool SDL_Interface::FullscreenAccepted(SCREEN_MODES F_ScreenMode, 
    bool F_Fullscreen, SCREEN_BITDEPTHS& M_Bitdepth)
{
    int i;
    uint Flags = SDL_HWSURFACE | SDL_DOUBLEBUF | (F_Fullscreen ? SDL_FULLSCREEN : 0);

    if (SDL_VideoModeOK(
        SCREEN_MODE_DETAILS[F_ScreenMode].Width, 
        SCREEN_MODE_DETAILS[F_ScreenMode].Height, 
        SCREEN_BITDEPTH_DETAILS[M_Bitdepth], Flags) 
        == SCREEN_BITDEPTH_DETAILS[M_Bitdepth]) {

        return true;
    } else {
        for (i = 0; i < NUM_SCREEN_BITDEPTHS; i++) {
            if (i != M_Bitdepth) {
                if (SDL_VideoModeOK(
                    SCREEN_MODE_DETAILS[F_ScreenMode].Width, 
                    SCREEN_MODE_DETAILS[F_ScreenMode].Height,
                    SCREEN_BITDEPTH_DETAILS[i], Flags) 
                    == SCREEN_BITDEPTH_DETAILS[i]) {

                    M_Bitdepth = (SCREEN_BITDEPTHS)i;
                    return true;
                }
            }
        }
    }
    return false;
}

//------------------------------------------------------------------------------
// Name: LoadScaledSample
// Desc: Loads a sample in from an external file and then scales it by a 
//       given scaling factor. The scaling factor must be positive.
//
// Note: Throughout this algorithm bits of sample data are (very inaccurately) 
//       refered to as pixels. This is because it helped me to think of a 
//       sample as a 1 dimensional image.
//------------------------------------------------------------------------------
Mix_Chunk *SDL_Interface::LoadScaledSample(const char *FileName, double SFactor)
{
    int BytesPerPixel = 2;
    int CurNewPixel;
    double CurPixel, NextPixel;
    int CurWholePixel, NextWholePixel;
    double NewPixelSize;
    bool ScalingCompleted = false;
    double TotalValue;
    double PixelDiff;
    short *OrigBuffer, *NewBuffer;
    int OrigSampleLength, NewSampleLength;
    Mix_Chunk *OrigSample = NULL;
    Mix_Chunk *NewSample = NULL;

    // Unfortunately, this function can only handle sapmles in the 
    // AUDIO_S16SYS format
    assert(FileName != NULL && SFactor >= 0.0);
    assert(FORMAT == AUDIO_S16SYS);

    // Handle the trivial cases where SFactor == 1.0 or 0.0
    if (SFactor == 1.0) {
        NewSample = Mix_LoadWAV(FileName);
        if (NewSample == NULL) ReturnWithMixError(NULL);
        return NewSample;
    } else if (SFactor == 0.0) {
        NewSample = (Mix_Chunk *)malloc(sizeof(Mix_Chunk));
        if (NewSample == NULL) 
            ReturnWithError(NULL, SDLI_ERROR_CANNOT_ALLOCATE_MEMORY);
        memset(NewSample, 0, sizeof(Mix_Chunk));
        return NewSample;
    }

    // Load in the original sample and allocate memory for the new one
    NewSample = (Mix_Chunk *)malloc(sizeof(Mix_Chunk));
    if (NewSample == NULL) ReturnWithError(NULL, SDLI_ERROR_CANNOT_ALLOCATE_MEMORY);
    OrigSample = Mix_LoadWAV(FileName);
    if (OrigSample == NULL) ReturnWithMixError(NULL);
    NewSample->allocated = 1;
    NewSample->volume = OrigSample->volume;
    NewSample->alen = (uint)ceil((double)OrigSample->alen * SFactor);
    if ((NewSample->alen - 1) / 2 == NewSample->alen / 2)
        NewSample->alen++;
    NewSample->abuf = (uchar *)malloc(NewSample->alen);
    if (NewSample->abuf == NULL) {
        Mix_FreeChunk(OrigSample);
        free(NewSample);
        ReturnWithError(NULL, SDLI_ERROR_CANNOT_ALLOCATE_MEMORY);
    }

    // Work out how many pixels in OrigSample a pixel in NewSample represents
    NewPixelSize = 1/SFactor;

    OrigSampleLength = OrigSample->alen / BytesPerPixel;
    NewSampleLength = NewSample->alen / BytesPerPixel;

    OrigBuffer = (short *)OrigSample->abuf;
    NewBuffer = (short *)NewSample->abuf;
    CurPixel = 0.0f;
    for (CurNewPixel = 0; CurNewPixel < NewSampleLength; CurNewPixel++) {
        TotalValue = 0.0f;
        NextPixel = CurPixel + NewPixelSize;
        do {
            CurWholePixel = (int)CurPixel;
            NextWholePixel = CurWholePixel + 1;
            if ((double)NextWholePixel >= NextPixel) {
                TotalValue += (NextPixel - CurPixel)*OrigBuffer[CurWholePixel];
                CurPixel = NextPixel;
            } else {
                PixelDiff = (double)NextWholePixel - CurPixel;
                TotalValue += PixelDiff*OrigBuffer[CurWholePixel];
                CurPixel += PixelDiff;

                // Check to see if we have run out of pixels in the original
                // sample. If we have then pretend that the last pixel 
                // continues for as long as is needed.
                if (NextWholePixel == OrigSampleLength) {
                    TotalValue += (NextPixel - CurPixel)*OrigBuffer[CurWholePixel];
                    CurPixel = NextPixel;
                }
            }
        } while (CurPixel < NextPixel);

        NewBuffer[CurNewPixel] = (short)(TotalValue / NewPixelSize);
    }

    return NewSample;
}

//------------------------------------------------------------------------------
// Name: SetSFX_Volume
// Desc: Sets the volume of the sound effects to a given value
//------------------------------------------------------------------------------
void SDL_Interface::SetSFX_Volume(int NewVolume)
{
    if (NewVolume < 0) {
        NewVolume = 0;
    } else if (NewVolume > MIX_MAX_VOLUME) {
        NewVolume = MIX_MAX_VOLUME;
    }
    SFX_Volume = NewVolume;
    G->SetOption(Game::SFX_VOLUME, NewVolume);
    Mix_Volume(-1, NewVolume);
}

//------------------------------------------------------------------------------
// Name: SetMusic_Volume
// Desc: Sets the volume of the in game music to a given value
//------------------------------------------------------------------------------
void SDL_Interface::SetMusic_Volume(int NewVolume)
{
    if (NewVolume < 0) {
        NewVolume = 0;
    } else if (NewVolume > MIX_MAX_VOLUME) {
        NewVolume = MIX_MAX_VOLUME;
    }
    Music_Volume = NewVolume;
    G->SetOption(Game::MUSIC_VOLUME, NewVolume);
    if (!MusicPaused) Mix_VolumeMusic(NewVolume);
}

//------------------------------------------------------------------------------
// Name: SetReverseStereo
// Desc: Sets whether reverse stereo is on or off
//------------------------------------------------------------------------------
void SDL_Interface::SetReverseStereo(bool ReverseStereo)
{
    bool ValueChanged = (ReverseStereo != this->ReverseStereo);

    this->ReverseStereo = ReverseStereo;
    G->SetOption(Game::REVERSE_STEREO, ReverseStereo);
    Mix_SetReverseStereo(MIX_CHANNEL_POST, ValueChanged);
}

//------------------------------------------------------------------------------
// Name: SW_FillRect
// Desc: Fills an area of 'dst' described by 'dstrect' with 'colour'. Use
//       SDL_MapRGB or SDL_MapRGBA to create colour
//------------------------------------------------------------------------------
bool SDL_Interface::SW_FillRect(SDL_Surface *dst, 
                                SDL_Rect *dstrect, Uint32 colour)
{
    uint i, j;
    uint *SurfacePtr;
    SDL_Rect DestRect;
    int Jump;

    assert(dst != NULL);
    assert(dst->format->BitsPerPixel == CAM_INTERNAL_SURFACE_BPP);

    if (!AdjustRect(dst, dstrect, &DestRect)) return true;

    if (SDL_MUSTLOCK(dst)) {
        if (SDL_FAILED(SDL_LockSurface(dst))) 
            SDF_SDLError(__FILE__, __LINE__);
    }

    SurfacePtr = (uint *)((char *)dst->pixels + DestRect.y*dst->pitch + 
        DestRect.x*dst->format->BytesPerPixel);
    Jump = (dst->pitch - DestRect.w*dst->format->BytesPerPixel) / dst->format->BytesPerPixel;
    assert(Jump >= 0);

    for (j = 0; j < DestRect.h; j++) {
        for (i = 0; i < DestRect.w; i++) {
            *SurfacePtr = colour;

            SurfacePtr++;
        }
        SurfacePtr += Jump;
    }
      
    SDL_UnlockSurface(dst);

    return true;
}

//------------------------------------------------------------------------------
// Name: SW_BlitSurface
// Desc: This is a very simple software blitter that copies the area of 'src' 
//       described by 'srcrect' to the area of 'dst' described by 'dstrect'. 
//       Like the rest of this program this function assumes that the surfaces
//       have a bit depth of 32 bits. It also just copies pixels regardless of 
//       the format so the channel masks for both surfaces must be the same.
//       The reason that this function exists is because I was unable to find
//       a way of copying the alpha channel using SDL.
//------------------------------------------------------------------------------
bool SDL_Interface::SW_BlitSurface(SDL_Surface *src, SDL_Rect *srcrect,
                                   SDL_Surface *dst, SDL_Rect *dstrect)
{
    uint i, j;
    uint *SrcSurfacePtr, *DstSurfacePtr;
    int SrcJump, DstJump;
    int BPP;
    uint Width, Height;
    SDL_Rect SourceRect, DestRect;

    assert(src != NULL && dst != NULL);
    assert(SameSurfaceFormat(src, dst));

    // Adjust and then validate the source and destination rectangles
    if (!AdjustRects(src, srcrect, &SourceRect, dst, dstrect, &DestRect)) return true;
    if (SourceRect.w != DestRect.w || SourceRect.h != DestRect.h)
        FailWithError(SDLI_ERROR_INVALID_PARAMETERS);

    // Lock the surfaces
    if (SDL_MUSTLOCK(src)) {
        if (SDL_FAILED(SDL_LockSurface(src))) 
            SDF_SDLError(__FILE__, __LINE__);
    }

    if (SDL_MUSTLOCK(dst)) {
        if (SDL_FAILED(SDL_LockSurface(dst))) {
            SDL_UnlockSurface(src);
            SDF_SDLError(__FILE__, __LINE__);
        }
    }

    Width = SourceRect.w;
    Height = SourceRect.h;
    BPP = src->format->BytesPerPixel;
    SrcSurfacePtr = (uint *)((char *)src->pixels + SourceRect.y*src->pitch + SourceRect.x*BPP);
    DstSurfacePtr = (uint *)((char *)dst->pixels + DestRect.y*dst->pitch + DestRect.x*BPP);
    SrcJump = (src->pitch - SourceRect.w*BPP) / BPP;
    DstJump = (dst->pitch - DestRect.w*BPP) / BPP;
    assert(SrcJump >= 0 && DstJump >= 0);

    for (j = 0; j < Height; j++) {
        for (i = 0; i < Width; i++) {
            *DstSurfacePtr = *SrcSurfacePtr;

            SrcSurfacePtr++;
            DstSurfacePtr++;
        }
        SrcSurfacePtr += SrcJump;
        DstSurfacePtr += DstJump;
    }
    
    SDL_UnlockSurface(dst);
    SDL_UnlockSurface(src);

    return true;
}

//------------------------------------------------------------------------------
// Name: AdjustRects
// Desc: Takes a pair of rectangles via the RectInfo variables and provides
//       modified source and destination rectangles that fit on the suface. If
//       it impossible to get valid rectangles because one or both of them are
//       completely outside the bounds of their surfaces clipping rectangle
//       then this function returns false.
//------------------------------------------------------------------------------
bool SDL_Interface::AdjustRects(SDL_Surface *SrcSurface, SDL_Rect *SrcRectInfo, 
                                SDL_Rect *SrcRect, SDL_Surface *DstSurface,
                                SDL_Rect *DstRectInfo, SDL_Rect *DstRect)
{
    assert (SrcSurface != NULL && SrcRect != NULL
        && DstSurface != NULL && DstRect != NULL);

    // Copy rectangles and perform initial validation
    if (SrcRectInfo == NULL) {
        SrcRect->x = 0;
        SrcRect->y = 0;
        SrcRect->w = SrcSurface->w;
        SrcRect->h = SrcSurface->h;
        // Note that the checking of the source rectangle is more strict than 
        // for the destination rectangle
    } else if (SrcRectInfo->x < 0 || SrcRectInfo->y < 0
        ||SrcRectInfo->x + SrcRectInfo->w > SrcSurface->w 
        || SrcRectInfo->y + SrcRectInfo->h > SrcSurface->h) {
        return false;
    } else {
        memcpy(SrcRect, SrcRectInfo, sizeof(SDL_Rect));
    }

    if (DstRectInfo == NULL) {
        memcpy(DstRect, &DstSurface->clip_rect, sizeof(SDL_Rect));
    } else if (DstRectInfo->x >= DstSurface->clip_rect.w 
        || DstRectInfo->y >= DstSurface->clip_rect.h
        || DstRectInfo->x + DstRectInfo->w <= DstSurface->clip_rect.x
        || DstRectInfo->y + DstRectInfo->h <= DstSurface->clip_rect.y) {
        return false;
    } else {
        memcpy(DstRect, DstRectInfo, sizeof(SDL_Rect));
    }

    // Clip the rectangles against the destination surfaces clip_rect
    if (DstRect->x < DstSurface->clip_rect.x) {
        SrcRect->x -= DstRect->x;
        SrcRect->w = DstRect->w = DstRect->w + DstRect->x;
        DstRect->x = 0;
    }
    if (DstRect->y < DstSurface->clip_rect.y) {
        SrcRect->y -= DstRect->y;
        SrcRect->h = DstRect->h = DstRect->h + DstRect->y; 
        DstRect->y = 0;
    }
    if (DstRect->x + DstRect->w > DstSurface->clip_rect.w) 
        SrcRect->w = DstRect->w = DstSurface->clip_rect.w - DstRect->x;
    if (DstRect->y + DstRect->h > DstSurface->clip_rect.h) 
        SrcRect->h = DstRect->h = DstSurface->clip_rect.h - DstRect->y;

    return true;
}

//------------------------------------------------------------------------------
// Name: AdjustRect
// Desc: A modified verion of AdjustRects that deals with the case where there
//       is just a destination rectangle. For example when using the 
//       SW_FillRect function.
//------------------------------------------------------------------------------
bool SDL_Interface::AdjustRect(SDL_Surface *DstSurface,
                               SDL_Rect *DstRectInfo, SDL_Rect *DstRect)
{
    assert (DstSurface != NULL && DstRect != NULL);

    if (DstRectInfo == NULL) {
        memcpy(DstRect, &DstSurface->clip_rect, sizeof(SDL_Rect));
    } else if (DstRectInfo->x >= DstSurface->clip_rect.w 
        || DstRectInfo->y >= DstSurface->clip_rect.h
        || DstRectInfo->x + DstRectInfo->w <= DstSurface->clip_rect.x
        || DstRectInfo->y + DstRectInfo->h <= DstSurface->clip_rect.y) {
        return false;
    } else {
        memcpy(DstRect, DstRectInfo, sizeof(SDL_Rect));
    }

    // Clip the rectangles against the destination surfaces clip_rect
    if (DstRect->x < DstSurface->clip_rect.x) {
        DstRect->w += DstRect->x;
        DstRect->x = 0;
    }
    if (DstRect->y < DstSurface->clip_rect.y) {
        DstRect->h += DstRect->y; 
        DstRect->y = 0;
    }
    if (DstRect->x + DstRect->w > DstSurface->clip_rect.w) 
        DstRect->w = DstSurface->clip_rect.w - DstRect->x;
    if (DstRect->y + DstRect->h > DstSurface->clip_rect.h) 
        DstRect->h = DstSurface->clip_rect.h - DstRect->y;

    return true;
}

//------------------------------------------------------------------------------
// Name: SameSurfaceFormat
// Desc: Tries to determine whether or not the two surfaces store pixels in
//       the same format
//------------------------------------------------------------------------------
bool SDL_Interface::SameSurfaceFormat(SDL_Surface *S1, SDL_Surface *S2)
{
    return (S1->format->BytesPerPixel == S2->format->BytesPerPixel);
}

//------------------------------------------------------------------------------
// Name: DrawLine
// Desc: This is a very rough and ready line drawing routine. It isn't really
//       optimised and does NO clipping so don't use it anywhere you need 
//       speed. The only reason that I use this routine is that I only need to
//       draw two lines in the game (the needles on the gauges) so thought that
//       it would be a bit wasteful to use something like SDL_draw. :P
// 
// Note: Only works for bitdepths of 8, 16 and 32 bits.
//------------------------------------------------------------------------------
void SDL_Interface::DrawLine(SDL_Surface *Surface, uint Colour, 
                             Point2D StartPos, Point2D EndPos)
{
    float dx, dy;
    float CurX, CurY;
    float Ratio;
    Point2D PixelPos;
    int XRange, YRange;
    bool AtLastX, AtLastY;
    int SurfacePitch;
    float (*RoundX)(float);
    float (*RoundY)(float);

    assert(Surface != NULL);
    assert(Surface->format->BitsPerPixel == 8 
        || Surface->format->BitsPerPixel == 16 
        || Surface->format->BitsPerPixel == 32);
    
    SurfacePitch = Surface->pitch / Surface->format->BytesPerPixel;

    // Calculate dx and dy
    XRange = EndPos.x - StartPos.x;
    YRange = EndPos.y - StartPos.y;

    if (abs(XRange) >= abs(YRange)) {
        dx = (float)(XRange >= 0 ? 1 : -1);
        Ratio = (float)abs(YRange) / abs(XRange);
        dy = (YRange >= 0 ? Ratio : -Ratio);
    } else {
        Ratio = (float)abs(XRange) / abs(YRange);
        dx = (XRange >= 0 ? Ratio : -Ratio);
        dy = (float)(YRange >= 0 ? 1 : -1);
    }
    
    // Lock the surface if necessary
    if (SDL_MUSTLOCK(Surface)) {
        if (SDL_FAILED(SDL_LockSurface(Surface))) {
            CU_SetLastSDLError(); 
            CU_PrintLastError(__FILE__, __LINE__); 
            return;
        }
    }

    // Work out what the rounding functions will be
    RoundX = (dx >= 0 ? floorf : ceilf);
    RoundY = (dy >= 0 ? floorf : ceilf);

    // Initialise status variables then draw the line
    CurX = (float)StartPos.x;
    CurY = (float)StartPos.y;
    AtLastX = false;
    AtLastY = false;
    do {
        PixelPos.x = (int)RoundX(CurX);
        PixelPos.y = (int)RoundY(CurY);

        switch (Surface->format->BitsPerPixel) {
            case 8:
                ((uchar *)Surface->pixels)[
                    PixelPos.y * SurfacePitch + PixelPos.x] = (uchar)Colour;
                break;
            case 16:
                ((ushort *)Surface->pixels)[
                    PixelPos.y * SurfacePitch + PixelPos.x] = (ushort)Colour;
                break;
            case 32:
                ((uint *)Surface->pixels)[
                    PixelPos.y * SurfacePitch + PixelPos.x] = Colour;
                break;
        }

        CurX += dx;
        CurY += dy;
        if (dx >= 0) {
            if (CurX >= EndPos.x) AtLastX = true;
        } else {
            if (CurX <= EndPos.x) AtLastX = true;
        }
        if (dy >= 0) {
            if (CurY >= EndPos.y) AtLastY = true;
        } else {
            if (CurY <= EndPos.y) AtLastY = true;
        }
    } while (!(AtLastX && AtLastY));

    SDL_UnlockSurface(Surface);
}

//------------------------------------------------------------------------------
// Name: PlayChannel
// Desc: This is simply a very thin wrapper around the Mix_PlayChannel
//       function that allows us to keep track of the number of channels in
//       use.
//------------------------------------------------------------------------------
int SDL_Interface::PlayChannel(int Channel, Mix_Chunk *Chunk, int Loops)
{
    int Result = Mix_PlayChannel(Channel, Chunk, Loops);

    if (Result >= 0) {
        CurChannelsUsed++;
        if (CurChannelsUsed > MaxChannelsUsed) MaxChannelsUsed = CurChannelsUsed;
    }
    return Result;
}
  
//------------------------------------------------------------------------------
// Name: ChannelFinished
// Desc: This is procedure is registered with SDL_mixer a called whenever a
//       channel finishes playing. Its sole purpose is to help track the number
//       of channels currently being used to play sound effects
//------------------------------------------------------------------------------
void SDL_Interface::ChannelFinished(int Channel)
{
    // Reset channel to the default volume and stereo panning
    Mix_Volume(Channel, Instance->GetSFX_Volume());
    Mix_SetPanning(Channel, CENTRAL_VOLUME, CENTRAL_VOLUME);

    Instance->CurChannelsUsed--;
    assert(Instance->CurChannelsUsed >= 0);
}

//------------------------------------------------------------------------------
// Name: VolumeOverDistance
// Desc: This works out what the volume of a sound effect should be over a 
//       given distance. The rule used is that the volume of a sound is 
//       proportional to the inverse square of the distance to the source of
//       a sound.
//------------------------------------------------------------------------------
int SDL_Interface::VolumeOverDistance(float Distance)
{
    // Deal with the trivial case where the volume would not be affected
    if (Distance <= SOUND_UNIT_OF_DISTANCE) return SFX_Volume;

    return (int)(SFX_Volume * 1 / powf(Distance / SOUND_UNIT_OF_DISTANCE, 2));
}

//------------------------------------------------------------------------------
// Name: GetStereoVolumes
// Desc: This function returns the volumes that should be set for each 
//       stereo channel in order for an appropriate stereo effect to be heard.
//       The arguments to the function (apart from pointers to variables that
//       will hold the output are the current viewports width divided
//       by 2, and the horizontal distance of the sound source from the centre
//       of the viewport. When a sound source is on or beyond the far right 
//       hand side of the viewport it is considered full right, and full left
//       for when the sound source is on or beyond the far left hand side of
//       the viewport.
//------------------------------------------------------------------------------
void SDL_Interface::GetStereoVolumes(int ViewportWidth_2, 
    int X_DistanceFromCentre, uchar *LeftChannel, uchar *RightChannel)
{
    uchar Offset;

    assert(LeftChannel != NULL && RightChannel != NULL);

    if (X_DistanceFromCentre == 0) {
        Offset = 0;
    } else if (abs(X_DistanceFromCentre) > ViewportWidth_2) {
        Offset = (X_DistanceFromCentre > 0 ? CENTRAL_VOLUME : -CENTRAL_VOLUME);
    } else {
        Offset = (uchar)((float)X_DistanceFromCentre/ViewportWidth_2 * CENTRAL_VOLUME);
    }
    *RightChannel = CENTRAL_VOLUME + Offset;
    *LeftChannel = 2*CENTRAL_VOLUME - *RightChannel;
}

//------------------------------------------------------------------------------
// Name: PlayMusic
// Desc: Plays a given piece of music
//------------------------------------------------------------------------------
void SDL_Interface::PlayMusic(Mix_Music *Melody, int Loops, int FadeIn)
{
    assert(Melody != NULL);

    if (FadeIn <= 0) {
        Mix_PlayMusic(Melody, Loops);
    } else {
        Mix_FadeInMusic(Melody, Loops, FadeIn);
    }
    Mix_VolumeMusic(Music_Volume);
    MusicPaused = false;
}

//------------------------------------------------------------------------------
// Name: PauseMusic
// Desc: This routine does not really pause the music, only lower the volume.
//       This is because MixPauseMusic does not work on MIDI files on the 
//       WIN32 platform.
//------------------------------------------------------------------------------
void SDL_Interface::PauseMusic()
{
    MusicPaused = true;
    Mix_VolumeMusic(0);
}

//------------------------------------------------------------------------------
// Name: ResumeMusic
// Desc: Resumes any music that may be playing
//------------------------------------------------------------------------------
void SDL_Interface::ResumeMusic()
{
    MusicPaused = false;
    Mix_VolumeMusic(Music_Volume);
}

//------------------------------------------------------------------------------
// Name: StopMusic
// Desc: Stops any music that may be playing.
//------------------------------------------------------------------------------
void SDL_Interface::StopMusic(int FadeOut)
{
    MusicPaused = false;

    if (FadeOut <= 0) {
        Mix_HaltMusic();
    } else {
        Mix_FadeOutMusic(FadeOut);
    }
}
