/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Sprites.cpp
// Desc: Functions that load in 1 or more bitmaps that contain the games
//       sprites and plot them to the screen
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <assert.h>
#include "CamelUtility.h"
#include "globals.h"
#include "Camel.h"
#include "SDL_Interface.h"
#include "Sprites.h"

//------------------------------------------------------------------------------
// Global Variables and Constants
//------------------------------------------------------------------------------

// The start frame for all of the available sprites
const int Sprites::PLAYER_PLANE_SPRITE = 0;
const int Sprites::SMOKEPARTICLE_SPRITE = 32;
const int Sprites::SMALLEXPLOSION_1_SPRITE = 33;
const int Sprites::SMALLEXPLOSION_2_SPRITE = 49;
const int Sprites::BOMB_SPRITE = 65;
const int Sprites::BULLET_SPRITE = 81;
const int Sprites::MOUSEPOINTER_SPRITE = 82;
const int Sprites::LARGEEXPLOSION_1_SPRITE = 83;
const int Sprites::LARGEEXPLOSION_2_SPRITE = 99;
const int Sprites::ENEMY_PLANE_SPRITE = 115;
const int Sprites::GOOD_SOLDIER_SPRITE = 147;
const int Sprites::EVIL_SOLDIER_SPRITE = 176;
const int Sprites::GOOD_CONTROLLED_ENTITY_FLAG_SPRITE = 205;
const int Sprites::EVIL_CONTROLLED_ENTITY_FLAG_SPRITE = 206;
const int Sprites::GOOD_PLATOON_FACTORY_SPRITE = 207;
const int Sprites::EVIL_PLATOON_FACTORY_SPRITE = 215;
const int Sprites::GOOD_FLAG_SPRITE = 223;
const int Sprites::EVIL_FLAG_SPRITE = 224;
const int Sprites::GAMEOVER_BANNER_SPRITE = 225;
const int Sprites::YOUWIN_BANNER_SPRITE = 226;
const int Sprites::UI_BULLET_SPRITE = 227;
const int Sprites::UI_BOMB_SPRITE = 228;
const int Sprites::UI_HEART_SPRITE = 229;
const int Sprites::UI_SPEEDOMETER_SPRITE = 230;
const int Sprites::UI_FUEL_GUAGE_SPRITE = 231;
const int Sprites::GOOD_TANK_SPRITE = 232;
const int Sprites::EVIL_TANK_SPRITE = 306;
const int Sprites::GOOD_PLATOON_ICON = 380;
const int Sprites::EVIL_PLATOON_ICON = 381;
const int Sprites::GOOD_TANKSQUAD_ICON = 382;
const int Sprites::EVIL_TANKSQUAD_ICON = 383;
const int Sprites::GOOD_TANKSQUAD_FACTORY_SPRITE = 384;
const int Sprites::EVIL_TANKSQUAD_FACTORY_SPRITE = 392;
const int Sprites::TESTFRAME_1 = 400;
const int Sprites::TESTFRAME_2 = 401;

const int Sprites::COLLISION_THRESHOLD = 80;


const int SPR_ERROR_BASE = 0x20000200;

const int SPR_ERROR_INCOMPATIBLE_SURFACES = SPR_ERROR_BASE;

const int SPR_NUM_ERRORS = 1;

static char *SPR_ErrorStrings[SPR_NUM_ERRORS] = {
    "SPR_ERROR_INCOMPATIBLE_SURFACES"
};

static CU_ERRORMESSAGES SPR_ErrorMessages = {
    SPR_ERROR_BASE, SPR_NUM_ERRORS, SPR_ErrorStrings
};

Sprites::FRAME Sprites::Frames[NUM_FRAMES] = {
//    x, y, w, h
    {{ 0, 0, 20, 20 }, false },         // 0: PPAC_0 (Player Plane anti-clockwise)
    {{ 20, 0, 20, 20 }, false },        // 1: PPAC_1
    {{ 40, 0, 20, 20 }, false },        // 2: PPAC_2
    {{ 60, 0, 20, 20 }, false },        // 3: PPAC_3
    {{ 80, 0, 20, 20 }, false },        // 4: PPAC_4
    {{ 100, 0, 20, 20 }, false },       // 5: PPAC_5
    {{ 120, 0, 20, 20 }, false },       // 6: PPAC_6
    {{ 140, 0, 20, 20 }, false },       // 7: PPAC_7
    {{ 0, 20, 20, 20 }, false },        // 8: PPAC_8
    {{ 20, 20, 20, 20 }, false },       // 9: PPAC_9
    {{ 40, 20, 20, 20 }, false },       // 10: PPAC_10
    {{ 60, 20, 20, 20 }, false },       // 11: PPAC_11
    {{ 80, 20, 20, 20 }, false },       // 12: PPAC_12
    {{ 100, 20, 20, 20 }, false },      // 13: PPAC_13
    {{ 120, 20, 20, 20 }, false },      // 14: PPAC_14
    {{ 140, 20, 20, 20 }, false },      // 15: PPAC_15
    {{ 0, 40, 20, 20 }, false },        // 16: PPC_0 (Player Plane clockwise)
    {{ 20, 40, 20, 20 }, false },       // 17: PPC_1
    {{ 40, 40, 20, 20 }, false },       // 18: PPC_2
    {{ 60, 40, 20, 20 }, false },       // 19: PPC_3
    {{ 80, 40, 20, 20 }, false },       // 20: PPC_4
    {{ 100, 40, 20, 20 }, false },      // 21: PPC_5
    {{ 120, 40, 20, 20 }, false },      // 22: PPC_6
    {{ 140, 40, 20, 20 }, false },      // 23: PPC_7
    {{ 0, 60, 20, 20 }, false },        // 24: PPC_8
    {{ 20, 60, 20, 20 }, false },       // 25: PPC_9
    {{ 40, 60, 20, 20 }, false },       // 26: PPC_10
    {{ 60, 60, 20, 20 }, false },       // 27: PPC_11
    {{ 80, 60, 20, 20 }, false },       // 28: PPC_12
    {{ 100, 60, 20, 20 }, false },      // 29: PPC_13
    {{ 120, 60, 20, 20 }, false },      // 30: PPC_14
    {{ 140, 60, 20, 20 }, false },      // 31: PPC_15
    {{ 240, 0, 2, 2 }, false },         // 32: Smoke particle
    {{ 0, 80, 32, 32 }, true },         // 33: Small Explosion 1 - F0
    {{ 32, 80, 32, 32 }, true },        // 34: Small Explosion 1 - F1
    {{ 64, 80, 32, 32 }, true },        // 35: Small Explosion 1 - F2
    {{ 96, 80, 32, 32 }, true },        // 36: Small Explosion 1 - F3
    {{ 0, 112, 32, 32 }, true },        // 37: Small Explosion 1 - F4
    {{ 32, 112, 32, 32 }, true },       // 38: Small Explosion 1 - F5
    {{ 64, 112, 32, 32 }, true },       // 39: Small Explosion 1 - F6
    {{ 96, 112, 32, 32 }, true },       // 40: Small Explosion 1 - F7
    {{ 0, 144, 32, 32 }, true },        // 41: Small Explosion 1 - F8
    {{ 32, 144, 32, 32 }, true },       // 42: Small Explosion 1 - F9
    {{ 64, 144, 32, 32 }, true },       // 43: Small Explosion 1 - F10
    {{ 96, 144, 32, 32 }, true },       // 44: Small Explosion 1 - F11
    {{ 0, 176, 32, 32 }, true },        // 45: Small Explosion 1 - F12
    {{ 32, 176, 32, 32 }, true },       // 46: Small Explosion 1 - F13
    {{ 64, 176, 32, 32 }, true },       // 47: Small Explosion 1 - F14
    {{ 96, 176, 32, 32 }, true },       // 48: Small Explosion 1 - F15
    {{ 128, 80, 32, 32 }, true },       // 49: Small Explosion 2 - F0
    {{ 160, 80, 32, 32 }, true },       // 50: Small Explosion 2 - F1
    {{ 192, 80, 32, 32 }, true },       // 51: Small Explosion 2 - F2
    {{ 224, 80, 32, 32 }, true },       // 52: Small Explosion 2 - F3
    {{ 128, 112, 32, 32 }, true },      // 53: Small Explosion 2 - F4
    {{ 160, 112, 32, 32 }, true },      // 54: Small Explosion 2 - F5
    {{ 192, 112, 32, 32 }, true },      // 55: Small Explosion 2 - F6
    {{ 224, 112, 32, 32 }, true },      // 56: Small Explosion 2 - F7
    {{ 128, 144, 32, 32 }, true },      // 57: Small Explosion 2 - F8
    {{ 160, 144, 32, 32 }, true },      // 58: Small Explosion 2 - F9
    {{ 192, 144, 32, 32 }, true },      // 59: Small Explosion 2 - F10
    {{ 224, 144, 32, 32 }, true },      // 60: Small Explosion 2 - F11
    {{ 128, 176, 32, 32 }, true },      // 61: Small Explosion 2 - F12
    {{ 160, 176, 32, 32 }, true },      // 62: Small Explosion 2 - F13
    {{ 192, 176, 32, 32 }, true },      // 63: Small Explosion 2 - F14
    {{ 224, 176, 32, 32 }, true },      // 64: Small Explosion 2 - F15
    {{ 160, 0, 10, 10 }, false },       // 65: Bomb - F0
    {{ 170, 0, 10, 10 }, false },       // 66: Bomb - F1
    {{ 180, 0, 10, 10 }, false },       // 67: Bomb - F2
    {{ 190, 0, 10, 10 }, false },       // 68: Bomb - F3
    {{ 200, 0, 10, 10 }, false },       // 69: Bomb - F4
    {{ 210, 0, 10, 10 }, false },       // 70: Bomb - F5
    {{ 220, 0, 10, 10 }, false },       // 71: Bomb - F6
    {{ 230, 0, 10, 10 }, false },       // 72: Bomb - F7
    {{ 160, 10, 10, 10 }, false },      // 73: Bomb - F8
    {{ 170, 10, 10, 10 }, false },      // 74: Bomb - F9
    {{ 180, 10, 10, 10 }, false },      // 75: Bomb - F10
    {{ 190, 10, 10, 10 }, false },      // 76: Bomb - F11
    {{ 200, 10, 10, 10 }, false },      // 77: Bomb - F12
    {{ 210, 10, 10, 10 }, false },      // 78: Bomb - F13
    {{ 220, 10, 10, 10 }, false },      // 79: Bomb - F14
    {{ 230, 10, 10, 10 }, false },      // 80: Bomb - F15
	{{ 250, 0, 2, 2 }, false },         // 81: Bullet
    {{ 160, 20, 16, 24 }, false },      // 82: Mouse Pointer
    {{ 256, 0, 64, 64 }, true },        // 83: Large Explosion 1 - F0
    {{ 320, 0, 64, 64 }, true },        // 84: Large Explosion 1 - F1
    {{ 384, 0, 64, 64 }, true },        // 85: Large Explosion 1 - F2
    {{ 448, 0, 64, 64 }, true },        // 86: Large Explosion 1 - F3
    {{ 256, 64, 64, 64 }, true },       // 87: Large Explosion 1 - F4
    {{ 320, 64, 64, 64 }, true },       // 88: Large Explosion 1 - F5
    {{ 384, 64, 64, 64 }, true },       // 89: Large Explosion 1 - F6
    {{ 448, 64, 64, 64 }, true },       // 90: Large Explosion 1 - F7
    {{ 256, 128, 64, 64 }, true },      // 91: Large Explosion 1 - F8
    {{ 320, 128, 64, 64 }, true },      // 92: Large Explosion 1 - F9
    {{ 384, 128, 64, 64 }, true },      // 93: Large Explosion 1 - F10
    {{ 448, 128, 64, 64 }, true },      // 94: Large Explosion 1 - F11
    {{ 256, 192, 64, 64 }, true },      // 95: Large Explosion 1 - F12
    {{ 320, 192, 64, 64 }, true },      // 96: Large Explosion 1 - F13
    {{ 384, 192, 64, 64 }, true },      // 97: Large Explosion 1 - F14
    {{ 448, 192, 64, 64 }, true },      // 98: Large Explosion 1 - F15
    {{ 0, 256, 64, 64 }, true },        // 99: Large Explosion 2 - F0
    {{ 64, 256, 64, 64 }, true },       // 100: Large Explosion 2 - F1
    {{ 128, 256, 64, 64 }, true },      // 101: Large Explosion 2 - F2
    {{ 192, 256, 64, 64 }, true },      // 102: Large Explosion 2 - F3
    {{ 0, 320, 64, 64 }, true },        // 103: Large Explosion 2 - F4
    {{ 64, 320, 64, 64 }, true },       // 104: Large Explosion 2 - F5
    {{ 128, 320, 64, 64 }, true },      // 105: Large Explosion 2 - F6
    {{ 192, 320, 64, 64 }, true },      // 106: Large Explosion 2 - F7
    {{ 0, 384, 64, 64 }, true },        // 107: Large Explosion 2 - F8
    {{ 64, 384, 64, 64 }, true },       // 108: Large Explosion 2 - F9
    {{ 128, 384, 64, 64 }, true },      // 109: Large Explosion 2 - F10
    {{ 192, 384, 64, 64 }, true },      // 110: Large Explosion 2 - F11
    {{ 0, 448, 64, 64 }, true },        // 111: Large Explosion 2 - F12
    {{ 64, 448, 64, 64 }, true },       // 112: Large Explosion 2 - F13
    {{ 128, 448, 64, 64 }, true },      // 113: Large Explosion 2 - F14
    {{ 192, 448, 64, 64 }, true },      // 114: Large Explosion 2 - F15
    {{ 256, 256, 20, 20 }, false },     // 115: EPAC_0 (Enemy Plane anti-clockwise)
    {{ 276, 256, 20, 20 }, false },     // 116: EPAC_1
    {{ 296, 256, 20, 20 }, false },     // 117: EPAC_2
    {{ 316, 256, 20, 20 }, false },     // 118: EPAC_3
    {{ 336, 256, 20, 20 }, false },     // 119: EPAC_4
    {{ 356, 256, 20, 20 }, false },     // 120: EPAC_5
    {{ 376, 256, 20, 20 }, false },     // 121: EPAC_6
    {{ 396, 256, 20, 20 }, false },     // 122: EPAC_7
    {{ 256, 276, 20, 20 }, false },     // 123: EPAC_8
    {{ 276, 276, 20, 20 }, false },     // 124: EPAC_9
    {{ 296, 276, 20, 20 }, false },     // 125: EPAC_10
    {{ 316, 276, 20, 20 }, false },     // 126: EPAC_11
    {{ 336, 276, 20, 20 }, false },     // 127: EPAC_12
    {{ 356, 276, 20, 20 }, false },     // 128: EPAC_13
    {{ 376, 276, 20, 20 }, false },     // 129: EPAC_14
    {{ 396, 276, 20, 20 }, false },     // 130: EPAC_15
    {{ 256, 296, 20, 20 }, false },     // 131: EPC_0 (Enemy Plane clockwise)
    {{ 276, 296, 20, 20 }, false },     // 132: EPC_1
    {{ 296, 296, 20, 20 }, false },     // 133: EPC_2
    {{ 316, 296, 20, 20 }, false },     // 134: EPC_3
    {{ 336, 296, 20, 20 }, false },     // 135: EPC_4
    {{ 356, 296, 20, 20 }, false },     // 136: EPC_5
    {{ 376, 296, 20, 20 }, false },     // 137: EPC_6
    {{ 396, 296, 20, 20 }, false },     // 138: EPC_7
    {{ 256, 316, 20, 20 }, false },     // 139: EPC_8
    {{ 276, 316, 20, 20 }, false },     // 140: EPC_9
    {{ 296, 316, 20, 20 }, false },     // 141: EPC_10
    {{ 316, 316, 20, 20 }, false },     // 142: EPC_11
    {{ 336, 316, 20, 20 }, false },     // 143: EPC_12
    {{ 356, 316, 20, 20 }, false },     // 144: EPC_13
    {{ 376, 316, 20, 20 }, false },     // 145: EPC_14
    {{ 396, 316, 20, 20 }, false },     // 146: EPC_15
    {{ 256, 336, 8, 16 }, false },      // 147: GSSS    (Good Soldier Standing Still)
    {{ 264, 336, 8, 16 }, false },      // 148: GSWR_0  (Good Soldier Walking Right)   
    {{ 272, 336, 8, 16 }, false },      // 149: GSWR_1
    {{ 280, 336, 8, 16 }, false },      // 150: GSWR_2
    {{ 288, 336, 8, 16 }, false },      // 151: GSWR_3
    {{ 296, 336, 8, 16 }, false },      // 152: GSWR_4
    {{ 304, 336, 8, 16 }, false },      // 153: GSWR_5
    {{ 312, 336, 8, 16 }, false },      // 154: GSWL_0  (Good Soldier Walking Left)   
    {{ 320, 336, 8, 16 }, false },      // 155: GSWL_1
    {{ 328, 336, 8, 16 }, false },      // 156: GSWL_2
    {{ 336, 336, 8, 16 }, false },      // 157: GSWL_3
    {{ 344, 336, 8, 16 }, false },      // 158: GSWL_4
    {{ 352, 336, 8, 16 }, false },      // 159: GSWL_5
    {{ 360, 336, 8, 16 }, false },      // 160: GSFG_0  (Good Soldier Firing Gun)
    {{ 368, 336, 8, 16 }, false },      // 161: GSFG_1
    {{ 376, 336, 8, 16 }, false },      // 162: GSFG_2
    {{ 384, 336, 8, 16 }, false },      // 163: GSFG_3
    {{ 392, 336, 8, 16 }, false },      // 164: GSFG_4
    {{ 400, 336, 8, 16 }, false },      // 165: GSFG_5
    {{ 408, 336, 8, 16 }, false },      // 166: GSFG_6
    {{ 416, 336, 8, 16 }, false },      // 167: GSFG_7
    {{ 424, 336, 8, 16 }, false },      // 168: GSFG_8
    {{ 432, 336, 8, 16 }, false },      // 169: GSFG_9
    {{ 440, 336, 8, 16 }, false },      // 170: GSFG_10
    {{ 448, 336, 8, 16 }, false },      // 171: GSFG_11
    {{ 456, 336, 8, 16 }, false },      // 172: GSFG_12
    {{ 464, 336, 8, 16 }, false },      // 173: GSFG_13
    {{ 472, 336, 8, 16 }, false },      // 174: GSFG_14
    {{ 480, 336, 8, 16 }, false },      // 175: GSFG_15
    {{ 256, 352, 8, 16 }, false },      // 176: ESSS    (Evil Soldier Standing Still)
    {{ 264, 352, 8, 16 }, false },      // 177: ESWR_0  (Evil Soldier Walking Right)   
    {{ 272, 352, 8, 16 }, false },      // 178: ESWR_1
    {{ 280, 352, 8, 16 }, false },      // 179: ESWR_2
    {{ 288, 352, 8, 16 }, false },      // 180: ESWR_3
    {{ 296, 352, 8, 16 }, false },      // 181: ESWR_4
    {{ 304, 352, 8, 16 }, false },      // 182: ESWR_5
    {{ 312, 352, 8, 16 }, false },      // 183: ESWL_0  (Evil Soldier Walking Left)   
    {{ 320, 352, 8, 16 }, false },      // 184: ESWL_1
    {{ 328, 352, 8, 16 }, false },      // 185: ESWL_2
    {{ 336, 352, 8, 16 }, false },      // 186: ESWL_3
    {{ 344, 352, 8, 16 }, false },      // 187: ESWL_4
    {{ 352, 352, 8, 16 }, false },      // 188: ESWL_5
    {{ 360, 352, 8, 16 }, false },      // 189: ESFG_0  (Evil Soldier Firing Gun)
    {{ 368, 352, 8, 16 }, false },      // 190: ESFG_1
    {{ 376, 352, 8, 16 }, false },      // 191: ESFG_2
    {{ 384, 352, 8, 16 }, false },      // 192: ESFG_3
    {{ 392, 352, 8, 16 }, false },      // 193: ESFG_4
    {{ 400, 352, 8, 16 }, false },      // 194: ESFG_5
    {{ 408, 352, 8, 16 }, false },      // 195: ESFG_6
    {{ 416, 352, 8, 16 }, false },      // 196: ESFG_7
    {{ 424, 352, 8, 16 }, false },      // 197: ESFG_8
    {{ 432, 352, 8, 16 }, false },      // 198: ESFG_9
    {{ 440, 352, 8, 16 }, false },      // 199: ESFG_10
    {{ 448, 352, 8, 16 }, false },      // 200: ESFG_11
    {{ 456, 352, 8, 16 }, false },      // 201: ESFG_12
    {{ 464, 352, 8, 16 }, false },      // 202: ESFG_13
    {{ 472, 352, 8, 16 }, false },      // 203: ESFG_14
    {{ 480, 352, 8, 16 }, false },      // 204: ESFG_15
    {{ 416, 256, 16, 32 }, false },     // 205: Good Platoon Flag
    {{ 432, 256, 16, 32 }, false },     // 206: Evil Platoon Flag
    {{ 256, 368, 24, 40 }, false },     // 207: GPF_0 (Good Platoon Factory)
    {{ 280, 368, 24, 40 }, false },     // 208: GPF_1
    {{ 304, 368, 24, 40 }, false },     // 209: GPF_2
    {{ 328, 368, 24, 40 }, false },     // 210: GPF_3
    {{ 352, 368, 24, 40 }, false },     // 211: GPF_4
    {{ 376, 368, 24, 40 }, false },     // 212: GPF_5
    {{ 400, 368, 24, 40 }, false },     // 213: GPF_6
    {{ 424, 368, 24, 40 }, false },     // 214: GPF_7
    {{ 256, 408, 24, 40 }, false },     // 215: EPF_0 (Evil Platoon Factory)
    {{ 280, 408, 24, 40 }, false },     // 216: EPF_1
    {{ 304, 408, 24, 40 }, false },     // 217: EPF_2
    {{ 328, 408, 24, 40 }, false },     // 218: EPF_3
    {{ 352, 408, 24, 40 }, false },     // 219: EPF_4
    {{ 376, 408, 24, 40 }, false },     // 220: EPF_5
    {{ 400, 408, 24, 40 }, false },     // 221: EPF_6
    {{ 424, 408, 24, 40 }, false },     // 222: EPF_7
    {{ 448, 256, 32, 64 }, false },     // 223: Good Flag
    {{ 480, 256, 32, 64 }, false },     // 224: Evil Flag
    {{ 512, 0, 496, 96 }, true },       // 225: Game Over Banner
    {{ 512, 96, 400, 96 }, true },      // 226: You Win! Banner
    {{ 512, 192, 30, 16 }, false },     // 227: User Interface Bullet Icon
    {{ 542, 192, 36, 30 }, false },     // 228: User Interface Bomb Icon
    {{ 578, 192, 25, 26 }, false },     // 229: User Interface Heart Icon
    {{ 912, 96, 80, 80 }, false },      // 230: User Interface Speedometer Icon
    {{ 0, 512, 80, 80 }, false },       // 231: User Interface Fuel Guage Icon
    {{ 512, 256, 32, 32 }, false },     // 232: GT_0 (Good Tank)
    {{ 544, 256, 32, 32 }, false },     // 233: GT_1
    {{ 576, 256, 32, 32 }, false },     // 234: GT_2
    {{ 608, 256, 32, 32 }, false },     // 235: GT_3
    {{ 640, 256, 32, 32 }, false },     // 236: GT_4
    {{ 672, 256, 32, 32 }, false },     // 237: GT_5
    {{ 704, 256, 32, 32 }, false },     // 238: GT_6
    {{ 736, 256, 32, 32 }, false },     // 239: GT_7
    {{ 768, 256, 32, 32 }, false },     // 240: GT_8
    {{ 800, 256, 32, 32 }, false },     // 241: GT_9
    {{ 832, 256, 32, 32 }, false },     // 242: GT_10
    {{ 864, 256, 32, 32 }, false },     // 243: GT_11
    {{ 896, 256, 32, 32 }, false },     // 244: GT_12
    {{ 928, 256, 32, 32 }, false },     // 245: GT_13
    {{ 960, 256, 32, 32 }, false },     // 246: GT_14
    {{ 992, 256, 32, 32 }, false },     // 247: GT_15
    {{ 512, 288, 32, 32 }, false },     // 248: GT_16
    {{ 544, 288, 32, 32 }, false },     // 249: GT_17
    {{ 576, 288, 32, 32 }, false },     // 250: GT_18
    {{ 608, 288, 32, 32 }, false },     // 251: GT_19
    {{ 640, 288, 32, 32 }, false },     // 252: GT_20
    {{ 672, 288, 32, 32 }, false },     // 253: GT_21
    {{ 704, 288, 32, 32 }, false },     // 254: GT_22
    {{ 736, 288, 32, 32 }, false },     // 255: GT_23
    {{ 768, 288, 32, 32 }, false },     // 256: GT_24
    {{ 800, 288, 32, 32 }, false },     // 257: GT_25
    {{ 832, 288, 32, 32 }, false },     // 258: GT_26
    {{ 864, 288, 32, 32 }, false },     // 259: GT_27
    {{ 896, 288, 32, 32 }, false },     // 260: GT_28
    {{ 928, 288, 32, 32 }, false },     // 261: GT_29
    {{ 960, 288, 32, 32 }, false },     // 262: GT_30
    {{ 992, 288, 32, 32 }, false },     // 263: GT_31
    {{ 512, 320, 32, 32 }, false },     // 264: GT_32
    {{ 544, 320, 32, 32 }, false },     // 265: GT_33
    {{ 576, 320, 32, 32 }, false },     // 266: GT_34
    {{ 608, 320, 32, 32 }, false },     // 267: GT_35
    {{ 640, 320, 32, 32 }, false },     // 268: GT_36
    {{ 672, 320, 32, 32 }, false },     // 269: GT_37
    {{ 704, 320, 32, 32 }, false },     // 270: GT_38
    {{ 736, 320, 32, 32 }, false },     // 271: GT_39
    {{ 768, 320, 32, 32 }, false },     // 272: GT_40
    {{ 800, 320, 32, 32 }, false },     // 273: GT_41
    {{ 832, 320, 32, 32 }, false },     // 274: GT_42
    {{ 864, 320, 32, 32 }, false },     // 275: GT_43
    {{ 896, 320, 32, 32 }, false },     // 276: GT_44
    {{ 928, 320, 32, 32 }, false },     // 277: GT_45
    {{ 960, 320, 32, 32 }, false },     // 278: GT_46
    {{ 992, 320, 32, 32 }, false },     // 279: GT_47
    {{ 512, 352, 32, 32 }, false },     // 280: GT_48
    {{ 544, 352, 32, 32 }, false },     // 281: GT_49
    {{ 576, 352, 32, 32 }, false },     // 282: GT_50
    {{ 608, 352, 32, 32 }, false },     // 283: GT_51
    {{ 640, 352, 32, 32 }, false },     // 284: GT_52
    {{ 672, 352, 32, 32 }, false },     // 285: GT_53
    {{ 704, 352, 32, 32 }, false },     // 286: GT_54
    {{ 736, 352, 32, 32 }, false },     // 287: GT_55
    {{ 768, 352, 32, 32 }, false },     // 288: GT_56
    {{ 800, 352, 32, 32 }, false },     // 289: GT_57
    {{ 832, 352, 32, 32 }, false },     // 290: GT_58
    {{ 864, 352, 32, 32 }, false },     // 291: GT_59
    {{ 896, 352, 32, 32 }, false },     // 292: GT_60
    {{ 928, 352, 32, 32 }, false },     // 293: GT_61
    {{ 960, 352, 32, 32 }, false },     // 294: GT_62
    {{ 992, 352, 32, 32 }, false },     // 295: GT_63
    {{ 512, 384, 32, 32 }, false },     // 296: GT_64
    {{ 544, 384, 32, 32 }, false },     // 297: GT_65
    {{ 576, 384, 32, 32 }, false },     // 298: GT_66
    {{ 608, 384, 32, 32 }, false },     // 299: GT_67
    {{ 640, 384, 32, 32 }, false },     // 300: GT_68
    {{ 672, 384, 32, 32 }, false },     // 301: GT_69
    {{ 704, 384, 32, 32 }, false },     // 302: GT_70
    {{ 736, 384, 32, 32 }, false },     // 303: GT_71
    {{ 768, 384, 32, 32 }, false },     // 304: GT_72
    {{ 800, 384, 32, 32 }, false },     // 305: GT_73
    {{ 832, 384, 32, 32 }, false },     // 306: ET_0 (Evil Tank)
    {{ 864, 384, 32, 32 }, false },     // 307: ET_1
    {{ 896, 384, 32, 32 }, false },     // 308: ET_2
    {{ 928, 384, 32, 32 }, false },     // 309: ET_3
    {{ 960, 384, 32, 32 }, false },     // 310: ET_4
    {{ 992, 384, 32, 32 }, false },     // 311: ET_5
    {{ 512, 416, 32, 32 }, false },     // 312: ET_6
    {{ 544, 416, 32, 32 }, false },     // 313: ET_7
    {{ 576, 416, 32, 32 }, false },     // 314: ET_8
    {{ 608, 416, 32, 32 }, false },     // 315: ET_9
    {{ 640, 416, 32, 32 }, false },     // 316: ET_10
    {{ 672, 416, 32, 32 }, false },     // 317: ET_11
    {{ 704, 416, 32, 32 }, false },     // 318: ET_12
    {{ 736, 416, 32, 32 }, false },     // 319: ET_13
    {{ 768, 416, 32, 32 }, false },     // 320: ET_14
    {{ 800, 416, 32, 32 }, false },     // 321: ET_15
    {{ 832, 416, 32, 32 }, false },     // 322: ET_16
    {{ 864, 416, 32, 32 }, false },     // 323: ET_17
    {{ 896, 416, 32, 32 }, false },     // 324: ET_18
    {{ 928, 416, 32, 32 }, false },     // 325: ET_19
    {{ 960, 416, 32, 32 }, false },     // 326: ET_20
    {{ 992, 416, 32, 32 }, false },     // 327: ET_21
    {{ 512, 448, 32, 32 }, false },     // 328: ET_22
    {{ 544, 448, 32, 32 }, false },     // 329: ET_23
    {{ 576, 448, 32, 32 }, false },     // 330: ET_24
    {{ 608, 448, 32, 32 }, false },     // 331: ET_25
    {{ 640, 448, 32, 32 }, false },     // 332: ET_26
    {{ 672, 448, 32, 32 }, false },     // 333: ET_27
    {{ 704, 448, 32, 32 }, false },     // 334: ET_28
    {{ 736, 448, 32, 32 }, false },     // 335: ET_29
    {{ 768, 448, 32, 32 }, false },     // 336: ET_30
    {{ 800, 448, 32, 32 }, false },     // 337: ET_31
    {{ 832, 448, 32, 32 }, false },     // 338: ET_32
    {{ 864, 448, 32, 32 }, false },     // 339: ET_33
    {{ 896, 448, 32, 32 }, false },     // 340: ET_34
    {{ 928, 448, 32, 32 }, false },     // 341: ET_35
    {{ 960, 448, 32, 32 }, false },     // 342: ET_36
    {{ 992, 448, 32, 32 }, false },     // 343: ET_37
    {{ 512, 480, 32, 32 }, false },     // 344: ET_38
    {{ 544, 480, 32, 32 }, false },     // 345: ET_39
    {{ 576, 480, 32, 32 }, false },     // 346: ET_40
    {{ 608, 480, 32, 32 }, false },     // 347: ET_41
    {{ 640, 480, 32, 32 }, false },     // 348: ET_42
    {{ 672, 480, 32, 32 }, false },     // 349: ET_43
    {{ 704, 480, 32, 32 }, false },     // 350: ET_44
    {{ 736, 480, 32, 32 }, false },     // 351: ET_45
    {{ 768, 480, 32, 32 }, false },     // 352: ET_46
    {{ 800, 480, 32, 32 }, false },     // 353: ET_47
    {{ 832, 480, 32, 32 }, false },     // 354: ET_48
    {{ 864, 480, 32, 32 }, false },     // 355: ET_49
    {{ 896, 480, 32, 32 }, false },     // 356: ET_50
    {{ 928, 480, 32, 32 }, false },     // 357: ET_51
    {{ 960, 480, 32, 32 }, false },     // 358: ET_52
    {{ 992, 480, 32, 32 }, false },     // 359: ET_53
    {{ 512, 512, 32, 32 }, false },     // 360: ET_54
    {{ 544, 512, 32, 32 }, false },     // 361: ET_55
    {{ 576, 512, 32, 32 }, false },     // 362: ET_56
    {{ 608, 512, 32, 32 }, false },     // 363: ET_57
    {{ 640, 512, 32, 32 }, false },     // 364: ET_58
    {{ 672, 512, 32, 32 }, false },     // 365: ET_59
    {{ 704, 512, 32, 32 }, false },     // 366: ET_60
    {{ 736, 512, 32, 32 }, false },     // 367: ET_61
    {{ 768, 512, 32, 32 }, false },     // 368: ET_62
    {{ 800, 512, 32, 32 }, false },     // 369: ET_63
    {{ 832, 512, 32, 32 }, false },     // 370: ET_64
    {{ 864, 512, 32, 32 }, false },     // 371: ET_65
    {{ 896, 512, 32, 32 }, false },     // 372: ET_66
    {{ 928, 512, 32, 32 }, false },     // 373: ET_67
    {{ 960, 512, 32, 32 }, false },     // 374: ET_68
    {{ 992, 512, 32, 32 }, false },     // 375: ET_69
    {{ 512, 544, 32, 32 }, false },     // 376: ET_70
    {{ 544, 544, 32, 32 }, false },     // 377: ET_71
    {{ 576, 544, 32, 32 }, false },     // 378: ET_72
    {{ 608, 544, 32, 32 }, false },     // 379: ET_73
    {{ 0, 208, 40, 40 }, false },       // 380: Good Platoon Icon
    {{ 40, 208, 40, 40 }, false },      // 381: Evil Platoon Icon
    {{ 80, 208, 40, 40 }, false },      // 382: Good TankSquad Icon
    {{ 120, 208, 40, 40 }, false },     // 383: Evil TankSquad Icon
    {{ 256, 448, 24, 40 }, false },     // 384: GTF_0 (Good TankSquad Factory)
    {{ 280, 448, 24, 40 }, false },     // 385: GTF_1
    {{ 304, 448, 24, 40 }, false },     // 386: GTF_2
    {{ 328, 448, 24, 40 }, false },     // 387: GTF_3
    {{ 352, 448, 24, 40 }, false },     // 388: GTF_4
    {{ 376, 448, 24, 40 }, false },     // 389: GTF_5
    {{ 400, 448, 24, 40 }, false },     // 390: GTF_6
    {{ 424, 448, 24, 40 }, false },     // 391: GTF_7
    {{ 256, 488, 24, 40 }, false },     // 392: ETF_0 (Evil TankSquad Factory)
    {{ 280, 488, 24, 40 }, false },     // 393: ETF_1
    {{ 304, 488, 24, 40 }, false },     // 394: ETF_2
    {{ 328, 488, 24, 40 }, false },     // 395: ETF_3
    {{ 352, 488, 24, 40 }, false },     // 396: ETF_4
    {{ 376, 488, 24, 40 }, false },     // 397: ETF_5
    {{ 400, 488, 24, 40 }, false },     // 398: ETF_6
    {{ 424, 488, 24, 40 }, false },     // 399: ETF_7
    {{ 240, 20, 8, 8 }, false },        // 400: Test Frame 1
    {{ 248, 20, 8, 12 }, false }        // 401: Test Frame 2
};

//------------------------------------------------------------------------------
// Sprites Implementation
//------------------------------------------------------------------------------
Sprites *Sprites::Instance = NULL;

//------------------------------------------------------------------------------
// Name: Sprites
// Desc: Loads in the sprites used within the game, then transfers them to a 
//       software surface. IMG_Load may return an image on a software surface
//       but the documentation does not say. A software surface is used as it
//       will not need to be locked. I don't know why but locking a surface 
//       and then blitting from it in the same frame seems to introduce
//       massive slowdown.
//------------------------------------------------------------------------------
Sprites::Sprites() 
{
    SDL_Surface *TempIMG;

    CU_AddErrorMessages(SPR_ErrorMessages);

    SDLI = SDL_Interface::GetInstance(NULL);
    
    TempIMG = IMG_Load(CAM_SPRITES_IMG);
    if (TempIMG == NULL) {
        SDE_SDLError(__FILE__, __LINE__);
    }

    // Set up the main sprites image
    SpritesIMG = SDL_ConvertSurface(TempIMG, TempIMG->format, SDL_SWSURFACE);
    if (SpritesIMG == NULL) SDE_SDLError(__FILE__, __LINE__);

    SDL_SetAlpha(SpritesIMG, 0, SDL_ALPHA_TRANSPARENT);
    SDL_SetColorKey(SpritesIMG, SDL_SRCCOLORKEY, 
        SDL_MapRGBA(SpritesIMG->format, 255, 0, 255, SDL_ALPHA_TRANSPARENT));
    
    SDL_FreeSurface(TempIMG);
}

//------------------------------------------------------------------------------
// Name: ~Sprites
// Desc: Shutsdown the Sprites class
//------------------------------------------------------------------------------
Sprites::~Sprites() 
{
    if (SpritesIMG != NULL) {
        SDL_FreeSurface(SpritesIMG);
        SpritesIMG = NULL;
    }
}

//------------------------------------------------------------------------------
// Name: GetInstance
// Desc: Returns or creates the single instance of this class
//------------------------------------------------------------------------------
Sprites *Sprites::GetInstance() 
{
    if (Instance == NULL) {
        Instance = new Sprites();
    }
    return Instance; 
}

//------------------------------------------------------------------------------
// Name: DrawFrame
// Desc: Draws the requested frame to an SDL_Surface at the xy co-ordinates
//       supplied. The user has the option of either just specifying target
//       coordinates or also specifying an area of the frame which is to be 
//       drawn.
//------------------------------------------------------------------------------
bool Sprites::DrawFrame(uint Frame, SDL_Surface *Surface, int x, int y)
{
    SDL_Rect TargetRect = { x, y, 0, 0 };

    assert(Surface != NULL);

    TargetRect.w = Frames[Frame].Rect.w;
    TargetRect.h = Frames[Frame].Rect.h;

    return InternalDrawFrame(Frame, Surface, &Frames[Frame].Rect, &TargetRect);
}

bool Sprites::DrawFrame(uint Frame, SDL_Surface *Surface, int x, int y, 
                        SDL_Rect FrameArea)
{
    SDL_Rect TargetRect = { x, y, 0, 0 };

    assert(Surface != NULL);

    // Handle the trivial cases
    if (FrameArea.x > Frames[Frame].Rect.w 
        || FrameArea.y > Frames[Frame].Rect.h) return false;

    // Adjust FrameArea so that the area it refers to is the actual Frame in
    // question
    FrameArea.x += Frames[Frame].Rect.x;
    FrameArea.y += Frames[Frame].Rect.y;
    if (FrameArea.x + FrameArea.w > Frames[Frame].Rect.x + Frames[Frame].Rect.w) {
        FrameArea.w = (Frames[Frame].Rect.x + Frames[Frame].Rect.w) - FrameArea.x;
    }
    if (FrameArea.y + FrameArea.h > Frames[Frame].Rect.y + Frames[Frame].Rect.h) {
        FrameArea.h = (Frames[Frame].Rect.y + Frames[Frame].Rect.h) - FrameArea.y;
    }

    TargetRect.w = FrameArea.w;
    TargetRect.h = FrameArea.h;

    return InternalDrawFrame(Frame, Surface, &FrameArea, &TargetRect);
}

//------------------------------------------------------------------------------
// Name: InternalDrawFrame
// Desc: This function blits an area from 'SpritesIMG' (supposedly from 
//       'Frame') onto 'Surface' in the area specified by 'TargetRect'. This
//       function makes no real checks to see if its arguments are sensible
//       or not so be careful.
//------------------------------------------------------------------------------
bool Sprites::InternalDrawFrame(uint Frame, SDL_Surface *Surface, 
        SDL_Rect *SourceRect, SDL_Rect *TargetRect)
{
    bool Result;

    if (Frames[Frame].UsesAlphaBlending) {

        // If a frame uses alpha blending then the flag SDL_SRCALPHA is set on
        // the SpritesIMG surface before the frame is drawn. After the frame 
        // is drawn SDL_SRCALPHA is removed and the colour key reset. This may
        // seem like a huge amount of overhead to draw a single frame but compared
        // to the overall time it takes to draw a frame with alpha blending it
        // is not that much at all. An alternative that was tried out was to have
        // a duplicate surface with SDL_SRCALPHA set. However, this uses twice the
        // memory for no noticeable increase in performance.
        CU_STARTFRAME(CAM_P_DRAW_ALPHA_ENTITIES);

        SDL_SetAlpha(SpritesIMG, SDL_SRCALPHA, SDL_ALPHA_TRANSPARENT);

        Result = SDL_SUCCEEDED(SDL_BlitSurface(SpritesIMG, SourceRect, 
            Surface, TargetRect));

        SDL_SetAlpha(SpritesIMG, 0, SDL_ALPHA_TRANSPARENT);
        SDL_SetColorKey(SpritesIMG, SDL_SRCCOLORKEY, 
            SDL_MapRGBA(SpritesIMG->format, 255, 0, 255, SDL_ALPHA_TRANSPARENT));

        CU_ENDFRAME(CAM_P_DRAW_ALPHA_ENTITIES);
    } else {
        CU_STARTFRAME(CAM_P_DRAW_NON_ALPHA_ENTITIES);

        Result = SDL_SUCCEEDED(SDL_BlitSurface(SpritesIMG, SourceRect, 
            Surface, TargetRect));

        CU_ENDFRAME(CAM_P_DRAW_NON_ALPHA_ENTITIES);
    }
    return Result;
}


//------------------------------------------------------------------------------
// Name: GetFrameDims
// Desc: Returns the width and height of the requested frame
//------------------------------------------------------------------------------
void Sprites::GetFrameDims(uint Frame, int *w, int *h)
{
    *w = Frames[Frame].Rect.w;
    *h = Frames[Frame].Rect.h;
}

//------------------------------------------------------------------------------
// Name: FindCollision
// Desc: This is a collection of functions all under the same name that are
//       used for collision detection. The method used to look for a collision
//       (i.e. a corresponding pair of non-transparent pixels) is the same in
//       all functions and the overloading is simply used to make it easier to
//       find collisions between frames. The areas compared in these functions
//       must be of the same size.
//
//       This function is used when you know that two areas of SDL_Surfaces
//       have a common area but want to see whether there is any collision at
//       a pixel level.
//------------------------------------------------------------------------------
bool Sprites::FindCollision(uint F1, SDL_Rect F1_Area, 
                            uint F2, SDL_Rect F2_Area)
{
    // Adjust the FrameArea rectangles so that their x and y 
    // co-ordinates are correct
    F1_Area.x += Frames[F1].Rect.x;
    F1_Area.y += Frames[F1].Rect.y;
    F2_Area.x += Frames[F2].Rect.x;
    F2_Area.y += Frames[F2].Rect.y;

    return FindCollision(SpritesIMG, &F1_Area, SpritesIMG, &F2_Area);
}

bool Sprites::FindCollision(uint Frame, SDL_Rect FrameArea,
                            SDL_Surface *Surface, SDL_Rect SurfaceArea)
{
    // Adjust the FrameArea Rect so that its x and y co-ordinates are correct
    FrameArea.x += Frames[Frame].Rect.x;
    FrameArea.y += Frames[Frame].Rect.y;

    return FindCollision(SpritesIMG, &FrameArea, Surface, &SurfaceArea);
}

bool Sprites::FindCollision(SDL_Surface *S1, const SDL_Rect *S1_Area,
                            SDL_Surface *S2, const SDL_Rect *S2_Area)
{
    bool retval = false;
    uint i, j;
   
    uint *S1_CurPixel, *S2_CurPixel;
    uint S1_Jump, S2_Jump;
    uint Width = S1_Area->w;
    uint Height = S1_Area->h;
    uint BytesPP = S1->format->BytesPerPixel;
    uchar Alpha1, Alpha2;

    // Check that the rectangles are the same size
    assert(S1_Area->w == S2_Area->w && S1_Area->h == S2_Area->h);

    // Check that the surfaces use an identical method for storing pixels
    if (!SDLI->SameSurfaceFormat(S1, S2))
        FailWithError(SPR_ERROR_INCOMPATIBLE_SURFACES);

    // Lock the surfaces
    if (SDL_MUSTLOCK(S1)) {
        if (SDL_FAILED(SDL_LockSurface(S1)))
            SDF_SDLError(__FILE__, __LINE__);
    }
    if (SDL_MUSTLOCK(S2)) {
        if (SDL_FAILED(SDL_LockSurface(S2))) {
            SDL_UnlockSurface(S1);
            SDF_SDLError(__FILE__, __LINE__);
        }
    }

    // Work out where the pixels we are interested in start and how far we need
    // to jump in between rows
    S1_CurPixel = (uint *)((uchar *)S1->pixels + S1->pitch*S1_Area->y 
        + S1_Area->x*BytesPP);
    S2_CurPixel = (uint *)((uchar *)S2->pixels + S2->pitch*S2_Area->y 
        + S2_Area->x*BytesPP);
    S1_Jump = S1->pitch / BytesPP - Width;
    S2_Jump = S2->pitch / BytesPP - Width;

    assert(S1_Jump >= 0 && S2_Jump >= 0);
    
    for (j = 0; j < Height; j++) {
        for (i = 0; i < Width; i++) {
            
            Alpha1 = ((*S1_CurPixel)&S1->format->Amask) >> S1->format->Ashift;
            Alpha2 = ((*S2_CurPixel)&S2->format->Amask) >> S2->format->Ashift;
            if (Alpha1 >= COLLISION_THRESHOLD && Alpha2 >= COLLISION_THRESHOLD) {
                retval = true;
                goto Finish;
            }
            
            S1_CurPixel++;
            S2_CurPixel++;
        }

        S1_CurPixel += S1_Jump;
        S2_CurPixel += S2_Jump;
    }
    
Finish:
    // Unlock the surfaces and return
    SDL_UnlockSurface(S1);
    SDL_UnlockSurface(S2);
    return retval;
}
