/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// Name: AutoPlane.cpp
// Desc: An AutoPlane is a plane that has the ability to fly automatically,
//       without the need for human intervention. This class provides
//       routines that let a plane fly towards a given point in the world 
//       whilst doing its best to avoid collisions with the landscape and
//       other planes. This class also provides routines that let a plane fly
//       home and refuel, these routines are used by both the PlayerPlane and
//       EnemyPlane classes.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <list>
#include "CamelMath.h"
#include "Game.h"
#include "SDL_Interface.h"
#include "InputHandler.h"
#include "Sprites.h"
#include "Level.h"
#include "Landscape.h"
#include "AutoPlane.h"

using namespace std;

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------
const int AutoPlane::NEAR_WAY_POINT = 20;
const int AutoPlane::COLLISION_RANGE = 100;

// Determines the rate at which the plane checks to see if it will
// collide with something when considering whether a given action
// is dangerous.
const int AutoPlane::COLLISION_CHECK_RATE = 30;

const Point2D AutoPlane::HOME_WAYPOINT_OFFSETS[NUM_WAYPOINTS_ON_ROUTE_HOME-1] = {
    Point2D(600, -190),
    Point2D(300, -90),
    Point2D(150, -30),
    Point2D(100, -20),
    Point2D(50, -10)
};

//------------------------------------------------------------------------------
// AutoPlane Implementation
//------------------------------------------------------------------------------
AutoPlane::AutoPlane(Level *L, PLANE_STATE State, Point2D Pos, 
                     bool Inverted, int Direction, ENT_SIDE Side) :
    Plane(L, State, Pos, Inverted, Direction, Side) 
{
    int i;
    int Multiplier;

    FlyingHome = false;
    CurWayPoint = 0;
    ResetAutopilot();

    // Calculate the route home given that Pos is the place we want to fly to
    Multiplier = (Side == GOOD ? 1 : -1);

    for (i = 0; i < NUM_WAYPOINTS_ON_ROUTE_HOME-1; i++) {
        RouteHome[i].Location.x = Pos.x + Multiplier*HOME_WAYPOINT_OFFSETS[i].x;
        RouteHome[i].Location.y = Pos.y + HOME_WAYPOINT_OFFSETS[i].y;
        RouteHome[i].NextPoint = i + 1;
    }
    RouteHome[NUM_WAYPOINTS_ON_ROUTE_HOME-1].Location = Pos;
    RouteHome[NUM_WAYPOINTS_ON_ROUTE_HOME-1].NextPoint = 0;
}

AutoPlane::~AutoPlane() {}

//------------------------------------------------------------------------------
// Name: FlyHome
// Desc: If this plane was created by a PlaneSpawnPoint then this routine
//       will try to fly it back to this spawn point so that it can refuel 
//       and rearm. The exact route home that this routine takes depends upon
//       the side that the plane is fighting for. GOOD planes always approach
//       from the right and EVIL planes always approach from the left.
//
// Note: When using this routine, you must call it until the plane starts
//       refuelling. If you want to abort flying home then you must call
//       'StopFlyingHome'.
//
//       You must also make sure that the 'Plane::Commands' array of 
//       booleans is set to 'false' before calling this routine as they may
//       otherwise interfere with the autopilot.
//------------------------------------------------------------------------------
void AutoPlane::FlyHome()
{
    bool ReachedHome = false;

    if (!FlyingHome) {
        CurWayPoint = 0;
        FlyingHome = true;
    }

    // Update the waypoint
    if (GetCentrePos().DistanceTo(
        RouteHome[CurWayPoint].Location) <= NEAR_WAY_POINT) {

        // Have we got home?
        if (CurWayPoint == NUM_WAYPOINTS_ON_ROUTE_HOME - 1) ReachedHome = true;

        CurWayPoint = RouteHome[CurWayPoint].NextPoint;
        // This is a bit of a botch but seems to work in most cases
        if (CurWayPoint > 0) EngineAcc = MAX_ENGINE_ACC / 2;
    }
    
    if (ReachedHome) {
        // If we have reached home then touch down and begin refuelling
        WorldPos = RouteHome[NUM_WAYPOINTS_ON_ROUTE_HOME-1].Location;
        RealX = (float)WorldPos.x;
        RealY = (float)WorldPos.y;

        State = LANDED;
        Velocity = 0;
        EngineAcc = 0;
        EngineOn = false;
        
        if (Side == GOOD) {
            Inverted = false;
            Direction = 0;
        } else {
            Inverted = true;
            Direction = 8;
        }
        Frame = CalcFrame(Direction, Inverted);
        Refuelling = true;

        // Reset the planes actions which are linked to motion
        Actions[CLIMB]->Reset();
        Actions[DIVE]->Reset();
        Actions[INVERT]->Reset();

        FlyingHome = false;
    } else {
        FlyTo(RouteHome[CurWayPoint].Location);
    }
}

//------------------------------------------------------------------------------
// Name: StopFlyingHome
// Desc: Stops the plane from flying home and resets the planes engine 
//       acceleration.
//------------------------------------------------------------------------------
void AutoPlane::StopFlyingHome() 
{
    FlyingHome = false;
    EngineAcc = MAX_ENGINE_ACC;
    ResetAutopilot();
}

//------------------------------------------------------------------------------
// Name: ResetAutopilot
// Desc: Simply resets the status variables used by the 'FlyTo' routines so
//       that they don't get confused.
//------------------------------------------------------------------------------
void AutoPlane::ResetAutopilot()
{
    OldTargetDirection = 0;
    ManoeuvreCompleted = true;
}

//------------------------------------------------------------------------------
// Name: FlyTo
// Desc: This subroutine is in charge of getting the plane from its current
//       position to a given point 'p'. It does this by issuing commands to the
//       underlying plane object such as CLIMB, DIVE and INVERT. Whilst
//       navigating to the target point it must also do a reasonable job of
//       keeping the plane out of harms way. The subroutines 3 priorities are
//       in order of importance
//
//       1. Keep the plane from hitting the Landscape and/or from stalling.
//       2. Keep the plane from hitting another plane. (This could be extended
//          to keeping it from hitting other entity types such as troops, but
//          we don't really want it dodging bullets).
//       3. Fly the plane in a fairly efficient and safe manner towards the
//          given point 'p'.
//
// Note: If you stop using this routine for a while (say to let the player fly 
//       the plane) and then come back to it. You must call 'ResetAutopilot'
//       first to avoid any possible strange behaviour.
//------------------------------------------------------------------------------
void AutoPlane::FlyTo(Point2D p)
{
    // Get the precise target direction and general directions for both the 
    // plane and the target
    CamelMath::GENERAL_DIRECTION GenDirection = CamelMath::GetGeneralDirection(Direction);
    int TargetDirection = 
        CamelMath::GetDirectionFromBearing(GetCentrePos().BearingTo(p));
    CamelMath::GENERAL_DIRECTION TargetGenDirection = 
        CamelMath::GetGeneralDirection(TargetDirection);
    int TempDir, TempTargetDir;
    list<Entity *>::iterator Iter;
    Point2D CollisionPoint(0, 0);
    int CollisionDirection;
    
    // Avoid landscape and stalling
    if (IsDangerous(NO_ACTION)) {
        Commands[CLIMB] = true;
        return;
    }

    // Avoid collison with other planes
    for (Iter = L->GameEntities[PLANE].begin();
        Iter != L->GameEntities[PLANE].end(); Iter++) {
        if (*Iter != this) {
            // Test to see if a collision is possible any time soon
            if (IsCollisionLikely((Plane *)*Iter, &CollisionPoint, &CollisionDirection)) {
                // Set the target direction directly away from the collision
                int DirMod = 2;//CamelMath::NUM_DIRECTIONS / 4;

                if (CollisionDirection == 0) {
                    DirMod = ((float)rand() / RAND_MAX < 0.5f ? DirMod : -DirMod);
                } else {
                    DirMod *= -CollisionDirection;
                }

                TargetDirection = CamelMath::NormaliseDirection(Direction + DirMod);

                if (Direction <= CamelMath::UP_DIRECTION + 1 && 
                    Direction >= CamelMath::UP_DIRECTION - 1) {
                    if ((*Iter)->GetWorldY() < WorldPos.y && (float)rand() / RAND_MAX < 0.5f) 

                        TargetDirection = CamelMath::DOWN_DIRECTION;
                }

                TargetGenDirection = CamelMath::GetGeneralDirection(TargetDirection);
                ManoeuvreCompleted = true;  // Forget whatever else we were doing
                break;
            }
        }
    }

    if (!ManoeuvreCompleted) {
        if (Direction == OldTargetDirection) {
            ManoeuvreCompleted = true;
        } else {
            Commands[CLIMB] = true;
            return;
        }
    }


    // Fly towards target
    if (TargetDirection == Direction) {
        // Everything else seems alright so check that we are upright
        if (GenDirection == CamelMath::LEFT) {
            // If travelling to the left you should be inverted to be upright
            if (!Inverted && !IsDangerous(INVERT)) Commands[INVERT] = true;
        } else if (GenDirection == CamelMath::RIGHT) {
            // If travelling to the right you should be non inverted to be upright
            if (Inverted && !IsDangerous(INVERT)) Commands[INVERT] = true;
        }
        return;
    } else if (GenDirection == CamelMath::UP) {
        if (TargetGenDirection == CamelMath::LEFT) {
            if (Inverted) {
                if (!IsDangerous(INVERT)) Commands[INVERT] = true;
            } else {
                if (!IsDangerous(CLIMB)) Commands[CLIMB] = true;
            }
        } else {
            if (!Inverted) {
                if (!IsDangerous(INVERT)) Commands[INVERT] = true;
            } else {
                if (!IsDangerous(CLIMB)) Commands[CLIMB] = true;
            }
        }
    } else if (GenDirection == CamelMath::DOWN) {
        if (TargetGenDirection == CamelMath::LEFT) {
            if (!Inverted) {
                if (!IsDangerous(INVERT)) Commands[INVERT] = true;
            } else {
                if (!IsDangerous(CLIMB)) Commands[CLIMB] = true;
            }
        } else {
            if (Inverted) {
                if (!IsDangerous(INVERT)) Commands[INVERT] = true;
            } else {
                if (!IsDangerous(CLIMB)) Commands[CLIMB] = true;
            }
        }
    } else if (GenDirection == CamelMath::LEFT) {
        if (TargetGenDirection != CamelMath::LEFT ) {
            if (!Inverted) {
                if (!IsDangerous(INVERT)) Commands[INVERT] = true;
            } else {
                // Here we need to go over the top and do not want to be 
                // interrupted till we are done.
                if (!IsDangerous(CLIMB)) {
                    Commands[CLIMB] = true;
                    OldTargetDirection = TargetDirection;
                    ManoeuvreCompleted = false;
                }
            }
        } else {
            if (TargetDirection < Direction) {
                if (!Inverted) {
                    if (!IsDangerous(INVERT)) Commands[INVERT] = true;
                } else {
                    if (!IsDangerous(CLIMB)) Commands[CLIMB] = true;
                }
            } else {
                if (!Inverted) {
                    if (!IsDangerous(INVERT)) Commands[INVERT] = true;
                } else {
                    if (!IsDangerous(DIVE)) Commands[DIVE] = true;
                }
            }
        }
    } else { // GenDirection == CamelMath::RIGHT
        if (TargetGenDirection != CamelMath::RIGHT) {
            if (Inverted) {
                if (!IsDangerous(INVERT)) Commands[INVERT] = true;
            } else {
                // Here we need to go over the top and do not want to be 
                // interrupted till we are done.
                if (!IsDangerous(CLIMB)) {
                    Commands[CLIMB] = true;
                    OldTargetDirection = TargetDirection;
                    ManoeuvreCompleted = false;
                }
            }
        } else {
            // Get rid of the big step between direction 15 and 0
            TempDir = (Direction < CamelMath::UP_DIRECTION ? 
                Direction + CamelMath::NUM_DIRECTIONS : Direction);
            TempTargetDir = (TargetDirection < CamelMath::UP_DIRECTION ? 
                TargetDirection + CamelMath::NUM_DIRECTIONS : TargetDirection);

            if (TempTargetDir > TempDir) {
                if (Inverted) {
                    if (!IsDangerous(INVERT)) Commands[INVERT] = true;
                } else {
                    if (!IsDangerous(CLIMB)) Commands[CLIMB] = true;
                }
            } else {
                if (Inverted) {
                    if (!IsDangerous(INVERT)) Commands[INVERT] = true;
                } else {
                    if (!IsDangerous(DIVE)) Commands[DIVE] = true;
                }
            }
        }
    }
}

//------------------------------------------------------------------------------
// Name: IsDangerous
// Desc: This function tries to answer the question that asks whether 
//       performing a certain action is dangerous or not. An action is
//       dangerous if it causes the plane to come in contact with the 
//       landscape (kaboom!), or if it could cause the plane to stall.
//
//       A brief explanation of why each action could be dangerous is given
//       at the appropriate clause in the switch statement below. Bear in
//       mind that this function does not take into account all possible 
//       actions and may indeed produce a false positive at some times as
//       safety margins have been incorporated in some tests.
//
//       FIXME: Needs to be improved a lot to take into account trees, and
//       also the fact that actions will probably take a bit longer than 
//       the specified rates.
//------------------------------------------------------------------------------
bool AutoPlane::IsDangerous(PLANE_ACTIONS AnAction)
{
    int i;
    uint dt;
    float SecondsPassed;
    int NumChecks;
    float LA_Velocity = Velocity;
    float LA_RealX = RealX;
    float LA_RealY = RealY;
    int LA_Direction = Direction;
    int LA_Frame = Frame;
    bool LA_Inverted = Inverted;

    switch (AnAction) {
        case NO_ACTION:
            // If NO_ACTION is specified then the function looks at whether 
            // the plane would be in danger from the landscape. The best action 
            // to avoid this danger is assumed to be to climb at least once and
            // in the worst case twice in a row. This is assuming that the 
            // plane is flying in an upright position and that the best way 
            // to avoid the landscape is to pull up. Two climbs may be needed 
            // as at the moment the land tends to rise in steps of 45 degrees 
            // or PI_4 radians.
            //
            // After the possibility of landscape collision has been checked
            // then the possibility of a stall is considered.

            // Calculate where the plane would be if it did not climb right now
            dt = CLIMBING_RATE + COLLISION_CHECK_RATE;
            SecondsPassed = (float)dt / 1000.0f;

            LA_Velocity = CalcNewVelocity(LA_Velocity, LA_Direction, dt);
            LA_RealX += CamelMath::XY_COMPONENTS[LA_Direction].x*LA_Velocity*SecondsPassed;
            LA_RealY += CamelMath::XY_COMPONENTS[LA_Direction].y*LA_Velocity*SecondsPassed;
       
            // Is it dangerous not to climb now?
            if (LS->HasCollided(LA_Frame, Point2D((int)LA_RealX, (int)LA_RealY), false)) return true;

            // Now simulate a climb and see if anything bad would happen before 
            // we were able to climb again.
            LA_Direction += (LA_Inverted ? -1 : 1);
            LA_Direction = CamelMath::NormaliseDirection(LA_Direction);
            LA_Frame = CalcFrame(LA_Direction, LA_Inverted);

            NumChecks = CLIMBING_RATE / COLLISION_CHECK_RATE;
            dt = COLLISION_CHECK_RATE;
            SecondsPassed = (float)dt / 1000.0f;

            for (i = 0; i < NumChecks; i++) {
                LA_Velocity = CalcNewVelocity(LA_Velocity, LA_Direction, dt);
                LA_RealX += CamelMath::XY_COMPONENTS[LA_Direction].x
                    * LA_Velocity*SecondsPassed;
                LA_RealY += CamelMath::XY_COMPONENTS[LA_Direction].y
                    * LA_Velocity*SecondsPassed;
                if (LS->HasCollided(LA_Frame, Point2D((int)LA_RealX, (int)LA_RealY), false)) 
                    return true;
            }

            // Could we stall if we took no action?
            if (IsStallLikely(Velocity, Direction, Inverted)) return true;

            break;
        case DIVE:
            // A DIVE might be dangerous if it put us on a collision course
            // with the landscape after we had just pulled up to avoid it.
            // It is assumed that a DIVE would never put us in danger of a 
            // stall as the default method of moving through the upward 
            // pointing directions is to climb through them.
            //
            // Here we are assuming that the action NO_ACTION has been checked 
            // and been found to be non dangerous. We are also assuming that
            // it takes longer to dive than it does to climb. It is safe to 
            // make these assumptions here as they have also been made 
            // throughout the rest of the code that controls this plane.
            
            // The first frame that IsDangerous(NO_ACTION) did not look at
            dt = CLIMBING_RATE + 2*COLLISION_CHECK_RATE;
            
            if ((uint)DIVING_RATE < dt) dt = DIVING_RATE;
            SecondsPassed = (float)dt / 1000.0f;

            // Get to the start frame
            LA_Velocity = CalcNewVelocity(LA_Velocity, LA_Direction, dt);
            LA_RealX += CamelMath::XY_COMPONENTS[LA_Direction].x*LA_Velocity*SecondsPassed;
            LA_RealY += CamelMath::XY_COMPONENTS[LA_Direction].y*LA_Velocity*SecondsPassed;
            
            if (dt != DIVING_RATE) {
                // If not at the diving frame then check for collisions up to
                // that point
                NumChecks = (DIVING_RATE - dt) / COLLISION_CHECK_RATE;
                dt = COLLISION_CHECK_RATE;
                SecondsPassed = (float)dt / 1000.0f;

                for (i = 0; i < NumChecks; i++) {
                    LA_Velocity = CalcNewVelocity(LA_Velocity, LA_Direction, dt);
                    LA_RealX += CamelMath::XY_COMPONENTS[LA_Direction].x
                        * LA_Velocity*SecondsPassed;
                    LA_RealY += CamelMath::XY_COMPONENTS[LA_Direction].y
                        *  LA_Velocity*SecondsPassed;
                    if (LS->HasCollided(LA_Frame, Point2D((int)LA_RealX, (int)LA_RealY), false)) 
                        return true;
                }
            }
            // Now simulate a dive and see if anything bad would happen before 
            // we were able to climb back.
            LA_Direction += (LA_Inverted ? 1 : -1);
            LA_Direction = CamelMath::NormaliseDirection(LA_Direction);
            LA_Frame = CalcFrame(LA_Direction, LA_Inverted);

            // Collsion in the diving frame?
            if (LS->HasCollided(LA_Frame, Point2D((int)LA_RealX, (int)LA_RealY), false)) 
                return true;

            dt = CLIMBING_RATE;
            SecondsPassed = (float)dt / 1000.0f;
            LA_Velocity = CalcNewVelocity(LA_Velocity, LA_Direction, dt);
            LA_RealX += CamelMath::XY_COMPONENTS[LA_Direction].x*LA_Velocity*SecondsPassed;
            LA_RealY += CamelMath::XY_COMPONENTS[LA_Direction].y*LA_Velocity*SecondsPassed;
            if (LS->HasCollided(LA_Frame, Point2D((int)LA_RealX, (int)LA_RealY), false)) 
                return true;

            break;
        case CLIMB:
            // The main worry with a CLIMB is that it could take the plane into
            // a position where a stall is likely

            dt = CLIMBING_RATE;
            SecondsPassed = (float)dt / 1000.0f;

            // Work out what the velocity would be before the climb took place
            LA_Velocity = CalcNewVelocity(LA_Velocity, LA_Direction, dt);
       
            // Simulate the climb
            LA_Direction += (LA_Inverted ? -1 : 1);
            LA_Direction = CamelMath::NormaliseDirection(LA_Direction);

            if (IsStallLikely(LA_Velocity, LA_Direction, LA_Inverted)) 
                return true;

            break;
        case INVERT:
            // Inverting may be dangerous if it stops us from pulling out of
            // a stall at a later point in time.
            dt = INVERTING_RATE;
            SecondsPassed = (float)dt / 1000.0f;
            LA_Velocity = CalcNewVelocity(LA_Velocity, LA_Direction, dt);
            LA_Inverted = !LA_Inverted;

            if (IsStallLikely(LA_Velocity, LA_Direction, LA_Inverted)) 
                return true;

            break;
    }

    return false;
}

//------------------------------------------------------------------------------
// Name: IsStallLikely
// Desc: This function carries out a check to make sure that the airspeed of 
//       the plane will not drop too low if we continue flying in the given
//       direction. It is only possible to stall when flying straight up or 
//       close to straight up. The solution to danger here would be
//------------------------------------------------------------------------------
bool AutoPlane::IsStallLikely(float CurVelocity, int CurDirection, 
                              bool CurInverted)
{
    uint dt;
    float SecondsPassed;
    const float SAFETY_MARGIN = 15.0f;

    float LA_Velocity = CurVelocity;
    int LA_Direction = CurDirection;

    // Calculate what the velocity of the plane would be if it had to climb
    // out of trouble right now. Well actually, not quite now. A safety margin
    // has been included to protect borderline cases.
    while (LA_Direction <= CamelMath::UP_DIRECTION + 1
        && LA_Direction >= CamelMath::UP_DIRECTION - 1) {
        
        dt = CLIMBING_RATE;
        SecondsPassed = (float)dt / 1000.0f;
        LA_Velocity = CalcNewVelocity(LA_Velocity, LA_Direction, dt);
        
        // Simulate a climb
        LA_Direction += (CurInverted ? -1 : 1);
        LA_Direction = CamelMath::NormaliseDirection(LA_Direction);
    }

    if (LA_Velocity <= MINIMUM_AIRSPEED + SAFETY_MARGIN) {
        return true;
    } else {
        return false;
    }
}

//------------------------------------------------------------------------------
// Name: IsCollisionLikely
// Desc: This is a very rough and ready function that could definitely do
//       with reworking, ecspecially if it is going to take into account 
//       entities other than planes. What it basically does is predict if a
//       collision might occur between this plane and another given plane.
//       This is very hard to do however as the movement of a plane can be
//       very complicated and unpredictable. Therefore this function basically
//       takes a rough guess by assuming that each plane will travel in a 
//       straight line in one of three directions. By checking all possible 
//       combinations of these routes the function can say whether a collision
//       is likely along with the point at which it will occur and the
//       direction from the current plane to the collision.
//------------------------------------------------------------------------------
bool AutoPlane::IsCollisionLikely(Plane *APlane, Point2D *CollisionPoint, int *CollisionDirection)
{
    const int LOOKAHEAD_DISTANCE = 20;
    
    int i, j, k;
    uint dt;
    float SecondsPassed;
    float LA_P1_Velocity;
    float LA_P1_RealX, LA_P1_RealY;
    int LA_P1_Frame, LA_P1_Direction[3];
    float LA_P2_Velocity;
    float LA_P2_RealX, LA_P2_RealY;
    int LA_P2_Frame, LA_P2_Direction[3];

    assert(APlane != NULL && CollisionPoint != NULL && CollisionDirection != NULL);

    LA_P1_Direction[0] = CamelMath::NormaliseDirection(Direction - 1);
    LA_P1_Direction[1] = Direction;
    LA_P1_Direction[2] = CamelMath::NormaliseDirection(Direction + 1);
    LA_P2_Direction[0] = CamelMath::NormaliseDirection(APlane->GetDirection() - 1);
    LA_P2_Direction[1] = APlane->GetDirection();
    LA_P2_Direction[2] = CamelMath::NormaliseDirection(APlane->GetDirection() + 1);

    dt = COLLISION_CHECK_RATE;
    SecondsPassed = (float)dt / 1000.0f;

    for (i = 0; i < 3; i++) {
        for (j = 0; j < 3; j++) {
            // Initialise variables
            LA_P1_Velocity = Velocity;
            LA_P1_RealX = RealX;
            LA_P1_RealY = RealY;
            LA_P1_Frame = Frame;
            LA_P2_Velocity = APlane->GetVelocity();
            LA_P2_RealX = (float)APlane->GetWorldX();  // Should be close enough as this is
            LA_P2_RealY = (float)APlane->GetWorldY();  // only a rough estimate
            LA_P2_Frame = APlane->GetFrame();

            for (k = 0; k < LOOKAHEAD_DISTANCE; k++) {
                LA_P1_Velocity = CalcNewVelocity(LA_P1_Velocity, LA_P1_Direction[i], dt);
                LA_P1_RealX += CamelMath::XY_COMPONENTS[LA_P1_Direction[i]].x*LA_P1_Velocity*SecondsPassed;
                LA_P1_RealY += CamelMath::XY_COMPONENTS[LA_P1_Direction[i]].y*LA_P1_Velocity*SecondsPassed;

                LA_P2_Velocity = CalcNewVelocity(LA_P2_Velocity, LA_P2_Direction[j], dt);
                LA_P2_RealX += CamelMath::XY_COMPONENTS[LA_P2_Direction[j]].x*LA_P2_Velocity*SecondsPassed;
                LA_P2_RealY += CamelMath::XY_COMPONENTS[LA_P2_Direction[j]].y*LA_P2_Velocity*SecondsPassed;

                if (TestForCollision(LA_P1_Frame, Point2D((int)LA_P1_RealX, (int)LA_P1_RealY),
                    LA_P2_Frame, Point2D((int)LA_P2_RealX, (int)LA_P2_RealY))) {

                    CollisionPoint->x = (int)LA_P1_RealX;
                    CollisionPoint->y = (int)LA_P1_RealY;
                    *CollisionDirection = i - 1;
                    return true;
                }
            }
        }
    }

    return false;
}
