/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Bomb.cpp
// Desc: This entity can be dropped by planes to cause mucho devastation
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <assert.h>
#include "Game.h"
#include "Level.h"
#include "Landscape.h"
#include "SDL_Interface.h"
#include "Sprites.h"
#include "Explosion.h"
#include "Bomb.h"

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------
const int Bomb::CHECKING_DIRECTION_RATE = 100;

//------------------------------------------------------------------------------
// Bomb Implementation
//------------------------------------------------------------------------------
Bomb::Bomb(Level *L, Point2D CentrePos, int Direction, float Velocity) : Entity(L)
{
    Type = BOMB;
    // Define the victims of a bomb
    NumVictims = 5;
    Victims[0] = PLANE;
    Victims[1] = FACTORY;
    Victims[2] = BOMB;
    Victims[3] = SOLDIER;
    Victims[4] = TANK;

    MaxCollisions = 1;

    this->Direction = Direction%CamelMath::NUM_DIRECTIONS;
    VelocityVector.x = CamelMath::XY_COMPONENTS[Direction].x*Velocity;
    VelocityVector.y = CamelMath::XY_COMPONENTS[Direction].y*Velocity;

    // Get the width and height of the bomb and then place it in 
    // the world based on its centre point
    SPR->GetFrameDims(Sprites::BOMB_SPRITE, &Width, &Height);
    WorldPos = Point2D(CentrePos.x - Width/2, CentrePos.y - Height/2);
	RealX = (float)WorldPos.x;
	RealY = (float)WorldPos.y;
    WorldPosLastDirectionCheck = WorldPos;

    // Add in a falling action to make stuff happen
    Actions[CHECK_DIRECTION] = new ACTION(0, CHECKING_DIRECTION_RATE,
        false, Check_Direction);

    CalcFrame();
}

Bomb::~Bomb() {}

void Bomb::CalcFrame()
{
    Frame = Sprites::BOMB_SPRITE + Direction;
}

void Bomb::Update(uint dt)
{
    XY_COMPONENT AccelerationVector;
    float SecondsPassed = (float)dt / 1000.0f;

    AccelerationVector.x = 
        (-L->GetAirResistance(Type, VelocityVector.x)) * SecondsPassed;
    AccelerationVector.y = 
        //CamelMath::MAX_GRAVITY_ACC * SecondsPassed;
        (CamelMath::MAX_GRAVITY_ACC - L->GetAirResistance(Type, VelocityVector.y)) * SecondsPassed;

    VelocityVector.x += AccelerationVector.x;
    VelocityVector.y += AccelerationVector.y;
	RealX += VelocityVector.x*SecondsPassed;
	RealY += VelocityVector.y*SecondsPassed;
	WorldPos = Point2D((int)RealX, (int)RealY);

    UpdateAction(Actions[CHECK_DIRECTION], true, dt, this);
	CalcFrame();
}

void Bomb::FindAndHandleCollisionWithLandscape()
{
    Explosion *Exp;

    if (LS->HasCollided(Frame, WorldPos, true)) {
        NeedsDeletion = true;

        // Create an explosion
        Exp = new Explosion(L, Explosion::SMALL, GetCentrePos());
        L->AddNewEntity(Exp);
    }
}

void Bomb::HandleCollision(Entity *AnEntity)
{
	Explosion *Exp;

	// If this bomb has already been marked for deletion (in a previous collision)
	// then do nothing
	if (!NeedsDeletion) {
		NeedsDeletion = true;

        // Create an explosion
        Exp = new Explosion(L, Explosion::SMALL, GetCentrePos());
        L->AddNewEntity(Exp);
	}
}

void Bomb::Check_Direction(void *Object)
{
    Bomb *ABomb = (Bomb *)Object;

    ABomb->Direction = CamelMath::GetDirectionFromBearing(
        ABomb->WorldPosLastDirectionCheck.BearingTo(ABomb->WorldPos));
    /*if (ABomb->Direction != 12) {
		if (ABomb->Direction == 4) {
			// If pointing straight up then topple either left or right
			if ((float)rand() / RAND_MAX < 0.5f) 
				ABomb->Direction = 5;
			else 
				ABomb->Direction = 3;
		} else if (ABomb->Direction >= 5 && ABomb->Direction <= 11) {
			ABomb->Direction++;
		} else {
			if (ABomb->Direction == 0) {
				ABomb->Direction = CamelMath::NUM_DIRECTIONS-1;
			} else {
				ABomb->Direction--;
			}
		}
	}*/
    ABomb->WorldPosLastDirectionCheck = ABomb->WorldPos;
}
