/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Bullet.cpp
// Desc: This entity is used to shoot down enemies (and maybe also allies).
//       Originally, this entity was used only for plane bullets, however it
//       has been extended so that it can be used for soldier bullets as well.
//
// Note: It would be cleaner and more OO to break this down into subclasses
//       for the different types of bullet. However this has not been done
//       yet because it was much quicker and simpler to leave it as is.
//
//       All bullet vectors must be unit vectors.
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <assert.h>
#include "Game.h"
#include "CamelMath.h"
#include "Landscape.h"
#include "SDL_Interface.h"
#include "Sprites.h"
#include "Explosion.h"
#include "Bullet.h"

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------
const float Bullet::BULLET_SPEED = 600.0f;   // Pixels per second

//------------------------------------------------------------------------------
// Bomb Implementation
//------------------------------------------------------------------------------
Bullet::Bullet(Level *L, Point2D CentrePos, XY_COMPONENT XY_Direction, 
        BULLET_TYPE BulletType, uint Range, ENT_SIDE Side) : Entity(L)
{
    Type = BULLET;
    this->BulletType = BulletType;
    // Define the victims of a bullet
    NumVictims = 5;
    Victims[0] = PLANE;
    Victims[1] = FACTORY;
    Victims[2] = BOMB;
    Victims[3] = SOLDIER;
    Victims[4] = TANK;

    // All bullet vectors must be unit vectors but we can't really test this
    // in an assertion as they are stored with floats, which may cause rounding 
    // errors

    // Not so much the side of the bullet as the side of its firer
    this->Side = Side;
    // Only want these bullets to be able to hit one entity at a time
    if (BulletType == SOLDIER_BULLET || BulletType == TANK_BULLET) {
        MaxCollisions = 1;
    }

    this->XY_Direction = XY_Direction;
    this->Range = Range;
    RangeTravelled = 0;

    // Get the width and height of the bullet and then place it in 
    // the world based on its centre point
    SPR->GetFrameDims(Sprites::BULLET_SPRITE, &Width, &Height);
    WorldPos = Point2D(CentrePos.x - Width/2, CentrePos.y - Height/2);
	RealX = (float)WorldPos.x;
	RealY = (float)WorldPos.y;
	Frame = Sprites::BULLET_SPRITE;
	Velocity = BULLET_SPEED;
}

Bullet::~Bullet() {}

void Bullet::Update(uint dt)
{
    float RangeIncrement;
    float SecondsPassed = (float)dt / 1000.0f;

    RangeIncrement = Velocity*SecondsPassed;
    RealX += XY_Direction.x*RangeIncrement;
	RealY += XY_Direction.y*RangeIncrement;
	WorldPos = Point2D((int)RealX, (int)RealY);
    RangeTravelled += RangeIncrement;

    if (RangeTravelled >= Range) NeedsDeletion = true;
}

//------------------------------------------------------------------------------
// Name: HandleCollision
// Desc: How a bullet reacts when it collides with another entity depends upon
//       the bullets type and what exactly the other entity is.
//------------------------------------------------------------------------------
void Bullet::HandleCollision(Entity *AnEntity)
{
    ENT_TYPE EType;
    Explosion *Exp;

    if (BulletType == SOLDIER_BULLET || BulletType == TANK_BULLET) {
        EType = AnEntity->GetType();
        if ((EType == SOLDIER || EType == TANK) 
            && AnEntity->GetSide() == Side) {
            return; // No friendly fire amongst ground units
        }

        if (BulletType == TANK_BULLET) {
            if (AnEntity->GetType() == EXPLOSION) return;
            // Create an explosion
			Exp = new Explosion(L, Explosion::SMALL, GetCentrePos());
			L->AddNewEntity(Exp);
        }
    }
    NeedsDeletion = true;
}
