/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: ControlledEntity.cpp
// Desc: A ControlledEntity is a FightingEntity that can be viewed through
//       the in game OrdersPanel. The ControlledEntity provides details of its
//       current orders, the number of entities it contains and an icon in the
//       form of a sprite. It must also respond to the predefined orders
//       ADVANCE, HOLD and WITHDRAW.
//
// Note: A ControlledEntity is usually viewed as a unit containing a number
//       of subordinate entities. Therefore it must also maintain a count of
//       the number of subordinate entities that it contains and how many of
//       them are left alive.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <string>
#include "Sprites.h"
#include "Level.h"
#include "Landscape.h"
#include "OrdersPanel.h"
#include "ControlledEntity.h"

using namespace std;

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------
const ControlledEntity::Order ControlledEntity::ADVANCE = ControlledEntity::Order("Advance");
const ControlledEntity::Order ControlledEntity::HOLD = ControlledEntity::Order("Hold");
const ControlledEntity::Order ControlledEntity::WITHDRAW = ControlledEntity::Order("Withdraw");

//------------------------------------------------------------------------------
// ControlledEntity::Order Implementation
//------------------------------------------------------------------------------
ControlledEntity::Order::Order(const char *OrderString)
{
    this->OrderString = string(OrderString);
}

const char *ControlledEntity::Order::GetOrderString() const
{
    return OrderString.c_str();
}

//------------------------------------------------------------------------------
// ControlledEntity Implementation
//------------------------------------------------------------------------------
ControlledEntity::ControlledEntity(Level *L, ENT_SIDE Side) : FightingEntity(L) 
{
    OP = OrdersPanel::GetInstance(NULL, NULL);

    this->Side = Side;
    IconFrame = 0;
    CurOrders = &HOLD;
    OrdersHaveChanged = false;
    NumSubordinatesAlive = 0;

    // Set up the flag used to display the platoons destination
    if (Side == GOOD) {
        FlagFrame = Sprites::GOOD_CONTROLLED_ENTITY_FLAG_SPRITE;
    } else {
        FlagFrame = Sprites::EVIL_CONTROLLED_ENTITY_FLAG_SPRITE;
    }
    SPR->GetFrameDims(FlagFrame, &FlagWidth, &FlagHeight);
}

ControlledEntity::~ControlledEntity() {}

//------------------------------------------------------------------------------
// Name: UpdateFlagPos
// Desc: A small helper function that makes sure that the flag used to
//       designate DestinationX is actually positioned at DestinationX.
//------------------------------------------------------------------------------
void ControlledEntity::UpdateFlagPos()
{
    FlagPos.x = DestinationX;
    FlagPos.y = LS->CalculatePixelHeight(FlagPos.x) - FlagHeight;
}

//------------------------------------------------------------------------------
// Name: UpdateDestinationX
// Desc: This function updates the destination of the ControlledEntity 
//       depending upon the value of 'NewOrders' and 'NewDestinationX'. It also
//       returns 'true' if the ControlledEntity's orders have changed and
//       false otherwise.
//------------------------------------------------------------------------------
bool ControlledEntity::UpdateDestinationX(const Order *NewOrders, int NewDestinationX)
{
    int OldDestinationX = DestinationX;

    if (NewOrders == &ADVANCE || NewOrders == &WITHDRAW) {
        DestinationX = NewDestinationX;
    } else if (NewOrders == &HOLD) {
        // Done to stop subordinates from re-arranging themselves after 
        // they have finished advancing or withdrawing.
        CurOrders = &HOLD;  
                                
        // Only need to change the destination if it is not near enough to
        // the ControlledEntity's current position
        if (abs(DestinationX - GetCentreX()) > GetNearDestinationX()) {
            DestinationX = GetCentreX();
        }
    } else {
        assert(false);      // Invalid order
    }

    return (CurOrders != NewOrders || DestinationX != OldDestinationX);
}

//------------------------------------------------------------------------------
// Name: SetOrders
// Desc: This routine sets te orders of the ControlledEntity. 
//------------------------------------------------------------------------------
void ControlledEntity::SetOrders(const Order *NewOrders, int NewDestinationX)
{
    // The following function updates the destination of the
    // ControlledEntity and decides whether or not the ControlledEntity's 
    // orders have actually changed.
    OrdersHaveChanged = UpdateDestinationX(NewOrders, NewDestinationX);
    if (OrdersHaveChanged) {
        
        // If the orders have changed then we record this fact and update the 
        // location of the destination flag.
        CurOrders = NewOrders;
        UpdateFlagPos();
    }
}

//------------------------------------------------------------------------------
// Name: Draw
// Desc: Draws a flag showing the destination of the ControlledEntity if the 
//       OrdersPanel is active and this is the selected unit.
//------------------------------------------------------------------------------
void ControlledEntity::Draw(SDL_Surface *ScreenBuffer)
{
    if (OP->GetActivated() && OP->GetCurUnit() == this) {
        SPR->DrawFrame(FlagFrame, ScreenBuffer, 
            LS->GetViewportX() + (FlagPos.x - LS->GetViewportWorldX()), 
            LS->GetViewportY() + (FlagPos.y - LS->GetViewportWorldY()));
    }
}
