/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: EnemyPlane.cpp
// Desc: Provides the functionality for the enemy planes (boo! hiss!) that will
//       try to thwart our valiant hero. 
//
//       At the moment, the plan is that the AI will be implemented as a 
//       finite state machine that moves from state to state according to
//       internal rules and external conditions. Wherever possible the plan is
//       to use simple rules that result in plausible emergent behaviour,
//       rather than try to account for every possible situation.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <list>
#include "CamelMath.h"
#include "Game.h"
#include "SDL_Interface.h"
#include "InputHandler.h"
#include "Sprites.h"
#include "Level.h"
#include "Landscape.h"
#include "EnemyPlane.h"

using namespace std;

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------
const int EnemyPlane::FIRING_RANGE = 150;
const int EnemyPlane::ATTACK_RANGE = 400;
const int EnemyPlane::BREAKOFF_RANGE = 500;
// This amount of fuel should give about 15 seconds to fly home to refuel,
// which should be more than enough.
const int EnemyPlane::BINGO_FUEL = 150;     

//------------------------------------------------------------------------------
// EnemyPlane Implementation
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Name: EnemyPlane
// Desc: Put the plane in a sensible state depending upon the initial 
//       parameters specified.
//------------------------------------------------------------------------------
EnemyPlane::EnemyPlane(Level *L, PLANE_STATE State, Point2D Pos, 
                       bool Inverted, int Direction, 
                       WAYPOINT PatrolPoints[MAX_NUM_PATROL_POINTS]) :
    AutoPlane(L, State, Pos, Inverted, Direction, EVIL) 
{
    switch (State) {
        case LANDED:
            AI_State = TAKING_OFF;
            break;
        case FLYING:
            AI_State = PATROLLING;
            break;
        default:
            AI_State = NO_STATE;
    }
    assert(PatrolPoints != NULL);
    memcpy(this->PatrolPoints,PatrolPoints, sizeof(WAYPOINT) * MAX_NUM_PATROL_POINTS);
    CurPatrolPoint = 0;
    CurTarget = NULL;

    OldTargetDirection = 0;
    ManoeuvreCompleted = true;
}

EnemyPlane::~EnemyPlane() 
{
    // Give the player some points
    L->AddToPlayerScore(POINTS_VALUE);
}

//------------------------------------------------------------------------------
// Name: Update
// Desc: Hopefully makes the EnemyPlane do something useful. At the moment its
//       ideal lifecycle is to takeoff and patrol along a predefined patrol
//       route. If a valid enemy target is spotted then it will try to attack
//       it though at the moment it cannot do bombing runs. Once an engaged
//       enemy is no more or it manages to get away, the EnemyPlane returns
//       to its apatrol route.
//------------------------------------------------------------------------------
void EnemyPlane::Update(uint dt)
{
    const float TAKEOFF_SAFETY_MARGIN = 60.0f;

    float TargetDistance;
    Point2D TargetLoc(0, 0);

    // Clear all previous commands
    memset(Commands, false, sizeof(bool)*NUM_ACTIONS);
    if (State == SPIRALING || State == CRASHING) {
        AI_State = NO_STATE;
    }

    switch (AI_State) {
        case NO_STATE:
            break;
        case REFUELLING:
            if (FuelLeft == STARTING_FUEL 
                && BulletsLeft == STARTING_BULLETS
                && BombsLeft == STARTING_BOMBS) {
                
                CurPatrolPoint = 0;
                CurTarget = NULL;
                AI_State = TAKING_OFF;
            }
            break;
        case TAKING_OFF:
            if (EngineOn) {
                // Safety margin included just in case
                if (Velocity > MINIMUM_AIRSPEED + TAKEOFF_SAFETY_MARGIN) 
                    AI_State = PATROLLING;
            } else {
                if (!Actions[START_ENGINE]->Active) Commands[START_ENGINE] = true;
            }
            break;
        case PATROLLING:
            // Update the patrol point
            if (GetCentrePos().DistanceTo(
                PatrolPoints[CurPatrolPoint].Location) <= NEAR_WAY_POINT) {
                CurPatrolPoint = PatrolPoints[CurPatrolPoint].NextPoint;
            }
            FlyTo(PatrolPoints[CurPatrolPoint].Location);

            // Test to see if we need to switch to a new state
            if (FuelLeft <= BINGO_FUEL || BulletsLeft == 0 || BombsLeft == 0) {
                AI_State = FLYING_HOME;
            } else {
                CurTarget = FindNewTarget();
                if (CurTarget != NULL) AI_State = ATTACKING;
            }
            break;
        case FLYING_HOME:
            FlyHome();
            if (!FlyingHome) AI_State = REFUELLING;
            break;
        case ATTACKING:
            assert(CurTarget != NULL);

            if (!L->IsEntityAlive(CurTarget)) {
                // Success! We have destroyed the enemy!
                CurTarget = NULL;
                AI_State = PATROLLING;
                break;
            }

            TargetLoc.x = CurTarget->GetCentreX();
            TargetLoc.y = CurTarget->GetCentreY();
            FlyTo(TargetLoc);
            
            if (FuelLeft <= BINGO_FUEL || BulletsLeft == 0 || BombsLeft == 0) {
                CurTarget = NULL;
                AI_State = FLYING_HOME;
            } else {
                TargetDistance = GetCentrePos().DistanceTo(TargetLoc);
                if (TargetDistance <= FIRING_RANGE) {
                    Commands[FIRE] = true;
                } else if (TargetDistance > BREAKOFF_RANGE) {
                    CurTarget = NULL;
                    AI_State = PATROLLING;
                }
            }
            break;
    }

    Plane::Update(dt);
}


//------------------------------------------------------------------------------
// Name: FindNewTarget
// Desc: This function loops through all of the planes possible targets and
//       tries to find a valid target. A valid target is one which is on the
//       side of 'Good' and is within the planes attack range. If a valid
//       target is found then a pointer to it is returned, if no valid target
//       is found then NULL is returned.
//------------------------------------------------------------------------------
Entity *EnemyPlane::FindNewTarget()
{
    list<Entity *>::iterator Iter;

    // FIXME: Should eventually target entities other than planes

    for (Iter = L->GameEntities[PLANE].begin();
        Iter != L->GameEntities[PLANE].end(); Iter++) {
            
        // Test to see if this entity is valid and within range           
        if (*Iter != this && (*Iter)->GetSide() == GOOD 
            && GetCentrePos().DistanceTo((*Iter)->GetCentrePos()) <= ATTACK_RANGE) {
            return *Iter;
        }
    }

    return NULL;
}
