/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Entity.cpp
// Desc: The abstract base class from which everything that moves, shoots 
//       and/or blows up in the game is derived.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <list>
#include <assert.h>
#include "CamelMath.h"
#include "Game.h"
#include "SDL_Interface.h"
#include "Sprites.h"
#include "Level.h"
#include "Landscape.h"
#include "InputHandler.h"
#include "Entity.h"

using namespace std;

//------------------------------------------------------------------------------
// struct ACTION Implementation
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Name: ACTION
// Desc: Constructs an ACTION, an action is is something that occurs at least
//       'Rate' milliseconds after it is activated.
//------------------------------------------------------------------------------
Entity::ACTION::ACTION(uint Delay, uint Rate, bool Active, 
                       void (*DoAction)(void *))
{
    assert(DoAction != NULL);

    this->Delay = Delay;
    this->Rate = Rate;
    this->Active = Active;
    this->DoAction = DoAction;
}

//------------------------------------------------------------------------------
// Name: Reset
// Desc: Resets an action so that its delay is set back to zero and it is
//       deactivated.
//------------------------------------------------------------------------------
void Entity::ACTION::Reset()
{
    Delay = 0;
    Active = false;
}

//------------------------------------------------------------------------------
// Entity Implementation
//------------------------------------------------------------------------------
const char *Entity::ENT_SIDE_STRINGS[NUM_SIDES] = {
    "GOOD",
    "EVIL"
};

//------------------------------------------------------------------------------
// Name: Entity
// Desc: An entity needs a few variables to be initialised every time it is set
//       up no matter what type of entity it is. This constructor takes care
//       of that and leaves the entity in a known state.
//------------------------------------------------------------------------------
Entity::Entity(Level *L) 
{
    assert(L != NULL);

    NeedsDeletion = false;
    G = Game::GetInstance();
    SDLI = SDL_Interface::GetInstance(G);
    IH = InputHandler::GetInstance(G);
    SPR = Sprites::GetInstance();
    this->L = L;
    LS = L->GetLandscape();

    NumVictims = 0;
    Side = NEUTRAL;
    Type = NO_TYPE;
    MaxCollisions = ABSOLUTE_MAX_COLLISIONS;
}

Entity::~Entity() {}

//------------------------------------------------------------------------------
// Name: UpdateAction
// Desc: This updates an action to take into account the passing of 'dt'
//       milliseconds. An action will take approximately its rate in 
//       milliseconds to occur once it has been started, although because 'dt'
//       can not be predicted it will probably take slightly longer.
//------------------------------------------------------------------------------
void Entity::UpdateAction(ACTION *AnAction, bool Command, uint dt, void *Object)
{
    if (Command) AnAction->Active = true;
    if (AnAction->Active) {
        AnAction->Delay += dt;  // Add on the time that has passed
        if (AnAction->Delay >= AnAction->Rate) {
            AnAction->DoAction(Object);
            AnAction->Delay = 0;
            AnAction->Active = false;
        }
    }
}

//------------------------------------------------------------------------------
// Name: HasCollided
// Desc: This function is used by an entity to answer the question of whether
//       or not it has collided with another entity. This virtual function
//       is only written in the most basic of terms and answers the question,
//       'If the other entity, consisted of this given frame at this given
//       point in the world, would you have collided with it?'. This means
//       that it can be used as is for single frame entities and expanded later 
//       for multiframe entities.
//------------------------------------------------------------------------------
bool Entity::HasCollided(int OtherEntityFrame, Point2D OtherEntityFramePos)
{
    return TestForCollision(Frame, WorldPos, OtherEntityFrame, OtherEntityFramePos);
}

//------------------------------------------------------------------------------
// Name: TestForCollision
// Desc: This function could probably be moved to the Sprites class but is kept
//       here for now because this is where it started. This is the fairly
//       high level part of collision detection where a test is made to see
//       if the frames that make up two entities overlap. If they do overlap
//       then the frames along with the rectangles (one for each frame) that
//       describe where they overlap are sent along to the Sprites object
//       for a more in depth pixel by pixel comparison.
//------------------------------------------------------------------------------
bool Entity::TestForCollision(int E1_Frame, Point2D E1Pos, int E2_Frame, Point2D E2Pos)
{
    bool XCollision, YCollision;
    int E1_Width, E1_Height;
    int E2_Width, E2_Height;
    int E1_RightX, E2_RightX;
    int E1_BottomY, E2_BottomY;
    SDL_Rect R1, R2;

    // Get frame dimensions
    SPR->GetFrameDims(E1_Frame, &E1_Width, &E1_Height);
    SPR->GetFrameDims(E2_Frame, &E2_Width, &E2_Height);

    E1_RightX = E1Pos.x + E1_Width;
    E2_RightX = E2Pos.x + E2_Width;
    if (E1Pos.x > E2Pos.x) {
        XCollision = E1Pos.x < E2Pos.x + E2_Width;
        if (XCollision) {
            R1.x = 0;
            R2.x = E1Pos.x - E2Pos.x;
            if (E1_RightX > E2_RightX) {
                R1.w = R2.w = E2_RightX - E1Pos.x;
            } else {
                R1.w = R2.w = E1_RightX - E1Pos.x;
            }
        }
    } else {
        XCollision = E2Pos.x < E1Pos.x + E1_Width;
        if (XCollision) {
            R2.x = 0;
            R1.x = E2Pos.x - E1Pos.x;
            if (E2_RightX > E1_RightX) {
                R1.w = R2.w = E1_RightX - E2Pos.x;
            } else {
                R1.w = R2.w = E2_RightX - E2Pos.x;
            }
        }
    }

    E1_BottomY = E1Pos.y + E1_Height;
    E2_BottomY = E2Pos.y + E2_Height;
    if (E1Pos.y > E2Pos.y) {
        YCollision = E1Pos.y < E2Pos.y + E2_Height;
        if (YCollision) {
            R1.y = 0;
            R2.y = E1Pos.y - E2Pos.y;
            if (E1_BottomY > E2_BottomY) {
                R1.h = R2.h = E2_BottomY - E1Pos.y;
            } else {
                R1.h = R2.h = E1_BottomY - E1Pos.y;
            }
        }
    } else {
        YCollision = E2Pos.y < E1Pos.y + E1_Height;
        if (YCollision) {
            R2.y = 0;
            R1.y = E2Pos.y - E1Pos.y;
            if (E2_BottomY > E1_BottomY) {
                R1.h = R2.h = E1_BottomY - E2Pos.y;
            } else {
                R1.h = R2.h = E2_BottomY - E2Pos.y;
            }
        }
    }

    if (XCollision && YCollision) {
        return SPR->FindCollision(E1_Frame, R1, E2_Frame, R2);
    } else {
        return false;
    }
}

//------------------------------------------------------------------------------
// Name: Draw
// Desc: Finds the entities screen position based upon the current viewport 
//       into the world and its world position and then draws its frame at
//       that position.
//------------------------------------------------------------------------------
void Entity::Draw(SDL_Surface *ScreenBuffer)
{
    SPR->DrawFrame(Frame, ScreenBuffer, 
        LS->GetViewportX() + (WorldPos.x - LS->GetViewportWorldX()), 
        LS->GetViewportY() + (WorldPos.y - LS->GetViewportWorldY()));
}

//------------------------------------------------------------------------------
// Name: FindCollisions
// Desc: When testing for collisions, we don't want every bullet testing to
//       see if it has collided with every plane and then every plane testing
//       to see if it has collided with every bullet as clearly this is a waste
//       of effort. Therefore every entity maintains an array of 'Victim Types'
//       which are the types of entities that it may collide with. An entity
//       then loops through all of its possible victims and asks them if it
//       has collided with them. If so then this fact is recorded, although
//       it must be noted that there is a maximum number of collisions that an 
//       entity can initiate in a frame and once this number is reached the
//       rest are ignored. 
//
// Note: This predator/prey approach to collision detection seems to work 
//       fairly well but there will still be duplication of effort between
//       entities of the same type. Fortunately, this has not been a problem
//       so far in the making of 'Camel'
//------------------------------------------------------------------------------
int Entity::FindCollisions(COLLISIONS_ARRAY *Collisions)
{
    int i;
    int NumCollisions = 0;
    list<Entity *>::iterator Iter;

    // Loop through all possible victims and see if there has been a collision
    for (i = 0; i < NumVictims; i++) {
        for (Iter = L->GameEntities[Victims[i]].begin();
            Iter != L->GameEntities[Victims[i]].end(); Iter++) {
            
            // Ask the victim if we have collided with them
            if (*Iter != this && (*Iter)->HasCollided(Frame, WorldPos)) {
                (*Collisions)[NumCollisions] = *Iter;
                NumCollisions++;
                if (NumCollisions >= MaxCollisions) return NumCollisions;
            }                        
        }
    }

    return NumCollisions;
}

//------------------------------------------------------------------------------
// Name: FindAndHandleCollisionWithLandscape
// Desc: Simply asks the landscape if the entities frame has collided with it
//       and marks the entity for deletion if it has.
//------------------------------------------------------------------------------
void Entity::FindAndHandleCollisionWithLandscape()
{
    if (LS->HasCollided(Frame, WorldPos, true)) NeedsDeletion = true;
}

//------------------------------------------------------------------------------
// Name: HandleCollision
// Desc: The default response to deletion is simply to disappear.
//------------------------------------------------------------------------------
void Entity::HandleCollision(Entity *AnEntity)
{
    NeedsDeletion = true;
}
