/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Explosion.cpp
// Desc: Things that make you go boom...
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <assert.h>
#include "CamelMath.h"
#include "SDL_Interface.h"
#include "Sprites.h"
#include "Landscape.h"
#include "Explosion.h"

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------
const int Explosion::NUM_FRAMES = 16;

const int Explosion::EXPLODING_RATE = 30;

//------------------------------------------------------------------------------
// Explosion Implementation
//------------------------------------------------------------------------------
Explosion::Explosion(Level *L, EXPLOSION_SIZE Size, Point2D CentrePos) : Entity(L)
{
    Type = EXPLOSION;
    TimePassed = 0;

    // Define the victims of an explosion
    NumVictims = 4;
    Victims[0] = PLANE;
    Victims[1] = BULLET;
    Victims[2] = BOMB;
    Victims[3] = SOLDIER;

    // Pick the exact explosion type at random
    if ((float)rand() / RAND_MAX < 0.5f) {
        ExplosionStartFrame = (Size == SMALL ?
            Sprites::SMALLEXPLOSION_1_SPRITE : Sprites::LARGEEXPLOSION_1_SPRITE);
    } else {
        ExplosionStartFrame = (Size == SMALL ?
            Sprites::SMALLEXPLOSION_2_SPRITE : Sprites::LARGEEXPLOSION_2_SPRITE);
    }
    Frame = ExplosionStartFrame;

    // Get the width and height of the explosion and then place it in 
    // the world based on its centre point
    SPR->GetFrameDims(ExplosionStartFrame, &Width, &Height);
    WorldPos = Point2D(CentrePos.x - Width/2, CentrePos.y - Height/2);

    // Add in an explode action to make stuff happen
    Actions[EXPLODE] = new ACTION(0, EXPLODING_RATE, false, Explode);

    // Start playing the sound
    ExplosionChannel = SDLI->PlayChannel(-1, 
        SDLI->Samples[SDL_Interface::BOMB_EXPLOSION], 0);

    // Calculate the sfx volume and stereo effect
    SFX_Volume = SDLI->VolumeOverDistance(
        GetCentrePos().DistanceTo(LS->GetViewportCentreWorldPos()));
    SDLI->GetStereoVolumes(LS->GetViewportWidth() / 2, 
        LS->GetViewportCentreWorldX() - GetCentreX(), 
        &LeftVolume, &RightVolume);

    // Apply the values just calculated to the ExplosionChannel
    if (ExplosionChannel != SDL_Interface::NO_CHANNEL) {
        Mix_Volume(ExplosionChannel, SFX_Volume);
        Mix_SetPanning(ExplosionChannel, LeftVolume, RightVolume);
    }
}

Explosion::~Explosion() {}

void Explosion::Update(uint dt)
{
    UpdateAction(Actions[EXPLODE], true, dt, this);
}

void Explosion::FindAndHandleCollisionWithLandscape() 
{
    // Do nothing
}

void Explosion::HandleCollision(Entity *AnEntity)
{
	// Do nothing
}

void Explosion::Explode(void *Object)
{
    Explosion *Exp = (Explosion *)Object;

    Exp->TimePassed++;
    Exp->Frame = Exp->ExplosionStartFrame + Exp->TimePassed;
    if (Exp->TimePassed >= NUM_FRAMES) Exp->NeedsDeletion = true;
}
