/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Factory.cpp
// Desc: This entity represents the small amount of resource management that 
//       there is in the game. The rate at which reinforcements arrive depends
//       upon the number of factories a side has standing.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <assert.h>
#include "Game.h"
#include "Landscape.h"
#include "SDL_Interface.h"
#include "Sprites.h"
#include "Level.h"
#include "Bullet.h"
#include "Explosion.h"
#include "Factory.h"

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------

// Points awarded per unit damage to the factory
const int Factory::POINTS_VALUE = 10;

// The hitpoints that a section starts with
const int Factory::INITIAL_SECTION_HITPOINTS = 20;
const int Factory::INITIAL_SECTION_HITPOINTS_2 = INITIAL_SECTION_HITPOINTS / 2;

// Damage taken from various objects
const int Factory::DAMAGE_TAKEN_FROM_BULLET = INITIAL_SECTION_HITPOINTS / 5;
const int Factory::DAMAGE_TAKEN_FROM_SOLDIERS_BULLET = 1;
const int Factory::DAMAGE_TAKEN_FROM_BOMB = INITIAL_SECTION_HITPOINTS_2;
const int Factory::DAMAGE_TAKEN_FROM_PLANE = INITIAL_SECTION_HITPOINTS;

// Constants to do with frames used for different parts of the factory
const int Factory::FRAME_OFFSETS[NUM_SECTIONS] = {
    0, 2, 4, 6
};

//------------------------------------------------------------------------------
// Factory Implementation
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Name: SECTION
// Desc: Sets up a SECTION.
//------------------------------------------------------------------------------
Factory::SECTION::SECTION() {}
Factory::SECTION::SECTION(Point2D WorldPos, int Frame)
{
    this->WorldPos = WorldPos;
    this->Frame = Frame;
    HitPoints = INITIAL_SECTION_HITPOINTS;
}

//------------------------------------------------------------------------------
// Name: Factory
// Desc: Creates a factory, this involves positioning the factory in the world
//       and making the various sections that make up the factory.
//------------------------------------------------------------------------------
Factory::Factory(Level *L, int x, FACTORY_TYPE FactoryType, ENT_SIDE Side) : Entity(L)
{
    this->Side = Side;
    Type = FACTORY;
    this->FactoryType = FactoryType;
    
    if (FactoryType == PLATOON_FACTORY) {
        if (Side == GOOD) {
            Frame = Sprites::GOOD_PLATOON_FACTORY_SPRITE;
        } else {
            Frame = Sprites::EVIL_PLATOON_FACTORY_SPRITE;
        }
    } else {
        if (Side == GOOD) {
            Frame = Sprites::GOOD_TANKSQUAD_FACTORY_SPRITE;
        } else {
            Frame = Sprites::EVIL_TANKSQUAD_FACTORY_SPRITE;
        }
    }
    SPR->GetFrameDims(Frame, &SectionWidth, &SectionHeight);
    Width = 3*SectionWidth;
    Height = 2*SectionHeight;

    // Position the factory in the world
    if (x < 0) {
        x = 0;
    } else if (x > LS->GetWorldPixelWidth() - Width) {
        x = LS->GetWorldPixelWidth() - Width;
    }
    WorldPos.x = x;
    WorldPos.y = LS->CalculatePixelHeight(x) - Height;

    // Create and position the sections that make up the factory
    Sections[CHIMNEYS] = SECTION(WorldPos, Frame + FRAME_OFFSETS[CHIMNEYS]);
    Sections[LEFT_SIDE] = SECTION(Point2D(WorldPos.x, WorldPos.y + SectionHeight), 
        Frame + FRAME_OFFSETS[LEFT_SIDE]);
    Sections[MIDDLE] = SECTION(Point2D(WorldPos.x + SectionWidth, WorldPos.y + SectionHeight), 
        Frame + FRAME_OFFSETS[MIDDLE]);
    Sections[RIGHT_SIDE] = SECTION(Point2D(WorldPos.x + 2*SectionWidth, WorldPos.y + SectionHeight), 
        Frame + FRAME_OFFSETS[RIGHT_SIDE]);

    SectionHit = NO_SECTION;
}

Factory::~Factory() {}

//------------------------------------------------------------------------------
// Name: Draw
// Desc: This involves drawing all of the sections that make up the factory
//------------------------------------------------------------------------------
void Factory::Draw(SDL_Surface *ScreenBuffer)
{
    int i;

    for (i = 0; i < NUM_SECTIONS; i++) {
        // Only draw the CHIMNEYS if they have the LEFT_SIDE to stand on
        if (i != CHIMNEYS 
            || (i == CHIMNEYS && Sections[LEFT_SIDE].HitPoints > INITIAL_SECTION_HITPOINTS_2)) {
            SPR->DrawFrame(Sections[i].Frame, ScreenBuffer, 
                LS->GetViewportX() + (Sections[i].WorldPos.x - LS->GetViewportWorldX()), 
                LS->GetViewportY() + (Sections[i].WorldPos.y - LS->GetViewportWorldY()));
        }
    }
}

void Factory::Update(uint dt)
{

}
    
//------------------------------------------------------------------------------
// Name: HasCollided
// Desc: A factory finds collisions by first checking against its outer 
//       bounds, and then if there has been a collsion there the factory 
//       checks each section individually to find out which one was hit.
//       As an entity will find all of its collisions and then handle them
//       all immediately, we know that when a call to HandleCollision next
//       comes it is from the entity that has hit us.
//------------------------------------------------------------------------------
bool Factory::HasCollided(int OtherEntityFrame, Point2D OtherEntityFramePos)
{
    int i;
    Point2D OE_Pos = OtherEntityFramePos;
    int OE_Width;
    int OE_Height;

    SPR->GetFrameDims(OtherEntityFrame, &OE_Width, &OE_Height);

    if (((WorldPos.x > OE_Pos.x && WorldPos.x < OE_Pos.x + OE_Width)
        || (OE_Pos.x >= WorldPos.x && OE_Pos.x < WorldPos.x + Width))
        &&
        ((WorldPos.y > OE_Pos.y && WorldPos.y < OE_Pos.y + OE_Height)
        || (OE_Pos.y >= WorldPos.y && OE_Pos.y < WorldPos.y + Height))) {
    
        // The other entity is within the factories space, but has there been a
        // collision?
        for (i = 0; i < NUM_SECTIONS; i++) {
            if (Sections[i].HitPoints > 0
                && TestForCollision(Sections[i].Frame, Sections[i].WorldPos, 
                    OtherEntityFrame, OtherEntityFramePos)) {
            
                SectionHit = (SECTION_TYPE)i;
                return true;
            }
        }
    }
    return false;
}

//------------------------------------------------------------------------------
// Name: HandleCollision
// Desc: Depending upon the type of entity that has collided with it, the 
//       section of the factory that has been hit. This damage may in turn
//       cause damage to adjacent sections or destroy the factory outright.
//------------------------------------------------------------------------------
void Factory::HandleCollision(Entity *AnEntity)
{
    int DamageTaken;

    assert(SectionHit != NO_SECTION && Sections[SectionHit].HitPoints > 0);

    switch (AnEntity->GetType()) {
        case BULLET:
            if (((Bullet *)AnEntity)->GetBulletType() == Bullet::SOLDIER_BULLET) {
                DamageTaken = DAMAGE_TAKEN_FROM_SOLDIERS_BULLET;
            } else {
                DamageTaken = DAMAGE_TAKEN_FROM_BULLET;
            }
            break;
        case BOMB:
            DamageTaken = DAMAGE_TAKEN_FROM_BOMB;
            break;
        case PLANE:
            DamageTaken = DAMAGE_TAKEN_FROM_PLANE;
            break;
        default:
            assert(false);      // Invallid Type
    }
    DamageSection(SectionHit, DamageTaken);
    SectionHit = NO_SECTION;
}

//------------------------------------------------------------------------------
// Name: DamageSection
// Desc: Inflicts a given amount of damage upon a given section.
//------------------------------------------------------------------------------
void Factory::DamageSection(SECTION_TYPE s, int Amount)
{
    Explosion *Exp = NULL;
    Point2D ExplosionPos;
    bool ExplosionGenerated;
    int NewHitPoints = Sections[s].HitPoints - Amount;
    bool SectionDestroyed = NewHitPoints <= 0;

    if (Sections[s].HitPoints <= 0) return;     // Can't do any more damage to this section

    // If the section takes enough damage an explosion is generated
    ExplosionGenerated = 
        ((Sections[s].HitPoints > INITIAL_SECTION_HITPOINTS_2
            && NewHitPoints <= INITIAL_SECTION_HITPOINTS_2)
        || SectionDestroyed);

    // Inflict the damage on this section and generate an explosion if need be
    Sections[s].HitPoints = NewHitPoints;
    if (ExplosionGenerated) {
        // Set frame to a more decayed look
        Sections[s].Frame = Frame + FRAME_OFFSETS[s]+1;

        // Position the explosion at the centre of the section
        ExplosionPos.x = Sections[s].WorldPos.x + SectionWidth / 2;
        ExplosionPos.y = Sections[s].WorldPos.y + SectionHeight / 2;
        Exp = new Explosion(L, Explosion::LARGE, ExplosionPos);
		L->AddNewEntity(Exp);
    }

    // Award the player points if this is an EVIL factory
    if (Side == EVIL) L->AddToPlayerScore(Amount*POINTS_VALUE);

    // See if this damage will effect other sections
    switch (s) {
        case MIDDLE:
            if (SectionDestroyed) {
                // Kaboom
                NeedsDeletion = true;
                DamageSection(LEFT_SIDE, INITIAL_SECTION_HITPOINTS);
                DamageSection(RIGHT_SIDE, INITIAL_SECTION_HITPOINTS);
            }
            break;
        case LEFT_SIDE:
            if (NewHitPoints <= INITIAL_SECTION_HITPOINTS_2) {
                DamageSection(CHIMNEYS, INITIAL_SECTION_HITPOINTS);
            }
            break;
    }
}
