/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// Name: FightingEntity.cpp
// Desc: A FightingEntity is an enity that can become involved in combat upon
//       the ground. A FightingEntity needs to be able to give some indication
//       of how far it can currently fire and whether or not it is fleeing
//       (fleeing uints cannot be engaged in combat).
//
//       This entity can be used to provide a subtle distinction between
//       entities that fighting units will attack and entities that will just
//       take damage as the result of a combat going on nearby. For example,
//       a Factory is not a FightingEntity (although it could be) and is 
//       therefore not targeted specifically by enemy units. However, AA guns
//       if they were to be implemented could be implemented as a 
//       FightingEntity. They would have a range of 0 and never flee, but 
//       enemy units would see them as a FightingEntity and specifically 
//       attack them.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <list>
#include "Level.h"
#include "Landscape.h"
#include "FightingEntity.h"

using namespace std;

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------
const float FightingEntity::MAX_MORALE = 100.0f;
const float FightingEntity::MIN_MORALE = 0.0f;

//------------------------------------------------------------------------------
// FightingEntity Implementation
//------------------------------------------------------------------------------
FightingEntity::FightingEntity(Level *L) : Entity(L) {}
FightingEntity::~FightingEntity() {}

//------------------------------------------------------------------------------
// Name: FindClosestEnemyInRange
// Desc: This function looks through all of the possible enities that could be
//       enemies and returns details about the closest one that is within range. 
//       If no valid enemy is found then the function returns NULL.
//
// Note: It may seem slightly wasteful not passing back the values that come
//       from 'FightingEntityInRange' but at the moment I believe that it makes
//       for a cleaner design.
//------------------------------------------------------------------------------
FightingEntity *FightingEntity::FindClosestEnemyInRange()
{
    int i;
    int CurDistance;
    int ShortestDistance = LS->GetWorldPixelWidth();
    FightingEntity *ClosestEnemy = NULL;
    list<Entity *>::iterator Iter;
    bool OnTheLeft; 
    int RelativeDistance; 
    int MaxRange; 
    int NearestX; 
    Point2D TargetPos;

    // Loop through all of the entity lists looking for lists that contain
    // FightingEntities
    for (i = 0; i < NUM_ENT_TYPES; i++) {
        if (L->GameEntities[i].size() > 0 
            && dynamic_cast<FightingEntity *>(*(L->GameEntities[i].begin()))) {
        
            // If the list is not empty and contains FightingEntities then loop
            // through the list and see if there is any chance of conflict.
            for (Iter = L->GameEntities[i].begin();
                Iter != L->GameEntities[i].end(); Iter++) {

                // Test to see if this FightingEntity is valid    
                if (*Iter != this && (*Iter)->GetSide() != Side
                    && !((FightingEntity *)*Iter)->IsFleeing()) {
            
                    // Is it closer than other enemies seen so far?
                    CurDistance = abs(GetCentreX() - (*Iter)->GetCentreX());
                    if (CurDistance < ShortestDistance) {
                        ClosestEnemy = (FightingEntity *)(*Iter);
                        ShortestDistance = CurDistance;
                    }
                }
            }
        }
    }
    
    if (ClosestEnemy != NULL 
        && FightingEntityInRange(ClosestEnemy, OnTheLeft, RelativeDistance, 
            MaxRange, NearestX, TargetPos)) {
        
        return ClosestEnemy;
    } else {
        return NULL;
    }
}

//------------------------------------------------------------------------------
// Name: FightingEntityInRange
// Desc: This function looks at another FightingEntity and determines whether 
//       or not it is within range of this FightingEntity. One FightingEntity 
//       will be to the left of the other so a return variable says whether or 
//       not this FightingEntity is on the left. Another return variable then 
//       provides the relative distance from the right hand side of the 
//       left FightingEntity to the left hand side of the right FightingEntity, 
//       (if this is -ve then the FightingEntities are entwined). The 
//       range of a FightingEntity's guns can vary depending upon whether or not 
//       they are on higher ground than their enemy. Therefore, this function 
//       also returns the extreme range at which this FightingEntity will be 
//       able to open fire. This is always +ve.
//       Finally, this function works out the x value of the closest point to
//       the enemy unit and the point on the enemy unit that should be shot
//       at.
//
// Note: You should not try to find out if a FightingEntity is within range
//       of itself as that is meaningless.
//------------------------------------------------------------------------------
bool FightingEntity::FightingEntityInRange(FightingEntity *EnemyFE, 
    bool& OnTheLeft, int& RelativeDistance, int& MaxRange, 
    int& NearestX, Point2D& TargetPos)
{
    int HeightDiff;
    int EnemyMaxRange;
    FightingEntity *FE1 = this;
    FightingEntity *FE2 = EnemyFE;

    assert(FE2 != NULL);
    assert(FE1 != FE2);

    OnTheLeft = (FE1->GetCentreX() <= FE2->GetCentreX());

    if (OnTheLeft) {
        NearestX = FE1->Right();
        TargetPos = GetShotPos(FE2, FE2->Left());
        RelativeDistance = TargetPos.x - NearestX;      
    } else {
        NearestX = FE1->Left();
        TargetPos = GetShotPos(FE2, FE2->Right());
        RelativeDistance = NearestX - TargetPos.x;
    } 
    HeightDiff = LS->CalculatePixelHeight(NearestX) - TargetPos.y;
    MaxRange = FE1->GetAvailableRange(HeightDiff);
    EnemyMaxRange = FE2->GetAvailableRange(-HeightDiff);

    // Work out whether the platoons are actually in range
    return (MaxRange > RelativeDistance || EnemyMaxRange > RelativeDistance);
}

//------------------------------------------------------------------------------
// Name: GetShotPos
// Desc: This function returns the point that bullets should either be aimed
//       at or emanate from when a given FightingEntity is shot at.
//------------------------------------------------------------------------------
Point2D FightingEntity::GetShotPos(FightingEntity *FE, int x)
{
    return Point2D(x, FE->LS->CalculatePixelHeight(x) - FE->GetTargetHeight());
}
