/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Plane.cpp
// Desc: The plane entity is vital to the whole game, it is used in the 
//       creation of the players plane. Also it is used to create the enemy 
//       planes that try to thwart the players actions.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <assert.h>
#include "Camel.h"
#include "Game.h"
#include "Sprites.h"
#include "Level.h"
#include "Landscape.h"
#include "InputHandler.h"
#include "Explosion.h"
#include "SmokeParticle.h"
#include "Bomb.h"
#include "Bullet.h"
#include "Plane.h"

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------
const int Plane::POINTS_VALUE = 500;

const int Plane::STARTING_BOMBS = 15;
const int Plane::STARTING_BULLETS = 150;
// With a fuel consumption rate of 1 per 100 milliseconds, 900 fuel gives a
// flight time of approximately 1 and 1/2 minutes
const int Plane::STARTING_FUEL = 900;

// Amount of time player must struggle with the stick before recovering 
// from a stall
const int Plane::STALL_RECOVERY_THRESHOLD = 900;

// Rates for various actions are measured in terms of milliseconds
// It is good to have these as multiples of EnemyPlane::COLLISION_CHECK_RATE
// as it make the AI's predictions more accurate.
// They should not be less than about 12-15 as slower computers will not
// give timeslices that small.
const int Plane::CLIMBING_RATE = 60;
const int Plane::DIVING_RATE = 150;
const int Plane::INVERTING_RATE = 120;
const int Plane::FIRING_RATE = 120;
const int Plane::HALTING_GUNFIRE_RATE = 150;
const int Plane::STARTING_ENGINE_RATE = 30;
const int Plane::SPIRALING_RATE = 150;
const int Plane::EMITING_SMOKE_RATE = 60;
const int Plane::DROPPING_BOMB_RATE = 30;
const int Plane::RESETTING_BOMB_RATE = 390;
const int Plane::CONSUMING_FUEL_RATE = 100;
// The whole reloading process should take about 5 seconds for bombs and bullets
// and 10 seconds for a full tank of fuel.
const int Plane::LOADING_FUEL_RATE = 30;  
const int Plane::LOADING_BULLETS_RATE = 5000 / STARTING_BULLETS;
const int Plane::LOADING_BOMBS_RATE = 5000 / STARTING_BOMBS;

const float Plane::MINIMUM_AIRSPEED = 70.0f;  // Pixels per second

// Acceleration values are given in pixels per second^-2
const float Plane::MAX_ENGINE_ACC = 485.0f;
const float Plane::ENGINE_ACC_ACC = MAX_ENGINE_ACC / 4.0f;

const int Plane::BULLET_RANGE = 400;

//------------------------------------------------------------------------------
// Plane Implementation
//------------------------------------------------------------------------------

Plane::Plane(Level *L, PLANE_STATE State, Point2D Pos, bool Inverted, int Direction,
             ENT_SIDE Side) : Entity(L)
{
    Type = PLANE;
    this->Side = Side;

    // Define the victims of a plane
    NumVictims = 3;
    Victims[0] = PLANE;
    Victims[1] = FACTORY;
    Victims[2] = TANK;

    Actions[CLIMB] = new ACTION(0, CLIMBING_RATE, false, Climb);
    Actions[DIVE] = new ACTION(0, DIVING_RATE, false, Dive);
    Actions[INVERT] = new ACTION(0, INVERTING_RATE, false, Invert);
    Actions[FIRE] = new ACTION(0, FIRING_RATE, false, Fire);
    Actions[HALT_GUNFIRE] = new ACTION(0, HALTING_GUNFIRE_RATE, false, HaltGunfire);
    Actions[START_ENGINE] = new ACTION(0, STARTING_ENGINE_RATE, false, StartEngine);
    Actions[SPIRAL] = new ACTION(0, SPIRALING_RATE, false, Spiral);
    Actions[EMIT_SMOKE] = new ACTION(0, EMITING_SMOKE_RATE, false, EmitSmoke);
	Actions[DROP_BOMB] = new ACTION(0, DROPPING_BOMB_RATE, false, DropBomb);
    Actions[RESET_BOMB] = new ACTION(0, RESETTING_BOMB_RATE, false, ResetBomb);
    Actions[CONSUME_FUEL] = new ACTION(0, CONSUMING_FUEL_RATE, false, ConsumeFuel);
    Actions[LOAD_FUEL] = new ACTION(0, LOADING_FUEL_RATE, false, LoadFuel);
    Actions[LOAD_BULLETS] = new ACTION(0, LOADING_BULLETS_RATE, false, LoadBullets);
    Actions[LOAD_BOMBS] = new ACTION(0, LOADING_BOMBS_RATE, false, LoadBombs);

    memset(Commands, false, sizeof(bool)*NUM_ACTIONS);

    if (Side == GOOD) {
        StartFrame = Sprites::PLAYER_PLANE_SPRITE;
    } else {
        StartFrame = Sprites::ENEMY_PLANE_SPRITE;
    }
    SPR->GetFrameDims(StartFrame, &Width, &Height);
    
    // Set the initial state of the plane
    this->State = State;
    if (State == LANDED) {
        Velocity = 0;
        EngineAcc = 0;
        EngineOn = false;
    } else {
        Velocity = 250.0f;
        EngineAcc = MAX_ENGINE_ACC;
        EngineOn = true;
    }

    BombsLeft = STARTING_BOMBS;
    BulletsLeft = STARTING_BULLETS;
    FuelLeft = STARTING_FUEL;

    CurEngineSample = SDL_Interface::NO_SAMPLE;
    Refuelling = false;
    HaltingGunfire = true;
    FiringMultipleShots = false;
    EngineChannel = SDL_Interface::NO_CHANNEL;
    GunfireChannel = SDL_Interface::NO_CHANNEL;
    WorldPos = Pos;
    RealX = (float)WorldPos.x;
    RealY = (float)WorldPos.y;
    this->Inverted = Inverted;
    this->Direction = CamelMath::NormaliseDirection(Direction);
    Frame = CalcFrame(this->Direction, this->Inverted);
}

Plane::~Plane() 
{
    if (EngineChannel != SDL_Interface::NO_CHANNEL) Mix_HaltChannel(EngineChannel);
    if (GunfireChannel != SDL_Interface::NO_CHANNEL) Mix_HaltChannel(GunfireChannel);
}

void Plane::Update(uint dt)
{
    float SecondsPassed = (float)dt / 1000.0f;
    int Plane_MinX = -LS->GetViewportWidth();
    int Plane_MaxX = LS->GetViewportMaxWorldX() + 2*LS->GetViewportWidth();
    int Plane_MinY = -LS->GetViewportHeight();
    int Plane_MaxY = LS->GetViewportMaxWorldY() + 2*LS->GetViewportHeight();

    switch (State) {
        case LANDED:
            Velocity = CalcNewVelocity(Velocity, Direction, dt);

            if (EngineOn) {
                if (EngineAcc < MAX_ENGINE_ACC) {
                    EngineAcc += ENGINE_ACC_ACC*SecondsPassed;
                }
            } else {
                if (Refuelling) {
                    UpdateAction(Actions[LOAD_FUEL], true, dt, this);
                    UpdateAction(Actions[LOAD_BULLETS], true, dt, this);
                    UpdateAction(Actions[LOAD_BOMBS], true, dt, this);
                }

                UpdateAction(Actions[START_ENGINE], Commands[START_ENGINE], dt, this);
            }

            if (Velocity > MINIMUM_AIRSPEED
                || !((Direction == 0 && !Inverted)
                || (Direction == 8 && Inverted))) {
                State = FLYING;
                EngineOn = true;
                EngineAcc = MAX_ENGINE_ACC;
            }

            break;
        case FLYING:
            Velocity = CalcNewVelocity(Velocity, Direction, dt);

            if (Velocity < MINIMUM_AIRSPEED) {
                State = SPIRALING;
                StallRecovery = 0;
            }

            break;
        case CRASHING:
            UpdateAction(Actions[EMIT_SMOKE], true, dt, this);
            UpdateAction(Actions[SPIRAL], true, dt, this);

            Velocity = CalcNewVelocity(Velocity, Direction, dt);
            break;
        case SPIRALING:
            UpdateAction(Actions[SPIRAL], true, dt, this);

            Velocity = CalcNewVelocity(Velocity, Direction, dt);

            // The pilot must constantly pull up in order to stand a chance of
            // recovering from a stall
            if (Commands[CLIMB]) {
                StallRecovery += dt;
                if (StallRecovery >= STALL_RECOVERY_THRESHOLD) State = FLYING;
            } else {
                StallRecovery = 0;
            }

            break;
        default:
            assert(false && "Error: Plane entity in invalid state");
    }
    if (FuelLeft <= 0) State = CRASHING;    // Hope you packed a parachute

    RealX += CamelMath::XY_COMPONENTS[Direction].x*Velocity*SecondsPassed;
    RealY += CamelMath::XY_COMPONENTS[Direction].y*Velocity*SecondsPassed;
    WorldPos = Point2D((int)RealX, (int)RealY);

    if (WorldPos.x < Plane_MinX || WorldPos.x > Plane_MaxX 
        || WorldPos.y > Plane_MaxY) {
        NeedsDeletion = true;
    } else if (WorldPos.y < Plane_MinY) {
        State = SPIRALING;
        StallRecovery = 0; 
    }

    if ((State == LANDED || State == FLYING) && EngineOn) {
        UpdateAction(Actions[CLIMB], Commands[CLIMB], dt, this);
        UpdateAction(Actions[DIVE], Commands[DIVE], dt, this);
        UpdateAction(Actions[INVERT], Commands[INVERT], dt, this);
        UpdateAction(Actions[FIRE], Commands[FIRE], dt, this);
		UpdateAction(Actions[DROP_BOMB], Commands[DROP_BOMB] && !ResetingBomb, dt, this);
        UpdateAction(Actions[RESET_BOMB], ResetingBomb, dt, this);
    }
    if (State != LANDED && EngineOn) {
        UpdateAction(Actions[CONSUME_FUEL], true, dt, this);
    }

    UpdateAction(Actions[HALT_GUNFIRE], HaltingGunfire, dt, this);

    Frame = CalcFrame(Direction, Inverted);

    if (EngineOn) {
        NewEngineSample = (int)(Velocity/33);

        if (NewEngineSample < (int)SDL_Interface::PLANE_ENGINE_SPEED_0) {
            NewEngineSample = SDL_Interface::PLANE_ENGINE_SPEED_0;
        } else if (NewEngineSample > (int)SDL_Interface::PLANE_ENGINE_SPEED_16) {
            NewEngineSample = SDL_Interface::PLANE_ENGINE_SPEED_16;
        }

        if (NewEngineSample != CurEngineSample ) {            
            // If the sample has changed or no sound is being played

            CurEngineSample = NewEngineSample;
            if (EngineChannel == SDL_Interface::NO_CHANNEL) {
                EngineChannel = 
                    SDLI->PlayChannel(-1, SDLI->Samples[CurEngineSample], -1);
            } else {
                SDLI->PlayChannel(EngineChannel, SDLI->Samples[CurEngineSample], -1);
            }
        }
    } else {
        if (EngineChannel != SDL_Interface::NO_CHANNEL) {
            Mix_HaltChannel(EngineChannel);
            EngineChannel = SDL_Interface::NO_CHANNEL;
        }
    }

    // Change the volume and stereo balance of the sound channels depending upon
    // where the plane is
    SFX_Volume = SDLI->VolumeOverDistance(
        Point2D(GetCentreX(), GetCentreY()).DistanceTo(
        Point2D(LS->GetViewportCentreWorldX(), LS->GetViewportCentreWorldY())));
    SDLI->GetStereoVolumes(LS->GetViewportWidth() / 2, 
        LS->GetViewportCentreWorldX() - GetCentreX(), 
        &LeftVolume, &RightVolume);

    if (EngineChannel != SDL_Interface::NO_CHANNEL) {
        Mix_Volume(EngineChannel, SFX_Volume);
        Mix_SetPanning(EngineChannel, LeftVolume, RightVolume);
    }
    if (GunfireChannel != SDL_Interface::NO_CHANNEL) {
        Mix_Volume(GunfireChannel, SFX_Volume);
        Mix_SetPanning(GunfireChannel, LeftVolume, RightVolume);
    }
}

void Plane::FindAndHandleCollisionWithLandscape()
{
    Explosion *Exp;

    if (LS->HasCollided(Frame, WorldPos, false)) {
        NeedsDeletion = true;

        // Create an explosion
        Exp = new Explosion(L, Explosion::LARGE, GetCentrePos());
        L->AddNewEntity(Exp);
    }
}

void Plane::HandleCollision(Entity *AnEntity)
{
	Explosion *Exp;

	if (AnEntity->GetType() == BULLET) {
		State = CRASHING;
	} else {
		// The response to anything other than a bullet is to explode
		// However, if the plane has already been deleted (exploded in a previous
		// collision) then do nothing
		if (!NeedsDeletion) {
			NeedsDeletion = true;

			// Create an explosion
			Exp = new Explosion(L, Explosion::LARGE, GetCentrePos());
			L->AddNewEntity(Exp);
		}
	}
	
}

int Plane::CalcFrame(int CurDirection, bool Inverted) 
{
    return (Inverted ? StartFrame + CurDirection + CamelMath::NUM_DIRECTIONS :
        StartFrame + CurDirection);
}

float Plane::CalcNewVelocity(float CurVelocity, int CurDirection, uint dt)
{
    float SecondsPassed = (float)dt / 1000.0f;
    float Acceleration;
    float NewVelocity;

    Acceleration = (EngineAcc + CamelMath::GRAVITY_ACC[CurDirection]
        - L->GetAirResistance(Type, CurVelocity)) * SecondsPassed;
    NewVelocity = CurVelocity + Acceleration;

    return (NewVelocity > 0.0f ? NewVelocity : 0.0f);
}

void Plane::Climb(void *Object)
{
    Plane *APlane = (Plane *)Object;

    APlane->Direction += (APlane->Inverted ? -1 : 1);
    APlane->Direction = CamelMath::NormaliseDirection(APlane->Direction);
}

void Plane::Dive(void *Object)
{
    Plane *APlane = (Plane *)Object;

    APlane->Direction += (APlane->Inverted ? 1 : -1);
    APlane->Direction = CamelMath::NormaliseDirection(APlane->Direction);
}

void Plane::Invert(void *Object)
{
    Plane *APlane = (Plane *)Object;

    APlane->Inverted = !APlane->Inverted;
}

void Plane::Fire(void *Object)
{
	Bullet *NewBullet;
	int DistanceFromPlane;
	Point2D CentrePos;
    Plane *APlane = (Plane *)Object;

    if (APlane->BulletsLeft <= 0) return;   // Nothing to fire

    if (APlane->GunfireChannel == APlane->SDLI->NO_CHANNEL) {
        APlane->GunfireChannel = APlane->SDLI->PlayChannel(-1, 
            APlane->SDLI->Samples[SDL_Interface::PLANE_GUNFIRE], -1);
    }
    
	// Create a new bullet
	DistanceFromPlane = APlane->Width;

    CentrePos = APlane->GetCentrePos();
	CentrePos.x += (int)(CamelMath::XY_COMPONENTS[APlane->Direction].x*DistanceFromPlane);
	CentrePos.y += (int)(CamelMath::XY_COMPONENTS[APlane->Direction].y*DistanceFromPlane);

    NewBullet = new Bullet(APlane->L, CentrePos, 
        CamelMath::XY_COMPONENTS[APlane->Direction], Bullet::PLANE_BULLET, 
        BULLET_RANGE, APlane->Side);
    APlane->L->AddNewEntity(NewBullet);

    // If already firing then we are firing multiple shots
    if (APlane->HaltingGunfire) {
        APlane->FiringMultipleShots = true;
    } else {
        APlane->HaltingGunfire = true;
    }
    APlane->BulletsLeft--;
}

void Plane::HaltGunfire(void *Object)
{
    Plane *APlane = (Plane *)Object;

    if (APlane->FiringMultipleShots) {
        APlane->FiringMultipleShots = false;
    } else {
        APlane->HaltingGunfire = false;
        if (APlane->GunfireChannel != APlane->SDLI->NO_CHANNEL) { 
            Mix_HaltChannel(APlane->GunfireChannel);
            APlane->GunfireChannel = APlane->SDLI->NO_CHANNEL;
        }
    }
}

void Plane::StartEngine(void *Object)
{
    Plane *APlane = (Plane *)Object;

    if (APlane->State == LANDED) {
        APlane->EngineOn = true;
        APlane->Refuelling = false;
    }
}

void Plane::Spiral(void *Object)
{
    // Speed at which the direction changes
    const int DIRCHANGE = 2;
    // Percentage ranges for deciding what to do randomly
    const float SP_LEFT = 1.0f/3.0f;
    const float SP_RIGHT = 2.0f/3.0f;
    const float INV_CHANCE = 0.5f;
    
    float RandomNumber;  // This will be picked to be a number between 0.0f and 1.0f
    Plane *APlane = (Plane *)Object;

    if (APlane->Direction >= 11 && APlane->Direction <= 13) {
        RandomNumber = (float)rand() / RAND_MAX;

        if (RandomNumber < SP_LEFT) {
            APlane->Direction = 11;
        } else if (RandomNumber >= SP_RIGHT) {
            APlane->Direction = 13;
        } else {
            APlane->Direction = 12;
        }
    } else if (APlane->Direction == 4) {
        if (APlane->Inverted) APlane->Direction -= DIRCHANGE;
        else APlane->Direction += DIRCHANGE;
    } else if (APlane->Direction > 4 && APlane->Direction < 11) {
        APlane->Direction += DIRCHANGE;
    } else {
        APlane->Direction -= DIRCHANGE;
    }

    // Is the plane inverted?
    RandomNumber = (float)rand() / RAND_MAX;
    APlane->Inverted = (RandomNumber >= INV_CHANCE);

    // Make sure that the planes direction is in range
    if (APlane->Direction < 0) {
        APlane->Direction = CamelMath::NUM_DIRECTIONS - 1;
    } else if (APlane->Direction >= CamelMath::NUM_DIRECTIONS) {
        APlane->Direction = 0;
    }
}

void Plane::EmitSmoke(void *Object)
{
    SmokeParticle *Smoke;
    Plane *APlane = (Plane *)Object;

    Smoke = new SmokeParticle(APlane->L, 10*30, APlane->GetCentrePos());
    APlane->L->AddNewEntity(Smoke);
}

void Plane::DropBomb(void *Object)
{
    Bomb *NewBomb;
    Plane *APlane = (Plane *)Object;
	int DistanceFromPlane;
	int XY_Component;
	Point2D CentrePos;

    if (APlane->BombsLeft <= 0) return;     // Nothing to drop

    // The distance is negated if the plane is inverted
	DistanceFromPlane = (APlane->Inverted ? -APlane->Height : APlane->Height);
	if (APlane->Direction < 4) {
		XY_Component = CamelMath::NUM_DIRECTIONS - (4 - APlane->Direction);
	} else {
		XY_Component = APlane->Direction - 4;
	}
    CentrePos = APlane->GetCentrePos();
    CentrePos.x += (int)(CamelMath::XY_COMPONENTS[XY_Component].x*DistanceFromPlane);
	CentrePos.y += (int)(CamelMath::XY_COMPONENTS[XY_Component].y*DistanceFromPlane);

	NewBomb = new Bomb(APlane->L, CentrePos, APlane->Direction, APlane->Velocity);
    APlane->L->AddNewEntity(NewBomb);

    APlane->BombsLeft--;
    APlane->ResetingBomb = true;
}

void Plane::ResetBomb(void *Object)
{
    Plane *APlane = (Plane *)Object;

    APlane->ResetingBomb = false;
}

//------------------------------------------------------------------------------
// Name: ConsumeFuel
// Desc: Reduces the fuel left by 1.
//------------------------------------------------------------------------------
void Plane::ConsumeFuel(void *Object)
{
    Plane *APlane = (Plane *)Object;

    if(APlane->FuelLeft > 0) APlane->FuelLeft--;
}

//------------------------------------------------------------------------------
// Name: LoadFuel
// Desc: Increases the amount of fuel the plane has by 3. This, combined with
//       the constant LOADING_FUEL_RATE should ensure that loading fuel takes
//       twice as long as loading bombs and bullets.
//------------------------------------------------------------------------------
void Plane::LoadFuel(void *Object)
{
    Plane *APlane = (Plane *)Object;

    if(APlane->FuelLeft < STARTING_FUEL) {
        APlane->FuelLeft += 3;
        if (APlane->FuelLeft > STARTING_FUEL) APlane->FuelLeft = STARTING_FUEL;
    }
}

//------------------------------------------------------------------------------
// Name: LoadBullets
// Desc: Increases the amount of bullets the plane has by 1.
//------------------------------------------------------------------------------
void Plane::LoadBullets(void *Object)
{
    Plane *APlane = (Plane *)Object;

    if(APlane->BulletsLeft < STARTING_BULLETS) APlane->BulletsLeft++;
}

//------------------------------------------------------------------------------
// Name: LoadBombs
// Desc: Increases the amount of bombs the plane has by 1.
//------------------------------------------------------------------------------
void Plane::LoadBombs(void *Object)
{
    Plane *APlane = (Plane *)Object;

    if(APlane->BombsLeft < STARTING_BOMBS) APlane->BombsLeft++;
}
