/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: PlaneSpawnPoint.cpp
// Desc: These entities are responsible for creating planes in the game world.
//       Each spawn point is responsible for one plane. When that plane dies
//       the spawn point waits for a certain amount of time before respawning
//       the plane.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <assert.h>
#include "Game.h"
#include "Level.h"
#include "Landscape.h"
#include "SDL_Interface.h"
#include "Sprites.h"
#include "Plane.h"
#include "PlayerPlane.h"
#include "EnemyPlane.h"
#include "PlaneSpawnPoint.h"

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------

// The runway needs to be raised to stop the plane from colliding 
// with the grass :P
const int PlaneSpawnPoint::HEIGHT_OF_RUNWAY = 3;

const int PlaneSpawnPoint::RESPAWNING_GOOD_PLANE_RATE = 5000;
const int PlaneSpawnPoint::RESPAWNING_EVIL_PLANE_RATE = 10000;

//------------------------------------------------------------------------------
// PlaneSpawnPoint Implementation
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Name: PlaneSpawnPoint
// Desc: Creates the plane spawn point and also creates the first plane.
//------------------------------------------------------------------------------
PlaneSpawnPoint::PlaneSpawnPoint(Level *L, int x, ENT_SIDE Side, 
    EnemyPlane::WAYPOINT PatrolPoints[EnemyPlane::MAX_NUM_PATROL_POINTS]) : Entity(L)
{
    Type = PLANESPAWNPOINT;
    this->Side = Side;
    
    // Set up the spawn point according to the entities side
    if (Side == GOOD) {
        // A PlaneSpawnPoint has no physical representation in the game world but
        // needs to be positioned so that a plane can be positioned there
        Frame = Sprites::PLAYER_PLANE_SPRITE;
        SPR->GetFrameDims(Frame, &Width, &Height);
        WorldPos.x = x;
        WorldPos.y = LS->CalculatePixelHeight(x) - (Height + HEIGHT_OF_RUNWAY);

        PatrolPoints = NULL;
        CurPlane = new PlayerPlane(L, Plane::LANDED, WorldPos, false, 0);
        Actions[RESPAWN_PLANE] = 
            new ACTION(0, RESPAWNING_GOOD_PLANE_RATE, false, RespawnPlane);
    } else {
        // A PlaneSpawnPoint has no physical representation in the game world but
        // needs to be positioned so that a plane can be positioned there
        Frame = Sprites::ENEMY_PLANE_SPRITE;
        SPR->GetFrameDims(Frame, &Width, &Height);
        WorldPos.x = x;
        WorldPos.y = LS->CalculatePixelHeight(x) - (Height + HEIGHT_OF_RUNWAY);

        assert(PatrolPoints != NULL);
        memcpy(this->PatrolPoints, PatrolPoints, 
            sizeof(EnemyPlane::WAYPOINT) * EnemyPlane::MAX_NUM_PATROL_POINTS);
        CurPlane = new EnemyPlane(L, Plane::LANDED, 
            WorldPos, true, 8, PatrolPoints);
        Actions[RESPAWN_PLANE] = 
            new ACTION(0, RESPAWNING_EVIL_PLANE_RATE, false, RespawnPlane);
    }
    // Add the plane to the level
    L->AddNewEntity(CurPlane);
}

PlaneSpawnPoint::~PlaneSpawnPoint() {}

//------------------------------------------------------------------------------
// Name: GetCurPlane
// Desc: Returns a pointer to the current plane that this spawn point has 
//       spawned.
//------------------------------------------------------------------------------
Plane *PlaneSpawnPoint::GetCurPlane()
{
    if (CurPlane != NULL && !L->IsEntityAlive(CurPlane)) {
        CurPlane = NULL;
    }
    return CurPlane;
}

//------------------------------------------------------------------------------
// Name: Update
// Desc: If for some reason the spawn points plane has been destroyed. The
//       spawn point waits for a predetermined amount of time and then 
//       respawns a new one.
//------------------------------------------------------------------------------
void PlaneSpawnPoint::Update(uint dt)
{
    if (CurPlane == NULL) {
        UpdateAction(Actions[RESPAWN_PLANE], true, dt, this);
    } else {
        if (!L->IsEntityAlive(CurPlane)) CurPlane = NULL;
    }
}

//------------------------------------------------------------------------------
// Name: RespawnPlane
// Desc: This routine respawns a plane of the same side as the PlaneSpawnPoint
//       at the position of the PlaneSpawnPoint.
//------------------------------------------------------------------------------
void PlaneSpawnPoint::RespawnPlane(void *Object)
{
    PlaneSpawnPoint *APSP = (PlaneSpawnPoint *)Object;

    if (APSP->GetSide() == GOOD) {
        if (APSP->L->GetPlayerLives() <= 0) return; // No more lives left
        APSP->CurPlane = new PlayerPlane(APSP->L, Plane::LANDED, 
            APSP->WorldPos, false, 0);
    } else {
        APSP->CurPlane = new EnemyPlane(APSP->L, Plane::LANDED, 
            APSP->WorldPos, true, 8, APSP->PatrolPoints);
    }
    APSP->L->AddNewEntity(APSP->CurPlane);
}
