/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Platoon.cpp
// Desc: This object provides a way to group together soldier entities. The
//       platoon can be given orders as a whole and will move and fight as one.
//       It also provides a group that can be affected by morale so they can
//       also run away as one. :P
//
//       This class has been made an entity so that it can be easily kept 
//       updated with the rest of the game. I am not sure whether this is
//       actually a good design decision or not however, as a platoon lacks
//       some of the elements of a traditional entity, such as a frame and a
//       width and height
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <limits>
#include <vector>
#include <stdlib.h>
#include <assert.h>
#include "Game.h"
#include "Sprites.h"
#include "Level.h"
#include "Landscape.h"
#include "ControlledEntity.h"
#include "Soldier.h"
#include "Platoon.h"

#undef max
#undef min

using namespace std;

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------

// The higher this number is, the more spread out the platoon will be be
// when it is placed
const int Platoon::SOLDIER_SPREAD = 5;
const int Platoon::SOLDIERS_PER_RANK = 4;
const int Platoon::DISENGAGE_RANGE = 50;
const int Platoon::MAX_ENTANGLEMENT = 50;

const int Platoon::NEAR_DESTINATION_X = 5;

// If morale drops to below this point then the platoon runs away
const float Platoon::BREAKPOINT = 20.0f;
// If morale rises back up above this point then the platoon rallies
const float Platoon::RALLYPOINT = 40.0f;
const int Platoon::MIN_NUM_SOLDIERS_NEEDED_TO_RALLY = 15;

const int Platoon::RANK_SEPERATION = 20;
const int Platoon::RANK_PLACEMENT_RANGE = SOLDIER_SPREAD * 2;
const int Platoon::RANK_PLACEMENT_RANGE_2 = RANK_PLACEMENT_RANGE / 2;

// This is the chance that a casualty in the front rank will be replaced by
// someone in the second rank as opposed to someone from the reserve.
const float Platoon::SECOND_RANK_REPLACEMENT_CHANCE = 1.0f / 3.0f;

const ControlledEntity::Order Platoon::FLEE = ControlledEntity::Order("Flee!");

// Rates for various actions are measured in milliseconds
const int Platoon::UPDATING_MORALE_RATE = 1000;

//------------------------------------------------------------------------------
// Platoon Implementation
//------------------------------------------------------------------------------
Platoon::Platoon(Level *L, int NumSoldiers, int x, ENT_SIDE Side) 
    : ControlledEntity(L, Side)
{
    int i;

    assert(NumSoldiers > 0 && NumSoldiers <= MAX_NUM_SOLDIERS);

    Type = PLATOON;
    UnitPlacementRange = NumSoldiers*SOLDIER_SPREAD;
    ReservePlacementRange = 0;

    this->NumSoldiers = NumSoldiers;
    NumSubordinatesAlive = NumSoldiers;
    SoldiersAliveLastMoraleCheck = NumSubordinatesAlive;
    Morale = MAX_MORALE;
    CurEnemy = NULL;

    // Create the soldiers
    for (i = 0; i < NumSoldiers; i++) {
        Soldiers[i] = new Soldier(L, x, i, this);
        L->AddNewEntity(Soldiers[i]);
    }

    // Place the soldiers
    PlaceAt(x);

    // Set current orders and find out where we should retreat to
    CurOrders = &HOLD;
    OrdersHaveChanged = false;
    DestinationX = x;
    UpdateFlagPos();
    if (Side == GOOD) {
        FallbackX = -2000;
    } else {
        FallbackX = LS->GetWorldPixelWidth() + 2000;
    }

    // Deal with the stuff that makes a platoon an entity
    if (Side == GOOD) {
        IconFrame = Sprites::GOOD_PLATOON_ICON;
    } else {
        IconFrame = Sprites::EVIL_PLATOON_ICON;
    }
    UpdateBoundingBox();
    Frame = 0;

    // Add in an action to keep the platoons morale up to date
    Actions[UPDATE_MORALE] = new ACTION(0, UPDATING_MORALE_RATE, false, UpdateMorale);
}

Platoon::~Platoon() {}

//------------------------------------------------------------------------------
// Name: PlaceAt
// Desc: Places the platoon at the given position 'x'
//------------------------------------------------------------------------------
void Platoon::PlaceAt(int x)
{
    int i;

    for (i = 0; i < NumSoldiers; i++) {
        if (Soldiers[i] != NULL) {
            Soldiers[i]->PlaceAt(GetPlacementPos(x, UnitPlacementRange));
        }
    }
}

//------------------------------------------------------------------------------
// Name: GetPlacementPos
// Desc: Returns a position scattered around the position 'x'. The range of the
//       scattering is determined by the 'PlacementRange' variable
//------------------------------------------------------------------------------
int Platoon::GetPlacementPos(int x, int PlacementRange)
{
    int Offset;

    Offset = (int)(((float)rand() / RAND_MAX)*PlacementRange 
        - (PlacementRange >> 1));
    return x + Offset;
}

//------------------------------------------------------------------------------
// Name: ChangeSoldiersOrders
// Desc: A very simple routine that tells all of the soldiers in the platoon
//       that their orders have changed.
//------------------------------------------------------------------------------
void Platoon::ChangeSoldiersOrders()
{
    int i;

    for (i = 0; i < NumSoldiers; i++) {
        if (Soldiers[i] != NULL) Soldiers[i]->OrdersHaveChanged = true;
    }
}

//------------------------------------------------------------------------------
// Name: Update
// Desc: Updates the state of the platoon given that 'dt' milliseconds have 
//       passed.
//------------------------------------------------------------------------------
void Platoon::Update(uint dt)
{
    FightingEntity *AlternativeEnemy;

    if (NumSubordinatesAlive <= 0) {
        NeedsDeletion = true;
        return;
    }
    UpdateBoundingBox();
    if (CurOrders == &ADVANCE || CurOrders == &WITHDRAW) {
        if (abs(DestinationX - GetCentreX()) <= GetNearDestinationX()) {
            // We have arrived
            CurOrders = &HOLD;
            DestinationX = GetCentreX();
            OrdersHaveChanged = true;
            UpdateFlagPos();
        }
    }

    // If not running away then check to see if there is anybody we can fight
    if (!IsFleeing()) {
        if (CurEnemy != NULL) {
            // If already fighting somebody then adjust the plan to match the
            // current situation.

            AlternativeEnemy = FindClosestEnemyInRange();
            if (AlternativeEnemy == CurEnemy) AlternativeEnemy = NULL;

            // Check for the situations where we would forget about the current enemy
            if (OrdersHaveChanged
                || !L->IsEntityAlive(CurEnemy) 
                || CurEnemy->IsFleeing()
                || (CurOrders != &ADVANCE && CurEnemy->IsWithdrawing())
                || (AlternativeEnemy != NULL    // There is an alternative enemy much closer to us
                    && abs(AlternativeEnemy->GetCentreX() 
                        - CurEnemy->GetCentreX()) > DISENGAGE_RANGE)) {
                
                CurEnemy = NULL;    // Forget about current enemy
                // Tell the soldiers that we are no longer engaged in combat with CurEnemy
                ChangeSoldiersOrders();
            } else {    // Update our battle plans
                
                // Find range to current enemy
                FightingEntityInRange(CurEnemy, OnTheLeft, RelativeDistance, 
                    MaxRange, NearestX, TargetPos);

                if (CurOrders == &HOLD) {
                    UpdatePlanHOLD();
                } else if (CurOrders == &ADVANCE) {
                    UpdatePlanADVANCE();
                } else if (CurOrders == &WITHDRAW) {
                    // Nothing to do really
                }
            }
        } else {
            CurEnemy = FindClosestEnemyInRange();
            // If an enemy was found then deal with it
            if (CurEnemy != NULL) {
                if (CurOrders != &ADVANCE && CurEnemy->IsWithdrawing()) {
                    // If we are not advancing then ignore withdrawing enemies
                    CurEnemy = NULL;
                } else {
                    DrawUpBattleGroups();
                    FormulateBattlePlan();
                    if (CurOrders != &WITHDRAW) ChangeSoldiersOrders();
                }
            }
        }
    }
    UpdateAction(Actions[UPDATE_MORALE], true, dt, this);
    if (OrdersHaveChanged) {
        ChangeSoldiersOrders();
        OrdersHaveChanged = false;
    }
}

//------------------------------------------------------------------------------
// Name: SoldierIsDying
// Desc: This routine is called whenever a soldier in the platoon is dying, 
//       i.e. they need deletion. If the platoon is currently engaged with the
//       enemy then a replacement may need to be found if the casualty 
//       occupied a place in the front or second rank.
//------------------------------------------------------------------------------
void Platoon::SoldierIsDying(int SoldierNum)
{
    BATTLE_POS ReplacementGroup;
    int ReplacementSoldier;

    // Make some checks to make sure that this procedure is not misused
    assert(SoldierNum >= 0 && SoldierNum < NumSoldiers);
    assert(Soldiers[SoldierNum] != NULL && Soldiers[SoldierNum]->GetNeedsDeletion());

    NumSubordinatesAlive--;
    UnitPlacementRange = NumSubordinatesAlive*SOLDIER_SPREAD;

    if (IsFighting()) {
        // Remove the dying soldier from its battle group
        RemoveSoldierFromBattleGroup(Soldiers[SoldierNum]->BattlePos, SoldierNum);

        // What should be done to replace the soldier?
        // Is there anyone to replace the soldier?
        if (BattleGroups[RESERVE].size() > 0) {

            ReplacementGroup = RESERVE;
            if (Soldiers[SoldierNum]->BattlePos == FRONT_RANK) {
                // replace with someone from the reserves or second rank
                if ((float)rand() / RAND_MAX < SECOND_RANK_REPLACEMENT_CHANCE) {
                    ReplacementGroup = SECOND_RANK;
                }
                // Get the replacement soldier and remove them from their group.
                ReplacementSoldier = BattleGroups[ReplacementGroup].back();
                BattleGroups[ReplacementGroup].pop_back();
                // Put the replacement soldier into the FRONT_RANK and tell it that
                // its orders have changed
                BattleGroups[FRONT_RANK].push_back(ReplacementSoldier);
                Soldiers[ReplacementSoldier]->BattlePos = FRONT_RANK;
                if (CurOrders != &WITHDRAW) {
                    Soldiers[ReplacementSoldier]->OrdersHaveChanged = true;
                }
            }

            // If we got a replacement for the front rank from the reserve ignore 
            // this part. Otherwise find a replacement for the second rank
            if (Soldiers[SoldierNum]->BattlePos == SECOND_RANK
                || (Soldiers[SoldierNum]->BattlePos == FRONT_RANK 
                    && ReplacementGroup == SECOND_RANK)) {
        
                // replace with someone from the reserves
                // Get the replacement soldier and remove them from their group.
                ReplacementSoldier = BattleGroups[RESERVE].back();
                BattleGroups[RESERVE].pop_back();
                // Put the replacement soldier into the SECOND_RANK and 
                // tell it that its orders have changed
                BattleGroups[SECOND_RANK].push_back(ReplacementSoldier);
                Soldiers[ReplacementSoldier]->BattlePos = SECOND_RANK;
                if (CurOrders != &WITHDRAW) {
                    Soldiers[ReplacementSoldier]->OrdersHaveChanged = true;
                }
            }

            // Do nothing a casualty from the reserves can not be replaced
        }
    }
    Soldiers[SoldierNum] = NULL;
}

//------------------------------------------------------------------------------
// Name: UpdateDestinationX
// Desc: This function updates the destination of the Platoon depending upon 
//       the value of 'NewOrders' and 'NewDestinationX'. If it does not 
//       recognise the order it has been given it passes it onto its parent 
//       'ControlledEntity' to get a decision.
//------------------------------------------------------------------------------
bool Platoon::UpdateDestinationX(const Order *NewOrders, int NewDestinationX)
{
    int OldDestinationX = DestinationX;

    if (NewOrders == &FLEE) {
        DestinationX = FallbackX;
        CurEnemy = NULL;    // Never mind the enemy, run away!
    } else {
        return ControlledEntity::UpdateDestinationX(NewOrders, NewDestinationX);
    }

    return (CurOrders != NewOrders || DestinationX != OldDestinationX);
}

//------------------------------------------------------------------------------
// Name: GetSubordinateDestinationX
// Desc: Returns the destination for a soldier. This is a position scattered
//       around the destination of the platoon.
//------------------------------------------------------------------------------
int Platoon::GetSubordinateDestinationX() 
{ 
    return GetPlacementPos(DestinationX, UnitPlacementRange);
}

//------------------------------------------------------------------------------
// Name: GetBattlePosX
// Desc: Returns the position that a soldier should be in dependant upon where
//       in the battle it is meant to be fighting.
//------------------------------------------------------------------------------
int Platoon::GetBattlePosX(BATTLE_POS BattlePos)
{
    if (BattlePos == RESERVE) {
        return GetPlacementPos(BattlePosX[RESERVE], ReservePlacementRange);
    } else {
        return GetPlacementPos(BattlePosX[BattlePos], RANK_PLACEMENT_RANGE);
    }
}

//------------------------------------------------------------------------------
// Name: UpdateBoundingBox
// Desc: The platoons position, width and height is determined by the location
//       of the soldiers which make up the platoon. This routine must be called
//       every frame so that the bounding box of the platoon remains accurate.
//------------------------------------------------------------------------------
void Platoon::UpdateBoundingBox()
{
    int i;
    int CurX, CurY;
    int MinX = numeric_limits<int>::max();
    int MaxX = numeric_limits<int>::min();
    int MinY = numeric_limits<int>::max();
    int MaxY = numeric_limits<int>::min();

    // Find the extremities
    for (i = 0; i < NumSoldiers; i++) {
        if (Soldiers[i] != NULL) {
            CurX = Soldiers[i]->GetWorldX();
            CurY = Soldiers[i]->GetWorldY();
            if (CurX < MinX) MinX = CurX;
            if (CurX > MaxX) MaxX = CurX;
            if (CurY < MinY) MinY = CurY;
            if (CurY > MaxY) MaxY = CurY;
        }
    }

    // Update the bounding box
    WorldPos.x = MinX;
    WorldPos.y = MinY;
    Width = MaxX - MinX;
    Height = MaxY - MinY;
}

//------------------------------------------------------------------------------
// Name: FormulateBattlePlan
// Desc: This routine formulates a new battle plan fo the platoon. Most of the
//       work is actually done by other routines which formulate plans for
//       specific orders. This routine handles general issues such as 
//       determining how far the reserves should be spread out and moving
//       entangled units apart.
//
// Note: This function assumes the CurEnemy != NULL, i.e. there is somebody to
//       fight. It then takes responsibility for making sure that the object
//       variables that relate to the enemy position and range are up to date.
//       Although it will not complain if the current enemy is out of range.
//------------------------------------------------------------------------------
void Platoon::FormulateBattlePlan()
{
    int SoldiersInReserve = NumSubordinatesAlive - 2*SOLDIERS_PER_RANK;

    assert(CurEnemy != NULL);

    // Update variables which refer to the enemies position and range
    FightingEntityInRange(CurEnemy, OnTheLeft, RelativeDistance, 
        MaxRange, NearestX, TargetPos);

    // Calculate the resrve placement range
    if (SoldiersInReserve > 0) {
        ReservePlacementRange = SoldiersInReserve*SOLDIER_SPREAD;
    } else {
        ReservePlacementRange = 0;
    }

    // The response to units that are entangled is simply to move them apart
    if (-RelativeDistance > MAX_ENTANGLEMENT) {
        // Move front rank to a safe place
        if (OnTheLeft) {
            BattlePosX[FRONT_RANK] = (NearestX + RelativeDistance) - RANK_PLACEMENT_RANGE;
            BattlePosX[SECOND_RANK] = BattlePosX[FRONT_RANK] - RANK_SEPERATION;
            BattlePosX[RESERVE] = BattlePosX[SECOND_RANK] - 
                (RANK_SEPERATION + ReservePlacementRange / 2);
        } else {
            BattlePosX[FRONT_RANK] = (NearestX - RelativeDistance) + RANK_PLACEMENT_RANGE;
            BattlePosX[SECOND_RANK] = BattlePosX[FRONT_RANK] + RANK_SEPERATION;
            BattlePosX[RESERVE] = BattlePosX[SECOND_RANK] + 
                (RANK_SEPERATION + ReservePlacementRange / 2);
        }
    } else {
        if (CurOrders == &HOLD) {
            FormulatePlanHOLD();
        } else if (CurOrders == &ADVANCE) {
            FormulatePlanADVANCE();
        } else if (CurOrders == &WITHDRAW) {
                // When withdrawing there is no real need to draw up a plan. If we have
                // superior range to the enemy we can keep on withdrawing. If they
                // have superior range to us then we need to withdraw faster!
                ;
        } else {
            assert(false);  // Inavlid orders are set
        }
    }
}

//------------------------------------------------------------------------------
// Name: FormulatePlanHOLD
// Desc: With a HOLD plan, the aim is to reduce casualties and move as little
//       as possible. The ideal place to HOLD is on higher ground than your
//       enemy.
//
// Note: This routine should ONLY be called from the FormulateBattlePlan
//       routine
//------------------------------------------------------------------------------
void Platoon::FormulatePlanHOLD()
{
    // NOTE: Relative distance is from the right side of the left hand platoon
    //       to the left side of the right hand platoon
    if (OnTheLeft) {
        if (MaxRange >= RelativeDistance) {
            // We are within range of the enemy.
            // Therefore place front rank just in range and let enemy come to us
            BattlePosX[FRONT_RANK] = NearestX + RANK_PLACEMENT_RANGE_2;
        } else {
            // The enemy have a greater range than us so we need to move the
            // front rank into range to stand a chance.
            BattlePosX[FRONT_RANK] = TargetPos.x - (MaxRange - RANK_PLACEMENT_RANGE_2);
        }
        // Calculate remaining rank positions
        BattlePosX[SECOND_RANK] = BattlePosX[FRONT_RANK] - RANK_SEPERATION;
        BattlePosX[RESERVE] = BattlePosX[SECOND_RANK] - 
            (RANK_SEPERATION + ReservePlacementRange / 2);
    } else {
        if (MaxRange >= RelativeDistance) {
            // We are within range of the enemy.
            // Therefore place front rank just in range and let enemy come to us
            BattlePosX[FRONT_RANK] = NearestX - RANK_PLACEMENT_RANGE_2;
        } else {
            // The enemy have a greater range than us so we need to move the
            // front rank into range to stand a chance.
            BattlePosX[FRONT_RANK] = TargetPos.x + (MaxRange - RANK_PLACEMENT_RANGE_2);
        }
        // Calculate remaining rank positions
        BattlePosX[SECOND_RANK] = BattlePosX[FRONT_RANK] + RANK_SEPERATION;
        BattlePosX[RESERVE] = BattlePosX[SECOND_RANK] + 
            (RANK_SEPERATION + ReservePlacementRange / 2);
    }
}

//------------------------------------------------------------------------------
// Name: FormulatePlanADVANCE
// Desc: When a unit is advancing the aim is to continually move the line 
//       forward, keeping the second rank in range. When attacking up hill, it
//       is also necessary to move the first and second ranks much further
//       forward so that some of the reserve are in range. This gives the
//       platoon greater firepower in order to counteract the range advantage 
//       of their opponents.
//
// Note: This routine should ONLY be called from the FormulateBattlePlan
//       routine
//------------------------------------------------------------------------------
void Platoon::FormulatePlanADVANCE()
{
    // Bunch up the reserves so that they can give more support to the attack
    ReservePlacementRange /= 2;

    // NOTE: Relative distance is from the right side of the left hand platoon
    //       to the left side of the right hand platoon
    if (OnTheLeft) {
        // Check first for range
        if (MaxRange >= RelativeDistance) {
            // We are within range of the enemy.
            // Therefore place second rank just in range to maximise firepower
            BattlePosX[SECOND_RANK] = NearestX + RANK_PLACEMENT_RANGE_2;
        } else {
            // The enemy have a greater range than us so we need to get both the
            // ranks and some of the reserve in range to stand a chance.
            BattlePosX[SECOND_RANK] = NearestX + (RelativeDistance - 
                (MaxRange/2 - (RANK_SEPERATION + RANK_PLACEMENT_RANGE_2)));
        }

        // Then check for LOS, moving the second rank forward till it can see 
        // the enemy
        if (!LS->LineOfSightExists(GetShotPos(this, BattlePosX[SECOND_RANK]), TargetPos)) {
            for(; BattlePosX[SECOND_RANK] < TargetPos.x; 
                BattlePosX[SECOND_RANK] += Landscape::TILE_WIDTH) {

                if (LS->LineOfSightExists(
                    GetShotPos(this, BattlePosX[SECOND_RANK]), TargetPos)) break;
            }
        }

        // Calculate remaining rank positions
        BattlePosX[FRONT_RANK] = BattlePosX[SECOND_RANK] + RANK_SEPERATION;
        BattlePosX[RESERVE] = BattlePosX[SECOND_RANK] - 
            (RANK_SEPERATION + ReservePlacementRange / 2);
    } else {
        // Check first for range
        if (MaxRange >= RelativeDistance) {
            // We are within range of the enemy.
            // Therefore place second rank just in range to maximise firepower
            BattlePosX[SECOND_RANK] = NearestX - RANK_PLACEMENT_RANGE_2;
        } else {
            // The enemy have a greater range than us so we need to get both the
            // ranks and some of the reserve in range to stand a chance.
            BattlePosX[SECOND_RANK] = NearestX - (RelativeDistance - 
                (MaxRange/2 - (RANK_SEPERATION + RANK_PLACEMENT_RANGE_2)));
        }

        // Then check for LOS, moving the second rank forward till it can see 
        // the enemy
        if (!LS->LineOfSightExists(GetShotPos(this, BattlePosX[SECOND_RANK]), TargetPos)) {
            for(; BattlePosX[SECOND_RANK] > TargetPos.x; 
                BattlePosX[SECOND_RANK] -= Landscape::TILE_WIDTH) {

                if (LS->LineOfSightExists(
                    GetShotPos(this, BattlePosX[SECOND_RANK]), TargetPos)) break;
            }
        }

        // Calculate remaining rank positions
        BattlePosX[FRONT_RANK] = BattlePosX[SECOND_RANK] - RANK_SEPERATION;
        BattlePosX[RESERVE] = BattlePosX[SECOND_RANK] + 
            (RANK_SEPERATION + ReservePlacementRange / 2);
    }
}

//------------------------------------------------------------------------------
// Name: UpdatePlanHOLD
// Desc: When a platoon is following HOLD orders its battle plan is to stand
//       still and keep firing on its enemy until the enemy is either dead,
//       running away or withdrawing. Its battle plan may need a slight 
//       adjustment however if ridges in the battle field obstruct its line of 
//       sight.
//------------------------------------------------------------------------------
void Platoon::UpdatePlanHOLD()
{
    int TestTargetX;
    int MaxTargetX;
    Point2D ShotStartPos, ShotEndPos;

    // Check for ridges that obstruct line of sight
    ShotStartPos = GetShotPos(this, BattlePosX[FRONT_RANK]);
    if (!LS->LineOfSightExists(ShotStartPos, TargetPos)) {
        // If there is no LOS then there must be an obstruction in the way.
        // Scan forward to see if there is a part of the enemy platoon that
        // we can hit.
        if (OnTheLeft) {
            MaxTargetX = BattlePosX[FRONT_RANK] + MaxRange;
            for (TestTargetX = TargetPos.x; TestTargetX < MaxTargetX; 
                TestTargetX += Landscape::TILE_WIDTH) {
                    
                ShotEndPos = GetShotPos(CurEnemy, TestTargetX);
                if (LS->LineOfSightExists(ShotStartPos, ShotEndPos)) {
                    TargetPos = ShotEndPos;
                    break;
                }
            }
            // If no suitable value for TargetPos has been found and we are higher up
            // we hedge our bets that somebody behind the front rank may be able to 
            // see someone past the obstruction but maybe not immediately behind it. 
            if (TargetPos.x != TestTargetX && ShotStartPos.y < TargetPos.y) {

                TargetPos = GetShotPos(CurEnemy, (TargetPos.x + MaxTargetX)/2);
            }
        } else {    // OnTheRight
            MaxTargetX = BattlePosX[FRONT_RANK] - MaxRange;
            for (TestTargetX = TargetPos.x; TestTargetX > MaxTargetX; 
                TestTargetX -= Landscape::TILE_WIDTH) {
                
                ShotEndPos = GetShotPos(CurEnemy, TestTargetX);
                if (LS->LineOfSightExists(ShotStartPos, ShotEndPos)) {
                    TargetPos = ShotEndPos;
                    break;
                }
            }
            // If no suitable value for TargetPos has been found and we are higher up
            // we hedge our bets that somebody behind the front rank may be able to 
            // see someone past the obstruction but maybe not immediately behind it. 
            if (TargetPos.x != TestTargetX && ShotStartPos.y < TargetPos.y) {

                TargetPos = GetShotPos(CurEnemy, (TargetPos.x + MaxTargetX)/2);
            }
        }
    }
}

//------------------------------------------------------------------------------
// Name: UpdatePlanADVANCE
// Desc: A platoon that is following ADVANCE orders keeps trying to move its
//       battle line forward. When a plan is initially drawn up it will rush 
//       forward to get its ranks in range and then keep on updating the plan 
//       so that it keeps its second rank in range of the enemy.
//------------------------------------------------------------------------------
void Platoon::UpdatePlanADVANCE()
{
    int HeightDiff;
    int SecondRankRange;
    Point2D SecondRankPos;

    SecondRankPos = GetShotPos(this, BattlePosX[SECOND_RANK]);
    HeightDiff = SecondRankPos.y - TargetPos.y;
    SecondRankRange = GetAvailableRange(HeightDiff);

    // Check that the second rank is within range and have a clear view to
    // the enemy
    // TODO: This is incorrect but doesn't break the game
    if (RelativeDistance > SecondRankRange  
        || !LS->LineOfSightExists(SecondRankPos, TargetPos)) {
        // Formulating a new battle plan will bring the second rank back into range
        FormulateBattlePlan();
        ChangeSoldiersOrders();
    }
}

//------------------------------------------------------------------------------
// Name: DrawUpBattleGroups
// Desc: Forms the soldiers in the current platoon up into 3 distinct groups
//       in which to fight. These are two ranks and a reserve.
//------------------------------------------------------------------------------
void Platoon::DrawUpBattleGroups()
{
    int CurSoldier = 0;
    int SoldiersAssigned;

    assert(NumSubordinatesAlive > 0);

    // Clear out the battle groups
    BattleGroups[FRONT_RANK].clear();
    BattleGroups[SECOND_RANK].clear();
    BattleGroups[RESERVE].clear();

    // Assign soldiers to the front rank
    for (SoldiersAssigned = 0; SoldiersAssigned < SOLDIERS_PER_RANK; SoldiersAssigned++) {
        // Find the next available soldier
        while (Soldiers[CurSoldier] == NULL) {
            CurSoldier++;
            if (CurSoldier >= NumSoldiers) return;
        }
        Soldiers[CurSoldier]->BattlePos = FRONT_RANK;
        BattleGroups[FRONT_RANK].push_back(CurSoldier);
        CurSoldier++;
        if (CurSoldier >= NumSoldiers) return;
    }

    // Assign soldiers to the second rank
    for (; SoldiersAssigned < 2*SOLDIERS_PER_RANK; SoldiersAssigned++) {
        // Find the next available soldier
        while (Soldiers[CurSoldier] == NULL) {
            CurSoldier++;
            if (CurSoldier >= NumSoldiers) return;
        }
        Soldiers[CurSoldier]->BattlePos = SECOND_RANK;
        BattleGroups[SECOND_RANK].push_back(CurSoldier);
        CurSoldier++;
        if (CurSoldier >= NumSoldiers) return;
    }

    // Assign the remainder to the reserve
    for (; CurSoldier < NumSoldiers; CurSoldier++) {
        if (Soldiers[CurSoldier] != NULL) {
            Soldiers[CurSoldier]->BattlePos = RESERVE;
            BattleGroups[RESERVE].push_back(CurSoldier);
        }
    }
}

//------------------------------------------------------------------------------
// Name: RemoveSoldierFromBattleGroup
// Desc: Does exactly what it says on the tin and removes a soldier from its
//       battle group.
//------------------------------------------------------------------------------
void Platoon::RemoveSoldierFromBattleGroup(BATTLE_POS BattleGroup, int SoldierNum)
{
    vector<int>::iterator Iter;

    for (Iter = BattleGroups[BattleGroup].begin();
        Iter != BattleGroups[BattleGroup].end(); Iter++) {
        if (*Iter == SoldierNum) {
            BattleGroups[BattleGroup].erase(Iter);
            return;
        }
    }
    assert(false);  // The soldier was not found in the battle group
}

//------------------------------------------------------------------------------
// Name: GetAvailableRange
// Desc: A Platoons range is determined by its Soldiers.
//------------------------------------------------------------------------------
int Platoon::GetAvailableRange(int HeightDiff)
{
    return Soldier::GetAvailableRange(HeightDiff);
}

//------------------------------------------------------------------------------
// Name: GetTargetHeight
// Desc: A Platoons target height is determined by its Soldiers.
//------------------------------------------------------------------------------
int Platoon::GetTargetHeight()
{
    return Soldier::GetTargetHeight();
}

//------------------------------------------------------------------------------
// Name: UpdateMorale
// Desc: Updates a platoons morale depending upon how many people are in the 
//       platoon and how many people have died since last the morale was
//       updated.
//------------------------------------------------------------------------------
void Platoon::UpdateMorale(void *Object)
{
    int DeadSoldiers;
    float PercentageDead;
    Platoon *APlatoon = (Platoon *)Object;

    DeadSoldiers = APlatoon->SoldiersAliveLastMoraleCheck - APlatoon->NumSubordinatesAlive;
    if (DeadSoldiers > 0) { // Morale falls
        PercentageDead = (float)DeadSoldiers / APlatoon->SoldiersAliveLastMoraleCheck * 100;
        APlatoon->Morale -= PercentageDead;
        if (APlatoon->Morale < 0) APlatoon->Morale = 0;
    } else { // Morale rises
        if (APlatoon->NumSubordinatesAlive > 100) {
            APlatoon->Morale += 5.0f;
        } else {
            APlatoon->Morale += (float)APlatoon->NumSubordinatesAlive / 20;
        }
        if (APlatoon->Morale > MAX_MORALE) APlatoon->Morale = MAX_MORALE;
    }

    if (APlatoon->IsFleeing()) {
        if (APlatoon->NumSubordinatesAlive >= MIN_NUM_SOLDIERS_NEEDED_TO_RALLY 
            && APlatoon->Morale >= RALLYPOINT) {
            APlatoon->SetOrders(&HOLD);
        }
    } else if (APlatoon->Morale <= BREAKPOINT) {
        APlatoon->SetOrders(&FLEE);        // Run away!
    }
    APlatoon->SoldiersAliveLastMoraleCheck = APlatoon->NumSubordinatesAlive;
}       
