/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: PlatoonSpawnPoint.cpp
// Desc: These entities are responsible for creating new platoons in the game 
//       world. The rate at which platoons are spwaned is dependant upon the 
//       number of factories that a side has alive. If none of a sides 
//       factories are left standing then no more platoons are created.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <list>
#include <stdlib.h>
#include <assert.h>
#include "Game.h"
#include "Level.h"
#include "Landscape.h"
#include "SDL_Interface.h"
#include "Sprites.h"
#include "Factory.h"
#include "Platoon.h"
#include "PlatoonSpawnPoint.h"

using namespace std;

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------

const int PlatoonSpawnPoint::BASIC_SPAWNING_PLATOON_RATE = 60000;

//------------------------------------------------------------------------------
// PlatoonSpawnPoint Implementation
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Name: PlatoonSpawnPoint
// Desc: Creates the platoon spawn point.
//------------------------------------------------------------------------------
PlatoonSpawnPoint::PlatoonSpawnPoint(Level *L, int x, ENT_SIDE Side) : Entity(L)
{
    Type = PLATOONSPAWNPOINT;
    this->Side = Side;
    
    OldNumFactories = NumFactories = 1;
    // A PlatooneSpawnPoint has no physical representation in the game world 
    // but needs to be positioned so that a platoon can be positioned there.
    if (Side == GOOD) {
        Frame = Sprites::GOOD_SOLDIER_SPRITE;
    } else {
        Frame = Sprites::EVIL_SOLDIER_SPRITE;
    }
    SPR->GetFrameDims(Frame, &Width, &Height);
    WorldPos.x = x;
    WorldPos.y = LS->CalculatePixelHeight(x) - Height;

    Actions[SPAWN_PLATOON] = 
        new ACTION(0, BASIC_SPAWNING_PLATOON_RATE, false, SpawnPlatoon);
}

PlatoonSpawnPoint::~PlatoonSpawnPoint() {}

//------------------------------------------------------------------------------
// Name: Update
// Desc: This routine checks to see how many factories its side has left. If
//       the number of factories has changed then it recalculates the rate
//       at which new platoons spawn.
//------------------------------------------------------------------------------
void PlatoonSpawnPoint::Update(uint dt)
{
    Factory *CurFactory;
    list<Entity *>::iterator Iter;
    
    NumFactories = 0;
    // Loop through all of the factories, counting the ones on our side
    for (Iter = L->GameEntities[FACTORY].begin();
        Iter != L->GameEntities[FACTORY].end(); Iter++) {
        
        CurFactory = (Factory *)*Iter;
        if (CurFactory->GetSide() == Side 
            && CurFactory->GetFactoryType() == Factory::PLATOON_FACTORY) NumFactories++;
    }
    
    if (NumFactories > 0) {
        if (NumFactories != OldNumFactories) {
            Actions[SPAWN_PLATOON]->Rate = BASIC_SPAWNING_PLATOON_RATE / NumFactories;
        }
        UpdateAction(Actions[SPAWN_PLATOON], true, dt, this);
        OldNumFactories = NumFactories;
    }
}

//------------------------------------------------------------------------------
// Name: SpawnPlatoon
// Desc: This routine spawns a platoon of the same side as the PlatoonSpawnPoint
//       at the position of the PlatoonSpawnPoint. If the platoon is GOOD, then
//       it is told to HOLD, if it is EVIL then it is told to advance to the
//       GOOD flag.
//------------------------------------------------------------------------------
void PlatoonSpawnPoint::SpawnPlatoon(void *Object)
{
    Platoon *NewPlatoon;
    PlatoonSpawnPoint *APSP = (PlatoonSpawnPoint *)Object;

    NewPlatoon = new Platoon(APSP->L, Level::NUM_SOLDIERS_IN_PLATOON, 
        APSP->WorldPos.x, APSP->Side);
    if (APSP->Side == GOOD) {
        APSP->L->NewTroopsReady = true;
    } else {
        NewPlatoon->SetOrders(&Platoon::ADVANCE, APSP->L->GetFlag(GOOD)->GetWorldX());
    }
    APSP->L->AddNewEntity(NewPlatoon);
}
