/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Soldier.cpp
// Desc: This is the entity that will do most of the fighting (and dying).
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <assert.h>
#include "CamelMath.h"
#include "Game.h"
#include "SDL_Interface.h"
#include "Level.h"
#include "Landscape.h"
#include "Sprites.h"
#include "Bullet.h"
#include "Platoon.h"
#include "Soldier.h"

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------

// The points earned for an EVIL soldiers demise
const int Soldier::POINTS_VALUE = 10;

// How high above the ground our enemies should aim
const int Soldier::HEIGHT_ABOVE_GROUND = 10;

// The speed of a soldier is measured in pixels per second. However, the 
// actual speed of a soldier is random and is in the range
//              SOLDIER_SPEED +/- SOLDIER_SPEED_RANGE/2
const int Soldier::SOLDIER_SPEED = 30;
const int Soldier::SOLDIER_SPEED_RANGE = 20;
const int Soldier::SOLDIER_SPEED_RANGE_2 = SOLDIER_SPEED_RANGE / 2;

const int Soldier::RANGE_OF_GUN = 150;

// Constants to do with the frames used to animate the soldier
const int Soldier::NUM_WALKING_FRAMES = 6;
const int Soldier::WALKING_RIGHT_OFFSET = 1;
const int Soldier::WALKING_LEFT_OFFSET = WALKING_RIGHT_OFFSET + NUM_WALKING_FRAMES;
const int Soldier::FIRING_OFFSET = WALKING_LEFT_OFFSET + NUM_WALKING_FRAMES;

// Rates for various actions are measured in milliseconds
const int Soldier::CHANGING_SPEED_RATE = 1000;
const int Soldier::WALKING_RATE = 125;
const int Soldier::FIRING_RATE = 1500;

//------------------------------------------------------------------------------
// Soldier Implementation
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Name: Soldier
// Desc: Constructs a soldier object. Places the soldier at position 'x'
//       and makes it soldier number 'SoldierNum' of a given platoon.
//------------------------------------------------------------------------------
Soldier::Soldier(Level *L, int x, int SoldierNum, Platoon *CurPlatoon) : Entity(L)
{
    Type = SOLDIER;

    assert(CurPlatoon != NULL);

    this->SoldierNum = SoldierNum;
    this->CurPlatoon = CurPlatoon;
    this->Side = CurPlatoon->GetSide();
    State = STANDING;
    WalkingFrame = 0;
    FiringDirection = 0;

    // Get the width and height of the soldier and then place it in 
    // the world 
    BaseFrame = (Side == GOOD ? Sprites::GOOD_SOLDIER_SPRITE : Sprites::EVIL_SOLDIER_SPRITE);
    SPR->GetFrameDims(BaseFrame, &Width, &Height);
    WorldPos.x = x;
    WorldPos.y = CalculateWorldY(WorldPos.x, Width, Height);
    RealX = (float)WorldPos.x;

    DestinationX = x;
    OrdersHaveChanged = false;

    // Set an initial speed
    ChangeSpeed(this);

    // Define the entities actions
    Actions[CHANGE_SPEED] = new ACTION(0, CHANGING_SPEED_RATE, false, ChangeSpeed);
    Actions[WALK] = new ACTION(0, WALKING_RATE, false, Walk);
    Actions[FIRE] = new ACTION(0, FIRING_RATE, false, Fire);

    CalcFrame();
}

Soldier::~Soldier() 
{
    // If the soldier was EVIL then increase the players score
    if (Side == EVIL) L->AddToPlayerScore(POINTS_VALUE);
}

//------------------------------------------------------------------------------
// Name: CalcFrame
// Desc: Works out the current frame of the soldier entity.
//------------------------------------------------------------------------------
void Soldier::CalcFrame()
{
    switch (State) {
        case STANDING:
            Frame = BaseFrame;
            break;
        case WALKING:
            if (Velocity >=0) {
                Frame = BaseFrame + WALKING_RIGHT_OFFSET + WalkingFrame;
            } else {
                Frame = BaseFrame + WALKING_LEFT_OFFSET + WalkingFrame;
            }
            break;
        case FIRING:
            Frame = BaseFrame + FIRING_OFFSET + FiringDirection;
    }
}

//------------------------------------------------------------------------------
// Name: Update
// Desc: Updates the state of the soldier given that 'dt' milliseconds have
//       passed.
//------------------------------------------------------------------------------
void Soldier::Update(uint dt)
{
    float SecondsPassed = (float)dt / 1000.0f;

    if (OrdersHaveChanged) {
        if (CurPlatoon->IsFighting() && !CurPlatoon->IsWithdrawing()) {
            DestinationX = CurPlatoon->GetBattlePosX(BattlePos);
        } else {
            DestinationX = CurPlatoon->GetSubordinateDestinationX();
        }
        OrdersHaveChanged = false;
        ChangeSpeed(this);
        // Reset the FIRE action to stop gun from firing
        Actions[FIRE]->Reset();
    }

    // Determine the current state of the soldier
    if (abs(DestinationX - WorldPos.x) > CurPlatoon->GetNearDestinationX()) {
        // Move if we are not yet at our target
        if (State != WALKING) {
            // Pick a random walking frame from 0 to NUM_WALKING_FRAMES-1
            WalkingFrame = CamelMath::Random(NUM_WALKING_FRAMES-1);
            State = WALKING;
        }
    } else if (CurPlatoon->IsFighting()) {
        // Lets cause a ruckus if we are not already
        if (State != FIRING) {
            ShotPos = Point2D(GetCentreX(), Bottom() - HEIGHT_ABOVE_GROUND);
            FiringDirection = CamelMath::GetDirectionFromBearing(
                ShotPos.BearingTo(CurPlatoon->GetTargetPos()));
            State = FIRING;
        }
    } else {
        State = STANDING;
    }
     
    // Act according to the soldiers current state
    switch (State) {
        case WALKING:
            RealX += Velocity*SecondsPassed;
            WorldPos.x = (int)RealX;
            WorldPos.y = CalculateWorldY(WorldPos.x, Width, Height);

            UpdateAction(Actions[CHANGE_SPEED], true, dt, this);
            UpdateAction(Actions[WALK], true, dt, this);

            // Check to see if the soldier has walked off screen
            if (WorldPos.x + Width <= 0 || WorldPos.x >= LS->GetWorldPixelWidth()) {
                NeedsDeletion = true;
                CurPlatoon->SoldierIsDying(SoldierNum);
            }
            break;
        case FIRING:
            UpdateAction(Actions[FIRE], true, dt, this);
    }

    // Calculate the new frame
    CalcFrame();
}

//------------------------------------------------------------------------------
// Name: FindAndHandleCollisionWithLandscape
// Desc: A soldier should not collide with the landscape.
//------------------------------------------------------------------------------
void Soldier::FindAndHandleCollisionWithLandscape()
{
    
}

//------------------------------------------------------------------------------
// Name: HandleCollision
// Desc: This routine determines how the soldier reacts to colliding with 
//       various entities.
//------------------------------------------------------------------------------
void Soldier::HandleCollision(Entity *AnEntity)
{
    // Variables used so that we can play the sound of a scream
    int ScreamSample;
    int ScreamChannel;
    int SFX_Volume;
    uchar LeftVolume, RightVolume;

    if (!NeedsDeletion) {

        if (AnEntity->GetType() == BULLET && AnEntity->GetSide() == Side 
            && ((Bullet *)AnEntity)->GetBulletType() != Bullet::PLANE_BULLET) {
            return; // We have been spared, no friendly fire from ground units
        } else if (AnEntity->GetType() == TANK && AnEntity->GetSide() == Side) {
            return;
        } else {
            NeedsDeletion = true;
            CurPlatoon->SoldierIsDying(SoldierNum);

            // Now play a sound to go with the fact that we are dying
            ScreamSample = CamelMath::Random(SDL_Interface::SOLDIER_SCREAM_BASE,
                SDL_Interface::SOLDIER_SCREAM_END);

            // Start playing the sound
            ScreamChannel = SDLI->PlayChannel(-1, SDLI->Samples[ScreamSample], 0);

            // Calculate the sfx volume and stereo effect
            SFX_Volume = SDLI->VolumeOverDistance(
                WorldPos.DistanceTo(LS->GetViewportCentreWorldPos()));
            SDLI->GetStereoVolumes(LS->GetViewportWidth() / 2, 
                LS->GetViewportCentreWorldX() - WorldPos.x, 
                &LeftVolume, &RightVolume);

            // Apply the values just calculated to the ScreamChannel
            if (ScreamChannel != SDL_Interface::NO_CHANNEL) {
                Mix_Volume(ScreamChannel, SFX_Volume);
                Mix_SetPanning(ScreamChannel, LeftVolume, RightVolume);
            }
        }
    }
}

//------------------------------------------------------------------------------
// Name: CalculateWorldY
// Desc: Given an 'x' coordinate along with the width and height of the
//       soldier. This routine works out what the coresponding 'y' should
//       be.
//------------------------------------------------------------------------------
int Soldier::CalculateWorldY(int x, int Width, int Height)
{
    int LeftY;
    int RightY;

    LeftY = LS->CalculatePixelHeight(x);
    RightY = LS->CalculatePixelHeight(x + Width);
    if (RightY > LeftY) {
        return RightY - Height;
    } else {
        return LeftY - Height;
    }
}

//------------------------------------------------------------------------------
// Name: PlaceAt
// Desc: Places the soldier at the given 'x' value and sets its target 'x'
//       to that value as well.
//------------------------------------------------------------------------------
void Soldier::PlaceAt(int x)
{
    DestinationX = x;
    WorldPos.x = x;
    WorldPos.y = CalculateWorldY(WorldPos.x, Width, Height);
    RealX = (float)WorldPos.x;
}

//------------------------------------------------------------------------------
// Name: GetAvailableRange
// Desc: Gets the available range of a soldiers gun depending upon the 
//       difference in height between the soldier and its enemy. 
//
// Note: Due to the fact that the y axis of the world is reversed (to match
//       the y axis of the screen). A -ve height difference indicates that
//       a soldier is higher than its enemy, and vice versa for a +ve
//       height difference.
//------------------------------------------------------------------------------
int Soldier::GetAvailableRange(int HeightDiff)
{
    return RANGE_OF_GUN + (HeightDiff < 0 ? -HeightDiff : 0);
}

//------------------------------------------------------------------------------
// Name: ChangeSpeed
// Desc: Sets the soldiers speed to a random value in the range 
//       SOLDIER_SPEED_RANGE +/- SOLDIER_SPEED_RANGE_2
//------------------------------------------------------------------------------
void Soldier::ChangeSpeed(void *Object)
{
    int Offset;
    Soldier *ASoldier = (Soldier *)Object;

    Offset = (int)(((float)rand() / RAND_MAX)*SOLDIER_SPEED_RANGE - SOLDIER_SPEED_RANGE_2);
    ASoldier->Velocity = (float)(SOLDIER_SPEED + Offset);

    // Reverse the speed if the target is behind us
    if (ASoldier->DestinationX < ASoldier->WorldPos.x) {
        ASoldier->Velocity = -ASoldier->Velocity;
    }
}

//------------------------------------------------------------------------------
// Name: Walk
// Desc: Updates the soldier's walking frame
//------------------------------------------------------------------------------
void Soldier::Walk(void *Object)
{
    Soldier *ASoldier = (Soldier *)Object;

    ASoldier->WalkingFrame = (ASoldier->WalkingFrame + 1)%NUM_WALKING_FRAMES;
}

//------------------------------------------------------------------------------
// Name: Fire
// Desc: Fires the soldiers gun if it can see the current enemy unit and is 
//       within range.
//------------------------------------------------------------------------------
void Soldier::Fire(void *Object)
{
    Bullet *NewBullet;
    Soldier *ASoldier = (Soldier *)Object;
    SDL_Interface *SDLI;
    Landscape *LS;
    XY_COMPONENT BulletVector;
    float BulletVectorLength;
    Point2D ShotPos, TargetPos;
    int BulletRange;

    // Variables used so that we can play the sound of a gunshot
    int GunshotSample;
    int GunshotChannel;
    int SFX_Volume;
    uchar LeftVolume, RightVolume;

    assert(ASoldier->CurPlatoon->IsFighting());

    SDLI = ASoldier->SDLI;
    LS = ASoldier->LS;

    // Find the points we are firing from and to
    ShotPos = ASoldier->ShotPos;
    TargetPos = ASoldier->CurPlatoon->GetTargetPos();

    BulletRange = GetAvailableRange(ShotPos.y - TargetPos.y);
    if (BulletRange > abs(ShotPos.x - TargetPos.x)
        && LS->LineOfSightExists(ShotPos, TargetPos)) {

        // Calculate the unit vector that the bullet will use to calculate its path
        BulletVector.x = (float)(TargetPos.x - ShotPos.x);
        BulletVector.y = (float)(TargetPos.y - ShotPos.y);
        BulletVectorLength = sqrtf(powf(BulletVector.x, 2) + powf(BulletVector.y, 2));
        BulletVector.x /= BulletVectorLength;
        BulletVector.y /= BulletVectorLength;

        NewBullet = new Bullet(ASoldier->L, ShotPos, BulletVector, 
            Bullet::SOLDIER_BULLET, BulletRange, ASoldier->Side);
        ASoldier->L->AddNewEntity(NewBullet);

        // Now create a sound to go with the firing of the gun
        GunshotSample = CamelMath::Random(SDL_Interface::SOLDIER_GUN_BASE,
            SDL_Interface::SOLDIER_GUN_END);

        // Start playing the sound
        GunshotChannel = SDLI->PlayChannel(-1, SDLI->Samples[GunshotSample], 0);

        // Calculate the sfx volume and stereo effect
        SFX_Volume = SDLI->VolumeOverDistance(
            ShotPos.DistanceTo(LS->GetViewportCentreWorldPos()));
        SDLI->GetStereoVolumes(LS->GetViewportWidth() / 2, 
            LS->GetViewportCentreWorldX() - ShotPos.x, 
            &LeftVolume, &RightVolume);

        // Apply the values just calculated to the GunshotChannel
        if (GunshotChannel != SDL_Interface::NO_CHANNEL) {
            Mix_Volume(GunshotChannel, SFX_Volume);
            Mix_SetPanning(GunshotChannel, LeftVolume, RightVolume);
        }
    }

    // Calculate the new soldier firing direction
    ASoldier->FiringDirection = CamelMath::GetDirectionFromBearing(ShotPos.BearingTo(TargetPos));
}
