/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: Tank.cpp
// Desc: Tanks are big metal monsters that should hopefully inspire fear and 
//       dread in the player when they face them, and joy and excitement when
//       they have them on their side. Tanks are constructed by a different
//       factory type and at a far slower rate than infantry platoons.
//       Having said that, tanks take a lot of effort to kill and can cause
//       a fair bit of damage. They fire explosive shells and can only be 
//       killed by a lot of bullets or a direct hit from a bomb. An explosion
//       will not cause any damage to tank.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <assert.h>
#include "CamelMath.h"
#include "Game.h"
#include "SDL_Interface.h"
#include "InputHandler.h"
#include "Level.h"
#include "Landscape.h"
#include "Sprites.h"
#include "Bullet.h"
#include "Explosion.h"
#include "TankSquad.h"
#include "Tank.h"

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------

// The points earned for an EVIL tanks demise
const int Tank::POINTS_VALUE = 400;

// How high above the ground our enemies should aim
const int Tank::HEIGHT_ABOVE_GROUND = 16;

// The hitpoints that a tank starts with
const int Tank::INITIAL_HITPOINTS = 100;
// The damage that various collisions cause
const int Tank::BOMB_DAMAGE = INITIAL_HITPOINTS;
const int Tank::PLANE_DAMAGE = INITIAL_HITPOINTS;
const int Tank::PLANE_BULLET_DAMAGE = 25;
const int Tank::TANK_BULLET_DAMAGE = 35;
const int Tank::SOLDIER_BULLET_DAMAGE = 1;

// The speed of a tank is measured in pixels per second. However, the 
// actual speed of a tank is random and is in the range
//              TANK_SPEED +/- TANK_SPEED_RANGE/2
const int Tank::TANK_SPEED = 40;
const int Tank::TANK_SPEED_RANGE = 20;
const int Tank::TANK_SPEED_RANGE_2 = TANK_SPEED_RANGE / 2;

// As this game is 2D, the tank probably 'flips' rather than turning in a 
// circle. However, this constant is used so that if a tank has to go 
// backwards a distance less than TURNING_CIRCLE it reverses rather than
// turning first.
const int Tank::TURNING_CIRCLE = 100;

const int Tank::RANGE_OF_GUN = 200;

// This is used to help determine when a tanks direction goes to PLUS_1
// or MINUS_1 as it travels between HORIZONTAL and PLUS_2 or MINUS_2
// and back again.
const int Tank::TILT_RANGE = 4;
const int Tank::TILT_RAISE = 3;

// Description of the tank's tracks for any given facing and offset
const Tank::TRACK_INFO Tank::TRACK_DETAILS[NUM_FACINGS][NUM_OFFSETS] = {
    {   // Facing LEFT
        { 26, 31, -22, 0 },     // -3
        { 31, 15, -16, 16 },    // -2
        { 29, 21, -21, 11 },    // -1
        { 26, 31, -22, 0 },     //  0
        { 21, 31, -21, -11 },   // +1
        { 16, 31, -16, -16 },   // +2
        { 26, 31, -22, 0 }      // +3
    },{  // Facing RIGHT
        { 5, 31, 22, 0 },       // -3
        { 0, 15, 16, 16 },      // -2
        { 2, 21, 21, 11 },      // -1
        { 5, 31, 22, 0 },       //  0
        { 10, 31, 21, -11 },    // +1
        { 15, 31, 16, -16 },    // +2
        { 5, 31, 22, 0 }        // +3
    }
};

// Constants to do with the frames used to animate the tank
const int Tank::NUM_DRIVING_FRAMES = 4;
const int Tank::FACING_OFFSET[NUM_FACINGS] = {
    37, // LEFT
    0   // RIGHT
};

// The direction offsets are taken from the Tanks 'BaseFrame' plus a facing offset.
const int Tank::DIRECTION_OFFSET[NUM_OFFSETS] = {
    0,  // -3
    29, // -2
    21, // -1
    0,  //  0
    8,  // +1
    15, // +2
    0   // +3
};
    
// The turret offset is taken from the 'BaseFrame' plus a facing offset 
// and a direction offset.
const int Tank::TURRET_OFFSET[NUM_OFFSETS] = {
    0,  // -3
    0,  // -2
    0,  // -1
    1,  //  0
    5,  // +1
    6,  // +2
    7   // +3
};

// Used to find if there is a valid turret offset for a given firing direction
const Tank::F_DO_TO Tank::TURRET_OFFSET_DIRECTORY[NUM_FACINGS][NUM_OFFSETS] = {
    {   // Facing LEFT
        { 0, NULL },                                                                                    // -3
        { 5, { { 11, MINUS_1 }, { 10, HORIZONTAL }, { 9, PLUS_1 }, { 8, PLUS_2 }, { 7, PLUS_3 } } },    // -2
        { 5, { { 10, MINUS_1 }, { 9, HORIZONTAL }, { 8, PLUS_1 }, { 7, PLUS_2 }, { 6, PLUS_3 } } },     // -1
        { 5, { { 9, MINUS_1 }, { 8, HORIZONTAL }, { 7, PLUS_1 }, { 6, PLUS_2 }, { 5, PLUS_3 } } },      //  0
        { 4, { { 8, MINUS_1 }, { 7, HORIZONTAL }, { 6, PLUS_1 }, { 5, PLUS_2 } } },                     // +1
        { 3, { { 7, MINUS_1 }, { 6, HORIZONTAL }, { 5, PLUS_1 } } },                                    // +2
        { 0, NULL }                                                                                     // +3
    },{ // Facing RIGHT
        { 0, NULL },                                                                                    // -3
        { 5, { { 13, MINUS_1 }, { 14, HORIZONTAL }, { 15, PLUS_1 }, { 0, PLUS_2 }, { 1, PLUS_3 } } },   // -2
        { 5, { { 14, MINUS_1 }, { 15, HORIZONTAL }, { 0, PLUS_1 }, { 1, PLUS_2 }, { 2, PLUS_3 } } },    // -1
        { 5, { { 15, MINUS_1 }, { 0, HORIZONTAL }, { 1, PLUS_1 }, { 2, PLUS_2 }, { 3, PLUS_3 } } },     //  0
        { 4, { { 0, MINUS_1 }, { 1, HORIZONTAL }, { 2, PLUS_1 }, { 3, PLUS_2 } } },                     // +1
        { 3, { { 1, MINUS_1 }, { 2, HORIZONTAL }, { 3, PLUS_1 } } },                                    // +2
        { 0, NULL }                                                                                     // +3
    }
};

// Rates for various actions are measured in milliseconds
const int Tank::CHANGING_SPEED_RATE = 1000;
const int Tank::MOVING_TRACKS_RATE = 150;
const int Tank::FIRING_RATE = 3000;
const int Tank::FIRING_RATE_VARIATION = 1000;

//------------------------------------------------------------------------------
// Tank Implementation
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Name: Tank
// Desc: Constructs a tank object. Places the tank at position 'x' and makes 
//       it tank number 'TankNum' of a given TankSquad.
//------------------------------------------------------------------------------
Tank::Tank(Level *L, int x, int TankNum, TankSquad *CurTankSquad) : Entity(L)
{
    assert(CurTankSquad != NULL);

    Type = TANK;

    // Define the victims of a tank
    NumVictims = 1;
    Victims[0] = SOLDIER;

    this->TankNum = TankNum;
    this->CurTankSquad = CurTankSquad;
    this->Side = CurTankSquad->GetSide();
    State = WAITING;
    TrackNoiseChannel = SDL_Interface::NO_CHANNEL;
    HitPoints = INITIAL_HITPOINTS;
    DrivingFrame = 0;
    TurretOffset = HORIZONTAL;

    // Get the width and height of the tank and then place it in 
    // the world 
    BaseFrame = (Side == GOOD ? Sprites::GOOD_TANK_SPRITE : Sprites::EVIL_TANK_SPRITE);
    SPR->GetFrameDims(BaseFrame, &Width, &Height);

    // Good tanks face right initially, and evil tanks face left initially
    Facing = (Side == GOOD ? RIGHT : LEFT);
    PlaceAt(x);
    OrdersHaveChanged = false;

    // Set an initial speed
    ChangeSpeed(this);

    // Define the entities actions
    Actions[CHANGE_SPEED] = new ACTION(0, CHANGING_SPEED_RATE, false, ChangeSpeed);
    Actions[MOVE_TRACKS] = new ACTION(0, MOVING_TRACKS_RATE, false, MoveTracks);
    Actions[FIRE] = new ACTION(0, FIRING_RATE, false, Fire);

    Actions[FIRE]->Rate = CamelMath::Random(FIRING_RATE - FIRING_RATE_VARIATION, 
        FIRING_RATE + FIRING_RATE_VARIATION);

    CalcFrame();
}

Tank::~Tank() 
{
    if (TrackNoiseChannel != SDL_Interface::NO_CHANNEL) Mix_HaltChannel(TrackNoiseChannel);
    // If the tank was EVIL then increase the players score
    if (Side == EVIL) L->AddToPlayerScore(POINTS_VALUE);
}

//------------------------------------------------------------------------------
// Name: CalcFrame
// Desc: Works out the current frame of the tank entity.
//------------------------------------------------------------------------------
void Tank::CalcFrame()
{
    switch (State) {
        case DRIVING:
            Frame = BaseFrame + FACING_OFFSET[Facing] + DIRECTION_OFFSET[DirectionOffset] 
                + TURRET_OFFSET[HORIZONTAL] + DrivingFrame;
            break;
        default:
            Frame = BaseFrame + FACING_OFFSET[Facing] + DIRECTION_OFFSET[DirectionOffset] 
                + TURRET_OFFSET[TurretOffset];
    }
}

//------------------------------------------------------------------------------
// Name: CalcTurretOffset
// Desc: This function tries to calculate the turret offset for a tank when
//       given details about its direction offset and facing along with the
//       direction in which it wants to fire. If it is unable to find a valid
//       turret offset then it sets 'TurretOffset' to HORIZONTAL and returns
//       'false'
//------------------------------------------------------------------------------
bool Tank::CalcTurretOffset(TANK_FACING Facing, 
    OFFSET_FROM_HORIZONTAL DirectionOffset, int FiringDirection,
    OFFSET_FROM_HORIZONTAL& TurretOffset)
{
    int i;
    bool TO_Found = false;

    TurretOffset = HORIZONTAL;
    for (i = 0; i < TURRET_OFFSET_DIRECTORY[Facing][DirectionOffset].NumAllowedFiringDirs; i++) {
        if (TURRET_OFFSET_DIRECTORY[Facing][DirectionOffset]
            .AllowedFiringDirs[i].FiringDirection == FiringDirection) {

            TurretOffset = TURRET_OFFSET_DIRECTORY[Facing][DirectionOffset]
                .AllowedFiringDirs[i].TurretOffset;
            TO_Found = true;
            break;
        }
    }
    return TO_Found;
}

//------------------------------------------------------------------------------
// Name: Update
// Desc: Updates the state of the tank given that 'dt' milliseconds have
//       passed.
//------------------------------------------------------------------------------
void Tank::Update(uint dt)
{
    int NewDestinationX;
    float SecondsPassed = (float)dt / 1000.0f;
    int TrackNoiseSample;
    int SFX_Volume;
    uchar LeftVolume, RightVolume;  // Stereo volumes

    if (OrdersHaveChanged) {
        if (CurTankSquad->IsFighting() && !CurTankSquad->IsWithdrawing()) {
            NewDestinationX = CurTankSquad->GetBattlePosX();
        } else {
            NewDestinationX = CurTankSquad->GetSubordinateDestinationX();
        }

        // Only bother turning around if there is a reasonable distance to go
        if (abs(NewDestinationX - GetCentreX()) >= TURNING_CIRCLE) {
            Facing = (NewDestinationX < GetCentreX() ? LEFT : RIGHT);
        }
        DestinationX = NewDestinationX;

        OrdersHaveChanged = false;
        ChangeSpeed(this);
        // Reset the FIRE action to stop gun from firing
        Actions[FIRE]->Reset();
        TurretOffset = HORIZONTAL;
    }

    // Determine the current state of the tank
    if (abs(DestinationX - GetCentreX()) > CurTankSquad->GetNearDestinationX()) {
        // Move if we are not yet at our target
        if (State != DRIVING) {
            // Pick a random driving frame from 0 to NUM_DRIVING_FRAMES-1
            DrivingFrame = CamelMath::Random(NUM_DRIVING_FRAMES-1);
            State = DRIVING;
        }
    } else if (CurTankSquad->IsFighting()) {
        // Lets cause a ruckus if we are not already
        if (State != FIRING) {
            State = FIRING;
            ShotPos = Point2D(GetCentreX(), Bottom() - HEIGHT_ABOVE_GROUND);
            UpdateTurretOffset();
        }
    } else {
        State = WAITING;
    }
     
    // Act according to the tanks current state
    switch (State) {
        case DRIVING:
            RealTrackX += Velocity*SecondsPassed;
            CalculatePosition();

            UpdateAction(Actions[CHANGE_SPEED], true, dt, this);
            UpdateAction(Actions[MOVE_TRACKS], true, dt, this);

            // Check to see if the tank has driven off screen
            if (WorldPos.x + Width <= 0 || WorldPos.x >= LS->GetWorldPixelWidth()) {
                NeedsDeletion = true;
                CurTankSquad->TankIsDying(TankNum);
            }
            break;
        case FIRING:
            UpdateAction(Actions[FIRE], true, dt, this);
    }

    // Update the noise that the tank may or may not be making
    if (State == DRIVING) {
        if (TrackNoiseChannel == SDL_Interface::NO_CHANNEL) {
            TrackNoiseSample = CamelMath::Random(
                SDL_Interface::TANK_LOOP_0,
                SDL_Interface::TANK_LOOP_2);

            TrackNoiseChannel = SDLI->PlayChannel(-1, 
                SDLI->Samples[TrackNoiseSample], -1);
        }

        // Change the volume and stereo balance of the sound channels 
        // depending upon where the tank is
        SFX_Volume = SDLI->VolumeOverDistance(
            GetCentrePos().DistanceTo(LS->GetViewportCentreWorldPos()));
        SDLI->GetStereoVolumes(LS->GetViewportWidth() / 2, 
            LS->GetViewportCentreWorldX() - GetCentreX(), 
            &LeftVolume, &RightVolume);

        if (TrackNoiseChannel != SDL_Interface::NO_CHANNEL) {
            Mix_Volume(TrackNoiseChannel, SFX_Volume);
            Mix_SetPanning(TrackNoiseChannel, LeftVolume, RightVolume);
        }
    } else {
        // Halt any TrackNoise if we have stopped DRIVING
        if (TrackNoiseChannel != SDL_Interface::NO_CHANNEL) {
            Mix_HaltChannel(TrackNoiseChannel);
            TrackNoiseChannel = SDL_Interface::NO_CHANNEL;
        }
    }

    // Calculate the new position and frame
    CalculatePosition();
    CalcFrame();
}

//------------------------------------------------------------------------------
// Name: UpdateTurretOffset
// Desc: Assuming that the tank is FIRING, this routine gets the tank facing
//       in the right direction and pointing its turret at the target. If it
//       is impossible (due to the tanks position) to get the turret pointing 
//       at the target, then the 'CanFire' status variable is set to 'false'.
//------------------------------------------------------------------------------
void Tank::UpdateTurretOffset()
{
    int FiringDirection;
    Point2D TargetPos;

    assert(State == FIRING);

    TargetPos = CurTankSquad->GetTargetPos();
    FiringDirection = CamelMath::GetDirectionFromBearing(ShotPos.BearingTo(TargetPos));

    Facing = (TargetPos.x < ShotPos.x ? LEFT : RIGHT);
    CanFire = CalcTurretOffset(Facing, DirectionOffset, FiringDirection, TurretOffset);

    // Try to edge the tank forward a bit to give it a better view
    if (!CanFire) DestinationX = CurTankSquad->GetBattlePosX() + 30;
}

//------------------------------------------------------------------------------
// Name: HandleCollision
// Desc: This routine determines how the tank reacts to colliding with 
//       various entities.
//------------------------------------------------------------------------------
void Tank::HandleCollision(Entity *AnEntity)
{
    Explosion *Exp;

    if (!NeedsDeletion) {

        if (AnEntity->GetType() == BULLET && AnEntity->GetSide() == Side 
            && ((Bullet *)AnEntity)->GetBulletType() != Bullet::PLANE_BULLET) {
            return; // We have been spared, no friendly fire from ground units
        } else {
            // Consider the type of entity which has hit the tank
            switch (AnEntity->GetType()) {
                case BOMB:
                    HitPoints -= BOMB_DAMAGE;
                    break;
                case PLANE:
                    HitPoints -= PLANE_DAMAGE;
                    break;
                case BULLET:
                    switch (((Bullet *)AnEntity)->GetBulletType()) {
                        case Bullet::PLANE_BULLET:
                            HitPoints -= PLANE_BULLET_DAMAGE;
                            break;
                        case Bullet::SOLDIER_BULLET:
                            HitPoints -= SOLDIER_BULLET_DAMAGE;
                            break;
                        case Bullet::TANK_BULLET:
                            HitPoints -= TANK_BULLET_DAMAGE;
                            break;
                    }
                    break;
            }
            // Destroy the tank if it's hitpoints have dropped below zero.
            if (HitPoints <= 0) {
                NeedsDeletion = true;
                CurTankSquad->TankIsDying(TankNum);

                // Create an explosion
			    Exp = new Explosion(L, Explosion::LARGE, GetCentrePos());
			    L->AddNewEntity(Exp);
            }
        }
    }
}

//------------------------------------------------------------------------------
// Name: PlaceAt
// Desc: When you place a tank at position 'x', what you are actually doing is
//       placing the centre of it's tracks at position 'x'. The WorldPos of the 
//       tank will then be worked out depending upon what frame is deemed most 
//       appropriate depending upon its facing and position on the landscape.
//------------------------------------------------------------------------------
void Tank::PlaceAt(int x)
{
    DestinationX = x;
    RealTrackX = (float)x;
    CalculatePosition();
}

//------------------------------------------------------------------------------
// Name: CalculatePosition
// Desc: This routine perfoms most of the grunt work involved in calculating
//       a tanks exact position as well its 'DirectionOffset'. Before this
//       routine is called you should have set 'RealTrackX' and the 'Facing'
//       of the tank. After this routine has finished it will have calculated
//       the exact 'TrackPos', 'WorldPos' and 'DirectionOffset' of the tank.
//
// Note: It only took a small amount of investigation to see that making the
//       tank tracks follow the contours of the landscape would be a fairly
//       hard thing to do, especially as there are only a very limited number
//       ways in which the tracks could be orientated. Therefore I opted
//       to implement this routine in the simplest way possible and constants
//       such as TILT_RANGE and TILT_RAISE were determined by trial and error.
//       The result is I think, reasonably good for the levels and tilests
//       included in the game but may not be so good levels created from other 
//       tilesets.
//       Ideally, it should be possible to have the tracks adopt any 
//       orientation but this would either require a very large number of 
//       frames or the ability to rotate the sprite on the fly.
//------------------------------------------------------------------------------
void Tank::CalculatePosition()
{
    TrackPos.x = (int)RealTrackX;
    TrackPos.y = LS->CalculatePixelHeight(TrackPos.x) - 1;

    switch(LS->GetPixelSlopeType(TrackPos.x)) {
        case Landscape::SLOPE_UP:
            if (LS->GetPixelSlopeType(TrackPos.x - TILT_RANGE) != Landscape::SLOPE_UP) {
                DirectionOffset = (Facing == LEFT ? MINUS_1 : PLUS_1);
            } else {
                DirectionOffset = (Facing == LEFT ? MINUS_2 : PLUS_2);
            }
            break;
        case Landscape::SLOPE_DOWN:
            if (LS->GetPixelSlopeType(TrackPos.x + TILT_RANGE) != Landscape::SLOPE_DOWN) {
                DirectionOffset = (Facing == LEFT ? PLUS_1 : MINUS_1);
            } else {
                DirectionOffset = (Facing == LEFT ? PLUS_2 : MINUS_2);
            }
            break;
        default:
            if (LS->GetPixelSlopeType(TrackPos.x - TILT_RANGE) == Landscape::SLOPE_DOWN) {
                DirectionOffset = (Facing == LEFT ? PLUS_1 : MINUS_1);
            } else if (LS->GetPixelSlopeType(TrackPos.x + TILT_RANGE) == Landscape::SLOPE_UP) {
                DirectionOffset = (Facing == LEFT ? MINUS_1 : PLUS_1);
            } else {
                DirectionOffset = HORIZONTAL;
            }
    }

    if (DirectionOffset == MINUS_1 || DirectionOffset == PLUS_1) TrackPos.y -= TILT_RAISE;

    WorldPos.x = TrackPos.x - TRACK_DETAILS[Facing][DirectionOffset].Width / 2 
        - TRACK_DETAILS[Facing][DirectionOffset].x;
    WorldPos.y = TrackPos.y - TRACK_DETAILS[Facing][DirectionOffset].Height / 2 
        - TRACK_DETAILS[Facing][DirectionOffset].y;
}

//------------------------------------------------------------------------------
// Name: GetAvailableRange
// Desc: Gets the available range of a tanks gun depending upon the 
//       difference in height between the tank and its enemy.
//
// Note: Due to the fact that the y axis of the world is reversed (to match
//       the y axis of the screen). A -ve height difference indicates that
//       a tank is higher than its enemy, and vice versa for a +ve
//       height difference.
//------------------------------------------------------------------------------
int Tank::GetAvailableRange(int HeightDiff)
{
    return RANGE_OF_GUN + (HeightDiff < 0 ? -HeightDiff : 0);
}

//------------------------------------------------------------------------------
// Name: ChangeSpeed
// Desc: Sets the tanks speed to a random value in the range 
//       TANK_SPEED +/- TANK_SPEED_RANGE_2
//------------------------------------------------------------------------------
void Tank::ChangeSpeed(void *Object)
{
    int Offset;
    Tank *ATank = (Tank *)Object;

    Offset = (int)(((float)rand() / RAND_MAX)*TANK_SPEED_RANGE - TANK_SPEED_RANGE_2);
    ATank->Velocity = (float)(TANK_SPEED + Offset);

    // Reverse the speed if the target is behind us
    if (ATank->DestinationX < ATank->WorldPos.x) {
        ATank->Velocity = -ATank->Velocity;
    }
}

//------------------------------------------------------------------------------
// Name: MoveTracks
// Desc: Updates the tank's frame to make it look as if it's tracks are moving.
//------------------------------------------------------------------------------
void Tank::MoveTracks(void *Object)
{
    Tank *ATank = (Tank *)Object;

    ATank->DrivingFrame = (ATank->DrivingFrame + 1)%NUM_DRIVING_FRAMES;
}

//------------------------------------------------------------------------------
// Name: Fire
// Desc: Fires the tanks gun if it can see the current enemy unit, point its
//       turret at it and is within range.
//------------------------------------------------------------------------------
void Tank::Fire(void *Object)
{
    Bullet *NewBullet;
    Tank *ATank = (Tank *)Object;
    SDL_Interface *SDLI;
    Landscape *LS;
    XY_COMPONENT BulletVector;
    float BulletVectorLength;
    Point2D ShotPos, TargetPos;
    int BulletRange;

    // Variables used so that we can play the sound of the tank firing
    int CannonChannel;
    int SFX_Volume;
    uchar LeftVolume, RightVolume;

    assert(ATank->State = FIRING);

    if (!ATank->CanFire) return;

    SDLI = ATank->SDLI;
    LS = ATank->LS;

    // Find the points we are firing from and to
    ShotPos = ATank->ShotPos;
    TargetPos = ATank->CurTankSquad->GetTargetPos();

    BulletRange = GetAvailableRange(ShotPos.y - TargetPos.y);
    if (BulletRange > abs(ShotPos.x - TargetPos.x)
        && LS->LineOfSightExists(ShotPos, TargetPos)) {

        // Calculate the unit vector that the bullet will use to calculate its path
        BulletVector.x = (float)(TargetPos.x - ShotPos.x);
        BulletVector.y = (float)(TargetPos.y - ShotPos.y);
        BulletVectorLength = sqrtf(powf(BulletVector.x, 2) + powf(BulletVector.y, 2));
        BulletVector.x /= BulletVectorLength;
        BulletVector.y /= BulletVectorLength;

        NewBullet = new Bullet(ATank->L, ShotPos, BulletVector, 
            Bullet::TANK_BULLET, BulletRange, ATank->Side);
        ATank->L->AddNewEntity(NewBullet);

        // Now create a sound to go with the firing of the gun
        // Start playing the sound
        CannonChannel = SDLI->PlayChannel(-1, SDLI->Samples[SDL_Interface::TANK_GUN], 0);

        // Calculate the sfx volume and stereo effect
        SFX_Volume = SDLI->VolumeOverDistance(
            ShotPos.DistanceTo(LS->GetViewportCentreWorldPos()));
        SDLI->GetStereoVolumes(LS->GetViewportWidth() / 2, 
            LS->GetViewportCentreWorldX() - ShotPos.x, 
            &LeftVolume, &RightVolume);

        // Apply the values just calculated to the GunshotChannel
        if (CannonChannel != SDL_Interface::NO_CHANNEL) {
            Mix_Volume(CannonChannel, SFX_Volume);
            Mix_SetPanning(CannonChannel, LeftVolume, RightVolume);
        }
    }

    // Make sure that the turret is still pointing in the right direction
    ATank->UpdateTurretOffset();

    ATank->Actions[FIRE]->Rate = CamelMath::Random(FIRING_RATE - FIRING_RATE_VARIATION, 
        FIRING_RATE + FIRING_RATE_VARIATION);
}
