/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: TankSquad.cpp
// Desc: A TankSquad is to a Tank, what a Platoon is to a Soldier. A tank 
//       squad contains far fewer tanks than a Platoon contains soldiers 
//       but then again tanks cause a bit more damage and are immune to 
//       psychology ;).
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <limits>
#include <stdlib.h>
#include <assert.h>
#include "Game.h"
#include "Sprites.h"
#include "Level.h"
#include "Landscape.h"
#include "ControlledEntity.h"
#include "Tank.h"
#include "TankSquad.h"

#undef max
#undef min

using namespace std;

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------

// This determines how spread out the tank squad is
const int TankSquad::UNIT_PLACEMENT_RANGE = 40;
const int TankSquad::BATTLE_PLACEMENT_RANGE = 20;
const int TankSquad::DISENGAGE_RANGE = 50;
const int TankSquad::MAX_ENTANGLEMENT = 50;

const int TankSquad::NEAR_DESTINATION_X = 20;

//------------------------------------------------------------------------------
// TankSquad Implementation
//------------------------------------------------------------------------------
TankSquad::TankSquad(Level *L, int InitialNumTanks, int x, ENT_SIDE Side) 
    : ControlledEntity(L, Side)
{
    int i;

    assert(InitialNumTanks > 0 && InitialNumTanks <= MAX_NUM_TANKS);

    Type = TANKSQUAD;

    this->InitialNumTanks = InitialNumTanks;
    NumSubordinatesAlive = InitialNumTanks;
    Morale = MAX_MORALE;
    CurEnemy = NULL;

    // Create the tanks
    for (i = 0; i < InitialNumTanks; i++) {
        Tanks[i] = new Tank(L, x, i, this);
        L->AddNewEntity(Tanks[i]);
    }

    // Place the tanks
    PlaceAt(x);

    // Set current orders
    CurOrders = &HOLD;
    OrdersHaveChanged = false;
    DestinationX = x;
    UpdateFlagPos();
    
    // Deal with the stuff that makes a TankSquad an entity
    if (Side == GOOD) {
        IconFrame = Sprites::GOOD_TANKSQUAD_ICON;
    } else {
        IconFrame = Sprites::EVIL_TANKSQUAD_ICON;
    }
    UpdateBoundingBox();
    Frame = 0;
}

TankSquad::~TankSquad() {}

//------------------------------------------------------------------------------
// Name: PlaceAt
// Desc: Places the TankSquad at the given position 'x'
//------------------------------------------------------------------------------
void TankSquad::PlaceAt(int x)
{
    int i;

    for (i = 0; i < InitialNumTanks; i++) {
        if (Tanks[i] != NULL) {
            Tanks[i]->PlaceAt(GetPlacementPos(x, UNIT_PLACEMENT_RANGE));
        }
    }
}

//------------------------------------------------------------------------------
// Name: GetPlacementPos
// Desc: Returns a position scattered around the position 'x'. The range of the
//       scattering is determined by the 'PlacementRange' variable
//------------------------------------------------------------------------------
int TankSquad::GetPlacementPos(int x, int PlacementRange)
{
    int Offset;

    Offset = (int)(((float)rand() / RAND_MAX)*PlacementRange 
        - (PlacementRange >> 1));
    return x + Offset;
}

//------------------------------------------------------------------------------
// Name: ChangeTanksOrders
// Desc: A very simple routine that tells all of the tanks in the TankSquad
//       that their orders have changed.
//------------------------------------------------------------------------------
void TankSquad::ChangeTanksOrders()
{
    int i;

    for (i = 0; i < InitialNumTanks; i++) {
        if (Tanks[i] != NULL) Tanks[i]->OrdersHaveChanged = true;
    }
}

//------------------------------------------------------------------------------
// Name: Update
// Desc: Updates the state of the TankSquad given that 'dt' milliseconds have 
//       passed.
//------------------------------------------------------------------------------
void TankSquad::Update(uint dt)
{
    FightingEntity *AlternativeEnemy;

    if (NumSubordinatesAlive <= 0) {
        NeedsDeletion = true;
        return;
    }
    UpdateBoundingBox();
    if (CurOrders == &ADVANCE || CurOrders == &WITHDRAW) {
        if (abs(DestinationX - GetCentreX()) <= GetNearDestinationX()) {
            // We have arrived
            CurOrders = &HOLD;
            DestinationX = GetCentreX();
            OrdersHaveChanged = true;
            UpdateFlagPos();
        }
    }

    // Check to see if there is anybody we can fight
    if (CurEnemy != NULL) {
        // If already fighting somebody then adjust the plan to match the
        // current situation.

        AlternativeEnemy = FindClosestEnemyInRange();
        if (AlternativeEnemy == CurEnemy) AlternativeEnemy = NULL;

        // Check for the situations where we would forget about the current enemy
        if (OrdersHaveChanged
            || !L->IsEntityAlive(CurEnemy) 
            || CurEnemy->IsFleeing()
            || (CurOrders != &ADVANCE && CurEnemy->IsWithdrawing())
            || (AlternativeEnemy != NULL    // There is an alternative enemy much closer to us
                && abs(AlternativeEnemy->GetCentreX() 
                    - CurEnemy->GetCentreX()) > DISENGAGE_RANGE)) {
                
            CurEnemy = NULL;    // Forget about current enemy
            // Tell the tanks that we are no longer engaged in combat with CurEnemy
            ChangeTanksOrders();
        } else {    // Update our battle plans
                
            // Find range to current enemy
            FightingEntityInRange(CurEnemy, OnTheLeft, RelativeDistance, 
                MaxRange, NearestX, TargetPos);

            if (CurOrders == &HOLD) {
                UpdatePlanHOLD();
            } else if (CurOrders == &ADVANCE) {
                UpdatePlanADVANCE();
            } else if (CurOrders == &WITHDRAW) {
                // Nothing to do really
            }
        }
    } else {
        CurEnemy = FindClosestEnemyInRange();
        // If an enemy was found then deal with it
        if (CurEnemy != NULL) {
            if (CurOrders != &ADVANCE && CurEnemy->IsWithdrawing()) {
                // If we are not advancing then ignore withdrawing enemies
                CurEnemy = NULL;
            } else {
                FormulateBattlePlan();
                if (CurOrders != &WITHDRAW) ChangeTanksOrders();
            }
        }
    }
    if (OrdersHaveChanged) {
        ChangeTanksOrders();
        OrdersHaveChanged = false;
    }
}

//------------------------------------------------------------------------------
// Name: TankIsDying
// Desc: This routine is called whenever a tank in the TankSquad is dying, 
//       i.e. they need deletion. 
//------------------------------------------------------------------------------
void TankSquad::TankIsDying(int TankNum)
{
    // Make some checks to make sure that this procedure is not misused
    assert(TankNum >= 0 && TankNum < InitialNumTanks);
    assert(Tanks[TankNum] != NULL && Tanks[TankNum]->GetNeedsDeletion());

    NumSubordinatesAlive--;
    Tanks[TankNum] = NULL;
}

//------------------------------------------------------------------------------
// Name: GetSubordinateDestinationX
// Desc: Returns the destination for a tank. This is a position scattered
//       around the destination of the TankSquad.
//------------------------------------------------------------------------------
int TankSquad::GetSubordinateDestinationX() 
{ 
    return GetPlacementPos(DestinationX, UNIT_PLACEMENT_RANGE);
}

//------------------------------------------------------------------------------
// Name: GetBattlePosX
// Desc: Returns the position that a tank should be when the TankSquad is
//       fighting.
//------------------------------------------------------------------------------
int TankSquad::GetBattlePosX()
{
    return GetPlacementPos(BattlePosX, BATTLE_PLACEMENT_RANGE);
}

//------------------------------------------------------------------------------
// Name: UpdateBoundingBox
// Desc: The TankSquad's position, width and height is determined by the 
//       location of the tanks which make up the TankSquad. This routine must 
//       be called every frame so that the bounding box of the TankSquad
//       remains accurate.
//------------------------------------------------------------------------------
void TankSquad::UpdateBoundingBox()
{
    int i;
    int CurTop, CurLeft, CurBottom, CurRight;
    int MinX = numeric_limits<int>::max();
    int MaxX = numeric_limits<int>::min();
    int MinY = numeric_limits<int>::max();
    int MaxY = numeric_limits<int>::min();

    // Find the extremities
    for (i = 0; i < InitialNumTanks; i++) {
        if (Tanks[i] != NULL) {
            // Get the bounds of the current tank and compare them to the
            // current bounds for X and Y.
            CurTop = Tanks[i]->Top();
            CurLeft = Tanks[i]->Left();
            CurBottom = Tanks[i]->Bottom();
            CurRight = Tanks[i]->Right();
            if (CurLeft < MinX) MinX = CurLeft;
            if (CurRight > MaxX) MaxX = CurRight;
            if (CurTop < MinY) MinY = CurTop;
            if (CurBottom > MaxY) MaxY = CurBottom;
        }
    }

    // Update the bounding box
    WorldPos.x = MinX;
    WorldPos.y = MinY;
    Width = MaxX - MinX;
    Height = MaxY - MinY;
}

//------------------------------------------------------------------------------
// Name: FormulateBattlePlan
// Desc: This routine formulates a new battle plan fo the TankSquad. Most of 
//       the work is actually done by other routines which formulate plans for
//       specific orders. This routine handles general issues such as 
//       moving entangled units apart.
//
// Note: This function assumes the CurEnemy != NULL, i.e. there is somebody to
//       fight. It then takes responsibility for making sure that the object
//       variables that relate to the enemy position and range are up to date.
//       Although it will not complain if the current enemy is out of range.
//------------------------------------------------------------------------------
void TankSquad::FormulateBattlePlan()
{
    assert(CurEnemy != NULL);

    // Update variables which refer to the enemies position and range
    FightingEntityInRange(CurEnemy, OnTheLeft, RelativeDistance, 
        MaxRange, NearestX, TargetPos);

    // The response to units that are entangled is simply to move them apart
    if (-RelativeDistance > MAX_ENTANGLEMENT) {
        // Move back to a safe place
        if (OnTheLeft) {
            BattlePosX = NearestX + RelativeDistance;
        } else {
            BattlePosX = NearestX - RelativeDistance;
        }
    } else {
        if (CurOrders == &HOLD) {
            FormulatePlanHOLD();
        } else if (CurOrders == &ADVANCE) {
            FormulatePlanADVANCE();
        } else if (CurOrders == &WITHDRAW) {
                // When withdrawing there is no real need to draw up a plan. If we have
                // superior range to the enemy we can keep on withdrawing. If they
                // have superior range to us then we need to withdraw faster!
                ;
        } else {
            assert(false);  // Inavlid orders are set
        }
    }
}

//------------------------------------------------------------------------------
// Name: FormulatePlanHOLD
// Desc: With a HOLD plan, the aim is to reduce casualties and move as little
//       as possible. The ideal place to HOLD is on higher ground than your
//       enemy.
//
// Note: This routine should ONLY be called from the FormulateBattlePlan
//       routine
//------------------------------------------------------------------------------
void TankSquad::FormulatePlanHOLD()
{
    // NOTE: Relative distance is from the right side of the left hand 
    //       TankSquad to the left side of the right hand TankSquad
    if (OnTheLeft) {
        if (MaxRange >= RelativeDistance) {
            // We are within range of the enemy.
            // Therefore stay where we are and let the enemy come to us
            BattlePosX = GetCentreX();
        } else {
            // The enemy have a greater range than us so we need to move 
            // into range to stand a chance.
            BattlePosX = TargetPos.x - MaxRange;
        }
    } else {
        if (MaxRange >= RelativeDistance) {
            // We are within range of the enemy.
            // Therefore stay where we are and let the enemy come to us
            BattlePosX = GetCentreX();
        } else {
            // The enemy have a greater range than us so we need to move 
            // into range to stand a chance.
            BattlePosX = TargetPos.x + MaxRange;
        }
    }
}

//------------------------------------------------------------------------------
// Name: FormulatePlanADVANCE
// Desc: When a unit is advancing the aim is to continually move the line 
//       forward, keeping the tanks in range. 

// Note: This routine should ONLY be called from the FormulateBattlePlan
//       routine
//------------------------------------------------------------------------------
void TankSquad::FormulatePlanADVANCE()
{
    // NOTE: Relative distance is from the right side of the left hand platoon
    //       to the left side of the right hand platoon
    if (OnTheLeft) {
        // Check first for range
        if (MaxRange >= RelativeDistance) {
            // We are within range of the enemy.
            // Place ourselves closer to get the advance going.
            BattlePosX = NearestX;
        } else {
            // The enemy have a greater range than us so we need to move
            // within range.
            BattlePosX = TargetPos.x - MaxRange;
        }
        BattlePosX += NEAR_DESTINATION_X;

        // Now check for LOS, moving BattlePosX forward till we can see the enemy
        if (!LS->LineOfSightExists(GetShotPos(this, BattlePosX), TargetPos)) {
            for(; BattlePosX < TargetPos.x; BattlePosX += Landscape::TILE_WIDTH) {

                if (LS->LineOfSightExists(
                    GetShotPos(this, BattlePosX), TargetPos)) break;
            }
        }
    } else {
        // Check first for range
        if (MaxRange >= RelativeDistance) {
            // We are within range of the enemy.
            // Place ourselves closer to get the advance going.
            BattlePosX = NearestX ;
        } else {
            // The enemy have a greater range than us so we need to move
            // within range.
            BattlePosX = TargetPos.x + MaxRange;
        }
        BattlePosX -= NEAR_DESTINATION_X;

        // Now check for LOS, moving BattlePosX forward till we can see the enemy
        if (!LS->LineOfSightExists(GetShotPos(this, BattlePosX), TargetPos)) {
            for(; BattlePosX > TargetPos.x; BattlePosX -= Landscape::TILE_WIDTH) {

                if (LS->LineOfSightExists(
                    GetShotPos(this, BattlePosX), TargetPos)) break;
            }
        }
    }
}

//------------------------------------------------------------------------------
// Name: UpdatePlanHOLD
// Desc: When a platoon is following HOLD orders its battle plan is to stand
//       still and keep firing on its enemy until the enemy is either dead,
//       running away or withdrawing. Its battle plan may need a slight 
//       adjustment however if ridges in the battle field obstruct its line of 
//       sight.
//------------------------------------------------------------------------------
void TankSquad::UpdatePlanHOLD()
{
    int TestTargetX;
    int MaxTargetX;
    Point2D ShotStartPos, ShotEndPos;

    // Check for ridges that obstruct line of sight
    ShotStartPos = GetShotPos(this, BattlePosX);
    if (!LS->LineOfSightExists(ShotStartPos, TargetPos)) {
        // If there is no LOS then there must be an obstruction in the way.
        // Scan forward to see if there is a part of the enemy that we can hit.
        if (OnTheLeft) {
            MaxTargetX = BattlePosX + MaxRange;
            for (TestTargetX = TargetPos.x; TestTargetX < MaxTargetX; 
                TestTargetX += Landscape::TILE_WIDTH) {
                    
                ShotEndPos = GetShotPos(CurEnemy, TestTargetX);
                if (LS->LineOfSightExists(ShotStartPos, ShotEndPos)) {
                    TargetPos = ShotEndPos;
                    break;
                }
            }
            // If no suitable value for TargetPos has been found and we are higher up
            // we hedge our bets that somebody behind the front tank may be able to 
            // see someone past the obstruction but maybe not immediately behind it. 
            if (TargetPos.x != TestTargetX && ShotStartPos.y < TargetPos.y) {

                TargetPos = GetShotPos(CurEnemy, (TargetPos.x + MaxTargetX)/2);
            }
        } else {    // OnTheRight
            MaxTargetX = BattlePosX - MaxRange;
            for (TestTargetX = TargetPos.x; TestTargetX > MaxTargetX; 
                TestTargetX -= Landscape::TILE_WIDTH) {
                
                ShotEndPos = GetShotPos(CurEnemy, TestTargetX);
                if (LS->LineOfSightExists(ShotStartPos, ShotEndPos)) {
                    TargetPos = ShotEndPos;
                    break;
                }
            }
            // If no suitable value for TargetPos has been found and we are higher up
            // we hedge our bets that somebody behind the front tank may be able to 
            // see someone past the obstruction but maybe not immediately behind it. 
            if (TargetPos.x != TestTargetX && ShotStartPos.y < TargetPos.y) {

                TargetPos = GetShotPos(CurEnemy, (TargetPos.x + MaxTargetX)/2);
            }
        }
    }
}

//------------------------------------------------------------------------------
// Name: UpdatePlanADVANCE
// Desc: A TankSquad that is following ADVANCE orders keeps trying to move its
//       battle line forward. 
//------------------------------------------------------------------------------
void TankSquad::UpdatePlanADVANCE()
{
    int HeightDiff;
    int TankRange;
    Point2D TankPos;

    if (OnTheLeft) {
        TankPos = GetShotPos(this, Left());
    } else {
        TankPos = GetShotPos(this, Right());
    }
    HeightDiff = TankPos.y - TargetPos.y;
    TankRange = GetAvailableRange(HeightDiff);

    // Check that the tanks are within range and have a clear view to the enemy
    if (abs(TankPos.x - TargetPos.x) > TankRange
        || !LS->LineOfSightExists(TankPos, TargetPos)) {
        // Formulating a new battle plan will bring the tanks back into range
        FormulateBattlePlan();
        ChangeTanksOrders();
    }
}

//------------------------------------------------------------------------------
// Name: GetAvailableRange
// Desc: A TankSquad's range is determined by its tanks.
//------------------------------------------------------------------------------
int TankSquad::GetAvailableRange(int HeightDiff)
{
    return Tank::GetAvailableRange(HeightDiff);
}

//------------------------------------------------------------------------------
// Name: GetTargetHeight
// Desc: A TankSquad's target height is determined by its tanks.
//------------------------------------------------------------------------------
int TankSquad::GetTargetHeight()
{
    return Tank::GetTargetHeight();
}
