/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: TankSquadSpawnPoint.h
// Desc: These entities are responsible for creating new TankSquads in the game 
//       world. The rate at which TankSquads are spwaned is dependant upon the 
//       number of tank factories that a side has alive. If none of a sides 
//       tank factories are left standing then no more TankSquads are created.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <list>
#include <stdlib.h>
#include <assert.h>
#include "Game.h"
#include "Level.h"
#include "Landscape.h"
#include "SDL_Interface.h"
#include "Sprites.h"
#include "Factory.h"
#include "TankSquad.h"
#include "TankSquadSpawnPoint.h"

using namespace std;

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------

const int TankSquadSpawnPoint::BASIC_SPAWNING_TANKSQUAD_RATE = 120000;

//------------------------------------------------------------------------------
// TankSquadSpawnPoint Implementation
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Name: TankSquadSpawnPoint
// Desc: Creates the TankSquad spawn point.
//------------------------------------------------------------------------------
TankSquadSpawnPoint::TankSquadSpawnPoint(Level *L, int x, ENT_SIDE Side) : Entity(L)
{
    Type = TANKSQUADSPAWNPOINT;
    this->Side = Side;
    
    OldNumFactories = NumFactories = 1;
    // A TankSquadSpawnPoint has no physical representation in the game world 
    // but needs to be positioned so that a TankSquad can be positioned there.
    if (Side == GOOD) {
        Frame = Sprites::GOOD_TANK_SPRITE;
    } else {
        Frame = Sprites::EVIL_TANK_SPRITE;
    }
    SPR->GetFrameDims(Frame, &Width, &Height);
    WorldPos.x = x;
    WorldPos.y = LS->CalculatePixelHeight(x) - Height;

    Actions[SPAWN_TANKSQUAD] = 
        new ACTION(0, BASIC_SPAWNING_TANKSQUAD_RATE, false, SpawnTankSquad);
}

TankSquadSpawnPoint::~TankSquadSpawnPoint() {}

//------------------------------------------------------------------------------
// Name: Update
// Desc: This routine checks to see how many factories its side has left. If
//       the number of factories has changed then it recalculates the rate
//       at which new TankSquads spawn.
//------------------------------------------------------------------------------
void TankSquadSpawnPoint::Update(uint dt)
{
    Factory *CurFactory;
    list<Entity *>::iterator Iter;
    
    NumFactories = 0;
    // Loop through all of the factories, counting the ones on our side
    for (Iter = L->GameEntities[FACTORY].begin();
        Iter != L->GameEntities[FACTORY].end(); Iter++) {
        
        CurFactory = (Factory *)*Iter;
        if (CurFactory->GetSide() == Side 
            && CurFactory->GetFactoryType() == Factory::TANKSQUAD_FACTORY) NumFactories++;
    }
    
    if (NumFactories > 0) {
        if (NumFactories != OldNumFactories) {
            Actions[SPAWN_TANKSQUAD]->Rate = BASIC_SPAWNING_TANKSQUAD_RATE / NumFactories;
        }
        UpdateAction(Actions[SPAWN_TANKSQUAD], true, dt, this);
        OldNumFactories = NumFactories;
    }
}

//------------------------------------------------------------------------------
// Name: SpawnTankSquad
// Desc: This routine spawns a TankSquad of the same side as the 
//       TankSquadSpawnPoint at the position of the TankSquadSpawnPoint. If the 
//       TankSquad is GOOD, then it is told to HOLD, if it is EVIL then it is 
//       told to advance to the GOOD flag.
//------------------------------------------------------------------------------
void TankSquadSpawnPoint::SpawnTankSquad(void *Object)
{
    TankSquad *NewTankSquad;
    TankSquadSpawnPoint *ATSSP = (TankSquadSpawnPoint *)Object;

    NewTankSquad = new TankSquad(ATSSP->L, Level::NUM_TANKS_IN_TANKSQUAD, 
        ATSSP->WorldPos.x, ATSSP->Side);
    if (ATSSP->Side == GOOD) {
        ATSSP->L->NewTroopsReady = true;
    } else {
        NewTankSquad->SetOrders(&TankSquad::ADVANCE, ATSSP->L->GetFlag(GOOD)->GetWorldX());
    }
    ATSSP->L->AddNewEntity(NewTankSquad);
}
