/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: CamelMath.h
// Desc: Provides any mathematical constants or objects that are needed
//       throughout the program. This is done mainly with the standard C math
//       library and also with a few custom objects.
//
//       Included in this library are
//
//       CamelMath: This object provides a number of useful static functions
//           that are maths based and used throughout Camel.
//       Point2D: Represents an integer point in 2D space.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------
#include <stdlib.h>
#include <assert.h>
#include "CamelMath.h"

//------------------------------------------------------------------------------
// Constants and Static Variables
//------------------------------------------------------------------------------
const float PI = (float)M_PI;
const float T_PI = 2.0f * (float)M_PI;
const float PI_2 = (float)M_PI_2;
const float T_PI_8 = 3.0f * (float)M_PI / 8.0f;
const float PI_4 = (float)M_PI_4;
const float PI_8 = (float)M_PI / 8.0f;

// Used to help calculate the general direction
// It is assumed that direction 0 always points directly RIGHT
const int CamelMath::UP_DIRECTION = 4;
const int CamelMath::DOWN_DIRECTION = 12;

const float CamelMath::HALF_DIRECTION_DIFF = T_PI / (2*NUM_DIRECTIONS);
// Angles given in world space, so PI_2 radians is straight up
const float CamelMath::DIRECTION_BEARINGS[NUM_DIRECTIONS] = {
    PI_2,           // 90           0
    5.0f * PI_8,    // 112.5        1
    3.0f * PI_4,    // 135          2
    7.0f * PI_8,    // 157.5        3 
    2.0f * PI_2,    // 180          4
    9.0f * PI_8,    // 202.5        5
    5.0f * PI_4,    // 225          6
    11.0f * PI_8,   // 247.5        7
    3.0f * PI_2,    // 270          8
    13.0f * PI_8,   // 292.5        9
    7.0f * PI_4,    // 315          10
    15.0f * PI_8,   // 337.5        11
    0,              // 0            12
    PI_8,           // 22.5         13
    PI_4,           // 45           14
    3.0f * PI_8,    // 67.5         15
};

const float CamelMath::AIR_THICKNESS = (1.0f/2.8f)*4.0f;

// Acceleration values are given in terms of pixels per second
// Gravity bears no relation to the real world value
const float CamelMath::MAX_GRAVITY_ACC = 425.0f;
const float CamelMath::GRAVITY_ACC[NUM_DIRECTIONS] = {
    sinf(0) * MAX_GRAVITY_ACC,
    -sinf(PI_8) * MAX_GRAVITY_ACC,
    -sinf(PI_4) * MAX_GRAVITY_ACC,
    -sinf(T_PI_8) * MAX_GRAVITY_ACC,
    -sinf(PI_2) * MAX_GRAVITY_ACC,
    -sinf(T_PI_8) * MAX_GRAVITY_ACC,
    -sinf(PI_4) * MAX_GRAVITY_ACC,
    -sinf(PI_8) * MAX_GRAVITY_ACC,
    sinf(0) * MAX_GRAVITY_ACC,
    sinf(PI_8) * MAX_GRAVITY_ACC,
    sinf(PI_4) * MAX_GRAVITY_ACC,
    sinf(T_PI_8) * MAX_GRAVITY_ACC,
    sinf(PI_2) * MAX_GRAVITY_ACC,
    sinf(T_PI_8) * MAX_GRAVITY_ACC,
    sinf(PI_4) * MAX_GRAVITY_ACC,
    sinf(PI_8) * MAX_GRAVITY_ACC
};

const XY_COMPONENT CamelMath::XY_COMPONENTS[NUM_DIRECTIONS] = {
    { cosf(0), sinf(0) },
    { cosf(PI_8), -sinf(PI_8) },
    { cosf(PI_4), -sinf(PI_4) },
    { cosf(T_PI_8), -sinf(T_PI_8) },
    { cosf(PI_2), -sinf(PI_2) },
    { -cosf(T_PI_8), -sinf(T_PI_8) },
    { -cosf(PI_4), -sinf(PI_4) },
    { -cosf(PI_8), -sinf(PI_8) },
    { -cosf(0), sinf(0) },
    { -cosf(PI_8), sinf(PI_8) },
    { -cosf(PI_4), sinf(PI_4) },
    { -cosf(T_PI_8), sinf(T_PI_8) },
    { cosf(PI_2), sinf(PI_2) },
    { cosf(T_PI_8), sinf(T_PI_8) },
    { cosf(PI_4), sinf(PI_4) },
    { cosf(PI_8), sinf(PI_8) }
};

//------------------------------------------------------------------------------
// Point2D Implementation
//------------------------------------------------------------------------------

Point2D::Point2D() {}

Point2D::Point2D(int x, int y) {
    this->x = x;
    this->y = y;
}

Point2D::~Point2D() {}

//------------------------------------------------------------------------------
// Name: DistanceTo
// Desc: Uses Pythagoras' Theorem to work out the distance to a given 
//       point p.
//------------------------------------------------------------------------------
float Point2D::DistanceTo(Point2D p)
{
    return sqrtf(powf((float)(p.x - x), 2) + powf((float)(p.y - y), 2));
}

//------------------------------------------------------------------------------
// Name: BearingTo
// Desc: This function returns the bearing to a given point p in radians.
//       Bearings are measured from the positive y-axis and range from 0 to
//       2PI radians.
//------------------------------------------------------------------------------
float Point2D::BearingTo(Point2D p)
{
    float Opp, Adj;     // Using ArcTan so need opposite and adjacent
    float Bearing;

    Opp = (float)(p.y - y);
    Adj = (float)(p.x - x);

    if (Adj == 0.0f) {  // Handle the trivial cases
        if (Opp >= 0) {
            return 0.0f;
        } else {
            return PI;
        }
    } else if (Opp == 0.0f) {
        if (Adj >= 0) {
            return PI_2;
        } else {
            return 3.0f * PI_2;
        }
    }

    Bearing = atanf(Opp / Adj);
    if (Adj < 0) Bearing += PI; // Correct angle when Adj is -ve

    Bearing = PI_2 - Bearing;   // Translate so that 0 radians is on +ve y-axis
    // Translate from range [-PI, PI] to [0, 2_PI]
    if (Bearing < 0) Bearing = T_PI + Bearing;

    assert (Bearing >= 0 && Bearing <= T_PI);
    return Bearing;
}

//------------------------------------------------------------------------------
// Name: Random
// Desc: This function returns a random number based upon certain criteria.
//       The underlying function it uses to generate random numbers is the 
//       rand() function found in the C library.
//
//       When only a (+ve) maximum is supplied, the function returns a number
//       in the range 0 to 'Max'. When both a start and end point are 
//       provided the function returns a number in the range 'Start' to 'End'.
//
// Note: You should make sure that the random number generator has been 
//       seeded (by calling 'srand') before calling this function.
//------------------------------------------------------------------------------
int CamelMath::Random(int Start, int End)
{
    int RandomNum = rand();
    // The number of possible numbers that can be picked
    int NumberOfNumbers = (End - Start) + 1;
    
    assert(Start <= End);

    // Ensure that RAND_MAX is not picked.
    // If RAND_MAX were allowed as a possible random number then there is a
    // 1 in RAND_MAX chance that End + 1 could be picked.
    while (RandomNum == RAND_MAX) RandomNum = rand();

    // Translate the random number into a number in the range 0 to
    // NumberOfNumbers-1, add it to 'Start' and return.
    return Start + (int)(((float)RandomNum / RAND_MAX) * NumberOfNumbers);
}

//------------------------------------------------------------------------------
// Name: NormaliseDirection
// Desc: Takes a number and transforms it so that it is a valid direction.
//       If the supplied direction is beyond the limits then it is wrapped
//       around to the other side.
//------------------------------------------------------------------------------
int CamelMath::NormaliseDirection(int Direction)
{
    if (Direction < 0) {
        return NormaliseDirection(Direction + NUM_DIRECTIONS);
    } else if (Direction >= NUM_DIRECTIONS) {
        return NormaliseDirection(Direction - NUM_DIRECTIONS);
    }
    return Direction;
}

//------------------------------------------------------------------------------
// Name: GetGeneralDirection
// Desc: This gets the general direction of travel without going into specifics
//       and allows entities to abstract away from the specific directions
//------------------------------------------------------------------------------
CamelMath::GENERAL_DIRECTION CamelMath::GetGeneralDirection(int Direction)
{
    assert(Direction >= 0 && Direction < NUM_DIRECTIONS);

    if (Direction == UP_DIRECTION) {
        return UP;
    } else if (Direction == DOWN_DIRECTION) {
        return DOWN;
    } else if (Direction > UP_DIRECTION && Direction < DOWN_DIRECTION) {
        return LEFT;
    } else {
        return RIGHT;
    }
}

//------------------------------------------------------------------------------
// Name: GetDirectionFromBearing
// Desc: This function takes a bearing in radians and returns one of the 
//       discrete directions used throughout the game for movement. The 
//       algorithm used isn't particularly elegant but tries to use the XP 
//       principles of being 'good enough' and as simple as possible.
//------------------------------------------------------------------------------
int CamelMath::GetDirectionFromBearing(float Bearing)
{
    int i;
    int Result = 0;
    float AbsDiffs[NUM_DIRECTIONS];

    assert(Bearing >= 0 && Bearing <= T_PI);

    // Handle the special case where the bearing is close to 2_PI
    if (T_PI - Bearing < HALF_DIRECTION_DIFF) Bearing = 0.0f;

    // Start off assuming that the result is 0
    AbsDiffs[0] = (float)fabs(Bearing - DIRECTION_BEARINGS[0]);
    for (i = 1; i < NUM_DIRECTIONS; i++) {
        // If we get any better values then adjust our answer
        AbsDiffs[i] = (float)fabs(Bearing - DIRECTION_BEARINGS[i]);
        if (AbsDiffs[i] < AbsDiffs[Result]) Result = i;
    }

    return Result;
}
