/*
    Copyright (C) 2003 Alan Broun

    This file is part of Camel.

    Camel is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Camel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Camel; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//------------------------------------------------------------------------------
// File: CamelUtility.cpp
// Desc: Utility functions that I can use when programming. At the moment it
//       consists of a mechanism for printing debug messages, a system for
//       reporting errors and for WIN32 only, a very basic profiler
//
//       Printing Debugging Messages: CU_Dbg_Printf can be used to output
//         messages to a logfile or debugger. Only works when _DEBUG is 
//         defined. CU_DBG must be placed at the beginning of any new lines
//         if the call to CU_Dbg_Printf spreads across multiple lines
//
//       Reporting Errors: Batches of user defined errors can be specified
//         by using CU_AddErrorMessages. Errors can then be set using
//         the CU_SetLast###Error error functions and printed using 
//         CU_PrintLastError. _DEBUG must be defined to use error reporting
//
//       Profiling: This works by defining a series of nested frames and then
//         using macros to indicate when the program enters and exits these
//         frames. At the end of the program, som profiling data can be 
//         printed out. Profilng only works when _DEBUG, WIN32 and 
//         _CU_PROFILING are defined
//
// 2/11/03: 
//       Added a function to trim the whitespace from a  standard C++ 
//       library string.
//
// 27/11/03:
//       I'm probably missing something blatantly obvious, but there doesn't 
//       seem to be any way of listing the files in a directory using either
//       the C runtime library or the STL. I have therefore provided a 
//       function called CU_ListFiles which when given the name of a directory
//       returns a std::list of strings which contain the names of all of 
//       the files in the directory.
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Include Files
//------------------------------------------------------------------------------

#include <string>
#include "CamelUtility.h"
#undef CU_Dbg_Printf
#undef CU_InitialiseProfiling
#undef CU_PrintProfilingResults
#undef CU_StartFrame
#undef CU_EndFrame

#ifdef WIN32

#include <windows.h>
#include <gdiplus.h>
#include <dxerr8.h>

using namespace Gdiplus;

#endif
#ifdef _SDL
#include "SDL.h"
#include "SDL_mixer.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>

using namespace std;

//------------------------------------------------------------------------------
// Defines
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
// Typedefs, Structs and Enums
//------------------------------------------------------------------------------

typedef struct _CU_ERRORMESSAGES_LIST {
    CU_ERRORMESSAGES *Slot;
    struct _CU_ERRORMESSAGES_LIST *Next;
} CU_ERRORMESSAGES_LIST;

enum LastErrorSource {
    NoError = 0,
#if defined(WIN32) && defined(_GDIPLUS)
    ErrorFromGdiPlus,
#endif
#if defined(WIN32) && defined(_DIRECTX)
    ErrorFromDirectX,
#endif
#ifdef _SDL
    ErrorFromSDL,
    ErrorFromMix,
#endif
    ErrorFromGeneral
};

//------------------------------------------------------------------------------
// Global Variables
//------------------------------------------------------------------------------
#if defined(WIN32) && defined(_GDIPLUS)
static char *GdiPlusStatusStrings[] = {
    "Ok",
    "GenericError",
    "InvalidParameter",
    "OutOfMemory",
    "ObjectBusy",
    "InsufficientBuffer",
    "NotImplemented",
    "Win32Error",
    "WrongState",
    "Aborted",
    "FileNotFound",
    "ValueOverflow",
    "AccessDenied",
    "UnknownImageFormat",
    "FontFamilyNotFound",
    "FontStyleNotFound",
    "NotTrueTypeFont",
    "UnsupportedGdiplusVersion",
    "GdiplusNotInitialized",
    "PropertyNotFound",
    "PropertyNotSupported"
};


static Status LastGdiPlusError;
#endif
#if defined(WIN32) && defined(_DIRECTX)
static HRESULT LastDirectXError;
#endif

static DWORD LastGeneralError;
static CU_ERRORMESSAGES_LIST *ErrorMessagesList = NULL;

static LastErrorSource LE_Source = NoError;

// Profiling
#ifdef WIN32
CU_TIMINGFRAME *TFArray = NULL;
static LARGE_INTEGER CounterFrequency;
int NumOfTFrames = 0;
#else 
// jamesc - not working yet
CU_TIMINGFRAME *TFArray = NULL;
static long CounterFrequency;
int NumOfTFrames = 0;
#endif

//------------------------------------------------------------------------------
// Name: CU_ShowMessage
// Desc: Displays a message that the user needs to see.
//------------------------------------------------------------------------------
void CU_ShowMessage(char *title, char *fmt, ...)
{
    va_list args;
    char msg[MAX_DEBUG_STRING_LENGTH];
    
    va_start(args, fmt);
    vsprintf(msg, fmt, args);
    va_end(args);
    
#ifdef WIN32
    MessageBox(NULL, msg, title, MB_OK);
#else
    fprintf(stderr, "%s: %s\n", title, msg);
#endif
}

//------------------------------------------------------------------------------
// Name: CU_TrimString
// Desc: Trims all of the white space from the front and back of the given
//       string. The string used to define whitespace is "\n\t ".
//------------------------------------------------------------------------------
string CU_TrimString(string AString)
{
    const string WHITESPACE = "\n\t ";
    string::size_type StringStart;
    string::size_type StringEnd;
    string::size_type StringLength;

    StringStart = AString.find_first_not_of(WHITESPACE);
    if (StringStart == string::npos) return "";     // Its all whitespace
    StringEnd = AString.find_last_not_of(WHITESPACE);
    StringLength = (StringEnd - StringStart) + 1;

    return AString.substr(StringStart, StringLength);
}

//------------------------------------------------------------------------------
// Name: CU_ListFiles
// Desc: This function returns a list of all the files that are in a given 
//       directory. It only returns the names of files, so no directories.
//
// Note: This is only a very quickly written function and is not exactly
//       general purpose. It is assumed that the directory name is either 
//       'absolute' or relative to the current path and that if is given in
//       the format '/foo/bar'. The string "/*" will be appended to it to
//       search for files so don't put on the trailing forward slash (although
//       the OS specific functions called may accept it anyway. 
//------------------------------------------------------------------------------
#if defined(WIN32)

void CU_ListFilesWIN32(const char *dir, std::list<std::string>& FileList)
{
    HANDLE SearchHandle;
    string SearchString;
    WIN32_FIND_DATA FileInfo;

    assert(dir != NULL);
    FileList.clear();   // Make sure that the filelist is empty

    SearchString = dir + string("/*");
    SearchHandle = FindFirstFile(SearchString.c_str(), &FileInfo);
    if (SearchHandle != INVALID_HANDLE_VALUE) {
        do {
            // Ignore directories
            if ((FileInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0) {
                FileList.push_back(string(FileInfo.cFileName));
            }
        } while (FindNextFile(SearchHandle, &FileInfo) != 0);
        FindClose(SearchHandle);
    }
}
#elif defined(LINUX)

#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>

void CU_ListFilesLINUX(const char *dir, std::list<std::string>& FileList)
{
    assert(dir != NULL);
    FileList.clear();   // Make sure that the filelist is empty

    DIR           *dir_p;
    struct dirent *dir_entry_p;
    dir_p = opendir(dir);

    while( NULL != (dir_entry_p = readdir(dir_p)))
    {
        struct stat fstatinfo;
        char pathname[534];
        sprintf(pathname,"%s/%s",dir,dir_entry_p->d_name);

        stat(pathname, &fstatinfo);
        //printf(" %s %d\n", dir_entry_p->d_name, fstatinfo.st_mode);
        if (S_ISREG(fstatinfo.st_mode)) {
            FileList.push_back(string(dir_entry_p->d_name));
            CU_Dbg_Printf(NULL," %s reg file \n", dir_entry_p->d_name);
        }

        //if (S_ISDIR(fstatinfo.st_mode)) 
        //printf(" %s dir \n", dir_entry_p->d_name);

    }
    closedir(dir_p);

    //CU_Dbg_Printf(NULL,"not done yet linux port - jamesc\n");
    //assert(0);
    //return Result;
}

#endif

#ifdef _DEBUG

//------------------------------------------------------------------------------
// Function Prototypes
//------------------------------------------------------------------------------
void DeleteErrorMessagesList();

//------------------------------------------------------------------------------
// Name: CU_Dbg_Printf
// Desc: Works in the style of printf to pass a string to a file. It only
//       works when the project is being compiled in the 'Debug' configuration.
//       If the statement stretches over 2 or mor lines then each extra line 
//       must be started with CU_DBG so that it can be removed in 'Release' 
//       mode. When the destination file is NULL, the function will write to
//       stderr or the debugger if compiling under WIN32
//------------------------------------------------------------------------------
void CU_Dbg_Printf(FILE *dst, char *fmt, ...)
{
    va_list args;
    char msg[MAX_DEBUG_STRING_LENGTH];
    
    va_start(args, fmt);
    vsprintf(msg, fmt, args);
    va_end(args);
    
    if (dst != NULL) {
        fprintf(dst, msg);
    } else {
#ifdef WIN32
        OutputDebugString(msg);
#else
        fprintf(stderr, msg);
#endif
    }
}

//------------------------------------------------------------------------------
// Name: CU_AddErrorMessages
// Desc: Adds a batch of user defined error messages to the Error Message 
//       list. This list is searched through when the user wants to print out
//       the last error message
//------------------------------------------------------------------------------
void CU_AddErrorMessages(CU_ERRORMESSAGES ErrorMessages)
{
    CU_ERRORMESSAGES_LIST *NewHead = NULL;
    CU_ERRORMESSAGES_LIST *CurList = ErrorMessagesList;

    // First try to find the ErrorBase in the current list
    while (CurList != NULL) {
        if (CurList->Slot->ErrorBase == ErrorMessages.ErrorBase) {
            *CurList->Slot = ErrorMessages;
            CU_Dbg_Printf(NULL, "Replacing error messages for error base %X\n", 
                ErrorMessages.ErrorBase);
            return;
        }
        CurList = CurList->Next;
    }

    // If it cannot be found, add a new item to the list
    NewHead = (CU_ERRORMESSAGES_LIST *)malloc(sizeof(CU_ERRORMESSAGES_LIST));
    if (NewHead == NULL) {
        CU_Dbg_Printf(NULL, "Error: Unable to allocate memory for error messages\n");
        return;
    } 
    NewHead->Slot = (CU_ERRORMESSAGES *)malloc(sizeof(CU_ERRORMESSAGES));
    if (NewHead->Slot == NULL) {
        CU_Dbg_Printf(NULL, "Error: Unable to allocate memory for error messages\n");
        return;
    } 

    *NewHead->Slot = ErrorMessages;
    NewHead->Next = ErrorMessagesList;
    ErrorMessagesList = NewHead;

    // If the Error Message list has just been created then arrange for its
    // future destruction
    if (NewHead->Next == NULL) atexit(DeleteErrorMessagesList);
}

//------------------------------------------------------------------------------
// Name: DeleteErrorMessagesList
// Desc: A utility function to delete the error message list when it is no
//       longer needed
//------------------------------------------------------------------------------
void DeleteErrorMessagesList()
{
    CU_ERRORMESSAGES_LIST *CurList = ErrorMessagesList;
    CU_ERRORMESSAGES_LIST *NextList;

    while (CurList != NULL) {
        if (CurList->Slot != NULL) free(CurList->Slot);
        NextList = CurList->Next;
        free(CurList);

        CurList = NextList;
    }
    ErrorMessagesList = NULL;
}

#if defined(WIN32) && defined(_GDIPLUS)
//------------------------------------------------------------------------------
// Name: CU_SetLastGdiPlusError
// Desc: Set the last error and record the fact that it came from GDI+
//------------------------------------------------------------------------------
void CU_SetLastGdiPlusError(Gdiplus::Status Err) 
{
    LE_Source = ErrorFromGdiPlus;
    LastGdiPlusError = Err;
    // Clear the Win32 error
    SetLastError(ERROR_SUCCESS);
}
#endif

#if defined(WIN32) && defined(_DIRECTX)
//------------------------------------------------------------------------------
// Name: CU_SetLastDirectXError
// Desc: Set the last error and record the fact that it came from DirectX
//------------------------------------------------------------------------------
void CU_SetLastDirectXError(HRESULT Err) 
{
    LE_Source = ErrorFromDirectX;
    LastDirectXError = Err;
    SetLastError(ERROR_SUCCESS);
}
#endif

#ifdef _SDL
//------------------------------------------------------------------------------
// Name: CU_SetLastSDLError
// Desc: Set the last error and record the fact that it came from SDL
//------------------------------------------------------------------------------
void CU_SetLastSDLError()
{
    LE_Source = ErrorFromSDL;
#ifdef WIN32
    SetLastError(ERROR_SUCCESS);
#endif
}

//------------------------------------------------------------------------------
// Name: CU_SetLastMixError
// Desc: Set the last error and record the fact that it came from SDL_mixer
//------------------------------------------------------------------------------
void CU_SetLastMixError()
{
    LE_Source = ErrorFromMix;
#ifdef WIN32
    SetLastError(ERROR_SUCCESS);
#endif
}
#endif //_SDL

//------------------------------------------------------------------------------
// Name: CU_SetLastError
// Desc: Set the last error and record the fact that it was a general error
//------------------------------------------------------------------------------
void CU_SetLastError(DWORD Err) 
{
    LE_Source = ErrorFromGeneral;
    LastGeneralError = Err;
#ifdef WIN32
    SetLastError(ERROR_SUCCESS);
#endif
}

//------------------------------------------------------------------------------
// Name: CU_PrintLastError
// Desc: Print out the name location and type of the last error to the 
//       attached debugger
//------------------------------------------------------------------------------
void CU_PrintLastError(char *Filename, DWORD LineNum)
{
    CU_ERRORMESSAGES_LIST *CurList = ErrorMessagesList;

#ifdef WIN32
    // Check for a predefined Win32 error                               
    if (GetLastError() != ERROR_SUCCESS) {
        char *msg = NULL;

        // Get message from Win32
        FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER 
            |  FORMAT_MESSAGE_FROM_SYSTEM 
            | FORMAT_MESSAGE_IGNORE_INSERTS,
            NULL, GetLastError(),
            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
            msg, 0,NULL);

        CU_Dbg_Printf(NULL, "%s(%i): Win32 Error - %s\n", Filename, LineNum, msg);

        // Free the memory allocated for the message
        LocalFree(msg);
        return;
    }
#endif

    switch (LE_Source) {
        case NoError:
            break;
#if defined(WIN32) && defined(_GDIPLUS)
        case ErrorFromGdiPlus:
            CU_Dbg_Printf(NULL, "%s(%i): GDI+ Error - %s\n",
                Filename, LineNum, GdiPlusStatusStrings[LastGdiPlusError]);
            break;
#endif
#if defined(WIN32) && defined(_DIRECTX)
        case ErrorFromDirectX:

            DXTrace(Filename, LineNum, LastDirectXError, NULL, FALSE);
            break;
#endif
#ifdef _SDL
        case ErrorFromSDL:
            CU_Dbg_Printf(NULL, "%s(%i): SDL Error - %s\n",
                Filename, LineNum, SDL_GetError());
            break;
        case ErrorFromMix:
            CU_Dbg_Printf(NULL, "%s(%i): SDL_mixer Error - %s\n",
                Filename, LineNum, Mix_GetError());
            break;
#endif
        case ErrorFromGeneral:
            if (LastGeneralError & USER_DEFINED_ERROR_MASK) {
                // User defined error
                while (CurList != NULL) {
                    assert(CurList->Slot != NULL);

                    if (LastGeneralError >= CurList->Slot->ErrorBase
                        && LastGeneralError < CurList->Slot->ErrorBase 
                        + CurList->Slot->NumErrors ) {

                        CU_Dbg_Printf(NULL, "%s(%i): Error - %s\n", Filename, LineNum, 
                            CurList->Slot->ErrorStrings[LastGeneralError 
                            - CurList->Slot->ErrorBase]);
                        return;
                    }

                    CurList = CurList->Next;
                }
                             
                CU_Dbg_Printf(NULL, "%s(%i): Unknown user defined error\n",
                        Filename, LineNum);
            } else {
#ifdef WIN32
                // predefined Win32 error       
                char *msg = NULL;

                // Get message from Win32
                FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER 
                    |  FORMAT_MESSAGE_FROM_SYSTEM 
                    | FORMAT_MESSAGE_IGNORE_INSERTS,
                    NULL, LastGeneralError,
                    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                    msg, 0,NULL);

                CU_Dbg_Printf(NULL, "%s(%i): Win32 Error - %s\n",
                        Filename, LineNum, msg);

                // Free the memory allocated for the message
                LocalFree(msg);                
#else
                CU_Dbg_Printf(NULL, "%s(%i): Unknown user defined error\n",
                        Filename, LineNum);
#endif
            }
            break;
    }
}

#endif

//jamesc #if defined(_DEBUG) && defined(_CU_PROFILING) && defined(WIN32)
#if defined(_DEBUG) && defined(_CU_PROFILING)

//------------------------------------------------------------------------------
// Name: NumFrames
// Desc: Returns the total number of frames in a list of CU_TIMINGFRAMEDATA
//       structs
//------------------------------------------------------------------------------
int NumFrames(CU_TIMINGFRAMEDATA *TFData, int Index)
{
    if (Index == -1) return 0;
    
    return 1 + NumFrames(TFData, TFData[Index].Children) 
        + NumFrames(TFData, TFData[Index].Next);
}

//------------------------------------------------------------------------------
// Name: InitialiseFrame
// Desc: Initialises a CU_TIMINGFRAME with data from the corresponding
//       CU_TIMINGFRAMEDATA data struct
//------------------------------------------------------------------------------
void InitialiseFrame(CU_TIMINGFRAME *TFrame, CU_TIMINGFRAMEDATA *TFData, int Index)
{
    TFrame->Name = TFData[Index].Name;
    
    if (TFData[Index].Children != -1) {
        TFrame->Children = TFrame + 1;
        InitialiseFrame(TFrame->Children, TFData, TFData[Index].Children);
    } else {
        TFrame->Children = NULL;
    }
    
    if (TFData[Index].Next != -1) {
        TFrame->Next = TFrame + NumFrames(TFData, TFData[Index].Children) + 1;
        InitialiseFrame(TFrame->Next, TFData, TFData[Index].Next);
    } else {
        TFrame->Next = NULL;
    }
}

//------------------------------------------------------------------------------
// Name: ShutdownProfiling
// Desc: Clean up after the profiler
//------------------------------------------------------------------------------
void ShutdownProfiling() 
{
    // Deallocate memory used for the array
    if (TFArray != NULL) {
        free(TFArray);
        TFArray = NULL;
    }
}

//------------------------------------------------------------------------------
// Name: PrintResults
// Desc: Recursively prints the results from a CU_TIMINGFRAME struct, 
//       the implementation is crude but it works
//------------------------------------------------------------------------------

#define INDENT          2
#define FRAME_NAME      44          // The name of the timing frame
#define HITCOUNT        10          // The number of times the frame was active
#define T_FRAMETIME     14          // The total amount of time spent in the frame
#define A_FRAMETIME     15          // The average amount of time spent in the frame
#define P_OF_PARENT     11          // The amount contributed to the parents frame time
#define P_NO_CHILDREN   11          // The percentage of the frame time not taken 
                                    // up by children

#define ADD_DIVIDER()   *(CurChar++) = '|'

#define HITCOUNT_S          "%10i"
#define T_FRAMETIME_S       "%14.3f"
#define A_FRAMETIME_S       "%15.3f"
#define P_OF_PARENT_S       "%11.1f"
#define P_NO_CHILDREN_S     "%11.1f"

void PrintResults(CU_TIMINGFRAME *TFrame, CU_TIMINGFRAME *Parent, 
    unsigned int Indent)
{
    char OutputString[200];
    char *CurChar = OutputString;
    __int64 ChildTicks = 0;
    CU_TIMINGFRAME *CurFrame;    
    double TotalFrameTime;
    double AverageFrameTime;

    // Return if the frame is empty
    if (TFrame == NULL) return;
    
    // Clear the output string
    memset(OutputString, ' ', 200);    
    // Calculate the number of ticks used in child frames
    CurFrame = TFrame->Children;
    while (CurFrame != NULL) {
        ChildTicks += CurFrame->TotalTicks;
        CurFrame = CurFrame->Next;
    }
    // Calculate the time spent in the frame
    TotalFrameTime = ((double)TFrame->TotalTicks / CounterFrequency.QuadPart);
    AverageFrameTime = (TotalFrameTime / TFrame->HitCount) * 1000;
    
    // Print the current row of the table
    ADD_DIVIDER();      // Frame Name    
    CurChar = OutputString + 1 + (Indent > FRAME_NAME ? FRAME_NAME : Indent);
    sprintf(CurChar, "%s                                             ", TFrame->Name);
    CurChar = OutputString + 1 + FRAME_NAME;
    ADD_DIVIDER();      // Hit Count
    sprintf(CurChar, HITCOUNT_S "                      ", TFrame->HitCount);
    CurChar += HITCOUNT;
    ADD_DIVIDER();      // Total FrameTime(s)
    sprintf(CurChar, T_FRAMETIME_S "                 ", TotalFrameTime);
    CurChar += T_FRAMETIME;
    ADD_DIVIDER();      // Average FrameTime(ms)
    sprintf(CurChar, A_FRAMETIME_S "                 ", AverageFrameTime);
    CurChar += A_FRAMETIME;
    ADD_DIVIDER();      // % of Parent
    sprintf(CurChar, P_OF_PARENT_S "                 ", 
        (float)(Parent == NULL ? 0 : ((float)TFrame->TotalTicks / Parent->TotalTicks) * 100));
    CurChar += P_OF_PARENT;
    ADD_DIVIDER();      // % no Children
    sprintf(CurChar, P_NO_CHILDREN_S "                 ", 
        (float)(((float)(TFrame->TotalTicks - ChildTicks) / TFrame->TotalTicks) * 100));
    CurChar += P_NO_CHILDREN;
    
    ADD_DIVIDER();
    *CurChar = '\0';
    CU_Dbg_Printf(NULL, "%s\n", OutputString);
 
    // Now print the rest of the table   
    PrintResults(TFrame->Children, TFrame, Indent + INDENT);
    PrintResults(TFrame->Next, Parent, Indent);    
}

//------------------------------------------------------------------------------
// Name: CU_InitialiseProfiling
// Desc: Allocates and initialises memory to be used in profiling
//------------------------------------------------------------------------------
void CU_InitialiseProfiling(CU_TIMINGFRAMEDATA *TFData)
{   
    assert(TFData != NULL);
    
    NumOfTFrames = NumFrames(TFData, 0);
    
    CU_Dbg_Printf(NULL, "Found %i frames\n", NumOfTFrames);
    
    // Get the frequency of the performance counter, if it doesn't exist then 
    //print out a debug message warning that results will be invalid.
    if (!QueryPerformanceFrequency(&CounterFrequency)) {
        CU_Dbg_Printf(NULL, "Warning: No performance counter available, "
            "no results will be produced\n");
    }

    // Allocate array for the timing frames and initialise frames
    TFArray = (CU_TIMINGFRAME *)malloc(NumOfTFrames * sizeof(CU_TIMINGFRAME));
    ZeroMemory(TFArray, NumOfTFrames * sizeof(CU_TIMINGFRAME));
    InitialiseFrame(&TFArray[0], TFData, 0);
        
    // Place the shutdown procedure in the atexit list
    atexit(ShutdownProfiling);
}

//------------------------------------------------------------------------------
// Name: CU_PrintProfilingResults
// Desc: Prints out the results obtained from profiling
//------------------------------------------------------------------------------
void CU_PrintProfilingResults()
{
    // No results if there is no performance counter
    if (CounterFrequency.QuadPart == 0) return;
    
    CU_Dbg_Printf(NULL, "\nProfiling Results\n\n");    
    
    CU_Dbg_Printf(NULL, "%I64i\n\n", CounterFrequency.QuadPart);
    
    CU_Dbg_Printf(NULL, "----------------------------------------------------------------------------------------------------------------\n");                
    CU_Dbg_Printf(NULL, "|                                            |    Hit   |   Total      |   Average     |   %% of    |  %% no     |\n");
    CU_Dbg_Printf(NULL, "|             Frame Name                     |   Count  | FrameTime(s) | FrameTime(ms) |  Parent   |  Children |\n");
    CU_Dbg_Printf(NULL, "----------------------------------------------------------------------------------------------------------------\n");                                                           
    PrintResults(&TFArray[0], NULL, 2);
    CU_Dbg_Printf(NULL, "----------------------------------------------------------------------------------------------------------------\n");
    CU_Dbg_Printf(NULL, "\n\n");
}

//------------------------------------------------------------------------------
// Name: CU_ResetFrames
// Desc: Clears the hits and ticks of all of the frames to 0
//------------------------------------------------------------------------------
void CU_ResetFrames()
{
    int i;
    
    for (i = 0; i < NumOfTFrames; i++) {
        TFArray[i].HitCount = 0;
        TFArray[i].TotalTicks = 0;
        TFArray[i].StartTicks = TFArray[i].EndTicks = 0;
    }
}

#endif // defined(_DEBUG) && defined(_CU_PROFILING) && defined(WIN32)
